# GUI Application automation and testing library
# Copyright (C) 2006-2018 Mark Mc Mahon and Contributors
# https://github.com/pywinauto/pywinauto/graphs/contributors
# http://pywinauto.readthedocs.io/en/latest/credits.html
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of pywinauto nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# generated by 'xml2py'
# flags 'commctrl.xml -d -w -kde -v -o win32defines.py'

"""Definition of Windows defines"""
from __future__ import unicode_literals
from ctypes import c_int

__revision__ = "$Revision$"



VFT_FONT = 4 # Variable c_long
WS_GROUP = 131072 # Variable c_long
REGCLS_SURROGATE = 8
VIEW_E_DRAW = -2147221184 # Variable c_long
TCM_GETITEMW = 4924 # Variable c_int
CBF_FAIL_POKES = 65536 # Variable c_int
WA_INACTIVE = 0 # Variable c_int
CTRY_IRELAND = 353 # Variable c_int
SECURITY_DESCRIPTOR_REVISION1 = 1 # Variable c_int
BAUD_57600 = 262144 # Variable c_ulong
IMAGE_REL_BASED_MIPS_JMPADDR = 5 # Variable c_int
LB_SETANCHORINDEX = 412 # Variable c_int
MM_WIM_OPEN = 958 # Variable c_int
FOURCC_LIST = 1414744396 # Variable c_ulong
SCS_32BIT_BINARY = 0 # Variable c_int
VFT_DLL = 2 # Variable c_long
ARW_BOTTOMRIGHT = 1 # Variable c_long
CBEM_GETITEMW = 1037 # Variable c_int
IMPORT_OBJECT_NAME_UNDECORATE = 3
TTM_HITTESTW = 1079 # Variable c_int
SND_RESOURCE = 262148 # Variable c_long
TBM_GETCHANNELRECT = 1050 # Variable c_int
IMAGE_SYM_TYPE_VOID = 1 # Variable c_int

_SECURITY_IMPERSONATION_LEVEL = c_int # enum
SecurityAnonymous = 0
SecurityIdentification = 1
SecurityImpersonation = 2
SecurityDelegation = 3
TMPF_TRUETYPE = 4 # Variable c_int
PAN_STROKE_GRADUAL_VERT = 4 # Variable c_int
WNNC_NET_AS400 = 720896 # Variable c_int
TC_CR_ANY = 16 # Variable c_int
PAN_STROKE_GRADUAL_TRAN = 3 # Variable c_int
cmb13 = 1148 # Variable c_int
ERROR_DEVICE_DOOR_OPEN = 1166 # Variable c_long
HKEY_CLASSES_ROOT = 2147483648 # Variable POINTER(HKEY__)
OLEOBJ_S_CANNOT_DOVERB_NOW = 262529 # Variable c_long
JobObjectBasicAccountingInformation = 1
SZDDE_ITEM_ITEMLIST = 'TopicItemList' # Variable POINTER(c_wchar)
CO_E_LOOKUPACCNAMEFAILED = -2147220977 # Variable c_long
TB_BUTTONSTRUCTSIZE = 1054 # Variable c_int
META_POLYLINE = 805 # Variable c_int
FS_CYRILLIC = 4 # Variable c_long
FUNC_STATIC = 3
stc5 = 1092 # Variable c_int
S_SERDSH = -11 # Variable c_int
DMPAPER_FANFOLD_STD_GERMAN = 40 # Variable c_int
SND_ASYNC = 1 # Variable c_int
MAX_DEFAULTCHAR = 2 # Variable c_int
PROCESS_VM_OPERATION = 8 # Variable c_int
TIME_SMPTE = 8 # Variable c_int
FOREGROUND_BLUE = 1 # Variable c_int
SELECTDIB = 41 # Variable c_int
JOB_OBJECT_UILIMIT_HANDLES = 1 # Variable c_int
MM_DRVM_ERROR = 979 # Variable c_int
TVM_SETSCROLLTIME = 4385 # Variable c_int
CF_SYLK = 4 # Variable c_int
WSAECONNREFUSED = 10061 # Variable c_int
HELP_QUIT = 2 # Variable c_long
STN_DBLCLK = 1 # Variable c_int
SBM_GETPOS = 225 # Variable c_int
DOWNLOADFACE = 514 # Variable c_int
ARW_STARTMASK = 3 # Variable c_long
LANG_PORTUGUESE = 22 # Variable c_int
STGFMT_ANY = 4 # Variable c_int
LVN_ITEMCHANGED = 4294967195 # Variable c_uint
WAVE_INVALIDFORMAT = 0 # Variable c_int
XST_EXECSENT = 9 # Variable c_int
COMPRESSION_ENGINE_MAXIMUM = 256 # Variable c_int
SC_MANAGER_MODIFY_BOOT_CONFIG = 32 # Variable c_int
RT_DLGINCLUDE = 23 # Variable POINTER(c_wchar)
HSHELL_ACTIVATESHELLWINDOW = 3 # Variable c_int
LOGPIXELSX = 88 # Variable c_int
TYPEFLAG_FRESTRICTED = 512
SERVICE_NO_CHANGE = 4294967295 # Variable c_uint
LVCOLUMNA_V1_SIZE = 24 # Variable c_uint
PIPE_WAIT = 0 # Variable c_int
ERROR_HOST_NODE_NOT_GROUP_OWNER = 5016 # Variable c_long
STARTF_USEHOTKEY = 512 # Variable c_int
EMR_HEADER = 1 # Variable c_int
MCI_VD_GETDEVCAPS_NORMAL_RATE = 16389 # Variable c_long
SPLREG_DEFAULT_SPOOL_DIRECTORY = 'DefaultSpoolDirectory' # Variable POINTER(c_wchar)
stc16 = 1103 # Variable c_int
LVNI_CUT = 4 # Variable c_int
LVM_SETHOTITEM = 4156 # Variable c_int
PAN_WEIGHT_BLACK = 10 # Variable c_int
CTRY_GREECE = 30 # Variable c_int
DT_DISPFILE = 6 # Variable c_int
PSINJECT_BOUNDINGBOX = 24 # Variable c_int
SB_SETMINHEIGHT = 1032 # Variable c_int
edt3 = 1154 # Variable c_int
NO_PROPAGATE_INHERIT_ACE = 4 # Variable c_int
ERROR_FLOPPY_WRONG_CYLINDER = 1123 # Variable c_long
DS_LOCALEDIT = 32 # Variable c_long
TBSTYLE_TRANSPARENT = 32768 # Variable c_int
lst7 = 1126 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_VOICEIN = 8 # Variable c_long
AF_PUP = 4 # Variable c_int
IMAGE_REL_SH3_DIRECT4_LONG = 8 # Variable c_int
VARFLAG_FDEFAULTBIND = 32
SM_CXMINTRACK = 34 # Variable c_int
R2_NOTMERGEPEN = 2 # Variable c_int
EM_GETWORDBREAKPROC = 209 # Variable c_int
RPC_C_PROTECT_LEVEL_PKT = 4 # Variable c_int
WGL_SWAP_UNDERLAY5 = 1048576 # Variable c_int
RGN_DIFF = 4 # Variable c_int
SPLDS_PRINT_OWNER = 'printOwner' # Variable POINTER(c_wchar)
PAN_STROKE_INSTANT_VERT = 8 # Variable c_int
JOY_BUTTON23 = 4194304 # Variable c_long
JOB_CONTROL_RESUME = 2 # Variable c_int
LVIF_NORECOMPUTE = 2048 # Variable c_int
MNC_SELECT = 3 # Variable c_int
SM_CYMIN = 29 # Variable c_int
LR_COPYFROMRESOURCE = 16384 # Variable c_int
DDE_FACKREQ = 32768 # Variable c_int
WSB_PROP_VBKGCOLOR = 64 # Variable c_long
NORM_IGNORESYMBOLS = 4 # Variable c_int
SORT_GEORGIAN_MODERN = 1 # Variable c_int
UI_CAP_2700 = 1 # Variable c_int
ico4 = 1087 # Variable c_int
TOKEN_WRITE = 131296 # Variable c_long
MCI_VD_FORMAT_TRACK = 16385 # Variable c_int
C3_NOTAPPLICABLE = 0 # Variable c_int
DMPAPER_B5 = 13 # Variable c_int
FADF_RESERVED = 61448 # Variable c_int
SPLREG_EVENT_LOG = 'EventLog' # Variable POINTER(c_wchar)
OFN_NONETWORKBUTTON = 131072 # Variable c_int
DRAGDROP_E_LAST = 2147746063 # Variable c_ulong
TBN_GETINFOTIPA = 4294966578 # Variable c_uint
IDHOT_SNAPWINDOW = -1 # Variable c_int
SORT_KOREAN_KSC = 0 # Variable c_int
JOY_BUTTON25 = 16777216 # Variable c_long
URLACTION_HTML_MIN = 5632 # Variable c_int
psh2 = 1025 # Variable c_int
PGK_CONTROL = 2 # Variable c_int
PST_MODEM = 6 # Variable c_ulong
DATA_S_LAST = 262463 # Variable c_long
LF_FULLFACESIZE = 64 # Variable c_int
ERROR_SET_NOT_FOUND = 1170 # Variable c_long
MIXER_OBJECTF_HANDLE = 2147483648 # Variable c_ulong
CBR_1200 = 1200 # Variable c_int
WNNC_NET_LANTASTIC = 655360 # Variable c_int
SO_DISCOPTLEN = 28679 # Variable c_int
CTRY_KUWAIT = 965 # Variable c_int
ILLUMINANT_A = 1 # Variable c_int
IMAGE_SYM_CLASS_END_OF_FUNCTION = 255 # Variable c_ubyte
OLEUPDATE_ALWAYS = 1
MM_WIM_CLOSE = 959 # Variable c_int
CONNECT_UPDATE_RECENT = 2 # Variable c_int
OF_PARSE = 256 # Variable c_int
NRC_NAMTFUL = 14 # Variable c_int
ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED = 3005 # Variable c_long
ERROR_INSTALL_FAILURE = 1603 # Variable c_long
AC_DST_NO_PREMULT_ALPHA = 16 # Variable c_int
UDN_FIRST = 4294966575 # Variable c_uint
MIDICAPS_VOLUME = 1 # Variable c_int
ERROR_UNKNOWN_PRINT_MONITOR = 3000 # Variable c_long
CS_DBLCLKS = 8 # Variable c_int
RPC_C_AUTHN_DCE_PUBLIC = 2 # Variable c_int
ERROR_DDE_FAIL = 1156 # Variable c_long
LANG_SPANISH = 10 # Variable c_int
ENHMETA_STOCK_OBJECT = 2147483648 # Variable c_uint
META_BITBLT = 2338 # Variable c_int
DBG_EXCEPTION_NOT_HANDLED = 2147549185 # Variable c_ulong
JOB_CONTROL_RESTART = 4 # Variable c_int
DROPEFFECT_MOVE = 2 # Variable c_int
IMAGE_REL_PPC_REFLO = 17 # Variable c_int
BINDSTATUS_BEGINUPLOADDATA = 17
SCREEN_FONTTYPE = 8192 # Variable c_int
MCI_OPEN = 2051 # Variable c_int
MCIERR_MULTIPLE = 280 # Variable c_int
SS_ETCHEDFRAME = 18 # Variable c_long
PAN_LETT_OBLIQUE_BOXED = 11 # Variable c_int
ERROR_CLUSTERLOG_CORRUPT = 5029 # Variable c_long
ico1 = 1084 # Variable c_int
MDMSPKRFLAG_ON = 4 # Variable c_int
CHANGEKIND_DELETEMEMBER = 1
BINDSTATUS_DOWNLOADINGDATA = 5
ENUMPAPERMETRICS = 34 # Variable c_int
LVS_SMALLICON = 2 # Variable c_int
CRYPT_E_NO_VERIFY_USAGE_DLL = -2146885593 # Variable c_long
ERROR_REPARSE_ATTRIBUTE_CONFLICT = 4391 # Variable c_long
LVM_GETITEMPOSITION = 4112 # Variable c_int
MDMVOLFLAG_MEDIUM = 2 # Variable c_int
MMIO_DIRTY = 268435456 # Variable c_int
PERF_DATA_REVISION = 1 # Variable c_int
TBCD_TICS = 1 # Variable c_int
ERROR_INVALID_EVENTNAME = 1211 # Variable c_long
WSAEMSGSIZE = 10040 # Variable c_int
IMAGE_REL_IA64_DIR32NB = 16 # Variable c_int
CB_GETCURSEL = 327 # Variable c_int
MM_STREAM_DONE = 982 # Variable c_int
MNC_EXECUTE = 2 # Variable c_int
C3_KASHIDA = 512 # Variable c_int
WS_CHILDWINDOW = 1073741824 # Variable c_long
ENUM_S_FIRST = 262576 # Variable c_long
PSH_USEHBMHEADER = 1048576 # Variable c_int
POSTSCRIPT_IDENTIFY = 4117 # Variable c_int
IMAGE_REL_IA64_SECREL22 = 12 # Variable c_int
TVN_FIRST = 4294966896 # Variable c_uint
NRC_OPENERR = 63 # Variable c_int
TTN_POP = 4294966774 # Variable c_uint
ERROR_CLASS_HAS_WINDOWS = 1412 # Variable c_long
TCM_INSERTITEMA = 4871 # Variable c_int
MIDIERR_NOTREADY = 67 # Variable c_int
GWL_HWNDPARENT = -8 # Variable c_int
CTRY_IRAN = 981 # Variable c_int
CDM_LAST = 1224 # Variable c_int
LOCALE_SABBREVMONTHNAME6 = 73 # Variable c_int
IMAGE_SYM_CLASS_ENUM_TAG = 15 # Variable c_int
PRINTER_NOTIFY_FIELD_DATATYPE = 11 # Variable c_int
EM_EMPTYUNDOBUFFER = 205 # Variable c_int
TVI_ROOT = 4294901760 # Variable POINTER(_TREEITEM)
PERF_RAW_FRACTION = 537003008 # Variable c_int
PRINTER_ENUM_ICON2 = 131072 # Variable c_int
CBEM_HASEDITCHANGED = 1034 # Variable c_int
QUERY_USES_NETWORK = 8
SE_SACL_AUTO_INHERIT_REQ = 512 # Variable c_int
HEAP_NO_SERIALIZE = 1 # Variable c_int
URLPOLICY_CREDENTIALS_SILENT_LOGON_OK = 0 # Variable c_int
PRINTER_HTML_VIEW_JOBPROPERTIES = 3 # Variable c_int
STUB_UNMARSHAL = 0
MMIO_FINDRIFF = 32 # Variable c_int
LOGON32_PROVIDER_WINNT35 = 1 # Variable c_int
MOVEFILE_REPLACE_EXISTING = 1 # Variable c_int
SECURITY_WORLD_RID = 0 # Variable c_long
TAPE_SPACE_SETMARKS = 8 # Variable c_long
CM_IN_GAMUT = 0 # Variable c_int
TBM_GETPOS = 1024 # Variable c_int
MCI_SEQ_MAPPER = 65535 # Variable c_int
IMAGE_REL_PPC_TOCREL14 = 9 # Variable c_int
IMAGE_FILE_MACHINE_ALPHA = 388 # Variable c_int
DCBA_FACEUPLEFT = 2 # Variable c_int
USERCLASSTYPE_APPNAME = 3
ISC_SHOWUICOMPOSITIONWINDOW = 2147483648 # Variable c_uint
RPC_C_AUTHN_DPA = 16 # Variable c_int
MF_MOUSESELECT = 32768 # Variable c_long
TBS_TOP = 4 # Variable c_int
MK_E_CANTOPENFILE = -2147221014 # Variable c_long
DCOM_NONE = 0
PIDSI_DOC_SECURITY = 19 # Variable c_long
DMPAPER_A3_EXTRA = 63 # Variable c_int
PARITY_MARK = 2048 # Variable c_ushort
CF_FIXEDPITCHONLY = 16384 # Variable c_long
VARCMP_EQ = 1 # Variable c_int
stc17 = 1104 # Variable c_int
PDERR_RETDEFFAILURE = 4099 # Variable c_int
ERROR_DECRYPTION_FAILED = 6001 # Variable c_long
RT_ACCELERATOR = 9 # Variable POINTER(c_wchar)
TYSPEC_CLSID = 0
LVNI_ABOVE = 256 # Variable c_int
CCM_GETUNICODEFORMAT = 8198 # Variable c_int
LVNI_DROPHILITED = 8 # Variable c_int
ERROR_ARITHMETIC_OVERFLOW = 534 # Variable c_long
SUBLANG_GERMAN_SWISS = 2 # Variable c_int
CHANGEKIND_MAX = 7
CBEN_BEGINEDIT = 4294966492 # Variable c_uint
cmb11 = 1146 # Variable c_int
_MAX_DIR = 256 # Variable c_int
HC_ACTION = 0 # Variable c_int
MK_ALT = 32 # Variable c_int
RPC_S_PROTSEQ_NOT_SUPPORTED = 1703 # Variable c_long
PSN_WIZBACK = 4294967090 # Variable c_uint
DRAGDROP_S_CANCEL = 262401 # Variable c_long
LANG_UKRAINIAN = 34 # Variable c_int
LC_NONE = 0 # Variable c_int
CO_E_INIT_UNACCEPTED_USER_ALLOCATOR = -2147467251 # Variable c_long
STATUS_NONCONTINUABLE_EXCEPTION = 3221225509 # Variable c_ulong
RTS_CONTROL_ENABLE = 1 # Variable c_int
NMTTDISPINFOW_V1_SIZE = 184 # Variable c_uint
IPM_CLEARADDRESS = 1124 # Variable c_int
WGL_FONT_LINES = 0 # Variable c_int
FADF_STATIC = 2 # Variable c_int
ERROR_INVALID_CMM = 2300 # Variable c_long
JOY_BUTTON32 = 2147483648 # Variable c_ulong
TVHT_ONITEMICON = 2 # Variable c_int
DUPLICATE_SAME_ACCESS = 2 # Variable c_int
SM_CYMINIMIZED = 58 # Variable c_int
stc8 = 1095 # Variable c_int
CIP_OLDER_VERSION_EXISTS = 3
MONTHCAL_CLASSA = 'SysMonthCal32' # Variable POINTER(c_char)
META_DIBCREATEPATTERNBRUSH = 322 # Variable c_int
IMAGE_FILE_MACHINE_R3000 = 354 # Variable c_int
ICM_OFF = 1 # Variable c_int
ERROR_INVALID_FORM_NAME = 1902 # Variable c_long
cmb16 = 1151 # Variable c_int
AUXCAPS_VOLUME = 1 # Variable c_int
MCI_DEVTYPE_OTHER = 521 # Variable c_int
TlbNamePath = 2
DMPAPER_FANFOLD_LGL_GERMAN = 41 # Variable c_int
MONTHCAL_CLASSW = 'SysMonthCal32' # Variable POINTER(c_wchar)
MAXCHAR = 127 # Variable c_int
LVIF_TEXT = 1 # Variable c_int
HEBREW_CHARSET = 177 # Variable c_int
CBEIF_IMAGE = 2 # Variable c_int
ERROR_NO_TRUST_LSA_SECRET = 1786 # Variable c_long
IDLFLAG_FRETVAL = 8 # Variable c_int

IMPORT_OBJECT_TYPE = c_int # enum
IMPORT_OBJECT_CODE = 0
IMPORT_OBJECT_DATA = 1
IMPORT_OBJECT_CONST = 2
XTYPF_NOBLOCK = 2 # Variable c_int
CDS_TEST = 2 # Variable c_int
CALLTYPE_ASYNC_CALLPENDING = 5
NTM_MM_INSTANCE = 524288 # Variable c_int
CBN_EDITCHANGE = 5 # Variable c_int
PRINTER_STATUS_WAITING = 8192 # Variable c_int
DOMAIN_GROUP_RID_GUESTS = 514 # Variable c_long
RPC_E_ATTEMPTED_MULTITHREAD = -2147417854 # Variable c_long
ERROR_CONNECTED_OTHER_PASSWORD = 2108 # Variable c_long
ERROR_CONTROL_ID_NOT_FOUND = 1421 # Variable c_long
NCBLANSTALERT = 115 # Variable c_int
ERROR_IO_DEVICE = 1117 # Variable c_long

tagSTGC = c_int # enum
STGC_DEFAULT = 0
STGC_OVERWRITE = 1
STGC_ONLYIFCURRENT = 2
STGC_DANGEROUSLYCOMMITMERELYTODISKCACHE = 4
STGC_CONSOLIDATE = 8
ACCESS_DS_SOURCE_W = 'DS' # Variable POINTER(c_wchar)
RPC_C_EP_MATCH_BY_IF = 1 # Variable c_int
R2_NOTCOPYPEN = 4 # Variable c_int
PRINTER_ENUM_CONTAINER = 32768 # Variable c_int
DISP_E_BADCALLEE = -2147352560 # Variable c_long
JobObjectBasicLimitInformation = 2
MF_POPUP = 16 # Variable c_long
META_SETBKMODE = 258 # Variable c_int
TOOLTIPS_CLASSW = 'tooltips_class32' # Variable POINTER(c_wchar)
SPI_SETDOUBLECLKHEIGHT = 30 # Variable c_int
FUNC_DISPATCH = 4
DISP_CHANGE_NOTUPDATED = -3 # Variable c_int
DLGC_BUTTON = 8192 # Variable c_int
RBBS_HIDDEN = 8 # Variable c_int
STARTF_USEPOSITION = 4 # Variable c_int
SECURITY_CREATOR_OWNER_SERVER_RID = 2 # Variable c_long
TVM_GETISEARCHSTRINGA = 4375 # Variable c_int
DEFAULT_PALETTE = 15 # Variable c_int
HELP_FINDER = 11 # Variable c_int
PAGE_GUARD = 256 # Variable c_int
EMARCH_ENC_I17_SIGN_INST_WORD_X = 3 # Variable c_int
CTRY_KENYA = 254 # Variable c_int
VFT2_DRV_DISPLAY = 4 # Variable c_long
SERVICE_DISABLED = 4 # Variable c_int
BF_DIAGONAL_ENDTOPRIGHT = 22 # Variable c_int
CERT_E_REVOKED = -2146762484 # Variable c_long
IMAGE_SYM_TYPE_UNION = 9 # Variable c_int
LANG_ROMANIAN = 24 # Variable c_int
FILE_TYPE_REMOTE = 32768 # Variable c_int
META_SETROP2 = 260 # Variable c_int
ERROR_LUIDS_EXHAUSTED = 1334 # Variable c_long
RPCFLG_AUTO_COMPLETE = 134217728 # Variable c_ulong
MCI_INFO_MEDIA_UPC = 1024 # Variable c_long
SOFTDIST_ADSTATE_INSTALLED = 3 # Variable c_int
HDM_SETIMAGELIST = 4616 # Variable c_int
LB_SETITEMDATA = 410 # Variable c_int
MCI_FROM = 4 # Variable c_long
CTRY_PERU = 51 # Variable c_int
BM_SETIMAGE = 247 # Variable c_int
SERVICE_CONFIG_FAILURE_ACTIONS = 2 # Variable c_int
SPLDS_PRINT_MEDIA_SUPPORTED = 'printMediaSupported' # Variable POINTER(c_wchar)
MCM_SETCOLOR = 4106 # Variable c_int
LOGPIXELSY = 90 # Variable c_int
TYPEFLAG_FAGGREGATABLE = 1024
IMAGE_REL_SH3_DIRECT32_NB = 16 # Variable c_int
MWMO_ALERTABLE = 2 # Variable c_int
MCI_DEVTYPE_VCR = 513 # Variable c_int
S_SERDLN = -6 # Variable c_int
BSM_INSTALLABLEDRIVERS = 4 # Variable c_int
LVM_SETBKIMAGEW = 4234 # Variable c_int
FILE_FLAG_DELETE_ON_CLOSE = 67108864 # Variable c_int
META_SCALEVIEWPORTEXT = 1042 # Variable c_int
EMR_FILLRGN = 71 # Variable c_int
ACCESS_OBJECT_GUID = 0 # Variable c_int
LVS_ICON = 0 # Variable c_int
HDS_HOTTRACK = 4 # Variable c_int
MMSYSERR_NOERROR = 0 # Variable c_int

_RPC_ASYNC_EVENT = c_int # enum
RpcCallComplete = 0
RpcSendComplete = 1
RpcReceiveComplete = 2
ERROR_NO_MORE_FILES = 18 # Variable c_long
IMAGE_REL_SH3_DIRECT8 = 3 # Variable c_int
TBSTYLE_GROUP = 4 # Variable c_int
TTDT_AUTOPOP = 2 # Variable c_int
GL_ID_NOMODULE = 1 # Variable c_int
SC_MANAGER_ALL_ACCESS = 983103 # Variable c_long
SPAPI_E_KEY_DOES_NOT_EXIST = -2146500092 # Variable c_long
DT_LEFT = 0 # Variable c_int
PCF_INTTIMEOUTS = 128 # Variable c_ulong
CTRY_UNITED_STATES = 1 # Variable c_int
SND_ALIAS_SYSTEMDEFAULT = 17491 # Variable c_ulong
CF_ENABLETEMPLATEHANDLE = 32 # Variable c_long
RPC_C_PROFILE_MATCH_BY_BOTH = 4 # Variable c_int
URLMON_OPTION_USERAGENT_REFRESH = 268435458 # Variable c_int
ACS_CENTER = 1 # Variable c_int
SEC_NOCACHE = 268435456 # Variable c_int
DC_ACTIVE = 1 # Variable c_int
HTBORDER = 18 # Variable c_int
VARFLAG_FHIDDEN = 64
PARITY_NONE = 256 # Variable c_ushort
DMPAPER_ENV_B4 = 33 # Variable c_int
LCMAP_KATAKANA = 2097152 # Variable c_int
BACKUP_INVALID = 0 # Variable c_int
BLACKNESS = 66 # Variable c_ulong
MCI_SYSINFO_NAME = 1024 # Variable c_long
ERROR_WMI_TRY_AGAIN = 4203 # Variable c_long
JOB_OBJECT_UILIMIT_WRITECLIPBOARD = 4 # Variable c_int
MCI_DELETE = 2134 # Variable c_int
DT_RIGHT = 2 # Variable c_int
PROPSHEETHEADERW_V1_SIZE = 40 # Variable c_uint
MCI_REALIZE = 2112 # Variable c_int
RDW_NOINTERNALPAINT = 16 # Variable c_int
PSH_NOAPPLYNOW = 128 # Variable c_int
LR_VGACOLOR = 128 # Variable c_int
RESOURCE_CONTEXT = 5 # Variable c_int
NMPWAIT_WAIT_FOREVER = 4294967295 # Variable c_uint
ERROR_WRITE_FAULT = 29 # Variable c_long
TVS_NONEVENHEIGHT = 16384 # Variable c_int
DTM_GETMCFONT = 4106 # Variable c_int
KEY_QUERY_VALUE = 1 # Variable c_int

tagREGKIND = c_int # enum
REGKIND_DEFAULT = 0
REGKIND_REGISTER = 1
REGKIND_NONE = 2
VOS_UNKNOWN = 0 # Variable c_long
BS_INDEXED = 4 # Variable c_int
SE_TCB_NAME = 'SeTcbPrivilege' # Variable POINTER(c_wchar)
SC_DEFAULT = 61792 # Variable c_int
ST_CLIENT = 16 # Variable c_int
TKF_HOTKEYSOUND = 16 # Variable c_int
CAL_SABBREVMONTHNAME11 = 44 # Variable c_int
HKEY_LOCAL_MACHINE = 2147483650 # Variable POINTER(HKEY__)
ERROR_DS_NO_MORE_RIDS = 1930 # Variable c_long
TRANSPARENT = 1 # Variable c_int
CTRY_COSTA_RICA = 506 # Variable c_int
PSU_DEFAULT = 1
CRYPT_E_HASH_VALUE = -2146889721 # Variable c_long
DRV_MCI_LAST = 6143 # Variable c_int
SPLREG_OS_VERSION = 'OSVersion' # Variable POINTER(c_wchar)
WSASYS_STATUS_LEN = 128 # Variable c_int
DBG_CONTROL_C = 1073807365 # Variable c_ulong
PRINTER_HTML_VIEW_PRINTPROPERTIES = 1 # Variable c_int
GROUP_SECURITY_INFORMATION = 2 # Variable c_long
ERROR_SET_POWER_STATE_VETOED = 1140 # Variable c_long
MCM_GETCURSEL = 4097 # Variable c_int
OLEUPDATE_ONCALL = 3
PORT_STATUS_OFFLINE = 1 # Variable c_int
PBS_SMOOTH = 1 # Variable c_int
EM_CHARFROMPOS = 215 # Variable c_int
TVHT_NOWHERE = 1 # Variable c_int
RPC_C_IMP_LEVEL_IDENTIFY = 2 # Variable c_int
SOCK_DGRAM = 2 # Variable c_int
NM_CLICK = 4294967294 # Variable c_uint
LANG_NORWEGIAN = 20 # Variable c_int
RPC_S_ENTRY_NOT_FOUND = 1761 # Variable c_long
VIEW_SORTTYPE = 7 # Variable c_int
EMR_PAINTRGN = 74 # Variable c_int
GCS_RESULTSTR = 2048 # Variable c_int
GMEM_DISCARDABLE = 256 # Variable c_int
PRINTER_STATUS_DOOR_OPEN = 4194304 # Variable c_int
ERROR_DUPLICATE_TAG = 2304 # Variable c_long
ERROR_PRINTER_ALREADY_EXISTS = 1802 # Variable c_long
ERROR_RING2SEG_MUST_BE_MOVABLE = 200 # Variable c_long
R2_WHITE = 16 # Variable c_int
ERROR_INVALID_EVENT_COUNT = 151 # Variable c_long
SERVICES_FAILED_DATABASEA = 'ServicesFailed' # Variable POINTER(c_char)
ERROR_UNABLE_TO_CLEAN = 4311 # Variable c_long
SERVICE_CONTROL_SHUTDOWN = 5 # Variable c_int
MCI_ANIM_GETDEVCAPS_NORMAL_RATE = 16388 # Variable c_long
MK_E_INTERMEDIATEINTERFACENOTSUPPORTED = -2147221017 # Variable c_long
AD_COUNTERCLOCKWISE = 1 # Variable c_int
LANG_THAI = 30 # Variable c_int
IMAGE_SCN_ALIGN_1BYTES = 1048576 # Variable c_int
ERROR_FLOPPY_ID_MARK_NOT_FOUND = 1122 # Variable c_long
DCX_LOCKWINDOWUPDATE = 1024 # Variable c_long
CE_OVERRUN = 2 # Variable c_int
RPC_C_VERS_MAJOR_ONLY = 4 # Variable c_int
PAN_FAMILY_PICTORIAL = 5 # Variable c_int
THREAD_PRIORITY_ERROR_RETURN = 2147483647 # Variable c_int
RPC_X_PIPE_DISCIPLINE_ERROR = 1917 # Variable c_long
IMAGE_SYM_TYPE_SHORT = 3 # Variable c_int
SM_CYVTHUMB = 9 # Variable c_int
CLIP_EMBEDDED = 128 # Variable c_int
GCPCLASS_POSTBOUNDRTL = 16 # Variable c_int
JOB_NOTIFY_FIELD_PRIORITY = 14 # Variable c_int
XMLELEMTYPE_COMMENT = 2
DC_PAPERSIZE = 3 # Variable c_int
IMAGE_REL_IA64_IMM64 = 3 # Variable c_int
GENERIC_READ = 2147483648 # Variable c_ulong
MMIOERR_CANNOTCLOSE = 260 # Variable c_int
HEAP_TAG_SHIFT = 18 # Variable c_int
ERROR_NOT_LOGGED_ON = 1245 # Variable c_long
IDLFLAG_FIN = 1 # Variable c_int
CB_SETCURSEL = 334 # Variable c_int
IMAGE_SYM_CLASS_EXTERNAL_DEF = 5 # Variable c_int
MK_S_HIM = 262629 # Variable c_long
TVIF_HANDLE = 16 # Variable c_int
SPAPI_E_INVALID_INF_LOGCONFIG = -2146500054 # Variable c_long
CAL_ICALINTVALUE = 1 # Variable c_int
IPPORT_LOGINSERVER = 513 # Variable c_int
TYPE_E_BUFFERTOOSMALL = -2147319786 # Variable c_long
IMAGE_ARCHIVE_END = '`\n' # Variable POINTER(c_char)
EMR_SETTEXTCOLOR = 24 # Variable c_int
SUBLANG_NORWEGIAN_BOKMAL = 1 # Variable c_int
EN_KILLFOCUS = 512 # Variable c_int
MMIOM_USER = 32768 # Variable c_int
XTYP_WILDCONNECT = 8418 # Variable c_int
ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT = 1808 # Variable c_long
SPI_GETANIMATION = 72 # Variable c_int
ERROR_MAX_THRDS_REACHED = 164 # Variable c_long
DDE_FDATRESERVED = -45057 # Variable c_int
lst8 = 1127 # Variable c_int
frm3 = 1078 # Variable c_int
CDERR_MEMLOCKFAILURE = 10 # Variable c_int
SCS_CAP_COMPSTR = 1 # Variable c_int
SP_PARITY = 1 # Variable c_ulong
TBN_DELETINGBUTTON = 4294966581 # Variable c_uint
MCI_RECORD_OVERWRITE = 512 # Variable c_long
LOCALE_SABBREVMONTHNAME7 = 74 # Variable c_int
LVKF_ALT = 1 # Variable c_int
FILE_ACTION_REMOVED = 2 # Variable c_int
UPDFCACHE_IFBLANK = 16 # Variable c_int
SM_CYMINTRACK = 35 # Variable c_int
SET_TAPE_MEDIA_INFORMATION = 0 # Variable c_int
UDM_GETPOS = 1128 # Variable c_int
ERROR_SHARING_BUFFER_EXCEEDED = 36 # Variable c_long
MF_POSTMSGS = 67108864 # Variable c_int
DV_E_DVTARGETDEVICE_SIZE = -2147221396 # Variable c_long
BACKUP_ALTERNATE_DATA = 4 # Variable c_int
PFD_SUPPORT_OPENGL = 32 # Variable c_int
INADDR_NONE = 4294967295 # Variable c_uint
TYSPEC_FILENAME = 6
SS_LEFT = 0 # Variable c_long
PSINJECT_COMMENTS = 2 # Variable c_int
JOB_EXECUTE = 131088 # Variable c_long
RBS_FIXEDORDER = 2048 # Variable c_int
DS_NOIDLEMSG = 256 # Variable c_long
STUB_CALL_SERVER = 1
rad4 = 1059 # Variable c_int
STM_SETICON = 368 # Variable c_int
SPI_SETSCREENSAVERRUNNING = 97 # Variable c_int
COINIT_APARTMENTTHREADED = 2
PF_COMPARE_EXCHANGE_DOUBLE = 2 # Variable c_int
MCI_FORMAT_SMPTE_24 = 4 # Variable c_int
WS_EX_RIGHT = 4096 # Variable c_long
NEWFILEOPENORD = 1547 # Variable c_int
MININT_PTR = 2147483648 # Variable c_ulong
STG_S_CANNOTCONSOLIDATE = 197126 # Variable c_long
LVCF_TEXT = 4 # Variable c_int
NI_OPENCANDIDATE = 16 # Variable c_int
PIPE_TYPE_MESSAGE = 4 # Variable c_int
FR_ENABLEHOOK = 256 # Variable c_int
SECURITY_NETWORK_RID = 2 # Variable c_long
ERROR_SEM_TIMEOUT = 121 # Variable c_long
CF_ANSIONLY = 1024 # Variable c_long
NLS_HIRAGANA = 262144 # Variable c_int
LCS_GM_GRAPHICS = 2 # Variable c_long
CBES_EX_NOEDITIMAGEINDENT = 2 # Variable c_int
DST_TEXT = 1 # Variable c_int

tagOLEWHICHMK = c_int # enum
OLEWHICHMK_CONTAINER = 1
OLEWHICHMK_OBJREL = 2
OLEWHICHMK_OBJFULL = 3
RPC_C_PROFILE_DEFAULT_ELT = 0 # Variable c_int
SPI_GETSCREENSAVEACTIVE = 16 # Variable c_int
MCI_WHERE = 2115 # Variable c_int
DMDUP_HORIZONTAL = 3 # Variable c_int
TRUST_E_FINANCIAL_CRITERIA = -2146869218 # Variable c_long
DMPAPER_A4 = 9 # Variable c_int
IA64_JMPE_MASK = 16777215 # Variable c_int
DCOM_CALL_COMPLETE = 1
FILE_GENERIC_WRITE = 1179926 # Variable c_long
NUMPRS_LEADING_MINUS = 16 # Variable c_int
FILE_UNICODE_ON_DISK = 4 # Variable c_int
TCM_DELETEITEM = 4872 # Variable c_int
DMBIN_AUTO = 7 # Variable c_int
RPC_C_MQ_TEMPORARY = 0 # Variable c_int
MCI_STATUS_ITEM = 256 # Variable c_long
TYSPEC_FILEEXT = 1
MCI_DEVTYPE_DIGITAL_VIDEO = 520 # Variable c_int
RPC_E_NO_GOOD_SECURITY_PACKAGES = -2147417830 # Variable c_long
PAN_XHEIGHT_CONSTANT_SMALL = 2 # Variable c_int
TRUETYPE_FONTTYPE = 4 # Variable c_int
IP_TTL = 7 # Variable c_int
IPM_SETFOCUS = 1128 # Variable c_int
WS_HSCROLL = 1048576 # Variable c_long
IMAGE_SIZEOF_SYMBOL = 18 # Variable c_int
TCM_GETITEMCOUNT = 4868 # Variable c_int
ERROR_SERVICE_LOGON_FAILED = 1069 # Variable c_long
MDITILE_VERTICAL = 0 # Variable c_int
HANGUP_PENDING = 4 # Variable c_int
PROCESS_HEAP_ENTRY_BUSY = 4 # Variable c_int
TKF_HOTKEYACTIVE = 4 # Variable c_int
SEE_MASK_FLAG_NO_UI = 1024 # Variable c_int
ERROR_NO_LOGON_SERVERS = 1311 # Variable c_long
RPC_C_NOTIFY_ON_SEND_COMPLETE = 1 # Variable c_int
MCI_WAVE_SET_FORMATTAG = 65536 # Variable c_long
PSP_USECALLBACK = 128 # Variable c_int
MIXERCONTROL_CONTROLTYPE_DECIBELS = 805568512 # Variable c_long
DPD_DELETE_UNUSED_FILES = 1 # Variable c_int

XLAT_SIDE = c_int # enum
XLAT_SERVER = 1
XLAT_CLIENT = 2
GMEM_ZEROINIT = 64 # Variable c_int
MMSYSERR_READERROR = 16 # Variable c_int
JOY_BUTTON31 = 1073741824 # Variable c_long
SEE_MASK_HOTKEY = 32 # Variable c_int
ERROR_INVALID_MENU_HANDLE = 1401 # Variable c_long
CPS_CANCEL = 4 # Variable c_int

tagDVASPECT = c_int # enum
DVASPECT_CONTENT = 1
DVASPECT_THUMBNAIL = 2
DVASPECT_ICON = 4
DVASPECT_DOCPRINT = 8
JOYCAPS_POV4DIR = 32 # Variable c_int
HSHELL_LANGUAGE = 8 # Variable c_int
PS_JOIN_ROUND = 0 # Variable c_int
SE_CREATE_PAGEFILE_NAME = 'SeCreatePagefilePrivilege' # Variable POINTER(c_wchar)
C2_RIGHTTOLEFT = 2 # Variable c_int
IPPROTO_IDP = 22 # Variable c_int
DDL_DIRECTORY = 16 # Variable c_int
EMARCH_ENC_I17_IMM9D_SIZE_X = 9 # Variable c_int
WSATRY_AGAIN = 11002 # Variable c_int
ERROR_MEDIA_OFFLINE = 4304 # Variable c_long
MMIO_CREATELIST = 64 # Variable c_int
NORM_IGNORENONSPACE = 2 # Variable c_int
LB_GETSELCOUNT = 400 # Variable c_int
PS_STYLE_MASK = 15 # Variable c_int
ERROR_BAD_REM_ADAP = 60 # Variable c_long
NCBUNLINK = 112 # Variable c_int
IMAGE_SYM_CLASS_EXTERNAL = 2 # Variable c_int
PERF_INVERSE_COUNTER = 16777216 # Variable c_int
SE_TAKE_OWNERSHIP_NAME = 'SeTakeOwnershipPrivilege' # Variable POINTER(c_wchar)
LVITEMW_V1_SIZE = 36 # Variable c_uint
URLTEMPLATE_MEDIUM = 69632
WNNC_NET_LOCUS = 393216 # Variable c_int
PROPSHEETPAGEW_V1_SIZE = 40 # Variable c_uint
PAN_SERIF_FLARED = 14 # Variable c_int

_EXCEPTION_DISPOSITION = c_int # enum
ExceptionContinueExecution = 0
ExceptionContinueSearch = 1
ExceptionNestedException = 2
ExceptionCollidedUnwind = 3
CTRY_MOROCCO = 212 # Variable c_int
ERROR_NO_SUCH_PRIVILEGE = 1313 # Variable c_long
TOOLTIPS_CLASSA = 'tooltips_class32' # Variable POINTER(c_char)
ERROR_JOIN_TO_JOIN = 138 # Variable c_long
IMAGE_REL_PPC_REL24 = 6 # Variable c_int
TC_HARDERR = 1 # Variable c_int
C2_COMMONSEPARATOR = 7 # Variable c_int
GROUP_NAME = 128 # Variable c_int
HDM_SETHOTDIVIDER = 4627 # Variable c_int
SETXOFF = 1 # Variable c_int
GMEM_VALID_FLAGS = 32626 # Variable c_int
TB_ADDSTRINGA = 1052 # Variable c_int
MWT_RIGHTMULTIPLY = 3 # Variable c_int
CF_SELECTSCRIPT = 4194304 # Variable c_long
CWP_ALL = 0 # Variable c_int
PGM_GETBUTTONSIZE = 5131 # Variable c_int
WH_JOURNALRECORD = 0 # Variable c_int
DT_NOCLIP = 256 # Variable c_int

tagCALLCONV = c_int # enum
CC_FASTCALL = 0
CC_CDECL = 1
CC_MSCPASCAL = 2
CC_PASCAL = 2
CC_MACPASCAL = 3
CC_STDCALL = 4
CC_FPFASTCALL = 5
CC_SYSCALL = 6
CC_MPWCDECL = 7
CC_MPWPASCAL = 8
CC_MAX = 9
R2_MERGEPENNOT = 14 # Variable c_int
SC_CONTEXTHELP = 61824 # Variable c_int
PSP_USEREFPARENT = 64 # Variable c_int
TBS_AUTOTICKS = 1 # Variable c_int
SS_PATHELLIPSIS = 32768 # Variable c_long
MSG_PEEK = 2 # Variable c_int
EMARCH_ENC_I17_SIGN_INST_WORD_POS_X = 27 # Variable c_int
CAL_SABBREVMONTHNAME13 = 46 # Variable c_int
LANG_RUSSIAN = 25 # Variable c_int
SETRGBSTRINGA = 'commdlg_SetRGBColor' # Variable POINTER(c_char)
MHDR_ISSTRM = 8 # Variable c_int
OFN_NODEREFERENCELINKS = 1048576 # Variable c_int
TB_GETBUTTONINFOA = 1089 # Variable c_int
MCIERR_SEQ_DIV_INCOMPATIBLE = 336 # Variable c_int
HDM_HITTEST = 4614 # Variable c_int
SEE_MASK_ICON = 16 # Variable c_int
MCI_ANIM_WINDOW_DISABLE_STRETCH = 2097152 # Variable c_long
DI_CHANNEL = 1 # Variable c_int
SIF_RANGE = 1 # Variable c_int
LANG_MALAY = 62 # Variable c_int
CALLBACK_FUNCTION = 196608 # Variable c_long
TOKEN_IMPERSONATE = 4 # Variable c_int
TVM_SETBKCOLOR = 4381 # Variable c_int
LR_SHARED = 32768 # Variable c_int
TCS_RIGHTJUSTIFY = 0 # Variable c_int
PSINJECT_ORIENTATION = 23 # Variable c_int
CAL_SABBREVDAYNAME1 = 14 # Variable c_int
IP_MULTICAST_TTL = 3 # Variable c_int
ATF_ONOFFFEEDBACK = 2 # Variable c_int
CF_ENHMETAFILE = 14 # Variable c_int
COLOR_HIGHLIGHT = 13 # Variable c_int
UDM_GETBUDDY = 1130 # Variable c_int
SEF_AVOID_PRIVILEGE_CHECK = 8 # Variable c_int
CBR_9600 = 9600 # Variable c_int
ODT_MENU = 1 # Variable c_int
TYPEFLAG_FREPLACEABLE = 2048
CTRY_SAUDI_ARABIA = 966 # Variable c_int
TBM_SETLINESIZE = 1047 # Variable c_int
PSP_USEHEADERSUBTITLE = 8192 # Variable c_int
SO_ERROR = 4103 # Variable c_int
LOCALE_SMONGROUPING = 24 # Variable c_int
PFD_DEPTH_DONTCARE = 536870912 # Variable c_int
SERVICE_CONTROL_CONTINUE = 3 # Variable c_int
NUMMARKERS = 20 # Variable c_int
JOYCAPS_HASR = 2 # Variable c_int
MMSYSERR_INVALPARAM = 11 # Variable c_int
SWP_NOOWNERZORDER = 512 # Variable c_int
TC_CP_STROKE = 4 # Variable c_int
ANYSIZE_ARRAY = 1 # Variable c_int
IMAGE_REL_ALPHA_LITUSE = 5 # Variable c_int
PSH_WIZARDCONTEXTHELP = 4096 # Variable c_int
SETLINECAP = 21 # Variable c_int
ERROR_DOMAIN_EXISTS = 1356 # Variable c_long
CF_NOVERTFONTS = 16777216 # Variable c_long
EC_LEFTMARGIN = 1 # Variable c_int
TC_SO_ABLE = 4096 # Variable c_int
edt8 = 1159 # Variable c_int
SO_SYNCHRONOUS_ALERT = 16 # Variable c_int
CP_ACP = 0 # Variable c_int
PAN_CONTRAST_NONE = 2 # Variable c_int
ERROR_UNKNOWN_PRODUCT = 1605 # Variable c_long
XTYP_MASK = 240 # Variable c_int
SYSTEM_ALARM_ACE_TYPE = 3 # Variable c_int
ERROR_PAGEFILE_QUOTA = 1454 # Variable c_long
CBEN_DRAGBEGINA = 4294966488 # Variable c_uint
TCM_SETPADDING = 4907 # Variable c_int
URLACTION_HTML_JAVA_RUN = 5637 # Variable c_int
MMIO_EXCLUSIVE = 16 # Variable c_int
DS_FIXEDSYS = 8 # Variable c_long
ERROR_USER_EXISTS = 1316 # Variable c_long
VARFLAG_FRESTRICTED = 128
INET_E_CANNOT_INSTANTIATE_OBJECT = -2146697200 # Variable c_long
STD_PRINTPRE = 9 # Variable c_int
ERROR_NETWORK_ACCESS_DENIED = 65 # Variable c_long
MCI_DEVTYPE_VIDEODISC = 514 # Variable c_int
CTRY_CZECH = 420 # Variable c_int
WGL_SWAP_UNDERLAY7 = 4194304 # Variable c_int
PS_GEOMETRIC = 65536 # Variable c_int
HTBOTTOM = 15 # Variable c_int
LANG_ARABIC = 1 # Variable c_int
CRYPT_E_STREAM_MSG_NOT_READY = -2146889712 # Variable c_long
GGO_METRICS = 0 # Variable c_int
FORMAT_MESSAGE_MAX_WIDTH_MASK = 255 # Variable c_int
FORMAT_MESSAGE_FROM_STRING = 1024 # Variable c_int
CF_INITTOLOGFONTSTRUCT = 64 # Variable c_long
WMSZ_LEFT = 1 # Variable c_int
SERKF_AVAILABLE = 2 # Variable c_int
PAN_BENT_ARMS_DOUBLE_SERIF = 11 # Variable c_int
TBN_DRAGOUT = 4294966582 # Variable c_uint
PSD_DISABLEPAPER = 512 # Variable c_int
ERROR_SERVICE_CANNOT_ACCEPT_CTRL = 1061 # Variable c_long
PAN_WEIGHT_BOOK = 5 # Variable c_int
IMAGE_REL_IA64_PCREL21B = 6 # Variable c_int
SUBLANG_ARABIC_JORDAN = 11 # Variable c_int
IO_REPARSE_TAG_DFS = 2147483656 # Variable c_uint
LZERROR_READ = -3 # Variable c_int
PSM_SETFINISHTEXTA = 1139 # Variable c_int
PROP_SM_CXDLG = 212 # Variable c_int
MCI_SET_ON = 8192 # Variable c_long
SM_CXSCREEN = 0 # Variable c_int
FS_LATIN1 = 1 # Variable c_long
MK_S_ME = 262628 # Variable c_long
HTSYSMENU = 3 # Variable c_int
TAPE_SPACE_FILEMARKS = 6 # Variable c_long
RESOURCE_REMEMBERED = 3 # Variable c_int
SUBLANG_ARABIC_SAUDI_ARABIA = 1 # Variable c_int
PSU_SECURITY_URL_ONLY = 2
ACE_INHERITED_OBJECT_TYPE_PRESENT = 2 # Variable c_int
XTYP_CONNECT_CONFIRM = 32882 # Variable c_int
LVIS_FOCUSED = 1 # Variable c_int
PATINVERT = 5898313 # Variable c_ulong
CREATE_NEW = 1 # Variable c_int
MB_RETRYCANCEL = 5 # Variable c_long
PRINTER_ATTRIBUTE_ENABLE_DEVQ = 128 # Variable c_int
URLACTION_ACTIVEX_CURR_MAX = 4613 # Variable c_int
chx12 = 1051 # Variable c_int
HDN_GETDISPINFOA = 4294966987 # Variable c_uint
RGN_COPY = 5 # Variable c_int
REG_REFRESH_HIVE = 2 # Variable c_long
CFS_CANDIDATEPOS = 64 # Variable c_int
RPC_C_AUTHN_GSS_KERBEROS = 18 # Variable c_int
EMR_ABORTPATH = 68 # Variable c_int
VOS_DOS = 65536 # Variable c_long
SORT_CHINESE_BIG5 = 0 # Variable c_int
RBBS_NOGRIPPER = 256 # Variable c_int
OBJ_METADC = 4 # Variable c_int
MIXERCONTROL_CONTROLTYPE_MULTIPLESELECT = 1895890944 # Variable c_long
STATUS_FLOAT_MULTIPLE_TRAPS = 3221226165 # Variable c_ulong
ERROR_DIFFERENT_SERVICE_ACCOUNT = 1079 # Variable c_long
SPI_GETSTICKYKEYS = 58 # Variable c_int
OLE_E_STATIC = -2147221493 # Variable c_long
RPC_C_EP_ALL_ELTS = 0 # Variable c_int
RBS_VERTICALGRIPPER = 16384 # Variable c_int
CF_RIFF = 11 # Variable c_int
DOMAIN_GROUP_RID_USERS = 513 # Variable c_long
RPC_S_INVALID_NET_ADDR = 1707 # Variable c_long
CPS_CONVERT = 2 # Variable c_int
IMAGE_ORDINAL_FLAG32 = 2147483648 # Variable c_uint
REG_WHOLE_HIVE_VOLATILE = 1 # Variable c_long
TB_SAVERESTOREA = 1050 # Variable c_int
BROADCAST_QUERY_DENY = 1112363332 # Variable c_int
IgnoreError = 0
LOGON32_LOGON_INTERACTIVE = 2 # Variable c_int
IN_CLASSA_NSHIFT = 24 # Variable c_int
KF_REPEAT = 16384 # Variable c_int
VER_PLATFORM_WIN32s = 0 # Variable c_int
scr6 = 1173 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT = 4104 # Variable c_long
LOCALE_IMEASURE = 13 # Variable c_int

tagOLELINKBIND = c_int # enum
OLELINKBIND_EVENIFCLASSDIFF = 1
IME_HOTKEY_DSWITCH_FIRST = 256 # Variable c_int
OFS_MAXPATHNAME = 128 # Variable c_int
NormalError = 1
VIF_DIFFLANG = 8 # Variable c_long
SPLDS_PRINT_RATE_UNIT = 'printRateUnit' # Variable POINTER(c_wchar)
SERVICE_AUTO_START = 2 # Variable c_int
LVM_GETCOLUMNORDERARRAY = 4155 # Variable c_int
TRUST_E_BAD_DIGEST = -2146869232 # Variable c_long
OLEOBJ_E_FIRST = 2147746176 # Variable c_ulong
CTRY_MACAU = 853 # Variable c_int
NULL_BRUSH = 5 # Variable c_int
BINDSTATUS_BEGINDOWNLOADCOMPONENTS = 7
NCBCHAINSENDNA = 114 # Variable c_int
PBM_SETBARCOLOR = 1033 # Variable c_int
RPC_S_PROTSEQ_NOT_FOUND = 1744 # Variable c_long
IS_TEXT_UNICODE_ODD_LENGTH = 512 # Variable c_int
CBEM_SETITEMA = 1029 # Variable c_int
XST_ADVSENT = 11 # Variable c_int
CFSTR_MIME_XBM = 'image/xbm' # Variable POINTER(c_wchar)
RPC_C_VERS_UPTO = 5 # Variable c_int
ico2 = 1085 # Variable c_int
LPD_TYPE_COLORINDEX = 1 # Variable c_int
C3_LEXICAL = 1024 # Variable c_int
IME_CMODE_HANJACONVERT = 64 # Variable c_int
PRINTER_ACCESS_USE = 8 # Variable c_int
SORT_KOREAN_UNICODE = 1 # Variable c_int
MCIERR_DUPLICATE_FLAGS = 295 # Variable c_int
END_PATH = 4098 # Variable c_int
GMEM_FIXED = 0 # Variable c_int
PIPE_ACCESS_DUPLEX = 3 # Variable c_int
MCIERR_FLAGS_NOT_COMPATIBLE = 284 # Variable c_int
LOCALE_IFIRSTWEEKOFYEAR = 4109 # Variable c_int
VT_UI2 = 18
ACCESS_MAX_LEVEL = 4 # Variable c_int
CTRY_LIECHTENSTEIN = 41 # Variable c_int
IMAGE_SYM_CLASS_WEAK_EXTERNAL = 105 # Variable c_int
REALTIME_PRIORITY_CLASS = 256 # Variable c_int
ERROR_INVALID_ORDINAL = 182 # Variable c_long
URLACTION_DOWNLOAD_UNSIGNED_ACTIVEX = 4100 # Variable c_int
ERROR_PRINT_MONITOR_IN_USE = 3008 # Variable c_long
LPD_TRANSPARENT = 4096 # Variable c_int
URLACTION_SCRIPT_MAX = 5631 # Variable c_int
VARIANT_NOVALUEPROP = 1 # Variable c_int
MAXPNAMELEN = 32 # Variable c_int
SND_ALIAS_SYSTEMHAND = 18515 # Variable c_ulong
DLGC_UNDEFPUSHBUTTON = 32 # Variable c_int
TCS_VERTICAL = 128 # Variable c_int
ERROR_TOO_MANY_MODULES = 214 # Variable c_long
FOF_ALLOWUNDO = 64 # Variable c_int
LOCALE_SABBREVMONTHNAME4 = 71 # Variable c_int
RBBIM_TEXT = 4 # Variable c_int
PSN_KILLACTIVE = 4294967095 # Variable c_uint
MF_HSZ_INFO = 16777216 # Variable c_int
SE_ERR_SHARE = 26 # Variable c_int
PROXY_CALCSIZE = 0
IME_CAND_RADICAL = 4 # Variable c_int
TYPE_E_NAMECONFLICT = -2147319763 # Variable c_long
GL_ID_CANNOTSAVE = 17 # Variable c_int
QUERY_IS_INSTALLEDENTRY = 10
IMAGE_REL_IA64_LTOFF64 = 15 # Variable c_int
CDM_SETCONTROLTEXT = 1128 # Variable c_int
TYSPEC_JAVACLASS = 7
HELPINFO_WINDOW = 1 # Variable c_int
SM_CXMENUSIZE = 54 # Variable c_int
ERROR_ALREADY_ASSIGNED = 85 # Variable c_long
LBN_DBLCLK = 2 # Variable c_int
MIXERCONTROL_CT_CLASS_MASK = 4026531840 # Variable c_ulong
rad5 = 1060 # Variable c_int
SPI_SETDRAGWIDTH = 76 # Variable c_int
SHGFI_SYSICONINDEX = 16384 # Variable c_int
BAUD_7200 = 1024 # Variable c_ulong
MCI_FORMAT_SMPTE_25 = 5 # Variable c_int
IMAGE_SEPARATE_DEBUG_FLAGS_MASK = 32768 # Variable c_int
ERROR_NOT_JOINED = 136 # Variable c_long
DM_GETDEFID = 1024 # Variable c_int
LIBFLAG_FHIDDEN = 4
TBN_GETBUTTONINFOA = 4294966596 # Variable c_uint
PRINTER_NOTIFY_FIELD_DRIVER_NAME = 4 # Variable c_int
IMAGE_SCN_ALIGN_2048BYTES = 12582912 # Variable c_int
ERROR_DISCARDED = 157 # Variable c_long
DMPAPER_LETTER_PLUS = 59 # Variable c_int
IMAGE_REL_PPC_TOCREL16 = 8 # Variable c_int
BS_NULL = 1 # Variable c_int
TB_GETTOOLTIPS = 1059 # Variable c_int

tagMKREDUCE = c_int # enum
MKRREDUCE_ONE = 196608
MKRREDUCE_TOUSER = 131072
MKRREDUCE_THROUGHUSER = 65536
MKRREDUCE_ALL = 0
MCI_SEQ_DIV_SMPTE_30DROP = 1219 # Variable c_int
TVN_GETDISPINFOA = 4294966893 # Variable c_uint
THREAD_BASE_PRIORITY_MIN = -2 # Variable c_int
TTS_ALWAYSTIP = 1 # Variable c_int
TVIF_IMAGE = 2 # Variable c_int
WNFMT_CONNECTION = 32 # Variable c_int
IMAGE_REL_PPC_REL14 = 7 # Variable c_int
GCP_USEKERNING = 8 # Variable c_int
IMAGE_REL_SH3_DIRECT8_WORD = 4 # Variable c_int
IMN_SETSTATUSWINDOWPOS = 12 # Variable c_int
JOB_OBJECT_LIMIT_PRIORITY_CLASS = 32 # Variable c_int
SCS_CAP_MAKEREAD = 2 # Variable c_int
CRYPT_E_REVOCATION_OFFLINE = -2146885613 # Variable c_long
WS_EX_LTRREADING = 0 # Variable c_long
CO_E_NETACCESSAPIFAILED = -2147220984 # Variable c_long
DMBIN_LARGECAPACITY = 11 # Variable c_int
EDGE_RAISED = 5 # Variable c_int
QS_SENDMESSAGE = 64 # Variable c_int
SMEXF_SERVER = 1
IMAGE_REL_ALPHA_REFHI = 10 # Variable c_int
ERROR_GROUP_NOT_AVAILABLE = 5012 # Variable c_long
R2_NOTMASKPEN = 8 # Variable c_int
SERVER_WRITE = 131075 # Variable c_long
MCI_STATUS = 2068 # Variable c_int
TCS_FIXEDWIDTH = 1024 # Variable c_int
LOCALE_STIMEFORMAT = 4099 # Variable c_int
URLACTION_INFODELIVERY_NO_ADDING_CHANNELS = 7424 # Variable c_int
IPPORT_RESERVED = 1024 # Variable c_int
ERROR_RESOURCE_DATA_NOT_FOUND = 1812 # Variable c_long
ERROR_INVALID_DRIVE = 15 # Variable c_long
IMAGE_FILE_LINE_NUMS_STRIPPED = 4 # Variable c_int
APPCLASS_MASK = 15 # Variable c_long
COMMON_LVB_GRID_RVERTICAL = 4096 # Variable c_int
DMPAPER_B5_TRANSVERSE = 62 # Variable c_int
MMIOERR_CANNOTWRITE = 262 # Variable c_int
CSTR_EQUAL = 2 # Variable c_int
PRINTER_STATUS_OUT_OF_MEMORY = 2097152 # Variable c_int
THREAD_SET_CONTEXT = 16 # Variable c_int
EVENTLOG_END_PAIRED_EVENT = 2 # Variable c_int
RBS_BANDBORDERS = 1024 # Variable c_int
PRINTER_NOTIFY_FIELD_DEFAULT_PRIORITY = 15 # Variable c_int
BACKUP_SPARSE_BLOCK = 9 # Variable c_int
WAVE_FORMAT_2S08 = 32 # Variable c_int
RPC_C_MQ_AUTHN_LEVEL_PKT_INTEGRITY = 8 # Variable c_int
GCS_CURSORPOS = 128 # Variable c_int
MS_CTS_ON = 16 # Variable c_ulong
TB_INSERTMARKHITTEST = 1105 # Variable c_int
BF_TOPLEFT = 3 # Variable c_int
HTCLIENT = 1 # Variable c_int
STGM_PRIORITY = 262144 # Variable c_long
CTLCOLOR_EDIT = 1 # Variable c_int
SUBLANG_ARABIC_MOROCCO = 6 # Variable c_int
EMR_WIDENPATH = 66 # Variable c_int
PROCESSOR_PPC_620 = 620 # Variable c_int
ERROR_MESSAGE_SYNC_ONLY = 1159 # Variable c_long
TV_FIRST = 4352 # Variable c_int
_MAX_DRIVE = 3 # Variable c_int
CTRY_RUSSIA = 7 # Variable c_int
SERVICE_ERROR_NORMAL = 1 # Variable c_int
Win32ServiceShareProcess = 32
IDC_UPARROW = 206102 # Variable POINTER(c_wchar)
DMLERR_NOTPROCESSED = 16393 # Variable c_int
WAVE_FORMAT_DIRECT_QUERY = 9 # Variable c_int
IMAGE_REL_ALPHA_GPRELLO = 22 # Variable c_int
SPAPI_E_NO_DEVICE_ICON = -2146500055 # Variable c_long
HDM_GETIMAGELIST = 4617 # Variable c_int
SWP_NOZORDER = 4 # Variable c_int
LEFT_CTRL_PRESSED = 8 # Variable c_int
URLPOLICY_CREDENTIALS_ANONYMOUS_ONLY = 196608 # Variable c_int
DRV_LOAD = 1 # Variable c_int
EMR_BEGINPATH = 59 # Variable c_int
DRIVE_FIXED = 3 # Variable c_int
IMAGE_FILE_MACHINE_POWERPC = 496 # Variable c_int
ES_MULTILINE = 4 # Variable c_long
EMR_POLYLINETO = 6 # Variable c_int
C3_HIRAGANA = 32 # Variable c_int

tagCALLTYPE = c_int # enum
CALLTYPE_TOPLEVEL = 1
CALLTYPE_NESTED = 2
CALLTYPE_ASYNC = 3
CALLTYPE_TOPLEVEL_CALLPENDING = 4
DTN_CLOSEUP = 4294966543 # Variable c_uint
TAPE_LOCK = 3 # Variable c_long
Win32ServiceOwnProcess = 16
LVGIT_UNFOLDED = 1 # Variable c_int
ABM_WINDOWPOSCHANGED = 9 # Variable c_int
HCBT_KEYSKIPPED = 7 # Variable c_int
GETSETPAPERMETRICS = 35 # Variable c_int
URLTEMPLATE_HIGH = 73728
PRINTER_STATUS_PROCESSING = 16384 # Variable c_int
SO_CONNECT_TIME = 28684 # Variable c_int
IMAGE_FILE_LOCAL_SYMS_STRIPPED = 8 # Variable c_int
DOF_EXECUTABLE = 32769 # Variable c_int
ERROR_NO_MORE_USER_HANDLES = 1158 # Variable c_long
AUXCAPS_AUXIN = 2 # Variable c_int
NOERROR = 0 # Variable c_int
ST_BLOCKED = 8 # Variable c_int
XTYP_ADVDATA = 16400 # Variable c_int
MMSYSERR_LASTERROR = 20 # Variable c_int
CFSTR_MIME_POSTSCRIPT = 'application/postscript' # Variable POINTER(c_wchar)
ES_SYSTEM_REQUIRED = 1 # Variable c_ulong
PRINTER_STATUS_PRINTING = 1024 # Variable c_int
MMIO_DENYREAD = 48 # Variable c_int
MIXERCONTROL_CT_CLASS_SLIDER = 1073741824 # Variable c_long
CCS_RIGHT = 131 # Variable c_long
ARABIC_CHARSET = 178 # Variable c_int
ERROR_CIRCULAR_DEPENDENCY = 1059 # Variable c_long
WSAEADDRINUSE = 10048 # Variable c_int
EWX_POWEROFF = 8 # Variable c_int
PRINTER_NOTIFY_FIELD_BYTES_PRINTED = 25 # Variable c_int
PRINTER_ENUM_ICON4 = 524288 # Variable c_int
PURGE_TXABORT = 1 # Variable c_int
NDR_VAX_FLOAT = 256 # Variable c_ulong
SECURITY_NULL_RID = 0 # Variable c_long
DSIG_FONTTYPE = 262144 # Variable c_int
HDI_ORDER = 128 # Variable c_int
META_DELETEOBJECT = 496 # Variable c_int
INET_E_DOWNLOAD_FAILURE = -2146697208 # Variable c_long
CAL_SABBREVMONTHNAME12 = 45 # Variable c_int
MCI_SEQ_STATUS_MASTER = 16392 # Variable c_long
TAPE_LOGICAL_BLOCK = 2 # Variable c_long
CDERR_DIALOGFAILURE = 65535 # Variable c_int
PSN_SETACTIVE = 4294967096 # Variable c_uint
TOKEN_READ = 131080 # Variable c_long
MCI_TO = 8 # Variable c_long
ERROR_NO_SUCH_USER = 1317 # Variable c_long
NETPROPERTY_PERSISTENT = 1 # Variable c_int
DMPAPER_A3_TRANSVERSE = 67 # Variable c_int
ERROR_NO_SUCH_PACKAGE = 1364 # Variable c_long
LCS_GM_BUSINESS = 1 # Variable c_long
KLF_NOTELLSHELL = 128 # Variable c_int
MMIO_COMPAT = 0 # Variable c_int
ERROR_RESOURCE_NAME_NOT_FOUND = 1814 # Variable c_long
RPC_C_MQ_CLEAR_ON_OPEN = 2 # Variable c_int
IMAGE_SIZEOF_SECTION_HEADER = 40 # Variable c_int
IMAGE_REL_SH3_ABSOLUTE = 0 # Variable c_int
SECTION_MAP_READ = 4 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_UNDEFINED = 0 # Variable c_long
SECURITY_RESTRICTED_CODE_RID = 12 # Variable c_long
LBS_MULTICOLUMN = 512 # Variable c_long
META_SETTEXTJUSTIFICATION = 522 # Variable c_int
FILE_GENERIC_EXECUTE = 1179808 # Variable c_long
IMN_CHANGECANDIDATE = 3 # Variable c_int
UDM_SETBUDDY = 1129 # Variable c_int
DCBA_FACEUPRIGHT = 3 # Variable c_int
AF_UNSPEC = 0 # Variable c_int
USER_MARSHAL_FC_ULONG = 9 # Variable c_int
TYPEFLAG_FDISPATCHABLE = 4096
JOB_OBJECT_MSG_END_OF_JOB_TIME = 1 # Variable c_int
IMAGE_SCN_ALIGN_32BYTES = 6291456 # Variable c_int
IMAGE_FILE_LARGE_ADDRESS_AWARE = 32 # Variable c_int
ERROR_NO_VOLUME_ID = 1173 # Variable c_long
DMPAPER_FANFOLD_US = 39 # Variable c_int
ERROR_WMI_SERVER_UNAVAILABLE = 4208 # Variable c_long
DRAGDROP_S_USEDEFAULTCURSORS = 262402 # Variable c_long
NTE_BAD_KEYSET_PARAM = -2146893793 # Variable c_long
JOYCAPS_HASV = 8 # Variable c_int
MCI_STATUS_MODE = 4 # Variable c_long
CB_GETCOUNT = 326 # Variable c_int
MB_ICONEXCLAMATION = 48 # Variable c_long
IMAGE_REL_ARM_ADDR32NB = 2 # Variable c_int
BAUD_300 = 16 # Variable c_ulong
RPC_C_MQ_EXPRESS = 0 # Variable c_int
TCIS_BUTTONPRESSED = 1 # Variable c_int
SPI_LANGDRIVER = 12 # Variable c_int
MM_WIM_DATA = 960 # Variable c_int
PSD_ENABLEPAGESETUPHOOK = 8192 # Variable c_int
MMSYSERR_ALLOCATED = 4 # Variable c_int
DESKTOP_SWITCHDESKTOP = 256 # Variable c_long
edt9 = 1160 # Variable c_int
RB_BEGINDRAG = 1048 # Variable c_int
LOCALE_IDAYLZERO = 38 # Variable c_int
EMR_FILLPATH = 62 # Variable c_int
PAN_LETT_NORMAL_CONTACT = 2 # Variable c_int
IMAGE_SCN_ALIGN_16BYTES = 5242880 # Variable c_int
LOCALE_ICOUNTRY = 5 # Variable c_int
RPC_C_AUTHN_LEVEL_PKT_INTEGRITY = 5 # Variable c_int
EVENTLOG_SEQUENTIAL_READ = 1 # Variable c_int
CFSTR_MIME_TIFF = 'image/tiff' # Variable POINTER(c_wchar)
SB_BOTH = 3 # Variable c_int
DETACHED_PROCESS = 8 # Variable c_int
MF_MENUBREAK = 64 # Variable c_long
CCH_MAX_PROPSTG_NAME = 31 # Variable c_int
VARFLAG_FDEFAULTCOLLELEM = 256
JobObjectBasicUIRestrictions = 4
ACL_REVISION2 = 2 # Variable c_int
BINDSTRING_UA_PIXELS = 7
WNNC_NET_POWERLAN = 983040 # Variable c_int
INFINITE = 4294967295 # Variable c_uint
INFOTIPSIZE = 1024 # Variable c_int
CC_SHOWHELP = 8 # Variable c_int
LCS_DEVICE_CMYK = 2 # Variable c_long
SE_CHANGE_NOTIFY_NAME = 'SeChangeNotifyPrivilege' # Variable POINTER(c_wchar)
ERROR_IS_SUBSTED = 135 # Variable c_long
DISPATCH_METHOD = 1 # Variable c_int
PGM_SETBUTTONSIZE = 5130 # Variable c_int
BINDSTRING_PASSWORD = 6
SOFTDIST_FLAG_DELETE_SUBSCRIPTION = 8 # Variable c_int
ERROR_HANDLE_DISK_FULL = 39 # Variable c_long
ERROR_ATOMIC_LOCKS_NOT_SUPPORTED = 174 # Variable c_long
TB_ISBUTTONENABLED = 1033 # Variable c_int
CTRY_GUATEMALA = 502 # Variable c_int
TAPE_DRIVE_FILEMARKS = 2147745792 # Variable c_uint
PSM_ADDPAGE = 1127 # Variable c_int
MCI_ANIM_PLAY_FAST = 262144 # Variable c_long
RTS_CONTROL_TOGGLE = 3 # Variable c_int
BI_RLE8 = 1 # Variable c_long
ONESTOPBIT = 0 # Variable c_int
IMAGE_REL_PPC_ADDR16 = 4 # Variable c_int
ERROR_PIPE_LISTENING = 536 # Variable c_long
RPC_E_REMOTE_DISABLED = -2147417828 # Variable c_long
STM_GETIMAGE = 371 # Variable c_int
WSAENETDOWN = 10050 # Variable c_int
MOD_IGNORE_ALL_MODIFIER = 1024 # Variable c_int
CB_GETITEMHEIGHT = 340 # Variable c_int
BINDINFOF_URLENCODESTGMEDDATA = 1
ACCESS_DS_SOURCE_A = 'DS' # Variable POINTER(c_char)
LVM_SETHOVERTIME = 4167 # Variable c_int
ERROR_NO_PROC_SLOTS = 89 # Variable c_long
MIDI_MAPPER = 4294967295 # Variable c_uint
N_BTMASK = 15 # Variable c_int
CO_E_RUNAS_CREATEPROCESS_FAILURE = -2147467239 # Variable c_long
ERROR_UNEXP_NET_ERR = 59 # Variable c_long
ODS_COMBOBOXEDIT = 4096 # Variable c_int

_tagPARSEACTION = c_int # enum
PARSE_CANONICALIZE = 1
PARSE_FRIENDLY = 2
PARSE_SECURITY_URL = 3
PARSE_ROOTDOCUMENT = 4
PARSE_DOCUMENT = 5
PARSE_ANCHOR = 6
PARSE_ENCODE = 7
PARSE_DECODE = 8
PARSE_PATH_FROM_URL = 9
PARSE_URL_FROM_PATH = 10
PARSE_MIME = 11
PARSE_SERVER = 12
PARSE_SCHEMA = 13
PARSE_SITE = 14
PARSE_DOMAIN = 15
PARSE_LOCATION = 16
PARSE_SECURITY_DOMAIN = 17
IME_CONFIG_GENERAL = 1 # Variable c_int
SSWF_CUSTOM = 4 # Variable c_int
GCP_GLYPHSHAPE = 16 # Variable c_int
OFN_SHARENOWARN = 1 # Variable c_int
RPC_S_DUPLICATE_ENDPOINT = 1740 # Variable c_long
SEC_FILE = 8388608 # Variable c_int
SZDDESYS_ITEM_FORMATS = 'Formats' # Variable POINTER(c_wchar)
JOB_NOTIFY_FIELD_PAGES_PRINTED = 21 # Variable c_int
GENERIC_ALL = 268435456 # Variable c_long
STM_GETICON = 369 # Variable c_int
HTRIGHT = 11 # Variable c_int
DDL_READONLY = 1 # Variable c_int
STD_PRINT = 14 # Variable c_int
RPC_E_CANTPOST_INSENDCALL = -2147418109 # Variable c_long
RPC_S_INVALID_NETWORK_OPTIONS = 1724 # Variable c_long
WSAEHOSTDOWN = 10064 # Variable c_int
SE_GROUP_USE_FOR_DENY_ONLY = 16 # Variable c_long
ERROR_CLASS_ALREADY_EXISTS = 1410 # Variable c_long
DRIVE_RAMDISK = 6 # Variable c_int
BF_SOFT = 4096 # Variable c_int
ABN_FULLSCREENAPP = 2 # Variable c_int
MCI_STEP = 2062 # Variable c_int
URLACTION_SHELL_CURR_MAX = 6149 # Variable c_int
ERROR_INVALID_DOMAIN_ROLE = 1354 # Variable c_long
SUBLANG_CHINESE_TRADITIONAL = 1 # Variable c_int
META_SETTEXTCOLOR = 521 # Variable c_int
ERROR_EMPTY = 4306 # Variable c_long
IMAGE_REL_SH3_STARTOF_SECTION = 12 # Variable c_int
SM_CYMAXTRACK = 60 # Variable c_int
META_CREATEBRUSHINDIRECT = 764 # Variable c_int
DMDUP_SIMPLEX = 1 # Variable c_int
ERROR_RESOURCE_LANG_NOT_FOUND = 1815 # Variable c_long
PBT_APMRESUMESUSPEND = 7 # Variable c_int
TVGN_CHILD = 4 # Variable c_int
DBG_CONTINUE = 65538 # Variable c_ulong
TT_PRIM_CSPLINE = 3 # Variable c_int
DESCKIND_NONE = 0
SE_SYSTEM_ENVIRONMENT_NAME = 'SeSystemEnvironmentPrivilege' # Variable POINTER(c_wchar)
EV_PERR = 512 # Variable c_int
PC_RESERVED = 1 # Variable c_int
URLPOLICY_AUTHENTICATE_MUTUAL_ONLY = 196608 # Variable c_int
ILD_ROP = 64 # Variable c_int
IMAGE_REL_SH3_PCREL8_WORD = 9 # Variable c_int
CALLBACK_WINDOW = 65536 # Variable c_long
CB_GETTOPINDEX = 347 # Variable c_int
DTM_SETFORMATA = 4101 # Variable c_int
PERF_COUNTER_LARGE_QUEUELEN_TYPE = 4523264 # Variable c_int
SYSPAL_STATIC = 1 # Variable c_int
MM_LOENGLISH = 4 # Variable c_int
TBN_GETINFOTIPW = 4294966577 # Variable c_uint
ERROR_INVALID_STATE = 5023 # Variable c_long
STG_E_READFAULT = -2147287010 # Variable c_long
TVHT_ONITEMINDENT = 8 # Variable c_int
MS_DSR_ON = 32 # Variable c_ulong
INET_E_AUTHENTICATION_REQUIRED = -2146697207 # Variable c_long
CTRY_UAE = 971 # Variable c_int
DMLERR_MEMORY_ERROR = 16392 # Variable c_int
BINDSTATUS_INSTALLINGCOMPONENTS = 8
BACKUP_SECURITY_DATA = 3 # Variable c_int
JOY_CAL_READYONLY = 2097152 # Variable c_long
HOTKEYF_ALT = 4 # Variable c_int
OBJ_REGION = 8 # Variable c_int
META_INVERTREGION = 298 # Variable c_int
ERROR_DISK_RECALIBRATE_FAILED = 1126 # Variable c_long
MCI_STATUS_TIME_FORMAT = 6 # Variable c_long
EMR_MOVETOEX = 27 # Variable c_int
IMAGE_REL_IA64_DIR64 = 5 # Variable c_int
CF_PENDATA = 10 # Variable c_int
MIXERCONTROL_CONTROLTYPE_CUSTOM = 0 # Variable c_long
EV_EVENT1 = 2048 # Variable c_int
MMIOERR_CHUNKNOTFOUND = 265 # Variable c_int
FIONREAD = 1074030207 # Variable c_long
ERROR_INVALID_SHOWWIN_COMMAND = 1449 # Variable c_long
CONNDLG_RO_PATH = 1 # Variable c_int
AUDIT_ALLOW_NO_PRIVILEGE = 1 # Variable c_int
EMR_SETBKMODE = 18 # Variable c_int
CF_DIF = 5 # Variable c_int
MMIO_REMOVEPROC = 131072 # Variable c_int
ADVF_NODATA = 1
EMR_MODIFYWORLDTRANSFORM = 36 # Variable c_int
VT_UI4 = 19
MCI_SET_VIDEO = 4096 # Variable c_long
SERVICE_KERNEL_DRIVER = 1 # Variable c_int
LOCALE_S2359 = 41 # Variable c_int
MF_DISABLED = 2 # Variable c_long
CW_USEDEFAULT = -2147483648 # Variable c_int
SPLREG_NET_POPUP = 'NetPopup' # Variable POINTER(c_wchar)
HWND_DESKTOP = 0 # Variable POINTER(HWND__)
DFCS_SCROLLDOWN = 1 # Variable c_int
CDM_GETFOLDERPATH = 1126 # Variable c_int
DDE_FACKRESERVED = -49408 # Variable c_int
PAN_CONTRAST_LOW = 4 # Variable c_int
MM_STREAM_ERROR = 983 # Variable c_int
PROCESSOR_ARCHITECTURE_UNKNOWN = 65535 # Variable c_int
DC_BINNAMES = 12 # Variable c_int
XST_EXECACKRCVD = 10 # Variable c_int
TIME_ZONE_ID_STANDARD = 1 # Variable c_int
LOCALE_SABBREVMONTHNAME5 = 72 # Variable c_int
LVS_EX_FLATSB = 256 # Variable c_int
MCIERR_UNRECOGNIZED_KEYWORD = 259 # Variable c_int
LB_RESETCONTENT = 388 # Variable c_int
IME_CHOTKEY_SHAPE_TOGGLE = 17 # Variable c_int
CO_E_APPSINGLEUSE = -2147221002 # Variable c_long
PORT_STATUS_OUT_OF_MEMORY = 9 # Variable c_int
FACILITY_CERT = 11 # Variable c_int
JOY_BUTTON14 = 8192 # Variable c_long
SM_CXMINSPACING = 47 # Variable c_int
MF_MENUBARBREAK = 32 # Variable c_long
LANG_SINDHI = 89 # Variable c_int
META_SETBKCOLOR = 513 # Variable c_int
QUERY_IS_CACHED_OR_MAPPED = 11
TYMED_HGLOBAL = 1
FOF_SILENT = 4 # Variable c_int
TYSPEC_PACKAGENAME = 8
WNNC_NET_TWINS = 2359296 # Variable c_int
REPARSE_GUID_DATA_BUFFER_HEADER_SIZE = 24 # Variable c_long
PERF_COUNTER_TIMER_INV = 557909248 # Variable c_int
PERF_AVERAGE_BASE = 1073939458 # Variable c_int
CLOSECHANNEL = 4112 # Variable c_int
STUB_CALL_SERVER_NO_HRESULT = 3
rad6 = 1061 # Variable c_int
CTRY_PANAMA = 507 # Variable c_int
EVENTLOG_FORWARDS_READ = 4 # Variable c_int
JOB_READ = 131088 # Variable c_long
MF_INSERT = 0 # Variable c_long
LOCALE_SYSTEM_DEFAULT = 2048 # Variable c_ulong
LVN_LAST = 4294967097 # Variable c_uint
SUBLANG_FRENCH = 1 # Variable c_int
ERROR_NO_SUCH_MEMBER = 1387 # Variable c_long
OLE_E_NOSTORAGE = -2147221486 # Variable c_long
TIME_NOSECONDS = 2 # Variable c_int
BN_SETFOCUS = 6 # Variable c_int
PGN_SCROLL = 4294966395 # Variable c_uint
ERROR_OUT_OF_PAPER = 28 # Variable c_long
SPAPI_E_FILEQUEUE_LOCKED = -2146500074 # Variable c_long
CBES_EX_PATHWORDBREAKPROC = 4 # Variable c_int
RPC_X_INVALID_PIPE_OBJECT = 1830 # Variable c_long
PSINJECT_PAGETRAILER = 13 # Variable c_int
NCBASTAT = 51 # Variable c_int
DECIMAL_NEG = 128 # Variable c_ubyte
RB_SETTOOLTIPS = 1042 # Variable c_int
PAN_SERIF_OBTUSE_COVE = 3 # Variable c_int
TAPE_DRIVE_GET_ABSOLUTE_BLK = 1048576 # Variable c_int
EMR_SELECTPALETTE = 48 # Variable c_int
PSD_DISABLEORIENTATION = 256 # Variable c_int
LVIF_STATE = 8 # Variable c_int
DST_COMPLEX = 0 # Variable c_int
MEM_E_INVALID_SIZE = -2146959343 # Variable c_long
TAPE_DRIVE_SEQUENTIAL_FMKS = 2148007936 # Variable c_uint
EM_SETRECTNP = 180 # Variable c_int
SKF_TRISTATE = 128 # Variable c_int
SIZE_MAXHIDE = 4 # Variable c_int
HELP_PARTIALKEY = 261 # Variable c_long
MCM_SETDAYSTATE = 4104 # Variable c_int
MIXERCONTROL_CONTROLTYPE_STEREOENH = 536936453 # Variable c_long
TVHT_TOLEFT = 2048 # Variable c_int
PROCESSOR_ARM920 = 2336 # Variable c_int
CO_E_APPDIDNTREG = -2147220994 # Variable c_long
SMEXF_HANDLER = 2
TAPE_SPACE_RELATIVE_BLOCKS = 5 # Variable c_long
PSWIZB_FINISH = 4 # Variable c_int
PIPE_CLIENT_END = 0 # Variable c_int
SPI_SETDEFAULTINPUTLANG = 90 # Variable c_int
OBJ_PAL = 5 # Variable c_int
FOURCC_DOS = 542330692 # Variable c_ulong
NRC_INVDDID = 59 # Variable c_int
N_TMASK1 = 192 # Variable c_int
REPARSE_DATA_BUFFER_HEADER_SIZE = 8 # Variable c_long
TCS_FLATBUTTONS = 8 # Variable c_int
ILC_COLOR32 = 32 # Variable c_int
SB_LINELEFT = 0 # Variable c_int
DISP_CHANGE_BADPARAM = -5 # Variable c_int
MIXER_OBJECTF_AUX = 1342177280 # Variable c_long
RPCFLG_NON_NDR = 2147483648 # Variable c_ulong
AF_ISO = 7 # Variable c_int
IMAGE_SUBSYSTEM_WINDOWS_CE_GUI = 9 # Variable c_int
TVIF_PARAM = 4 # Variable c_int
MCI_RECORD = 2063 # Variable c_int
EC_ENABLEALL = 0 # Variable c_int
OLECLOSE_SAVEIFDIRTY = 0
HDI_DI_SETITEM = 64 # Variable c_int
RESOURCEUSAGE_RESERVED = 2147483648 # Variable c_uint
JOY_BUTTON3 = 4 # Variable c_int
MB_ICONASTERISK = 64 # Variable c_long
VFT2_FONT_TRUETYPE = 3 # Variable c_long
TPM_VCENTERALIGN = 16 # Variable c_long
STREAM_CONTAINS_PROPERTIES = 4 # Variable c_int
SPLDS_PRINT_MIN_X_EXTENT = 'printMinXExtent' # Variable POINTER(c_wchar)
ERROR_PARTITION_FAILURE = 1105 # Variable c_long
psh8 = 1031 # Variable c_int
PC_WIDE = 16 # Variable c_int
TVM_ENSUREVISIBLE = 4372 # Variable c_int
CRYPT_E_OSS_ERROR = -2146881536 # Variable c_long
S_NORMAL = 0 # Variable c_int
AdapterType = 4
MCI_SEQ_STATUS_TEMPO = 16386 # Variable c_long
ERROR_BUFFER_OVERFLOW = 111 # Variable c_long
SUBLANG_SPANISH_VENEZUELA = 8 # Variable c_int
IOC_IN = 2147483648 # Variable c_uint
TOKEN_ADJUST_GROUPS = 64 # Variable c_int
PAN_CONTRAST_MEDIUM_LOW = 5 # Variable c_int
TVM_INSERTITEMW = 4402 # Variable c_int
SERVICE_ACTIVE = 1 # Variable c_int
ERROR_POPUP_ALREADY_ACTIVE = 1446 # Variable c_long
COMMON_LVB_SBCSDBCS = 768 # Variable c_int
WAVERR_BASE = 32 # Variable c_int
ERROR_INVALID_SERVICE_ACCOUNT = 1057 # Variable c_long
PC_INTERIORS = 128 # Variable c_int
DIALOPTION_DIALTONE = 256 # Variable c_int
URLACTION_SCRIPT_CURR_MAX = 5125 # Variable c_int
MK_E_UNAVAILABLE = -2147221021 # Variable c_long
WMSZ_RIGHT = 2 # Variable c_int

tagDISCARDCACHE = c_int # enum
DISCARDCACHE_SAVEIFDIRTY = 0
DISCARDCACHE_NOSAVE = 1
LVKF_SHIFT = 4 # Variable c_int
TAPE_LONG_FILEMARKS = 3 # Variable c_long
TIME_SAMPLES = 2 # Variable c_int
PORT_STATUS_WARMING_UP = 11 # Variable c_int
MB_TOPMOST = 262144 # Variable c_long
URLTEMPLATE_PREDEFINED_MAX = 131072
WC_COMBOBOXEXA = 'ComboBoxEx32' # Variable POINTER(c_char)
IME_SMODE_PHRASEPREDICT = 8 # Variable c_int
ACTIVEOBJECT_WEAK = 1 # Variable c_int
WSB_PROP_MASK = 4095 # Variable c_long
LB_SELITEMRANGE = 411 # Variable c_int
CTRY_HONDURAS = 504 # Variable c_int
STN_DISABLE = 3 # Variable c_int
STREAM_SPARSE_ATTRIBUTE = 8 # Variable c_int
VARCMP_NULL = 3 # Variable c_int
SS_REALSIZEIMAGE = 2048 # Variable c_long
UI_CAP_ROTANY = 4 # Variable c_int
ERROR_REGISTRY_RECOVERED = 1014 # Variable c_long
C2_SEGMENTSEPARATOR = 9 # Variable c_int
ICC_TAB_CLASSES = 8 # Variable c_int
STG_E_PATHNOTFOUND = -2147287037 # Variable c_long
IME_ESC_RESERVED_FIRST = 4 # Variable c_int
PBTF_APMRESUMEFROMFAILURE = 1 # Variable c_int
IMAGE_SYM_CLASS_BIT_FIELD = 18 # Variable c_int
DT_MODIFYSTRING = 65536 # Variable c_int
RPC_E_CANTCALLOUT_ININPUTSYNCCALL = -2147417843 # Variable c_long
RPC_E_CANTCALLOUT_INEXTERNALCALL = -2147418107 # Variable c_long
REG_DWORD_LITTLE_ENDIAN = 4 # Variable c_int
SRCAND = 8913094 # Variable c_ulong
PERF_100NSEC_MULTI_TIMER = 575735040 # Variable c_int
WH_DEBUG = 9 # Variable c_int
SPAPI_E_INVALID_COINSTALLER = -2146500057 # Variable c_long
DS_MODALFRAME = 128 # Variable c_long
PRINTER_ENUM_ICON7 = 4194304 # Variable c_int
WSB_PROP_CYVSCROLL = 1 # Variable c_long
PAN_STRAIGHT_ARMS_DOUBLE_SERIF = 6 # Variable c_int
STGFMT_DOCUMENT = 0 # Variable c_int
MS_RING_ON = 64 # Variable c_ulong
EXCEPTION_DEBUG_EVENT = 1 # Variable c_int
stc20 = 1107 # Variable c_int
EWX_LOGOFF = 0 # Variable c_int
CF_SHOWHELP = 4 # Variable c_long
MK_E_NOTBOUND = -2147221015 # Variable c_long
CO_E_BAD_PATH = -2146959356 # Variable c_long
ERROR_INVALID_CATEGORY = 117 # Variable c_long
SEE_MASK_DOENVSUBST = 512 # Variable c_int
RBHT_CAPTION = 2 # Variable c_int
MCI_VD_OFFSET = 1024 # Variable c_int
chx4 = 1043 # Variable c_int
EMR_SETMETARGN = 28 # Variable c_int
BAUD_2400 = 256 # Variable c_ulong
LANG_GEORGIAN = 55 # Variable c_int
ERROR_ALIAS_EXISTS = 1379 # Variable c_long
ERROR_LISTBOX_ID_NOT_FOUND = 1416 # Variable c_long
HANDLE_FLAG_PROTECT_FROM_CLOSE = 2 # Variable c_int
CONTEXT_SEGMENTS = 65540 # Variable c_long
MIXER_GETLINEINFOF_QUERYMASK = 15 # Variable c_long
MCIERR_MISSING_STRING_ARGUMENT = 269 # Variable c_int
PBT_APMRESUMEAUTOMATIC = 18 # Variable c_int
DOMAIN_GROUP_RID_CONTROLLERS = 516 # Variable c_long
DMPAPER_ENV_9 = 19 # Variable c_int
URLACTION_SCRIPT_SAFE_ACTIVEX = 5125 # Variable c_int
IMC_GETCOMPOSITIONWINDOW = 11 # Variable c_int
VFT2_DRV_INSTALLABLE = 8 # Variable c_long
KEY_EXECUTE = 131097 # Variable c_long
DDE_FADVRESERVED = -49153 # Variable c_int
MOD_SQSYNTH = 3 # Variable c_int
NIM_MODIFY = 1 # Variable c_int
DMPAPER_A2 = 66 # Variable c_int
IDHOT_SNAPDESKTOP = -2 # Variable c_int
TYPEFLAG_FREVERSEBIND = 8192
SPI_GETKEYBOARDDELAY = 22 # Variable c_int
SM_DBCSENABLED = 42 # Variable c_int
DC_DRIVER = 11 # Variable c_int
STG_E_INVALIDFLAG = -2147286785 # Variable c_long
EVENTLOG_WARNING_TYPE = 2 # Variable c_int
VAR_TIMEVALUEONLY = 1 # Variable c_ulong
MIDIERR_UNPREPARED = 64 # Variable c_int
COLOR_INACTIVECAPTION = 3 # Variable c_int
EMARCH_ENC_I17_IMM9D_INST_WORD_POS_X = 18 # Variable c_int
ENUM_CURRENT_SETTINGS = 4294967295 # Variable c_ulong
JOYCAPS_HASU = 4 # Variable c_int
CT_CTYPE2 = 2 # Variable c_int
LANG_CROATIAN = 26 # Variable c_int
HDN_ENDTRACKA = 4294966989 # Variable c_uint
SE_DACL_DEFAULTED = 8 # Variable c_int
SND_PURGE = 64 # Variable c_int
STD_DELETE = 5 # Variable c_int
EMARCH_ENC_I17_IMM5C_VAL_POS_X = 16 # Variable c_int
LANG_JAPANESE = 17 # Variable c_int
ERROR_SEVERITY_INFORMATIONAL = 1073741824 # Variable c_int
CTRY_INDONESIA = 62 # Variable c_int
SW_MAXIMIZE = 3 # Variable c_int

tagSF_TYPE = c_int # enum
SF_ERROR = 10
SF_I1 = 16
SF_I2 = 2
SF_I4 = 3
SF_I8 = 20
SF_BSTR = 8
SF_UNKNOWN = 13
SF_DISPATCH = 9
SF_VARIANT = 12
SF_RECORD = 36
SF_HAVEIID = 32781
CO_E_OBJNOTCONNECTED = -2147220995 # Variable c_long
MCI_DEVTYPE_SCANNER = 518 # Variable c_int
SAVE_CTM = 4101 # Variable c_int
WSAENOBUFS = 10055 # Variable c_int
MSGF_COMMCTRL_DRAGSELECT = 16898 # Variable c_int
JOB_OBJECT_UILIMIT_ALL = 255 # Variable c_int
PROGRESS_CLASSA = 'msctls_progress32' # Variable POINTER(c_char)
UPDFCACHE_ONLYIFBLANK = 2147483648 # Variable c_uint
ERROR_EAS_DIDNT_FIT = 275 # Variable c_long
TA_LEFT = 0 # Variable c_int
JOY_POVFORWARD = 0 # Variable c_int
LVM_GETBKIMAGEA = 4165 # Variable c_int
ERROR_BAD_COMMAND = 22 # Variable c_long
MCI_LAST = 4095 # Variable c_int
LBN_SELCHANGE = 1 # Variable c_int
LOCALE_SDAYNAME2 = 43 # Variable c_int
PAN_FAMILY_SCRIPT = 3 # Variable c_int
VARFLAG_FUIDEFAULT = 512
ERROR_NO_SUCH_LOGON_SESSION = 1312 # Variable c_long
TA_NOUPDATECP = 0 # Variable c_int
MIXERLINE_TARGETTYPE_MIDIIN = 4 # Variable c_int
LB_GETLOCALE = 422 # Variable c_int
WGL_SWAP_UNDERLAY1 = 65536 # Variable c_int
WSAEINVAL = 10022 # Variable c_int
LB_SETCOLUMNWIDTH = 405 # Variable c_int
RPC_S_NO_PROTSEQS = 1719 # Variable c_long
RPC_S_INVALID_TIMEOUT = 1709 # Variable c_long
RPC_E_VERSION_MISMATCH = -2147417840 # Variable c_long
EMARCH_ENC_I17_IMM41a_INST_WORD_X = 1 # Variable c_int
DDE_FPOKRESERVED = -8193 # Variable c_int
RPC_S_INVALID_STRING_BINDING = 1700 # Variable c_long
SYSTEM_FLAG_REMOTE_BOOT_CLIENT = 1 # Variable c_int
URLOSTRM_GETNEWESTVERSION = 3 # Variable c_int
ERROR_SHARING_VIOLATION = 32 # Variable c_long
LVM_SETTEXTCOLOR = 4132 # Variable c_int
HDI_IMAGE = 32 # Variable c_int
PRINTER_ENUM_DEFAULT = 1 # Variable c_int
DMPAPER_NOTE = 18 # Variable c_int
MM_JOY2MOVE = 929 # Variable c_int
OLEIVERB_HIDE = -3 # Variable c_long
MA_ACTIVATEANDEAT = 2 # Variable c_int
CDM_HIDECONTROL = 1129 # Variable c_int
BINDINFOF_URLENCODEDEXTRAINFO = 2
CF_NULL = 0 # Variable c_int
RAND_MAX = 32767 # Variable c_int
ERROR_CLUSTER_SHUTTING_DOWN = 5022 # Variable c_long
MMSYSERR_NOTSUPPORTED = 8 # Variable c_int
ERROR_REDIRECTOR_HAS_OPEN_HANDLES = 1794 # Variable c_long
PBT_APMSUSPEND = 4 # Variable c_int
edt4 = 1155 # Variable c_int
SPLDS_PRIORITY = 'priority' # Variable POINTER(c_wchar)
STG_E_INVALIDHEADER = -2147286789 # Variable c_long
PRINTER_CHANGE_CONFIGURE_PORT = 2097152 # Variable c_int
FILE_PERSISTENT_ACLS = 8 # Variable c_int
RPC_C_QOS_CAPABILITIES_DEFAULT = 0 # Variable c_int
PDERR_PARSEFAILURE = 4098 # Variable c_int
VIF_CANNOTDELETE = 4096 # Variable c_long
TC_SIGNAL = 3 # Variable c_int
ERROR_POSSIBLE_DEADLOCK = 1131 # Variable c_long
SERVICE_CONTROL_INTERROGATE = 4 # Variable c_int
WH_CALLWNDPROCRET = 12 # Variable c_int
ERROR_INVALID_PRINT_MONITOR = 3007 # Variable c_long
DCTT_SUBDEV = 4 # Variable c_long
chx10 = 1049 # Variable c_int
SB_GETBORDERS = 1031 # Variable c_int
MB_YESNO = 4 # Variable c_long
OF_SHARE_EXCLUSIVE = 16 # Variable c_int
SO_DONTLINGER = 4294967167 # Variable c_uint
NDR_IEEE_FLOAT = 0 # Variable c_ulong
META_EXTFLOODFILL = 1352 # Variable c_int
SHGNLI_NOUNIQUE = 4 # Variable c_int
MCIERR_HARDWARE = 262 # Variable c_int
ERROR_INVALID_BLOCK = 9 # Variable c_long
IMAGE_CURSOR = 2 # Variable c_int
ODS_DEFAULT = 32 # Variable c_int
SUBLANG_CHINESE_SINGAPORE = 4 # Variable c_int
HICF_ENTERING = 16 # Variable c_int
SPI_SETKEYBOARDPREF = 69 # Variable c_int
CF_NOSIZESEL = 2097152 # Variable c_long
EM_LINEINDEX = 187 # Variable c_int
PRINTER_NOTIFY_FIELD_SHARE_NAME = 2 # Variable c_int
HICF_LEAVING = 32 # Variable c_int
UOI_TYPE = 3 # Variable c_int
RPC_X_SS_IN_NULL_CONTEXT = 1775 # Variable c_long
NM_CHAR = 4294967278 # Variable c_uint
_HEX = 128 # Variable c_int
TAPE_DRIVE_COMPRESSION = 131072 # Variable c_int
AC_LINE_ONLINE = 1 # Variable c_int
ERROR_NON_MDICHILD_WINDOW = 1445 # Variable c_long
ERROR_EA_ACCESS_DENIED = 994 # Variable c_long
LANG_USER_DEFAULT = 1024 # Variable c_int
MMIOERR_CANNOTOPEN = 259 # Variable c_int
QUERYESCSUPPORT = 8 # Variable c_int
MAXIMUM_SUPPORTED_EXTENSION = 512 # Variable c_int
OR_INVALID_SET = 1912 # Variable c_long
RPC_C_PROTECT_LEVEL_PKT_PRIVACY = 6 # Variable c_int
HELP_FORCEFILE = 9 # Variable c_long
BS_DIBPATTERN = 5 # Variable c_int
ERROR_REMOTE_SESSION_LIMIT_EXCEEDED = 1220 # Variable c_long
PGM_GETBUTTONSTATE = 5132 # Variable c_int
MIXERCONTROL_CT_UNITS_MASK = 16711680 # Variable c_long
PSINJECT_BEGINPAGESETUP = 10 # Variable c_int
DTM_SETSYSTEMTIME = 4098 # Variable c_int
MCI_GETDEVCAPS_CAN_SAVE = 9 # Variable c_long
PAN_CONTRAST_VERY_HIGH = 9 # Variable c_int
DFCS_BUTTONRADIO = 4 # Variable c_int
TTN_LAST = 4294966747 # Variable c_uint
BINDSTATUS_ENDDOWNLOADCOMPONENTS = 9
PRINTER_ENUM_ICON5 = 1048576 # Variable c_int
SZ_URLCONTEXT = 'URL Context' # Variable POINTER(c_wchar)
PSD_INHUNDREDTHSOFMILLIMETERS = 8 # Variable c_int
CCS_NOPARENTALIGN = 8 # Variable c_long
META_SETDIBTODEV = 3379 # Variable c_int
LVM_GETTOPINDEX = 4135 # Variable c_int
CD_LBSELSUB = 1 # Variable c_int
MOUSEEVENTF_MIDDLEDOWN = 32 # Variable c_int
MAP_FOLDCZONE = 16 # Variable c_int
WSASYSNOTREADY = 10091 # Variable c_int
PAN_MIDLINE_STANDARD_POINTED = 3 # Variable c_int
MDITILE_SKIPDISABLED = 2 # Variable c_int
JOY_BUTTON2 = 2 # Variable c_int
WC_TABCONTROLA = 'SysTabControl32' # Variable POINTER(c_char)
PRINTER_CHANGE_ADD_PRINT_PROCESSOR = 16777216 # Variable c_int
CBR_56000 = 56000 # Variable c_int
LOGON32_LOGON_SERVICE = 5 # Variable c_int
STG_E_INVALIDPOINTER = -2147287031 # Variable c_long
BSF_NOTIMEOUTIFNOTHUNG = 64 # Variable c_int
CBN_EDITUPDATE = 6 # Variable c_int
MIXERCONTROL_CT_UNITS_BOOLEAN = 65536 # Variable c_long
IMC_SETCOMPOSITIONWINDOW = 12 # Variable c_int
ADVF_PRIMEFIRST = 2
ERROR_CANTWRITE = 1013 # Variable c_long
MCI_GETDEVCAPS_CAN_EJECT = 7 # Variable c_long
COMMON_LVB_GRID_LVERTICAL = 2048 # Variable c_int
MCIERR_CANNOT_USE_ALL = 279 # Variable c_int
MIDIPROP_SET = 2147483648 # Variable c_ulong
_MAX_PATH = 260 # Variable c_int
VT_EMPTY = 0
ERROR_BADKEY = 1010 # Variable c_long
GCL_WNDPROC = -24 # Variable c_int
HDS_DRAGDROP = 64 # Variable c_int
NUMLOCK_ON = 32 # Variable c_int
PROCESSOR_STRONGARM = 2577 # Variable c_int
IMAGE_DIRECTORY_ENTRY_EXPORT = 0 # Variable c_int
MCI_STRING_OFFSET = 512 # Variable c_int
SC_ACTION_NONE = 0
CF_SCALABLEONLY = 131072 # Variable c_long
MMIOM_CLOSE = 4 # Variable c_int
LVN_BEGINDRAG = 4294967187 # Variable c_uint
URLACTION_ACTIVEX_OVERRIDE_DATA_SAFETY = 4610 # Variable c_int
GCL_HBRBACKGROUND = -10 # Variable c_int
RPC_X_WRONG_PIPE_VERSION = 1832 # Variable c_long
ERROR_BAD_NETPATH = 53 # Variable c_long
IS_TEXT_UNICODE_REVERSE_SIGNATURE = 128 # Variable c_int
SO_SNDBUF = 4097 # Variable c_int
STD_REDOW = 4 # Variable c_int
SIOCSHIWAT = 2147775232 # Variable c_ulong
ABE_TOP = 1 # Variable c_int
QUERY_USES_CACHE = 12
TYMED_FILE = 2
CF_WAVE = 12 # Variable c_int
MAXDWORD = 4294967295 # Variable c_uint
LVM_GETITEMSTATE = 4140 # Variable c_int
ENUM_S_LAST = 262591 # Variable c_long
AUXCAPS_CDAUDIO = 1 # Variable c_int
ERROR_DS_CANT_ON_NON_LEAF = 1934 # Variable c_long
DM_DUPLEX = 4096 # Variable c_long
STOPBITS_10 = 1 # Variable c_ushort
FILE_ATTRIBUTE_NORMAL = 128 # Variable c_int
ERROR_NO_SITENAME = 1919 # Variable c_long
EMR_SCALEWINDOWEXTEX = 32 # Variable c_int
SIZE_MAXSHOW = 3 # Variable c_int
LVM_GETISEARCHSTRINGW = 4213 # Variable c_int
MCI_ANIM_STATUS_SPEED = 16385 # Variable c_long
OR_INVALID_OID = 1911 # Variable c_long
ERROR_SERVER_HAS_OPEN_HANDLES = 1811 # Variable c_long
LVSICF_NOSCROLL = 2 # Variable c_int
STD_INPUT_HANDLE = 4294967286 # Variable c_ulong
MCIERR_BAD_CONSTANT = 290 # Variable c_int
CAL_SMONTHNAME3 = 23 # Variable c_int
MCIERR_WAVE_SETOUTPUTINUSE = 321 # Variable c_int
CONTEXT_EXTENDED_REGISTERS = 65568 # Variable c_long

tagBINDSPEED = c_int # enum
BINDSPEED_INDEFINITE = 1
BINDSPEED_MODERATE = 2
BINDSPEED_IMMEDIATE = 3
ERROR_INVALID_PRINTER_COMMAND = 1803 # Variable c_long
FADF_EMBEDDED = 4 # Variable c_int
RPC_S_INVALID_BINDING = 1702 # Variable c_long
JOB_STATUS_DELETING = 4 # Variable c_int
MNC_IGNORE = 0 # Variable c_int
DMPAPER_A3 = 8 # Variable c_int
PATPAINT = 16452105 # Variable c_ulong
NRC_CMDTMO = 5 # Variable c_int
SERVICE_WIN32_OWN_PROCESS = 16 # Variable c_int
RPC_C_PROTECT_LEVEL_PKT_INTEGRITY = 5 # Variable c_int
REG_CREATED_NEW_KEY = 1 # Variable c_long
UDM_SETPOS = 1127 # Variable c_int
VAR_DATEVALUEONLY = 2 # Variable c_ulong
WC_TREEVIEWA = 'SysTreeView32' # Variable POINTER(c_char)
SP_RLSD = 64 # Variable c_ulong
MAX_NATURAL_ALIGNMENT = 4 # Variable c_uint
RPC_S_TYPE_ALREADY_REGISTERED = 1712 # Variable c_long

tagCOINIT = c_int # enum
META_SETMAPMODE = 259 # Variable c_int
HOTKEYF_SHIFT = 1 # Variable c_int
AD_CLOCKWISE = 2 # Variable c_int
TBS_BOTH = 8 # Variable c_int
URLACTION_HTML_MAX = 6143 # Variable c_int
PBT_APMRESUMESTANDBY = 8 # Variable c_int
CFERR_MAXLESSTHANMIN = 8194 # Variable c_int
CTRY_ESTONIA = 372 # Variable c_int
LVSIL_NORMAL = 0 # Variable c_int
EMR_POLYBEZIERTO = 5 # Variable c_int
LVM_INSERTITEMA = 4103 # Variable c_int
ACS_TRANSPARENT = 2 # Variable c_int
NEWFRAME = 1 # Variable c_int
ERROR_NO_QUOTAS_FOR_ACCOUNT = 1302 # Variable c_long
SUBLANG_SPANISH_HONDURAS = 18 # Variable c_int
META_OFFSETVIEWPORTORG = 529 # Variable c_int
IME_CAND_STROKE = 5 # Variable c_int
ACCESS_PROPERTY_SET_GUID = 1 # Variable c_int
TBN_ENDADJUST = 4294966592 # Variable c_uint
PAN_MIDLINE_CONSTANT_SERIFED = 10 # Variable c_int
TBSTATE_MARKED = 128 # Variable c_int
OF_DELETE = 512 # Variable c_int
PD_ALLPAGES = 0 # Variable c_int
OLECLOSE_NOSAVE = 1
PROCESS_TERMINATE = 1 # Variable c_int
CCS_NORESIZE = 4 # Variable c_long
VS_FF_PATCHED = 4 # Variable c_long
MB_DEFMASK = 3840 # Variable c_long
LVCF_SUBITEM = 8 # Variable c_int
MK_E_FIRST = 2147746272 # Variable c_ulong
DTN_FORMATQUERYW = 4294966554 # Variable c_uint
DM_MEDIATYPE = 33554432 # Variable c_long
IMAGE_REL_ALPHA_HINT = 8 # Variable c_int
EMR_SETPOLYFILLMODE = 19 # Variable c_int
TVM_GETSCROLLTIME = 4386 # Variable c_int
DSPRINT_PUBLISH = 1 # Variable c_int
RecognizerType = 8
PSCB_INITIALIZED = 1 # Variable c_int
MM_JOY1BUTTONUP = 951 # Variable c_int
TAPE_DRIVE_SET_BLOCK_SIZE = 2147483664 # Variable c_uint
MK_S_FIRST = 262624 # Variable c_long
BSF_FLUSHDISK = 4 # Variable c_int
MCI_WAVE_STATUS_CHANNELS = 16386 # Variable c_long
CB_GETLBTEXT = 328 # Variable c_int
TVM_EDITLABELA = 4366 # Variable c_int
SPAPI_E_EXPECTED_SECTION_NAME = -2146500608 # Variable c_long
IP_OPTIONS = 1 # Variable c_int
CTRY_BRAZIL = 55 # Variable c_int
LMEM_DISCARDABLE = 3840 # Variable c_int
ERROR_LOGON_NOT_GRANTED = 1380 # Variable c_long
ERROR_ENCRYPTION_FAILED = 6000 # Variable c_long
ERROR_SPL_NO_ADDJOB = 3004 # Variable c_long
STGFMT_FILE = 3 # Variable c_int
MIXERR_INVALCONTROL = 1025 # Variable c_int
URLOSTRM_USECACHEDCOPY = 2 # Variable c_int
AF_CCITT = 10 # Variable c_int
META_SETSTRETCHBLTMODE = 263 # Variable c_int
WS_MINIMIZE = 536870912 # Variable c_long
CREATE_FOR_DIR = 2 # Variable c_int
OLEIVERB_DISCARDUNDOSTATE = -6 # Variable c_long

tagURLZONE = c_int # enum
URLZONE_PREDEFINED_MIN = 0
URLZONE_LOCAL_MACHINE = 0
URLZONE_INTRANET = 1
URLZONE_TRUSTED = 2
URLZONE_INTERNET = 3
URLZONE_UNTRUSTED = 4
URLZONE_PREDEFINED_MAX = 999
URLZONE_USER_MIN = 1000
URLZONE_USER_MAX = 10000
NI_SELECTCANDIDATESTR = 18 # Variable c_int
ERROR_BAD_CONFIGURATION = 1610 # Variable c_long
XCLASS_MASK = 64512 # Variable c_int
LANG_ICELANDIC = 15 # Variable c_int
TCS_FORCEICONLEFT = 16 # Variable c_int
MCIERR_DRIVER = 278 # Variable c_int
RBN_CHILDSIZE = 4294966457 # Variable c_uint
NRC_REMTFUL = 18 # Variable c_int
CREATE_SEPARATE_WOW_VDM = 2048 # Variable c_int
EMR_ANGLEARC = 41 # Variable c_int
SE_SECURITY_NAME = 'SeSecurityPrivilege' # Variable POINTER(c_wchar)
DISP_E_MEMBERNOTFOUND = -2147352573 # Variable c_long
VIEW_E_LAST = 2147746127 # Variable c_ulong
TIMERR_BASE = 96 # Variable c_int
DD_DEFDRAGDELAY = 200 # Variable c_int
SPI_SETICONTITLELOGFONT = 34 # Variable c_int
VOS_NT_WINDOWS32 = 262148 # Variable c_long
DMMEDIA_USER = 256 # Variable c_int
TPM_TOPALIGN = 0 # Variable c_long
MDMSPKRFLAG_OFF = 1 # Variable c_int
DTN_USERSTRINGW = 4294966551 # Variable c_uint
PARAMFLAG_FHASCUSTDATA = 64 # Variable c_int
DCBA_FACEDOWNNONE = 256 # Variable c_int
stc4 = 1091 # Variable c_int
PRINTER_ENUM_ICON6 = 2097152 # Variable c_int
ERROR_SET_POWER_STATE_FAILED = 1141 # Variable c_long
RB_IDTOINDEX = 1040 # Variable c_int
TTM_SETMAXTIPWIDTH = 1048 # Variable c_int
PSD_MARGINS = 2 # Variable c_int
ODT_LISTVIEW = 102 # Variable c_int
CAL_SMONTHNAME4 = 24 # Variable c_int
ERROR_EXE_MARKED_INVALID = 192 # Variable c_long
stc21 = 1108 # Variable c_int
BIND_MAYBOTHERUSER = 1
MMIO_PARSE = 256 # Variable c_int
IDB_HIST_SMALL_COLOR = 8 # Variable c_int
URLPOLICY_CREDENTIALS_CONDITIONAL_PROMPT = 131072 # Variable c_int
IMAGE_FILE_MACHINE_SH3 = 418 # Variable c_int
MA_NOACTIVATE = 3 # Variable c_int
BS_PUSHLIKE = 4096 # Variable c_long
LANG_NEUTRAL = 0 # Variable c_int
FKF_CLICKON = 64 # Variable c_int
IMAGE_DLLCHARACTERISTICS_WDM_DRIVER = 8192 # Variable c_int
FS_VIETNAMESE = 256 # Variable c_long
MCI_VD_SPIN_UP = 65536 # Variable c_long
ST_CONNECTED = 1 # Variable c_int
ERROR_PRODUCT_UNINSTALLED = 1614 # Variable c_long
STM_MSGMAX = 372 # Variable c_int
DMPAPER_ENV_C4 = 30 # Variable c_int
PAN_MIDLINE_CONSTANT_TRIMMED = 8 # Variable c_int
SSGF_DISPLAY = 3 # Variable c_int
MINHALF_PTR = 32768 # Variable c_int
NI_IMEMENUSELECTED = 24 # Variable c_int
SERVICE_DRIVER = 11 # Variable c_int
SPI_GETACCESSTIMEOUT = 60 # Variable c_int
STREAM_NORMAL_ATTRIBUTE = 0 # Variable c_int
IMAGE_SYM_TYPE_DWORD = 15 # Variable c_int
STGM_SIMPLE = 134217728 # Variable c_long
WNNC_NET_LANMAN = 131072 # Variable c_int
HWND_TOP = 0 # Variable POINTER(HWND__)
PRINTER_ATTRIBUTE_DEFAULT = 4 # Variable c_int
FILE_NOTIFY_CHANGE_CREATION = 64 # Variable c_int
FO_DELETE = 3 # Variable c_int
MIXERCONTROL_CONTROLTYPE_SLIDER = 1073872896 # Variable c_long
WININETINFO_OPTION_LOCK_HANDLE = 65534 # Variable c_int
PST_X25 = 259 # Variable c_ulong
JOB_OBJECT_LIMIT_VALID_FLAGS = 127 # Variable c_int
NTE_PROV_DLL_NOT_FOUND = -2146893794 # Variable c_long
HDM_GETORDERARRAY = 4625 # Variable c_int
SUBLANG_LITHUANIAN_CLASSIC = 2 # Variable c_int
DATADIR_SET = 2
MOD_FMSYNTH = 4 # Variable c_int
CONNECT_TEMPORARY = 4 # Variable c_int
BN_CLICKED = 0 # Variable c_int
LANG_SLOVENIAN = 36 # Variable c_int
PGF_INVISIBLE = 0 # Variable c_int
IMAGE_SYM_TYPE_NULL = 0 # Variable c_int
GCP_ERROR = 32768 # Variable c_int
RPCFLG_MESSAGE = 16777216 # Variable c_ulong
PRINTER_ENUM_SHARED = 32 # Variable c_int
IMAGE_SIZEOF_AUX_SYMBOL = 18 # Variable c_int
PRINTER_CHANGE_JOB = 65280 # Variable c_int
MCI_SEQ_DIV_SMPTE_30 = 1220 # Variable c_int
CCS_ADJUSTABLE = 32 # Variable c_long
RESOURCE_CONNECTED = 1 # Variable c_int
MF_SYSMENU = 8192 # Variable c_long
LVNI_TORIGHT = 2048 # Variable c_int
WGL_SWAP_OVERLAY10 = 1024 # Variable c_int
SERVICE_ERROR_CRITICAL = 3 # Variable c_int
LOCALE_SDAYNAME3 = 44 # Variable c_int
RESOURCEDISPLAYTYPE_DOMAIN = 1 # Variable c_int
EPT_S_CANT_CREATE = 1899 # Variable c_long
VARFLAG_FNONBROWSABLE = 1024
WS_DISABLED = 134217728 # Variable c_long
AC_SRC_OVER = 0 # Variable c_int
ACCESS_SYSTEM_SECURITY = 16777216 # Variable c_long
WGL_SWAP_UNDERLAY2 = 131072 # Variable c_int
JOY_POVLEFT = 27000 # Variable c_int
DS_CENTERMOUSE = 4096 # Variable c_long
ERROR_TLW_WITH_WSCHILD = 1406 # Variable c_long
PAN_XHEIGHT_CONSTANT_LARGE = 4 # Variable c_int
WSAEDESTADDRREQ = 10039 # Variable c_int
GWL_USERDATA = -21 # Variable c_int
MOD_MAPPER = 5 # Variable c_int
IMAGE_REL_ARM_ADDR32 = 1 # Variable c_int
DC_TRUETYPE = 15 # Variable c_int
TVSIL_NORMAL = 0 # Variable c_int
SET_SPREAD = 4106 # Variable c_int
SS_ELLIPSISMASK = 49152 # Variable c_long
SPLDS_PRINT_COLOR = 'printColor' # Variable POINTER(c_wchar)
DISP_E_BADVARTYPE = -2147352568 # Variable c_long
IMAGE_SCN_ALIGN_512BYTES = 10485760 # Variable c_int
RC_SCALING = 4 # Variable c_int
MHDR_PREPARED = 2 # Variable c_int
SCS_CHANGECLAUSE = 36 # Variable c_int
CONTEXT_FULL = 65543 # Variable c_long
LVIS_SELECTED = 2 # Variable c_int
HDN_GETDISPINFOW = 4294966967 # Variable c_uint
HDN_ITEMDBLCLICKA = 4294966993 # Variable c_uint
OF_READWRITE = 2 # Variable c_int
CDDS_ITEM = 65536 # Variable c_int
CFS_RECT = 1 # Variable c_int
RB_GETPALETTE = 1062 # Variable c_int
BINDSTATUS_ENDDOWNLOADDATA = 6
EMR_SETROP2 = 20 # Variable c_int
ASPECTY = 42 # Variable c_int
SB_LINERIGHT = 1 # Variable c_int
TBTS_TOP = 0 # Variable c_int
CBS_DISABLENOSCROLL = 2048 # Variable c_long
MIXER_GETLINEINFOF_COMPONENTTYPE = 3 # Variable c_long
EV_RXFLAG = 2 # Variable c_int
MIXER_OBJECTF_MIDIOUT = 805306368 # Variable c_long
LB_GETCURSEL = 392 # Variable c_int
MNC_CLOSE = 1 # Variable c_int
CLASS_E_NOTLICENSED = -2147221230 # Variable c_long
DOF_DIRECTORY = 32771 # Variable c_int
CA_NEGATIVE = 1 # Variable c_int
PAN_PROP_CONDENSED = 6 # Variable c_int
NM_NCHITTEST = 4294967282 # Variable c_uint
TB_AUTOSIZE = 1057 # Variable c_int
chx11 = 1050 # Variable c_int
FILE_ACTION_RENAMED_NEW_NAME = 5 # Variable c_int
RPC_E_CALL_CANCELED = -2147418110 # Variable c_long
TBCDRF_NOMARK = 524288 # Variable c_int
DM_PAPERWIDTH = 8 # Variable c_long
RBN_HEIGHTCHANGE = 4294966465 # Variable c_uint
OFN_SHOWHELP = 16 # Variable c_int
VTBIT_R4 = 16 # Variable c_int
ES_LEFT = 0 # Variable c_long
SO_SYNCHRONOUS_NONALERT = 32 # Variable c_int
FILE_ATTRIBUTE_HIDDEN = 2 # Variable c_int
psh14 = 1037 # Variable c_int
MCI_MODE_RECORD = 527 # Variable c_int
SERVICE_CONTROL_NETBINDDISABLE = 10 # Variable c_int
SERVICES_ACTIVE_DATABASEA = 'ServicesActive' # Variable POINTER(c_char)
WINSTA_READATTRIBUTES = 2 # Variable c_long
IO_COMPLETION_ALL_ACCESS = 2031619 # Variable c_long
SSWF_DISPLAY = 3 # Variable c_int
SE_GROUP_OWNER = 8 # Variable c_long
NRC_ENVNOTDEF = 52 # Variable c_int
SE_ASSIGNPRIMARYTOKEN_NAME = 'SeAssignPrimaryTokenPrivilege' # Variable POINTER(c_wchar)
MCI_DEVTYPE_ANIMATION = 519 # Variable c_int
MCIERR_NO_ELEMENT_ALLOWED = 301 # Variable c_int
MCI_VD_MEDIA_CAV = 1027 # Variable c_int
ERROR_INFLOOP_IN_RELOC_CHAIN = 202 # Variable c_long
CBR_38400 = 38400 # Variable c_int
FR_WHOLEWORD = 2 # Variable c_int
GGO_GLYPH_INDEX = 128 # Variable c_int
HOVER_DEFAULT = 4294967295 # Variable c_uint
PSM_CANCELTOCLOSE = 1131 # Variable c_int
META_CREATEPATTERNBRUSH = 505 # Variable c_int
DC_PAPERNAMES = 16 # Variable c_int
EXIT_FAILURE = 1 # Variable c_int
LVM_FIRST = 4096 # Variable c_int
LVIS_OVERLAYMASK = 3840 # Variable c_int
OLECREATE_LEAVERUNNING = 1 # Variable c_int
SO_MAXPATHDG = 28682 # Variable c_int
ERROR_CANTREAD = 1012 # Variable c_long
CACHE_E_NOCACHE_UPDATED = -2147221136 # Variable c_long
BF_TOPRIGHT = 6 # Variable c_int
FS_CHINESETRAD = 1048576 # Variable c_long
RPC_S_FP_OVERFLOW = 1771 # Variable c_long
ERROR_INVALID_GROUPNAME = 1209 # Variable c_long
TCS_RAGGEDRIGHT = 2048 # Variable c_int
SB_THUMBTRACK = 5 # Variable c_int
DMPAPER_A5_TRANSVERSE = 61 # Variable c_int
BST_INDETERMINATE = 2 # Variable c_int
CHANGEKIND_SETNAMES = 2
ERROR_BAD_FORMAT = 11 # Variable c_long
BINDSTATUS_USINGCACHEDCOPY = 10
NTM_NONNEGATIVE_AC = 65536 # Variable c_int
PERF_COUNTER_RAWCOUNT_HEX = 0 # Variable c_int
CB_LIMITTEXT = 321 # Variable c_int
GETSETPAPERBINS = 29 # Variable c_int
META_TEXTOUT = 1313 # Variable c_int
MIIM_SUBMENU = 4 # Variable c_int
NUMRESERVED = 106 # Variable c_int
ASYNC_MODE_DEFAULT = 0 # Variable c_long
OPEN_ALWAYS = 4 # Variable c_int
LANG_KASHMIRI = 96 # Variable c_int
PIDSI_CHARCOUNT = 16 # Variable c_long
BINDSTATUS_MIMETYPEAVAILABLE = 13
JOYERR_NOCANDO = 166 # Variable c_int
DDL_POSTMSGS = 8192 # Variable c_int
MIXER_SETCONTROLDETAILSF_VALUE = 0 # Variable c_long
PAN_PROP_VERY_CONDENSED = 8 # Variable c_int
CO_E_OBJISREG = -2147220996 # Variable c_long
IMAGE_REL_IA64_ADDEND = 31 # Variable c_int
CO_E_INCOMPATIBLESTREAMVERSION = -2147220968 # Variable c_long
IMAGE_FILE_MACHINE_ALPHA64 = 644 # Variable c_int
CF_DIB = 8 # Variable c_int
CDERR_INITIALIZATION = 2 # Variable c_int
ADVF_ONLYONCE = 4
HGDI_ERROR = 4294967295 # Variable c_void_p
MCI_ANIM_GETDEVCAPS_FAST_RATE = 16386 # Variable c_long
VT_UI8 = 21
SUBLANG_SPANISH_URUGUAY = 14 # Variable c_int
IME_CMODE_ROMAN = 16 # Variable c_int
QID_SYNC = 4294967295 # Variable c_uint
CONVERT10_E_OLESTREAM_GET = -2147221056 # Variable c_long
PRINTER_CHANGE_ADD_PRINTER = 1 # Variable c_int
SPAPI_E_DI_BAD_PATH = -2146500076 # Variable c_long
LOCALE_STIME = 30 # Variable c_int
RPC_C_SECURITY_QOS_VERSION = 1 # Variable c_long
stc22 = 1109 # Variable c_int
REG_OPTION_CREATE_LINK = 2 # Variable c_long
HC_SKIP = 2 # Variable c_int
OLE_E_INVALIDHWND = -2147221489 # Variable c_long
BINDSTATUS_CACHEFILENAMEAVAILABLE = 14
SC_ACTION_RESTART = 1
ABM_GETTASKBARPOS = 5 # Variable c_int
IME_ESC_IME_NAME = 4102 # Variable c_int
MCI_SEQ_FORMAT_SONGPTR = 16385 # Variable c_int
E_FAIL = -2147467259 # Variable c_long
MAX_PRIORITY = 99 # Variable c_int
DMICM_ABS_COLORIMETRIC = 4 # Variable c_int
SPAPI_E_BAD_INTERFACE_INSTALLSECT = -2146500067 # Variable c_long
TVM_GETCOUNT = 4357 # Variable c_int
ERROR_VC_DISCONNECTED = 240 # Variable c_long
ERROR_RESOURCE_NOT_AVAILABLE = 5006 # Variable c_long
PRINTER_ATTRIBUTE_DIRECT = 2 # Variable c_int
WS_VSCROLL = 2097152 # Variable c_long

_tagPSUACTION = c_int # enum
TYMED_ISTREAM = 4
TVI_FIRST = 4294901761 # Variable POINTER(_TREEITEM)
PSINJECT_VMRESTORE = 18 # Variable c_int
MAXSHORT = 32767 # Variable c_int
LCID_INSTALLED = 1 # Variable c_int
PBT_APMRESUMECRITICAL = 6 # Variable c_int
IA64_JMPE_MARKER = 3473423 # Variable c_int
SP_PARITY_CHECK = 32 # Variable c_ulong
RBN_DELETINGBAND = 4294966459 # Variable c_uint
SPI_SETBORDER = 6 # Variable c_int
TB_SETINDENT = 1071 # Variable c_int
IE_DEFAULT = -5 # Variable c_int
RPC_E_TOO_LATE = -2147417831 # Variable c_long
IMAGE_SYM_CLASS_STATIC = 3 # Variable c_int
DCX_INTERSECTRGN = 128 # Variable c_long
PSINJECT_ENDDEFAULTS = 4 # Variable c_int
GGO_GRAY2_BITMAP = 4 # Variable c_int
ERROR_BUS_RESET = 1111 # Variable c_long
ERROR_DOMAIN_TRUST_INCONSISTENT = 1810 # Variable c_long
CAL_SMONTHNAME2 = 22 # Variable c_int
AF_IPX = 6 # Variable c_int
LB_CTLCODE = 0 # Variable c_long
IDH_OK = 28443 # Variable c_int
CRYPT_E_INVALID_NUMERIC_STRING = -2146885600 # Variable c_long
FILE_SUPPORTS_REPARSE_POINTS = 128 # Variable c_int
CRYPT_E_UNKNOWN_ALGO = -2146889726 # Variable c_long
NIM_ADD = 0 # Variable c_int
NONZEROLHND = 2 # Variable c_int
RESOURCEDISPLAYTYPE_SHAREADMIN = 8 # Variable c_int
OBJ_ENHMETADC = 12 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED = 4096 # Variable c_long
OLE_S_LAST = 262399 # Variable c_long
MDM_CELLULAR = 8 # Variable c_int
SHOW_FULLSCREEN = 3 # Variable c_int
ERROR_OLD_WIN_VERSION = 1150 # Variable c_long
USER_MARSHAL_FC_SMALL = 3 # Variable c_int
LOCALE_SENGLANGUAGE = 4097 # Variable c_int
RPC_C_OPT_MQ_ACKNOWLEDGE = 4 # Variable c_int
IME_CONFIG_SELECTDICTIONARY = 3 # Variable c_int
LB_ADDFILE = 406 # Variable c_int
BF_RECT = 15 # Variable c_int
SPI_GETBEEP = 1 # Variable c_int
CF_FORCEFONTEXIST = 65536 # Variable c_long
stc1 = 1088 # Variable c_int
ERROR_CONNECTION_ACTIVE = 1230 # Variable c_long
TBM_SETPAGESIZE = 1045 # Variable c_int
DRIVE_NO_ROOT_DIR = 1 # Variable c_int
ENABLE_LINE_INPUT = 2 # Variable c_int
_OUT_TO_MSGBOX = 2 # Variable c_int
DFCS_CHECKED = 1024 # Variable c_int
STD_FILESAVE = 8 # Variable c_int
DMPAPER_A5_EXTRA = 64 # Variable c_int
IMAGE_SIZEOF_STD_OPTIONAL_HEADER = 28 # Variable c_int
MCI_STATUS_READY = 7 # Variable c_long
META_SETWINDOWORG = 523 # Variable c_int
COLOR_ADJ_MIN = -100 # Variable c_short
WT_EXECUTEINWAITTHREAD = 4 # Variable c_int
TTM_SETTIPBKCOLOR = 1043 # Variable c_int
CONNECT_LOCALDRIVE = 256 # Variable c_int
BINDSTATUS_VERIFIEDMIMETYPEAVAILABLE = 22
HKCOMB_SCA = 128 # Variable c_int
SPLDS_SPOOLER_KEY = 'DsSpooler' # Variable POINTER(c_wchar)
AUXCAPS_LRVOLUME = 2 # Variable c_int
LVM_GETWORKAREAS = 4166 # Variable c_int
OLECLOSE_PROMPTSAVE = 2
RTL_CRITSECT_TYPE = 0 # Variable c_int
ERROR_CRC = 23 # Variable c_long
CBEM_GETEXTENDEDSTYLE = 1033 # Variable c_int
_PUNCT = 16 # Variable c_int
TBM_SETRANGEMIN = 1031 # Variable c_int
BINDSTATUS_CLASSINSTALLLOCATION = 23
SORT_DEFAULT = 0 # Variable c_int
LOCALE_SSHORTDATE = 31 # Variable c_int
WGL_SWAP_UNDERLAY6 = 2097152 # Variable c_int
SPI_SETTOGGLEKEYS = 53 # Variable c_int
DD_DEFSCROLLDELAY = 50 # Variable c_int
SHUTDOWN_NORETRY = 1 # Variable c_int
DEFAULT_GUI_FONT = 17 # Variable c_int
PURGE_RXABORT = 2 # Variable c_int
SPAPI_E_DEVINFO_LIST_LOCKED = -2146500078 # Variable c_long
PATCOPY = 15728673 # Variable c_ulong
MCI_WAVE_SET_ANYINPUT = 67108864 # Variable c_long
GPTR = 64 # Variable c_int
ERROR_ILL_FORMED_PASSWORD = 1324 # Variable c_long
LOCALE_SABBREVLANGNAME = 3 # Variable c_int
MIXER_GETLINECONTROLSF_QUERYMASK = 15 # Variable c_long
MB_TYPEMASK = 15 # Variable c_long
RPC_E_ACCESS_DENIED = -2147417829 # Variable c_long
PR_JOBSTATUS = 0 # Variable c_int
ABN_WINDOWARRANGE = 3 # Variable c_int
JOY_CAL_READZONLY = 16777216 # Variable c_long
ERROR_NEGATIVE_SEEK = 131 # Variable c_long
ERROR_SERIAL_NO_DEVICE = 1118 # Variable c_long
REGDB_E_CLASSNOTREG = -2147221164 # Variable c_long
CUDR_NEGATIVECOORDS = 8 # Variable c_int
TBM_GETPAGESIZE = 1046 # Variable c_int
MCIERR_EXTENSION_NOT_FOUND = 281 # Variable c_int
FROM_LEFT_2ND_BUTTON_PRESSED = 4 # Variable c_int
EMARCH_ENC_I17_IMM7B_VAL_POS_X = 0 # Variable c_int
PRINTER_CHANGE_SET_JOB = 512 # Variable c_int
SPI_SETHIGHCONTRAST = 67 # Variable c_int
SM_CYSMCAPTION = 51 # Variable c_int
NTM_DSIG = 2097152 # Variable c_int
TCN_FIRST = 4294966746 # Variable c_uint
MMIO_SHAREMODE = 112 # Variable c_int
ERROR_INVALID_FILTER_PROC = 1427 # Variable c_long
PSP_DLGINDIRECT = 1 # Variable c_int
IMAGE_REL_PPC_SECREL = 11 # Variable c_int
ERROR_DEVICE_ALREADY_REMEMBERED = 1202 # Variable c_long
TBCD_THUMB = 2 # Variable c_int
MMIOERR_INVALIDFILE = 272 # Variable c_int
LVFI_WRAP = 32 # Variable c_int
ERROR_MEMBERS_PRIMARY_GROUP = 1374 # Variable c_long
GCP_NUMERICSLATIN = 67108864 # Variable c_long
VT_DATE = 7
TTM_RELAYEVENT = 1031 # Variable c_int
MMSYSERR_NOTENABLED = 3 # Variable c_int
RPC_C_PROFILE_MATCH_BY_IF = 2 # Variable c_int
IMAGE_SCN_LNK_COMDAT = 4096 # Variable c_int
IME_ESC_PRIVATE_LAST = 4095 # Variable c_int
DEVICE_FONTTYPE = 2 # Variable c_int
SPI_SETANIMATION = 73 # Variable c_int
IMAGE_SYM_CLASS_REGISTER_PARAM = 17 # Variable c_int
STG_E_FILEALREADYEXISTS = -2147286960 # Variable c_long
PRF_ERASEBKGND = 8 # Variable c_long
MAX_SIZE_SECURITY_ID = 512 # Variable c_int
RPC_C_VERS_COMPATIBLE = 2 # Variable c_int
IMAGE_ARCHIVE_LINKER_MEMBER = '/               ' # Variable POINTER(c_char)
PAGE_WRITECOPY = 8 # Variable c_int
NRC_NORESOURCES = 56 # Variable c_int
DEREGISTERED = 5 # Variable c_int
FR_ENABLETEMPLATEHANDLE = 8192 # Variable c_int
STDOLE_MINORVERNUM = 0 # Variable c_int
BIND_JUSTTESTEXISTENCE = 2
OBJECT_INHERIT_ACE = 1 # Variable c_int
CDERR_STRUCTSIZE = 1 # Variable c_int
SPI_ICONVERTICALSPACING = 24 # Variable c_int
IMAGE_REL_SH3_DIRECT4_WORD = 7 # Variable c_int
CAL_SABBREVMONTHNAME1 = 34 # Variable c_int
ERROR_DS_OBJ_CLASS_VIOLATION = 1933 # Variable c_long
SBM_SETRANGE = 226 # Variable c_int
UNIQUE_NAME = 0 # Variable c_int
CBN_DBLCLK = 2 # Variable c_int
REBARCLASSNAMEA = 'ReBarWindow32' # Variable POINTER(c_char)
BACKUP_DATA = 1 # Variable c_int
IMAGE_COMDAT_SELECT_SAME_SIZE = 3 # Variable c_int
LB_GETITEMHEIGHT = 417 # Variable c_int
TVS_EDITLABELS = 8 # Variable c_int
TVN_SELCHANGINGW = 4294966846 # Variable c_uint
SYSTEM_FIXED_FONT = 16 # Variable c_int
MCI_BREAK_OFF = 1024 # Variable c_long
URLACTION_INFODELIVERY_NO_EDITING_SUBSCRIPTIONS = 7428 # Variable c_int
USER_MARSHAL_FC_SHORT = 6 # Variable c_int
SE_DACL_AUTO_INHERITED = 1024 # Variable c_int
BSCF_LASTDATANOTIFICATION = 4
CRYPT_E_SELF_SIGNED = -2146885625 # Variable c_long
URLACTION_INFODELIVERY_NO_REMOVING_SUBSCRIPTIONS = 7429 # Variable c_int
TVIS_SELECTED = 2 # Variable c_int
TVM_GETITEMSTATE = TV_FIRST + 39
ERROR_TOO_MANY_TCBS = 155 # Variable c_long
ERROR_DISK_OPERATION_FAILED = 1127 # Variable c_long
CTRY_SOUTH_AFRICA = 27 # Variable c_int
IMAGE_SCN_CNT_INITIALIZED_DATA = 64 # Variable c_int
QUERY_IS_CACHED = 9
MEM_COMMIT = 4096 # Variable c_int
MM_JOY1MOVE = 928 # Variable c_int
CBS_SIMPLE = 1 # Variable c_long
PIDSI_LASTSAVE_DTM = 13 # Variable c_long
C2_NOTAPPLICABLE = 0 # Variable c_int
TB_SETBUTTONSIZE = 1055 # Variable c_int
MCI_OVLY_WINDOW_STATE = 262144 # Variable c_long
IP_MAX_MEMBERSHIPS = 20 # Variable c_int

_RPC_NOTIFICATION_TYPES = c_int # enum
RpcNotificationTypeNone = 0
RpcNotificationTypeEvent = 1
RpcNotificationTypeApc = 2
RpcNotificationTypeIoc = 3
RpcNotificationTypeHwnd = 4
RpcNotificationTypeCallback = 5
JOY_BUTTON7 = 64 # Variable c_long
IPPROTO_TCP = 6 # Variable c_int
ERROR_NO_SPOOL_SPACE = 62 # Variable c_long
TBS_VERT = 2 # Variable c_int
MCM_SETSELRANGE = 4102 # Variable c_int
DISPID_VALUE = 0 # Variable c_int
URLACTION_NETWORK_MAX = 7167 # Variable c_int
SKF_HOTKEYSOUND = 16 # Variable c_int
RPC_S_INVALID_OBJECT = 1900 # Variable c_long
IDC_ICON = 206401 # Variable POINTER(c_wchar)
SERVICE_WIN32_SHARE_PROCESS = 32 # Variable c_int
RC_OP_DX_OUTPUT = 16384 # Variable c_int
PRINTER_HTML_VIEW_QUEUE = 0 # Variable c_int
DELETE = 65536 # Variable c_long
MDMSPKR_DIAL = 1 # Variable c_int
RPC_C_NS_DEFAULT_EXP_AGE = -1 # Variable c_int
IDNO = 7 # Variable c_int
JOY_BUTTON22 = 2097152 # Variable c_long
MK_S_ASYNCHRONOUS = 262632 # Variable c_long
HDN_ENDDRAG = 4294966985 # Variable c_uint
IMAGE_REL_BASED_IA64_IMM64 = 9 # Variable c_int
LOCALE_SDAYNAME1 = 42 # Variable c_int
LOGON32_PROVIDER_DEFAULT = 0 # Variable c_int
MCIERR_FILE_NOT_FOUND = 275 # Variable c_int
VARFLAG_FREPLACEABLE = 2048
TCIF_TEXT = 1 # Variable c_int
PBT_APMBATTERYLOW = 9 # Variable c_int
CBN_ERRSPACE = -1 # Variable c_int
BLACK_BRUSH = 4 # Variable c_int
MM_DRVM_CLOSE = 977 # Variable c_int
PAN_ARMSTYLE_INDEX = 6 # Variable c_int
LPD_SUPPORT_OPENGL = 32 # Variable c_int
OPAQUE = 2 # Variable c_int
ERROR_LOG_FILE_FULL = 1502 # Variable c_long
PID_FIRST_NAME_DEFAULT = 4095 # Variable c_int
DRV_CLOSE = 4 # Variable c_int
RPC_C_PROTECT_LEVEL_CONNECT = 2 # Variable c_int
ERROR_WAIT_NO_CHILDREN = 128 # Variable c_long
EMR_SETDIBITSTODEVICE = 80 # Variable c_int
IPN_LAST = 4294966417 # Variable c_uint
FORMAT_MESSAGE_FROM_SYSTEM = 4096 # Variable c_int
MIIM_ID = 2 # Variable c_int
LVM_SETEXTENDEDLISTVIEWSTYLE = 4150 # Variable c_int
ILLUMINANT_D50 = 4 # Variable c_int
ERROR_INVALID_KEYBOARD_HANDLE = 1457 # Variable c_long
OF_SHARE_DENY_READ = 48 # Variable c_int
CTRY_ICELAND = 354 # Variable c_int
ERROR_NO_UNICODE_TRANSLATION = 1113 # Variable c_long

tagSTGTY = c_int # enum
STGTY_STORAGE = 1
STGTY_STREAM = 2
STGTY_LOCKBYTES = 3
STGTY_PROPERTY = 4
TCS_BOTTOM = 2 # Variable c_int
DT_RASDISPLAY = 1 # Variable c_int
GETDEVICEUNITS = 42 # Variable c_int
OLE_E_CANTCONVERT = -2147221487 # Variable c_long
TTF_ABSOLUTE = 128 # Variable c_int
CF_GDIOBJLAST = 1023 # Variable c_int
MCHT_TODAYLINK = 196608 # Variable c_int
MIXERCONTROL_CONTROLTYPE_MILLITIME = 1627586560 # Variable c_long
RPC_C_VERS_EXACT = 3 # Variable c_int
GL_ID_UNKNOWN = 0 # Variable c_int
ODT_TAB = 101 # Variable c_int
SBN_SIMPLEMODECHANGE = 4294966416 # Variable c_uint
CO_E_IIDSTRING = -2147221004 # Variable c_long
SE_AUDIT_NAME = 'SeAuditPrivilege' # Variable POINTER(c_wchar)
E_PENDING = -2147483638 # Variable c_long
PRINTER_CHANGE_DELETE_PRINTER_DRIVER = 1073741824 # Variable c_int
DOF_PROGMAN = 1 # Variable c_int
MKSYS_POINTERMONIKER = 5
chx16 = 1055 # Variable c_int
HCBT_CREATEWND = 3 # Variable c_int
SPI_SETHANDHELD = 78 # Variable c_int
DCX_CACHE = 2 # Variable c_long
URLACTION_HTML_CURR_MAX = 5637 # Variable c_int
GWL_HINSTANCE = -6 # Variable c_int
SW_SHOWMINIMIZED = 2 # Variable c_int
URLACTION_JAVA_MIN = 7168 # Variable c_int
LVS_EDITLABELS = 512 # Variable c_int
LMEM_LOCKCOUNT = 255 # Variable c_int
MF_BITMAP = 4 # Variable c_long
IMAGE_FILE_MACHINE_R4000 = 358 # Variable c_int
DFCS_INACTIVE = 256 # Variable c_int
MCIERR_UNNAMED_RESOURCE = 298 # Variable c_int

__MIDL_IBindStatusCallback_0003 = c_int # enum
BINDF_ASYNCHRONOUS = 1
BINDF_ASYNCSTORAGE = 2
BINDF_NOPROGRESSIVERENDERING = 4
BINDF_OFFLINEOPERATION = 8
BINDF_GETNEWESTVERSION = 16
BINDF_NOWRITECACHE = 32
BINDF_NEEDFILE = 64
BINDF_PULLDATA = 128
BINDF_IGNORESECURITYPROBLEM = 256
BINDF_RESYNCHRONIZE = 512
BINDF_HYPERLINK = 1024
BINDF_NO_UI = 2048
BINDF_SILENTOPERATION = 4096
BINDF_PRAGMA_NO_CACHE = 8192
BINDF_FREE_THREADED = 65536
BINDF_DIRECT_READ = 131072
BINDF_FORMS_SUBMIT = 262144
BINDF_GETFROMCACHE_IF_NET_FAIL = 524288
LB_INITSTORAGE = 424 # Variable c_int
STARTF_FORCEONFEEDBACK = 64 # Variable c_int
HICF_OTHER = 0 # Variable c_int
ERROR_DEVICE_REQUIRES_CLEANING = 1165 # Variable c_long
PERF_COUNTER_RATE = 65536 # Variable c_int
WGL_SWAP_OVERLAY8 = 256 # Variable c_int
ERROR_CANT_ACCESS_FILE = 1920 # Variable c_long
SPAPI_E_INVALID_CLASS = -2146500090 # Variable c_long
BAUD_19200 = 8192 # Variable c_ulong
MIXER_OBJECTF_MIXER = 0 # Variable c_long
STG_S_RETRYNOW = 197122 # Variable c_long
psh3 = 1026 # Variable c_int
HKCOMB_CA = 64 # Variable c_int
MCI_CUE = 2096 # Variable c_int
ERROR_TAG_NOT_PRESENT = 2303 # Variable c_long
ONE5STOPBITS = 1 # Variable c_int
ERROR_TRUSTED_RELATIONSHIP_FAILURE = 1789 # Variable c_long
LVM_SETTEXTBKCOLOR = 4134 # Variable c_int
WNNC_NET_SUN_PC_NFS = 458752 # Variable c_int
PSINJECT_APPEND = 0 # Variable c_int
MM_WOM_DONE = 957 # Variable c_int
META_PATBLT = 1565 # Variable c_int
FILE_TYPE_CHAR = 2 # Variable c_int
MIDICAPS_STREAM = 8 # Variable c_int
REG_LINK = 6 # Variable c_int
ERROR_RXACT_INVALID_STATE = 1369 # Variable c_long
XMLELEMTYPE_DOCUMENT = 3
BINDSTATUS_SENDINGREQUEST = 11
STANDARD_RIGHTS_ALL = 2031616 # Variable c_long
EMARCH_ENC_I17_IMM41b_VAL_POS_X = 32 # Variable c_int
SM_CXHTHUMB = 10 # Variable c_int
ERROR_REQ_NOT_ACCEP = 71 # Variable c_long
ERROR_INSUFFICIENT_BUFFER = 122 # Variable c_long
PROFILE_KERNEL = 536870912 # Variable c_int
SUBLANG_NEUTRAL = 0 # Variable c_int
MEVT_NOP = 2 # Variable c_ubyte
TAPE_ERASE_LONG = 1 # Variable c_long
ERROR_INVALID_SUB_AUTHORITY = 1335 # Variable c_long
N_TMASK2 = 240 # Variable c_int
PS_ENDCAP_FLAT = 512 # Variable c_int
PSD_DISABLEMARGINS = 16 # Variable c_int
JOB_STATUS_SPOOLING = 8 # Variable c_int
FILE_ATTRIBUTE_OFFLINE = 4096 # Variable c_int
PRINTER_STATUS_OUTPUT_BIN_FULL = 2048 # Variable c_int
ERROR_INVALID_OWNER = 1307 # Variable c_long
PERF_SIZE_VARIABLE_LEN = 768 # Variable c_int
DM_COPY = 2 # Variable c_int
MB_USEGLYPHCHARS = 4 # Variable c_int
IMAGE_SCN_ALIGN_8192BYTES = 14680064 # Variable c_int
PFD_GENERIC_FORMAT = 64 # Variable c_int
DRAGDROP_S_LAST = 262415 # Variable c_long
VT_INT = 22
SOFTDIST_FLAG_USAGE_AUTOINSTALL = 4 # Variable c_int
MCI_SEQ_STATUS_COPYRIGHT = 16396 # Variable c_long
CDM_GETFOLDERIDLIST = 1127 # Variable c_int
NDR_CHAR_REP_MASK = 15 # Variable c_ulong
SBT_TOOLTIPS = 2048 # Variable c_int
NORM_IGNOREKANATYPE = 65536 # Variable c_int
MK_E_CONNECTMANUALLY = -2147221024 # Variable c_long
S_SERDFQ = -13 # Variable c_int
PS_ENDCAP_ROUND = 0 # Variable c_int
DMLERR_INVALIDPARAMETER = 16390 # Variable c_int
PRSPEC_INVALID = 4294967295 # Variable c_uint
SC_ACTION_REBOOT = 2
TVHT_ONITEM = 70 # Variable c_int
R2_MERGENOTPEN = 12 # Variable c_int
MIXERLINE_LINEF_SOURCE = 2147483648 # Variable c_ulong
GMDI_USEDISABLED = 1 # Variable c_long
MCI_GETDEVCAPS_HAS_AUDIO = 2 # Variable c_long
PARITY_ODD = 512 # Variable c_ushort
ERROR_INVALID_SERVER_STATE = 1352 # Variable c_long
MF_BYPOSITION = 1024 # Variable c_long
MIXERCONTROL_CONTROLTYPE_MUTE = 536936450 # Variable c_long
TB_SETBITMAPSIZE = 1056 # Variable c_int
WS_EX_MDICHILD = 64 # Variable c_long
CTRY_PAKISTAN = 92 # Variable c_int
OLE_E_NOCONNECTION = -2147221500 # Variable c_long
DCB_ENABLE = 4 # Variable c_int
UNLOAD_DLL_DEBUG_EVENT = 7 # Variable c_int
PGS_HORZ = 1 # Variable c_int
PFD_TYPE_COLORINDEX = 1 # Variable c_int
JOY_RETURNR = 8 # Variable c_long
RPC_E_INVALID_EXTENSION = -2147417838 # Variable c_long
STOPBITS_15 = 2 # Variable c_ushort
STG_E_UNIMPLEMENTEDFUNCTION = -2147286786 # Variable c_long
VIEW_S_LAST = 262479 # Variable c_long
CTRY_NETHERLANDS = 31 # Variable c_int
TIMER_QUERY_STATE = 1 # Variable c_int
UDM_GETRANGE32 = 1136 # Variable c_int
HCBT_DESTROYWND = 4 # Variable c_int
BS_MULTILINE = 8192 # Variable c_long
DSS_UNION = 16 # Variable c_int
PROPSETFLAG_NONSIMPLE = 1 # Variable c_int
VERTSIZE = 6 # Variable c_int
RC_STRETCHBLT = 2048 # Variable c_int
FILE_ATTRIBUTE_SPARSE_FILE = 512 # Variable c_int
ERROR_MEMBER_IN_ALIAS = 1378 # Variable c_long
PRINTER_NOTIFY_FIELD_PARAMETERS = 10 # Variable c_int
ERROR_INVALID_FLAGS = 1004 # Variable c_long
CTRY_CROATIA = 385 # Variable c_int
CBEM_GETEDITCONTROL = 1031 # Variable c_int

tagOLEGETMONIKER = c_int # enum
OLEGETMONIKER_ONLYIFTHERE = 1
OLEGETMONIKER_FORCEASSIGN = 2
OLEGETMONIKER_UNASSIGN = 3
OLEGETMONIKER_TEMPFORUSER = 4
META_FILLREGION = 552 # Variable c_int
OFN_FILEMUSTEXIST = 4096 # Variable c_int
META_PIE = 2074 # Variable c_int
URLACTION_ACTIVEX_OVERRIDE_OBJECT_SAFETY = 4609 # Variable c_int
CDERR_MEMALLOCFAILURE = 9 # Variable c_int
EVENT_ALL_ACCESS = 2031619 # Variable c_long
WAVECAPS_LRVOLUME = 8 # Variable c_int
CERT_E_CN_NO_MATCH = -2146762481 # Variable c_long
MCIERR_DEVICE_ORD_LENGTH = 311 # Variable c_int
EMARCH_ENC_I17_IMM7B_INST_WORD_POS_X = 4 # Variable c_int
RC_BIGFONT = 1024 # Variable c_int
ERROR_NO_SCROLLBARS = 1447 # Variable c_long
CFSTR_MIME_AIFF = 'audio/aiff' # Variable POINTER(c_wchar)
IMAGE_REL_BASED_HIGHADJ = 4 # Variable c_int
CBF_FAIL_CONNECTIONS = 8192 # Variable c_int
DFCS_SCROLLSIZEGRIPRIGHT = 16 # Variable c_int
TVN_GETINFOTIPW = 4294966882 # Variable c_uint
TVCDRF_NOIMAGES = 65536 # Variable c_int
BI_BITFIELDS = 3 # Variable c_long
CDIS_INDETERMINATE = 256 # Variable c_int
LVCFMT_IMAGE = 2048 # Variable c_int
SPI_SETSHOWSOUNDS = 57 # Variable c_int
IMAGE_REL_ALPHA_GPDISP = 6 # Variable c_int
CLIP_DEFAULT_PRECIS = 0 # Variable c_int
CCHDEVICENAME = 32 # Variable c_int
IDIGNORE = 5 # Variable c_int
TCM_GETTOOLTIPS = 4909 # Variable c_int
SBS_VERT = 1 # Variable c_long
DMPAPER_ISO_B4 = 42 # Variable c_int
STATFLAG_DEFAULT = 0
DM_DISPLAYFREQUENCY = 4194304 # Variable c_long
ERROR_INVALID_PIXEL_FORMAT = 2000 # Variable c_long
STATUS_ARRAY_BOUNDS_EXCEEDED = 3221225612 # Variable c_ulong
SECURITY_CREATOR_OWNER_RID = 0 # Variable c_long
VFF_CURNEDEST = 1 # Variable c_int
ABS_ALWAYSONTOP = 2 # Variable c_int
VS_FFI_FILEFLAGSMASK = 63 # Variable c_long
MIXERCONTROL_CT_SC_TIME_MICROSECS = 0 # Variable c_long
TB_SETIMAGELIST = 1072 # Variable c_int

tagOLEMISC = c_int # enum
OLEMISC_RECOMPOSEONRESIZE = 1
OLEMISC_ONLYICONIC = 2
OLEMISC_INSERTNOTREPLACE = 4
OLEMISC_STATIC = 8
OLEMISC_CANTLINKINSIDE = 16
OLEMISC_CANLINKBYOLE1 = 32
OLEMISC_ISLINKOBJECT = 64
OLEMISC_INSIDEOUT = 128
OLEMISC_ACTIVATEWHENVISIBLE = 256
OLEMISC_RENDERINGISDEVICEINDEPENDENT = 512
OLEMISC_INVISIBLEATRUNTIME = 1024
OLEMISC_ALWAYSRUN = 2048
OLEMISC_ACTSLIKEBUTTON = 4096
OLEMISC_ACTSLIKELABEL = 8192
OLEMISC_NOUIACTIVATE = 16384
OLEMISC_ALIGNABLE = 32768
OLEMISC_SIMPLEFRAME = 65536
OLEMISC_SETCLIENTSITEFIRST = 131072
OLEMISC_IMEMODE = 262144
OLEMISC_IGNOREACTIVATEWHENVISIBLE = 524288
OLEMISC_WANTSTOMENUMERGE = 1048576
OLEMISC_SUPPORTSMULTILEVELUNDO = 2097152
COLOR_INFOTEXT = 23 # Variable c_int
LVS_NOCOLUMNHEADER = 16384 # Variable c_int
URLACTION_INFODELIVERY_MAX = 7679 # Variable c_int
ERROR_NONE_MAPPED = 1332 # Variable c_long
NRC_CANOCCR = 36 # Variable c_int
IPPROTO_UDP = 17 # Variable c_int
BS_CENTER = 768 # Variable c_long
URLACTION_HTML_SUBMIT_FORMS_TO = 5635 # Variable c_int
OBJ_BRUSH = 2 # Variable c_int
SUCCESSFUL_ACCESS_ACE_FLAG = 64 # Variable c_int
CBES_EX_NOSIZELIMIT = 8 # Variable c_int
SO_UPDATE_ACCEPT_CONTEXT = 28683 # Variable c_int
TBCDRF_HILITEHOTTRACK = 131072 # Variable c_int
RPC_E_SYS_CALL_FAILED = -2147417856 # Variable c_long
DLL_PROCESS_ATTACH = 1 # Variable c_int
FILE_SUPPORTS_OBJECT_IDS = 65536 # Variable c_int
WIZ_CXDLG = 276 # Variable c_int
ERROR_DIR_NOT_EMPTY = 145 # Variable c_long
LBS_DISABLENOSCROLL = 4096 # Variable c_long
NCBNAMSZ = 16 # Variable c_int
ERROR_BAD_QUERY_SYNTAX = 1615 # Variable c_long
MMIO_FHOPEN = 16 # Variable c_int
ILC_COLOR24 = 24 # Variable c_int
SPLDS_PRINTER_CLASS = 'printQueue' # Variable POINTER(c_wchar)
PSH_RTLREADING = 2048 # Variable c_int
ERROR_CONNECTION_INVALID = 1229 # Variable c_long
ERROR_SERVICE_DISABLED = 1058 # Variable c_long
MCS_NOTODAYCIRCLE = 8 # Variable c_int
NO_ERROR = 0 # Variable c_long
IMAGE_FILE_MACHINE_WCEMIPSV2 = 361 # Variable c_int
SPI_SETACCESSTIMEOUT = 61 # Variable c_int
DRAGDROP_E_FIRST = 2147746048 # Variable c_ulong
WSAEPROTOTYPE = 10041 # Variable c_int
IMAGE_REL_BASED_HIGHLOW = 3 # Variable c_int
CBN_CLOSEUP = 8 # Variable c_int
_BLANK = 64 # Variable c_int
CO_E_FAILEDTOCLOSEHANDLE = -2147220971 # Variable c_long
HSHELL_REDRAW = 6 # Variable c_int
SB_GETTIPTEXTA = 1042 # Variable c_int
stc13 = 1100 # Variable c_int
CFSTR_MIME_RAWDATA = 'application/octet-stream' # Variable POINTER(c_wchar)
LVIS_CUT = 4 # Variable c_int
RBBS_FIXEDSIZE = 2 # Variable c_int
WS_EX_DLGMODALFRAME = 1 # Variable c_long
CE_RXPARITY = 4 # Variable c_int
SELECT_CAP_CONVERSION = 1 # Variable c_int
DLL_THREAD_ATTACH = 2 # Variable c_int
WHDR_DONE = 1 # Variable c_int
MDMSPKR_CALLSETUP = 3 # Variable c_int
PSPCB_CREATE = 2 # Variable c_int
LPD_SUPPORT_GDI = 16 # Variable c_int
TAPE_DRIVE_ECC = 65536 # Variable c_int
DOF_SHELLDATA = 2 # Variable c_int
KEY_WRITE = 131078 # Variable c_long
HDM_INSERTITEMA = 4609 # Variable c_int
SND_ALIAS_SYSTEMSTART = 21331 # Variable c_ulong
RPC_S_INTERNAL_ERROR = 1766 # Variable c_long
PORT_STATUS_OUTPUT_BIN_FULL = 4 # Variable c_int
MCI_SEQ_FILE = 16386 # Variable c_int
BN_HILITE = 2 # Variable c_int
CRYPT_E_CONTROL_TYPE = -2146889716 # Variable c_long
MIXERCONTROL_CT_SC_LIST_SINGLE = 0 # Variable c_long
RPC_E_INVALID_CALLDATA = -2147417844 # Variable c_long
GETVECTORPENSIZE = 26 # Variable c_int
PRINTER_ENUM_ICON8 = 8388608 # Variable c_int
CTRY_UKRAINE = 380 # Variable c_int
SCS_WOW_BINARY = 2 # Variable c_int
SP_OUTOFDISK = -4 # Variable c_int
WSAEPROTONOSUPPORT = 10043 # Variable c_int
MCI_NOTIFY_SUPERSEDED = 2 # Variable c_int
PENDINGTYPE_NESTED = 2
SO_DEBUG = 1 # Variable c_int
LANG_SWEDISH = 29 # Variable c_int
OF_VERIFY = 1024 # Variable c_int
BS_AUTO3STATE = 6 # Variable c_long
stc23 = 1110 # Variable c_int
LBS_NOSEL = 16384 # Variable c_long
BST_UNCHECKED = 0 # Variable c_int
CO_E_CLASSSTRING = -2147221005 # Variable c_long
IMAGE_REL_MIPS_REFHI = 4 # Variable c_int
EMR_POLYPOLYLINE = 7 # Variable c_int
DCB_DISABLE = 8 # Variable c_int
LCMAP_UPPERCASE = 512 # Variable c_int
WNNC_NET_BWNFS = 1048576 # Variable c_int
JOB_OBJECT_SET_ATTRIBUTES = 2 # Variable c_int
VOS_DOS_WINDOWS32 = 65540 # Variable c_long
SW_OTHERZOOM = 2 # Variable c_int
PIDSI_TEMPLATE = 7 # Variable c_long
VS_FFI_STRUCVERSION = 65536 # Variable c_long
TBN_GETBUTTONINFOW = 4294966576 # Variable c_uint
CLSCTX_ALL = 7 # Variable c_int
CTRY_SINGAPORE = 65 # Variable c_int
CAL_SABBREVMONTHNAME9 = 42 # Variable c_int
DISP_CHANGE_FAILED = -1 # Variable c_int
CF_METAFILEPICT = 3 # Variable c_int
SO_OOBINLINE = 256 # Variable c_int
ERROR_NOACCESS = 998 # Variable c_long
S_SERDVNA = -1 # Variable c_int
CS_BYTEALIGNWINDOW = 8192 # Variable c_int
SBS_SIZEGRIP = 16 # Variable c_long
TPM_RIGHTBUTTON = 2 # Variable c_long
ENABLE_WRAP_AT_EOL_OUTPUT = 2 # Variable c_int
CBS_DROPDOWNLIST = 3 # Variable c_long
CAL_SDAYNAME1 = 7 # Variable c_int
ERROR_BAD_VALIDATION_CLASS = 1348 # Variable c_long
DMORIENT_PORTRAIT = 1 # Variable c_int
ERROR_INVALID_DATA = 13 # Variable c_long
WGL_SWAP_UNDERLAY9 = 16777216 # Variable c_int
VIF_CANNOTREADSRC = 65536 # Variable c_long
MWMO_INPUTAVAILABLE = 4 # Variable c_int
EXCEPTION_MAXIMUM_PARAMETERS = 15 # Variable c_int
SPLDS_DRIVER_NAME = 'driverName' # Variable POINTER(c_wchar)
REG_BINARY = 3 # Variable c_int
MB_APPLMODAL = 0 # Variable c_long
VIEW_LIST = 2 # Variable c_int
LOCALE_SABBREVDAYNAME5 = 53 # Variable c_int
CBEMAXSTRLEN = 260 # Variable c_int
TAPE_DRIVE_REPORT_SMKS = 524288 # Variable c_int
LVNI_FOCUSED = 1 # Variable c_int
SERKF_INDICATOR = 4 # Variable c_int
PORT_STATUS_TONER_LOW = 10 # Variable c_int
ERROR_TOO_MANY_CMDS = 56 # Variable c_long
SW_ERASE = 4 # Variable c_int
SEE_MASK_CONNECTNETDRV = 128 # Variable c_int
VS_FFI_SIGNATURE = 4277077181 # Variable c_ulong
XTYPF_ACKREQ = 8 # Variable c_int
CO_E_INIT_MEMORY_ALLOCATOR = -2147467256 # Variable c_long
ODDPARITY = 1 # Variable c_int
GCPCLASS_LATINNUMERICTERMINATOR = 6 # Variable c_int
WGL_FONT_POLYGONS = 1 # Variable c_int
MCI_NOTIFY_FAILURE = 8 # Variable c_int
FS_JISJAPAN = 131072 # Variable c_long
TOKEN_ADJUST_SESSIONID = 256 # Variable c_int
TB_ISBUTTONHIGHLIGHTED = 1038 # Variable c_int
ERROR_REQUEST_REFUSED = 4320 # Variable c_long
JOB_NOTIFY_FIELD_PRINTER_NAME = 0 # Variable c_int
CO_E_LOOKUPACCSIDFAILED = -2147220979 # Variable c_long
SECURITY_CREATOR_GROUP_RID = 1 # Variable c_long
PGM_FORWARDMOUSE = 5123 # Variable c_int
VARFLAG_FIMMEDIATEBIND = 4096
DCX_WINDOW = 1 # Variable c_long
TAPE_DRIVE_SET_COMPRESSION = 2147484160 # Variable c_uint
SUBLANG_NORWEGIAN_NYNORSK = 2 # Variable c_int
OF_SHARE_DENY_WRITE = 32 # Variable c_int
MCIERR_SEQ_PORT_INUSE = 337 # Variable c_int
PM_NOYIELD = 2 # Variable c_int
PRINTER_CHANGE_PORT = 7340032 # Variable c_int
PSWIZB_NEXT = 2 # Variable c_int
RBHT_GRABBER = 4 # Variable c_int
TCP_NODELAY = 1 # Variable c_int
MK_CONTROL = 8 # Variable c_int
ARW_TOPLEFT = 2 # Variable c_long
ERROR_BAD_UNIT = 20 # Variable c_long
SERVICE_ACCEPT_PAUSE_CONTINUE = 2 # Variable c_int
IP_DEFAULT_MULTICAST_LOOP = 1 # Variable c_int
SKF_AVAILABLE = 2 # Variable c_int
ARW_LEFT = 0 # Variable c_long
LVS_SHAREIMAGELISTS = 64 # Variable c_int
SEE_MASK_HMONITOR = 2097152 # Variable c_int
STGTY_REPEAT = 256 # Variable c_long
TKIND_ENUM = 0
STGM_TRANSACTED = 65536 # Variable c_long
_REPORT_ERRMODE = 3 # Variable c_int
LB_SETCURSEL = 390 # Variable c_int
SO_RCVLOWAT = 4100 # Variable c_int
RGB_GAMMA_MAX = 65000 # Variable c_ushort
IMAGE_FILE_MACHINE_IA64 = 512 # Variable c_int
ERROR_NO_IMPERSONATION_TOKEN = 1309 # Variable c_long
NI_COMPOSITIONSTR = 21 # Variable c_int
GETTRACKKERNTABLE = 259 # Variable c_int
PROCESSOR_PPC_604 = 604 # Variable c_int
BINDVERB_GET = 0
RB_SHOWBAND = 1059 # Variable c_int
RPC_S_NO_BINDINGS = 1718 # Variable c_long
BS_TEXT = 0 # Variable c_long
BS_COMMANDLINK = 14
BSM_NETDRIVER = 2 # Variable c_int
ERROR_NO_SIGNAL_SENT = 205 # Variable c_long
LVBKIF_SOURCE_MASK = 3 # Variable c_int
MCIERR_SEQ_PORTUNSPECIFIED = 342 # Variable c_int
NM_RDBLCLK = 4294967290 # Variable c_uint
CTLCOLOR_STATIC = 6 # Variable c_int
MCI_UNFREEZE = 2117 # Variable c_int
ERROR_LM_CROSS_ENCRYPTION_REQUIRED = 1390 # Variable c_long
MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY = 4105 # Variable c_long
MKSYS_CLASSMONIKER = 7
chx14 = 1053 # Variable c_int
PSH_WIZARD = 32 # Variable c_int
CO_E_INIT_ONLY_SINGLE_THREADED = -2147467246 # Variable c_long
BS_PATTERN = 3 # Variable c_int
ODT_STATIC = 5 # Variable c_int
HDM_INSERTITEMW = 4618 # Variable c_int
IMAGE_BITMAP = 0 # Variable c_int
LANG_FINNISH = 11 # Variable c_int
GETSETSCREENPARAMS = 3072 # Variable c_int
DMPAPER_B5_EXTRA = 65 # Variable c_int
MK_E_ENUMERATION_FAILED = -2147221009 # Variable c_long
ERROR_DS_UNAVAILABLE = 1928 # Variable c_long
BITSPIXEL = 12 # Variable c_int
GL_ID_NOCONVERT = 32 # Variable c_int
PERF_DETAIL_NOVICE = 100 # Variable c_int
ERROR_SERVICE_NEVER_STARTED = 1077 # Variable c_long
PAN_MIDLINE_HIGH_POINTED = 6 # Variable c_int
DUPLICATE = 6 # Variable c_int
PBT_APMQUERYSUSPENDFAILED = 2 # Variable c_int
ERROR_IS_SUBST_PATH = 146 # Variable c_long
MSG_MAXIOVLEN = 16 # Variable c_int
WC_HEADERW = 'SysHeader32' # Variable POINTER(c_wchar)
NRC_LOCTFUL = 17 # Variable c_int
GCS_COMPREADCLAUSE = 4 # Variable c_int
MIXERCONTROL_CT_SC_LIST_MULTIPLE = 16777216 # Variable c_long
WRITE_OWNER = 524288 # Variable c_long
TBN_QUERYINSERT = 4294966590 # Variable c_uint
IMAGE_MIX_ISA_LIMIT = 2 # Variable c_int
CDN_SHAREVIOLATION = 4294966692 # Variable c_uint
CDRF_NOTIFYPOSTPAINT = 16 # Variable c_int
TB_GETBITMAPFLAGS = 1065 # Variable c_int
LB_SELITEMRANGEEX = 387 # Variable c_int
STGM_NOSNAPSHOT = 2097152 # Variable c_long
LVM_SETITEMSTATE = 4139 # Variable c_int
META_CREATEPALETTE = 247 # Variable c_int
LOCALE_SMONTHNAME5 = 60 # Variable c_int
SWP_ASYNCWINDOWPOS = 16384 # Variable c_int
VFF_BUFFTOOSMALL = 4 # Variable c_int
BAUD_14400 = 4096 # Variable c_ulong
RPC_BUFFER_NONOTIFY = 65536 # Variable c_int
URLACTION_SCRIPT_JAVA_USE = 5122 # Variable c_int
LB_GETHORIZONTALEXTENT = 403 # Variable c_int
PRINTER_CHANGE_ADD_PRINTER_DRIVER = 268435456 # Variable c_int
ERROR_HOST_UNREACHABLE = 1232 # Variable c_long
LANG_MANIPURI = 88 # Variable c_int
LVM_GETHOTCURSOR = 4159 # Variable c_int
BINDSTATUS_CLASSIDAVAILABLE = 12
C1_UPPER = 1 # Variable c_int
chx9 = 1048 # Variable c_int
GCL_HCURSOR = -12 # Variable c_int
LVS_EX_SUBITEMIMAGES = 2 # Variable c_int
DEBUG_PROCESS = 1 # Variable c_int
PST_TCPIP_TELNET = 258 # Variable c_ulong
RESOURCETYPE_DISK = 1 # Variable c_int
OFN_ENABLESIZING = 8388608 # Variable c_int
TC_SA_DOUBLE = 64 # Variable c_int
MCHT_TITLEBTNPREV = 33619971 # Variable c_int
NCBDGSEND = 32 # Variable c_int
AF_DECnet = 12 # Variable c_int
edt10 = 1161 # Variable c_int
MB_RIGHT = 524288 # Variable c_long
SSWF_TITLE = 1 # Variable c_int
RPC_C_QOS_IDENTITY_DYNAMIC = 1 # Variable c_int
GCP_REORDER = 2 # Variable c_int
ERROR_SERVICE_MARKED_FOR_DELETE = 1072 # Variable c_long
SC_RESTORE = 61728 # Variable c_int
URLACTION_SHELL_WEBVIEW_VERB = 6149 # Variable c_int
TBM_GETTIC = 1027 # Variable c_int
PIDSI_LASTAUTHOR = 8 # Variable c_long
DCX_NORESETATTRS = 4 # Variable c_long
DTN_DATETIMECHANGE = 4294966537 # Variable c_uint
CO_E_FAILEDTOGETWINDIR = -2147220975 # Variable c_long
VT_UINT = 23
MCM_GETTODAY = 4109 # Variable c_int
SPLDS_PRINT_NUMBER_UP = 'printNumberUp' # Variable POINTER(c_wchar)
IE_BYTESIZE = -11 # Variable c_int
ERROR_DESTINATION_ELEMENT_FULL = 1161 # Variable c_long
EVENTLOG_ERROR_TYPE = 1 # Variable c_int
IS_TEXT_UNICODE_DBCS_LEADBYTE = 1024 # Variable c_int
NUMPRS_LEADING_PLUS = 4 # Variable c_int
MB_ICONQUESTION = 32 # Variable c_long
SC_ACTION_RUN_COMMAND = 3
HDN_ITEMCLICKW = 4294966974 # Variable c_uint
MDM_SPEED_ADJUST = 128 # Variable c_int
EPT_S_NOT_REGISTERED = 1753 # Variable c_long
GMEM_MODIFY = 128 # Variable c_int
LOCALE_ITLZERO = 37 # Variable c_int
TB_LINEDOWN = 1 # Variable c_int
SECURITY_DESCRIPTOR_MIN_LENGTH = 20 # Variable c_uint
TBTS_RIGHT = 3 # Variable c_int
MIXERLINE_LINEF_DISCONNECTED = 32768 # Variable c_long
OFN_ENABLEINCLUDENOTIFY = 4194304 # Variable c_int
SE_GROUP_MANDATORY = 1 # Variable c_long
MCI_INFO = 2058 # Variable c_int
ERROR_IOPL_NOT_ENABLED = 197 # Variable c_long
JOB_OBJECT_LIMIT_PRESERVE_JOB_TIME = 64 # Variable c_int
NRC_BUFLEN = 1 # Variable c_int
JOY_RETURNPOV = 64 # Variable c_long
MB_YESNOCANCEL = 3 # Variable c_long
LVM_SETITEMW = 4172 # Variable c_int
TBM_SETSEL = 1034 # Variable c_int
SPLDS_PRINT_MEMORY = 'printMemory' # Variable POINTER(c_wchar)

__MIDL_IBindStatusCallback_0001 = c_int # enum
BINDVERB_POST = 1
BINDVERB_PUT = 2
BINDVERB_CUSTOM = 3
CO_E_WRONGTRUSTEENAMESYNTAX = -2147220983 # Variable c_long
SUBLANG_PORTUGUESE = 2 # Variable c_int
BACKGROUND_BLUE = 16 # Variable c_int
HDN_ITEMCHANGINGA = 4294966996 # Variable c_uint
LIBFLAG_FHASDISKIMAGE = 8
SC_VSCROLL = 61552 # Variable c_int
RDW_ALLCHILDREN = 128 # Variable c_int
JOY_RETURNZ = 4 # Variable c_long
ERROR_INVALID_MEDIA = 4300 # Variable c_long
R2_NOP = 11 # Variable c_int
LANG_ARMENIAN = 43 # Variable c_int
IMAGE_DIRECTORY_ENTRY_DEBUG = 6 # Variable c_int
TBM_CLEARSEL = 1043 # Variable c_int
DOMAIN_ALIAS_RID_ADMINS = 544 # Variable c_long
PC_POLYPOLYGON = 256 # Variable c_int
ERROR_BEGINNING_OF_MEDIA = 1102 # Variable c_long
BEGIN_PATH = 4096 # Variable c_int
SidTypeDeletedAccount = 6
chx8 = 1047 # Variable c_int
MAX_MONITORS = 4 # Variable c_int
RPC_C_AUTHN_MQ = 100 # Variable c_int
CFSTR_MIME_FRACTALS = 'application/fractals' # Variable POINTER(c_wchar)
WAVE_FORMAT_2M08 = 16 # Variable c_int
PAN_LETT_OBLIQUE_FLATTENED = 12 # Variable c_int
MCI_MODE_OPEN = 530 # Variable c_int
DCOM_CALL_CANCELED = 2
IMAGE_SYM_CLASS_TYPE_DEFINITION = 13 # Variable c_int
MSGF_USER = 4096 # Variable c_int
EM_SETREADONLY = 207 # Variable c_int
NIF_MESSAGE = 1 # Variable c_int
GMEM_NOCOMPACT = 16 # Variable c_int
HELP_HELPONHELP = 4 # Variable c_long
WAVE_ALLOWSYNC = 2 # Variable c_int
ADVFCACHE_ONSAVE = 32
BN_DOUBLECLICKED = 5 # Variable c_int
ODA_FOCUS = 4 # Variable c_int
MCI_WAVE_SET_BITSPERSAMPLE = 2097152 # Variable c_long
EMR_ELLIPSE = 42 # Variable c_int
SBS_SIZEBOXTOPLEFTALIGN = 2 # Variable c_long
INET_E_UNKNOWN_PROTOCOL = -2146697203 # Variable c_long
BDR_RAISEDINNER = 4 # Variable c_int
EPT_S_INVALID_ENTRY = 1751 # Variable c_long
RPC_C_MQ_JOURNAL_DEADLETTER = 1 # Variable c_int
PFD_SWAP_EXCHANGE = 512 # Variable c_int
KEY_NOTIFY = 16 # Variable c_int
ERROR_INVALID_SEGMENT_NUMBER = 180 # Variable c_long
MCIERR_NO_CLOSING_QUOTE = 294 # Variable c_int
SPAPI_E_DEVINFO_DATA_LOCKED = -2146500077 # Variable c_long
CBENF_DROPDOWN = 4 # Variable c_int
VOS_DOS_WINDOWS16 = 65537 # Variable c_long
FOF_NOERRORUI = 1024 # Variable c_int
FR_DOWN = 1 # Variable c_int
CFSTR_MIME_BASICAUDIO = 'audio/basic' # Variable POINTER(c_wchar)
GCS_DELTASTART = 256 # Variable c_int
LVFI_PARAM = 1 # Variable c_int
GGO_GRAY8_BITMAP = 6 # Variable c_int
EDGE_SUNKEN = 10 # Variable c_int
MMIOM_RENAME = 6 # Variable c_int
EMR_ENDPATH = 60 # Variable c_int
MF_DEFAULT = 4096 # Variable c_long
NRC_PENDING = 255 # Variable c_int
FILE_CASE_SENSITIVE_SEARCH = 1 # Variable c_int
VOS_OS232_PM32 = 196611 # Variable c_long
DMPAPER_QUARTO = 15 # Variable c_int
RB_DELETEBAND = 1026 # Variable c_int
DI_APPBANDING = 1 # Variable c_int
IPPORT_ROUTESERVER = 520 # Variable c_int
DFCS_CAPTIONRESTORE = 3 # Variable c_int
STREAM_MODIFIED_WHEN_READ = 1 # Variable c_int
TYPE_E_LIBNOTREGISTERED = -2147319779 # Variable c_long
MM_JOY2BUTTONUP = 952 # Variable c_int
EMR_SETWINDOWORGEX = 10 # Variable c_int
EM_GETMODIFY = 184 # Variable c_int
LVM_GETITEMRECT = 4110 # Variable c_int
RBN_LAYOUTCHANGED = 4294966463 # Variable c_uint
CTRY_SLOVAK = 421 # Variable c_int
IMAGE_REL_SH3_PCREL12_WORD = 11 # Variable c_int
TVM_EDITLABELW = 4417 # Variable c_int
ERROR_MAGAZINE_NOT_PRESENT = 1163 # Variable c_long
CAL_SDAYNAME5 = 11 # Variable c_int
DRV_OPEN = 3 # Variable c_int
FO_MOVE = 1 # Variable c_int
MCI_GETDEVCAPS_DEVICE_TYPE = 4 # Variable c_long
SPI_SETICONMETRICS = 46 # Variable c_int
SPLREG_SCHEDULER_THREAD_PRIORITY = 'SchedulerThreadPriority' # Variable POINTER(c_wchar)
ERROR_IRQ_BUSY = 1119 # Variable c_long
RPC_C_AUTHN_DEFAULT = 4294967295 # Variable c_ulong
DT_INTERNAL = 4096 # Variable c_int
PGF_SCROLLDOWN = 2 # Variable c_int
ERROR_HOOK_NEEDS_HMOD = 1428 # Variable c_long
SHGFI_LINKOVERLAY = 32768 # Variable c_int
DTS_SHOWNONE = 2 # Variable c_int
LVN_INSERTITEM = 4294967194 # Variable c_uint
MIXERLINE_TARGETTYPE_WAVEIN = 2 # Variable c_int
HIST_FORWARD = 1 # Variable c_int
DSS_RIGHT = 32768 # Variable c_int
STATUSCLASSNAMEW = 'msctls_statusbar32' # Variable POINTER(c_wchar)
ERROR_MOD_NOT_FOUND = 126 # Variable c_long
GCL_HICONSM = -34 # Variable c_int
QS_MOUSEMOVE = 2 # Variable c_int
CB_SETEDITSEL = 322 # Variable c_int
META_SCALEWINDOWEXT = 1040 # Variable c_int
PROCESSOR_HITACHI_SH3 = 10003 # Variable c_int
DT_TOP = 0 # Variable c_int
WSB_PROP_VSTYLE = 256 # Variable c_long
DMPAPER_A4_EXTRA = 53 # Variable c_int
EMR_ROUNDRECT = 44 # Variable c_int
ERROR_SPOOL_FILE_NOT_FOUND = 3002 # Variable c_long
ERROR_NOT_LOCKED = 158 # Variable c_long
RBBIM_IDEALSIZE = 512 # Variable c_int
TOKEN_QUERY = 8 # Variable c_int
PAN_WEIGHT_HEAVY = 9 # Variable c_int
SB_LINEDOWN = 1 # Variable c_int
TOKEN_DUPLICATE = 2 # Variable c_int
CRYPT_E_NO_PROVIDER = -2146885626 # Variable c_long
MCIERR_BASE = 256 # Variable c_int
MCIERR_SET_DRIVE = 309 # Variable c_int
OLE_S_USEREG = 262144 # Variable c_long
DCX_EXCLUDEUPDATE = 256 # Variable c_long
IMAGE_NT_OPTIONAL_HDR64_MAGIC = 523 # Variable c_int
TTF_IDISHWND = 1 # Variable c_int
IMAGE_FILE_AGGRESIVE_WS_TRIM = 16 # Variable c_int
PFD_DRAW_TO_BITMAP = 8 # Variable c_int
MIDIPATCHSIZE = 128 # Variable c_int
stc24 = 1111 # Variable c_int
EM_GETPASSWORDCHAR = 210 # Variable c_int
TB_BOTTOM = 7 # Variable c_int
TB_GETBUTTONINFOW = 1087 # Variable c_int
GMEM_NODISCARD = 32 # Variable c_int
VFT2_DRV_KEYBOARD = 2 # Variable c_long
JOYERR_UNPLUGGED = 167 # Variable c_int
SECURITY_DESCRIPTOR_REVISION = 1 # Variable c_int
S_SERDTP = -8 # Variable c_int
PERF_DETAIL_ADVANCED = 200 # Variable c_int
WNNC_NET_FJ_REDIR = 2228224 # Variable c_int
PSNRET_INVALID_NOCHANGEPAGE = 2 # Variable c_int
AF_UNKNOWN1 = 20 # Variable c_int
ERROR_PROC_NOT_FOUND = 127 # Variable c_long
IPPROTO_ND = 77 # Variable c_int
CAL_SABBREVMONTHNAME8 = 41 # Variable c_int
MMIO_FINDCHUNK = 16 # Variable c_int
IMAGE_REL_SH3_DIRECT4 = 6 # Variable c_int
EMARCH_ENC_I17_IMM41b_SIZE_X = 8 # Variable c_int
TYPE_E_INVDATAREAD = -2147319784 # Variable c_long
ENABLEPAIRKERNING = 769 # Variable c_int
CF_PRINTERFONTS = 2 # Variable c_int
ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED = 1925 # Variable c_long
MM_TWIPS = 6 # Variable c_int
NRC_SNUMOUT = 8 # Variable c_int
CONVERT10_E_OLESTREAM_FMT = -2147221054 # Variable c_long
PROPSETFLAG_UNBUFFERED = 4 # Variable c_int
SE_ERR_DDETIMEOUT = 28 # Variable c_int
HIGH_PRIORITY_CLASS = 128 # Variable c_int
SPI_GETICONTITLEWRAP = 25 # Variable c_int
PSINJECT_PAGEBBOX = 26 # Variable c_int
SB_SETTIPTEXTA = 1040 # Variable c_int
SND_NODEFAULT = 2 # Variable c_int
NRC_SYSTEM = 64 # Variable c_int
WSAETIMEDOUT = 10060 # Variable c_int
RPC_E_NOT_REGISTERED = -2147417853 # Variable c_long
I_INDENTCALLBACK = -1 # Variable c_int
ERROR_UNRECOGNIZED_VOLUME = 1005 # Variable c_long
CF_TTONLY = 262144 # Variable c_long
ERROR_BROKEN_PIPE = 109 # Variable c_long
PAN_SERIF_EXAGGERATED = 9 # Variable c_int
MF_HELP = 16384 # Variable c_long
BF_DIAGONAL_ENDBOTTOMRIGHT = 28 # Variable c_int
CAL_SMONTHNAME11 = 31 # Variable c_int
MAXWORD = 65535 # Variable c_int
TB_ISBUTTONINDETERMINATE = 1037 # Variable c_int
DS_SETFONT = 64 # Variable c_long
XTYP_REQUEST = 8368 # Variable c_int
IO_REPARSE_TAG_SIS = 2147483655 # Variable c_uint
MSHCTX_DIFFERENTMACHINE = 2
FS_CHINESESIMP = 262144 # Variable c_long
MWT_IDENTITY = 1 # Variable c_int
SUBLANG_SPANISH_PUERTO_RICO = 20 # Variable c_int
PAN_WEIGHT_MEDIUM = 6 # Variable c_int
MF_ERRORS = 268435456 # Variable c_int
R2_MASKNOTPEN = 3 # Variable c_int
LOCALE_SDAYNAME7 = 48 # Variable c_int
rad2 = 1057 # Variable c_int

tagFUNCFLAGS = c_int # enum
FUNCFLAG_FRESTRICTED = 1
FUNCFLAG_FSOURCE = 2
FUNCFLAG_FBINDABLE = 4
FUNCFLAG_FREQUESTEDIT = 8
FUNCFLAG_FDISPLAYBIND = 16
FUNCFLAG_FDEFAULTBIND = 32
FUNCFLAG_FHIDDEN = 64
FUNCFLAG_FUSESGETLASTERROR = 128
FUNCFLAG_FDEFAULTCOLLELEM = 256
FUNCFLAG_FUIDEFAULT = 512
FUNCFLAG_FNONBROWSABLE = 1024
FUNCFLAG_FREPLACEABLE = 2048
FUNCFLAG_FIMMEDIATEBIND = 4096
JobObjectSecurityLimitInformation = 5
SUBLANG_ENGLISH_EIRE = 6 # Variable c_int
PAN_LETT_NORMAL_FLATTENED = 5 # Variable c_int
JOY_RETURNU = 16 # Variable c_long
LCMAP_BYTEREV = 2048 # Variable c_int
MINLONG = 2147483648 # Variable c_uint
PS_DASHDOTDOT = 4 # Variable c_int
CC_ELLIPSES = 8 # Variable c_int
WAVERR_STILLPLAYING = 33 # Variable c_int
TRUST_E_NOSIGNATURE = -2146762496 # Variable c_long
MOD_ALT = 1 # Variable c_int
DL_DRAGGING = 1158 # Variable c_int
NUMPRS_INEXACT = 131072 # Variable c_int
DESKTOP_ENUMERATE = 64 # Variable c_long
MCIERR_NO_WINDOW = 346 # Variable c_int
THREAD_ALL_ACCESS = 2032639 # Variable c_long
HDM_DELETEITEM = 4610 # Variable c_int
IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR = 14 # Variable c_int
RPC_S_INVALID_TAG = 1733 # Variable c_long
ERROR_HOOK_NOT_INSTALLED = 1431 # Variable c_long
EMR_SCALEVIEWPORTEXTEX = 31 # Variable c_int
ILD_OVERLAYMASK = 3840 # Variable c_int
PAN_PROP_EVEN_WIDTH = 4 # Variable c_int
TKIND_RECORD = 1
WNCON_FORNETCARD = 1 # Variable c_int
MF_BYCOMMAND = 0 # Variable c_long
ILC_MASK = 1 # Variable c_int
JOB_NOTIFY_FIELD_PRINT_PROCESSOR = 6 # Variable c_int
__REQUIRED_RPCNDR_H_VERSION__ = 440 # Variable c_int
MIXER_GETLINEINFOF_DESTINATION = 0 # Variable c_long
SUBLANG_ARABIC_QATAR = 16 # Variable c_int
TBIF_STYLE = 8 # Variable c_int
JOB_OBJECT_MSG_ACTIVE_PROCESS_LIMIT = 3 # Variable c_int
CDN_HELP = 4294966691 # Variable c_uint
LVN_MARQUEEBEGIN = 4294967140 # Variable c_uint
TYPE_E_OUTOFBOUNDS = -2147316575 # Variable c_long
GETSETPRINTORIENT = 30 # Variable c_int
MM_MIM_ERROR = 965 # Variable c_int
EMR_EXTSELECTCLIPRGN = 75 # Variable c_int
TCS_EX_REGISTERDROP = 2 # Variable c_int
MIXERCONTROL_CT_UNITS_SIGNED = 131072 # Variable c_long
SS_RIGHTJUST = 1024 # Variable c_long
MCI_ANIM_GETDEVCAPS_MAX_WINDOWS = 16392 # Variable c_long
PRF_OWNED = 32 # Variable c_long
SERVICE_PAUSE_CONTINUE = 64 # Variable c_int
PGM_GETBKCOLOR = 5125 # Variable c_int
CLIP_CHARACTER_PRECIS = 1 # Variable c_int
chx15 = 1054 # Variable c_int
SUBLANG_ARABIC_OMAN = 8 # Variable c_int
ICC_COOL_CLASSES = 1024 # Variable c_int
DESKTOP_CREATEMENU = 4 # Variable c_long
DM_PAPERSIZE = 2 # Variable c_long
BST_CHECKED = 1 # Variable c_int
EV_ERR = 128 # Variable c_int
TBSTYLE_AUTOSIZE = 16 # Variable c_int
FOF_NOCONFIRMMKDIR = 512 # Variable c_int
MF_MASK = 4278190080 # Variable c_uint
RPC_S_PROTOCOL_ERROR = 1728 # Variable c_long
VT_I8 = 20
WHITE_PEN = 6 # Variable c_int
NEWTRANSPARENT = 3 # Variable c_int
RPC_E_FAULT = -2147417852 # Variable c_long
META_ARC = 2071 # Variable c_int
SPECIFIC_RIGHTS_ALL = 65535 # Variable c_long
PS_COSMETIC = 0 # Variable c_int
BAUD_4800 = 512 # Variable c_ulong
DRVCNF_OK = 1 # Variable c_int
SM_SECURE = 44 # Variable c_int
SECURITY_AUTHENTICATED_USER_RID = 11 # Variable c_long
LMEM_VALID_FLAGS = 3954 # Variable c_int
PDEVICESIZE = 26 # Variable c_int
MMIO_GETTEMP = 131072 # Variable c_int
XTYP_ADVREQ = 8226 # Variable c_int
CFSTR_MIME_TEXT = 'text/plain' # Variable POINTER(c_wchar)
DC_DATATYPE_PRODUCED = 21 # Variable c_int
FILE_FLAG_POSIX_SEMANTICS = 16777216 # Variable c_int
DISP_E_UNKNOWNINTERFACE = -2147352575 # Variable c_long
ACL_REVISION = 2 # Variable c_int
GCP_DISPLAYZWG = 4194304 # Variable c_long
TVGN_NEXTVISIBLE = 6 # Variable c_int
TTTOOLINFOW_V1_SIZE = 40 # Variable c_uint
SUBLANG_NEPALI_INDIA = 2 # Variable c_int
PBM_GETPOS = 1032 # Variable c_int
GW_CHILD = 5 # Variable c_int
ODT_HEADER = 100 # Variable c_int
FILE_FLAG_SEQUENTIAL_SCAN = 134217728 # Variable c_int
HELP_CONTEXT = 1 # Variable c_long
CRYPT_E_INVALID_PRINTABLE_STRING = -2146885599 # Variable c_long
GL_ID_PRIVATE_LAST = 65535 # Variable c_int
SM_CYFULLSCREEN = 17 # Variable c_int
MCIERR_GET_CD = 307 # Variable c_int
RPC_S_FP_UNDERFLOW = 1770 # Variable c_long
IMAGE_REL_ALPHA_PAIR = 12 # Variable c_int
ACN_START = 1 # Variable c_int
SPAPI_E_REMOTE_COMM_FAILURE = -2146500063 # Variable c_long
DESKTOP_JOURNALRECORD = 16 # Variable c_long
MM_MOM_CLOSE = 968 # Variable c_int
EMARCH_ENC_I17_IMM41c_VAL_POS_X = 40 # Variable c_int
MCI_SEQ_SET_TEMPO = 65536 # Variable c_long
PGS_DRAGNDROP = 4 # Variable c_int
MIXERCONTROL_CT_CLASS_NUMBER = 805306368 # Variable c_long
MOVEFILE_COPY_ALLOWED = 2 # Variable c_int
IMAGE_SCN_ALIGN_256BYTES = 9437184 # Variable c_int
LCID_ALTERNATE_SORTS = 4 # Variable c_int
OLE_E_WRONGCOMPOBJ = -2147221490 # Variable c_long
DPD_DELETE_SPECIFIC_VERSION = 2 # Variable c_int
DV_E_NOIVIEWOBJECT = -2147221395 # Variable c_long
PROCESS_QUERY_INFORMATION = 1024 # Variable c_int
LVM_GETNUMBEROFWORKAREAS = 4169 # Variable c_int
SERVICE_CONTROL_STOP = 1 # Variable c_int
OLE_E_NOT_INPLACEACTIVE = -2147221488 # Variable c_long
MCI_MODE_PLAY = 526 # Variable c_int
RT_GROUP_CURSOR = 18 # Variable POINTER(c_wchar)
IMAGE_SYM_TYPE_INT = 4 # Variable c_int
GL_ID_TYPINGERROR = 33 # Variable c_int
ERROR_CONTINUE = 1246 # Variable c_long
INVALID_FILE_SIZE = 4294967295 # Variable c_ulong
OLEOBJ_S_FIRST = 262528 # Variable c_long
CFSTR_MIME_X_SGI_MOVIE = 'video/x-sgi-movie' # Variable POINTER(c_wchar)
TCM_INSERTITEMW = 4926 # Variable c_int
SPLDS_PRINT_DUPLEX_SUPPORTED = 'printDuplexSupported' # Variable POINTER(c_wchar)

tagBINDSTATUS = c_int # enum
BINDSTATUS_FINDINGRESOURCE = 1
BINDSTATUS_CONNECTING = 2
BINDSTATUS_REDIRECTING = 3
BINDSTATUS_BEGINDOWNLOADDATA = 4
BINDSTATUS_BEGINSYNCOPERATION = 15
BINDSTATUS_ENDSYNCOPERATION = 16
BINDSTATUS_UPLOADINGDATA = 18
BINDSTATUS_ENDUPLOADDATA = 19
BINDSTATUS_PROTOCOLCLASSID = 20
BINDSTATUS_ENCODING = 21
BINDSTATUS_DECODING = 24
BINDSTATUS_LOADINGMIMEHANDLER = 25
ERROR_DISK_FULL = 112 # Variable c_long
DFCS_SCROLLSIZEGRIP = 8 # Variable c_int
RDW_INTERNALPAINT = 2 # Variable c_int
WS_EX_LEFTSCROLLBAR = 16384 # Variable c_long
META_SETRELABS = 261 # Variable c_int
SW_SHOWMINNOACTIVE = 7 # Variable c_int
IME_ESC_RESERVED_LAST = 2047 # Variable c_int
CDRF_NEWFONT = 2 # Variable c_int
FRAME_TSS = 2 # Variable c_int
XTYP_XACT_COMPLETE = 32896 # Variable c_int
MCI_CLOSE = 2052 # Variable c_int
SC_MOVE = 61456 # Variable c_int
THREAD_BASE_PRIORITY_MAX = 2 # Variable c_int
MIXERCONTROL_CT_UNITS_CUSTOM = 0 # Variable c_long
SECURITY_ANONYMOUS = 0 # Variable c_int
PERF_DETAIL_WIZARD = 400 # Variable c_int
HEAP_FREE_CHECKING_ENABLED = 64 # Variable c_int
SIZE_RESTORED = 0 # Variable c_int
CSTR_LESS_THAN = 1 # Variable c_int
LANG_CHINESE = 4 # Variable c_int
ACTFLG_SystemWide = 4 # Variable c_int
SS_NOPREFIX = 128 # Variable c_long
RPC_S_NOT_CANCELLED = 1826 # Variable c_long
ERROR_BAD_PROFILE = 1206 # Variable c_long
IMAGE_SUBSYSTEM_UNKNOWN = 0 # Variable c_int
MCM_GETMINREQRECT = 4105 # Variable c_int
BDR_INNER = 12 # Variable c_int
ERROR_SERVICE_START_HANG = 1070 # Variable c_long
FORMAT_MESSAGE_FROM_HMODULE = 2048 # Variable c_int
MCI_OPEN_ELEMENT = 512 # Variable c_long
PERF_COUNTER_HISTOGRAM = 393216 # Variable c_int
WNNC_NET_PATHWORKS = 851968 # Variable c_int
SW_MAX = 11 # Variable c_int
VTBIT_I2 = 4 # Variable c_int
DRAFT_QUALITY = 1 # Variable c_int
CAL_SMONTHNAME6 = 26 # Variable c_int
IMAGE_SYM_CLASS_UNDEFINED_LABEL = 7 # Variable c_int
S_SEROFM = -2 # Variable c_int
CO_E_DECODEFAILED = -2147220966 # Variable c_long
SPI_SETPENWINDOWS = 49 # Variable c_int
RPC_C_VERS_ALL = 1 # Variable c_int
HKL_PREV = 0 # Variable c_int
MCI_FORMAT_SAMPLES = 9 # Variable c_int
APPCMD_FILTERINITS = 32 # Variable c_long
STG_E_BADBASEADDRESS = -2147286768 # Variable c_long
ERROR_SHUTDOWN_CLUSTER = 5008 # Variable c_long
RPC_S_INVALID_NAF_ID = 1763 # Variable c_long
CALLBACK_TASK = 131072 # Variable c_long
ERROR_OPERATION_ABORTED = 995 # Variable c_long
RPC_S_INVALID_NAME_SYNTAX = 1736 # Variable c_long
PCF_TOTALTIMEOUTS = 64 # Variable c_ulong
MCI_VD_PLAY_FAST = 131072 # Variable c_long
scr4 = 1171 # Variable c_int
IPPORT_ECHO = 7 # Variable c_int
SSGF_NONE = 0 # Variable c_int
stc14 = 1101 # Variable c_int
DM_PELSWIDTH = 524288 # Variable c_long
PRINTER_ERROR_OUTOFPAPER = 1 # Variable c_int
TVM_GETIMAGELIST = 4360 # Variable c_int
GCP_NUMERICOVERRIDE = 16777216 # Variable c_long
CTRY_BOLIVIA = 591 # Variable c_int
IS_TEXT_UNICODE_SIGNATURE = 8 # Variable c_int
RPC_E_INVALID_OBJREF = -2147417827 # Variable c_long
RBBS_GRIPPERALWAYS = 128 # Variable c_int
PERF_COUNTER_LARGE_DELTA = 4195584 # Variable c_int
PS_DASHDOT = 3 # Variable c_int
ABORTDOC = 2 # Variable c_int
STATFLAG_NOOPEN = 2
CTRY_SPAIN = 34 # Variable c_int
DFCS_MENUARROW = 0 # Variable c_int
SND_SYNC = 0 # Variable c_int
CO_E_ERRORINDLL = -2147220999 # Variable c_long
EPSPRINTING = 33 # Variable c_int
LOCALE_SMONTHNAME11 = 66 # Variable c_int
DT_CHARSTREAM = 4 # Variable c_int
ERROR_CANT_ACCESS_DOMAIN_INFO = 1351 # Variable c_long
WC_HEADERA = 'SysHeader32' # Variable POINTER(c_char)
IMAGE_SUBSYSTEM_NATIVE = 1 # Variable c_int
SM_CXMENUCHECK = 71 # Variable c_int
CONSOLE_TEXTMODE_BUFFER = 1 # Variable c_int
MCIERR_SET_CD = 308 # Variable c_int
ETO_NUMERICSLOCAL = 1024 # Variable c_int
LCMAP_SORTKEY = 1024 # Variable c_int
PGS_AUTOSCROLL = 2 # Variable c_int
DF_ALLOWOTHERACCOUNTHOOK = 1 # Variable c_long
PGF_CALCHEIGHT = 2 # Variable c_int
LB_GETTEXTLEN = 394 # Variable c_int
SETDTR = 5 # Variable c_int
ILD_NORMAL = 0 # Variable c_int
CAL_SMONTHNAME9 = 29 # Variable c_int
ERROR_CANNOT_OPEN_PROFILE = 1205 # Variable c_long
BAUD_1200 = 64 # Variable c_ulong
EVENPARITY = 2 # Variable c_int
MIXER_OBJECTF_HMIXER = 2147483648 # Variable c_ulong
HEAP_GROWABLE = 2 # Variable c_int
ODS_GRAYED = 2 # Variable c_int
EMR_EXTCREATEPEN = 95 # Variable c_int
VIF_SRCOLD = 4 # Variable c_long
LOCALE_IPOSSIGNPOSN = 82 # Variable c_int
TVM_GETVISIBLECOUNT = 4368 # Variable c_int
MIDI_CACHE_BESTFIT = 2 # Variable c_int
ERROR_READ_FAULT = 30 # Variable c_long
ERROR_INVALID_PARAMETER = 87 # Variable c_long
ERROR_NOT_READY = 21 # Variable c_long
RP_INIFILE = 2 # Variable c_int
QS_HOTKEY = 128 # Variable c_int
LVSCW_AUTOSIZE_USEHEADER = -2 # Variable c_int
LR_LOADFROMFILE = 16 # Variable c_int
FOF_NOCONFIRMATION = 16 # Variable c_int
CFSTR_MIME_WAV = 'audio/wav' # Variable POINTER(c_wchar)
PD_ENABLEPRINTTEMPLATE = 16384 # Variable c_int
ERROR_INVALID_DLL = 1154 # Variable c_long
LOGON32_LOGON_BATCH = 4 # Variable c_int
DTN_FORMATA = 4294966540 # Variable c_uint
DST_BITMAP = 4 # Variable c_int
FRAME_NONFPO = 3 # Variable c_int
WAVECAPS_DIRECTSOUND = 64 # Variable c_int
RPC_C_OPT_MAX_OPTIONS = 10 # Variable c_int
SERVICE_FILE_SYSTEM_DRIVER = 2 # Variable c_int
OFN_NOREADONLYRETURN = 32768 # Variable c_int
TC_RESERVED = 32768 # Variable c_int
META_RESTOREDC = 295 # Variable c_int
LPD_SHARE_ACCUM = 256 # Variable c_int
LVM_SETITEMPOSITION32 = 4145 # Variable c_int
SUBLANG_ARABIC_TUNISIA = 7 # Variable c_int
STUB_MARSHAL = 2
MCI_OVLY_GETDEVCAPS_CAN_FREEZE = 16386 # Variable c_long
stc25 = 1112 # Variable c_int
EMARCH_ENC_I17_IMM7B_SIZE_X = 7 # Variable c_int
PAGE_READONLY = 2 # Variable c_int
JOY_BUTTON2CHG = 512 # Variable c_int
MF_LINKS = 536870912 # Variable c_int
BS_NOTIFY = 16384 # Variable c_long
CTRY_TUNISIA = 216 # Variable c_int
DC_ENUMRESOLUTIONS = 13 # Variable c_int
PROCESS_SET_QUOTA = 256 # Variable c_int
ERROR_PARTIAL_COPY = 299 # Variable c_long
FOF_MULTIDESTFILES = 1 # Variable c_int
VIF_ACCESSVIOLATION = 512 # Variable c_long
DMPAPER_LEDGER = 4 # Variable c_int
MOD_WIN = 8 # Variable c_int
TVE_EXPANDPARTIAL = 16384 # Variable c_int
MCI_VD_MEDIA_CLV = 1026 # Variable c_int
SPLDS_SERVER_NAME = 'serverName' # Variable POINTER(c_wchar)
PROCESSOR_ARCHITECTURE_ARM = 5 # Variable c_int
MCIERR_DRIVER_INTERNAL = 272 # Variable c_int
SPLREG_PORT_THREAD_PRIORITY = 'PortThreadPriority' # Variable POINTER(c_wchar)
PSM_APPLY = 1134 # Variable c_int
MIXERCONTROL_CONTROLTYPE_TREBLE = 1342373891 # Variable c_long
CO_E_INIT_RPC_CHANNEL = -2147467254 # Variable c_long
IMAGE_FILE_MACHINE_SH4 = 422 # Variable c_int
LVHT_TOLEFT = 64 # Variable c_int
FROM_LEFT_3RD_BUTTON_PRESSED = 8 # Variable c_int
TVM_CREATEDRAGIMAGE = 4370 # Variable c_int
WHDR_INQUEUE = 16 # Variable c_int
SERVICE_TYPE_ALL = 319 # Variable c_int
LZERROR_WRITE = -4 # Variable c_int
KF_EXTENDED = 256 # Variable c_int
TTM_UPDATETIPTEXTW = 1081 # Variable c_int
MIXER_SETCONTROLDETAILSF_QUERYMASK = 15 # Variable c_long
BAUD_1800 = 128 # Variable c_ulong
SC_HOTKEY = 61776 # Variable c_int
MM_HIMETRIC = 3 # Variable c_int
CF_BITMAP = 2 # Variable c_int
SM_CXICON = 11 # Variable c_int
HKCOMB_C = 4 # Variable c_int
INET_E_RESOURCE_NOT_FOUND = -2146697211 # Variable c_long
TVC_BYMOUSE = 1 # Variable c_int
MKF_HOTKEYACTIVE = 4 # Variable c_int
MSHCTX_INPROC = 3
JOB_OBJECT_UILIMIT_DESKTOP = 64 # Variable c_int
TVN_ITEMEXPANDEDW = 4294966841 # Variable c_uint
TVM_SETITEMA = 4365 # Variable c_int
RBBIM_HEADERSIZE = 2048 # Variable c_int
MCIERR_PARSER_INTERNAL = 271 # Variable c_int
BINDSTRING_POST_DATA_MIME = 13
PRINTER_CHANGE_SET_PRINTER_DRIVER = 536870912 # Variable c_int
MIXERCONTROL_CT_CLASS_METER = 268435456 # Variable c_long
BF_LEFT = 1 # Variable c_int
PDERR_CREATEICFAILURE = 4106 # Variable c_int
MCHT_TITLE = 65536 # Variable c_int
IMAGE_SYM_TYPE_LONG = 5 # Variable c_int
IME_SMODE_PLAURALCLAUSE = 1 # Variable c_int
SS_NOTIFY = 256 # Variable c_long
IME_CMODE_NOCONVERSION = 256 # Variable c_int
UDS_NOTHOUSANDS = 128 # Variable c_int
NTE_BAD_HASH_STATE = -2146893812 # Variable c_long
IMAGE_REL_PPC_GPREL = 21 # Variable c_int
ILLUMINANT_D55 = 5 # Variable c_int
DESKTOP_JOURNALPLAYBACK = 32 # Variable c_long
TKIND_MODULE = 2
MCI_VD_GETDEVCAPS_FAST_RATE = 16387 # Variable c_long
NUMPRS_CURRENCY = 1024 # Variable c_int
PID_ILLEGAL = 4294967295 # Variable c_uint
RB_MAXIMIZEBAND = 1055 # Variable c_int
MIXERR_LASTERROR = 1026 # Variable c_int
TBM_GETRANGEMAX = 1026 # Variable c_int
NETINFO_DLL16 = 1 # Variable c_int
ICM_DELETEPROFILE = 2 # Variable c_int
TVM_GETTEXTCOLOR = 4384 # Variable c_int
PERF_MULTI_COUNTER = 33554432 # Variable c_int
RPC_C_AUTHN_LEVEL_PKT = 4 # Variable c_int
TVHT_BELOW = 512 # Variable c_int
MCI_VD_STATUS_MEDIA_TYPE = 16388 # Variable c_long
TA_RIGHT = 2 # Variable c_int
XTYP_MONITOR = 33010 # Variable c_int
ACTFLG_UserInstall = 32 # Variable c_int
WSAEREMOTE = 10071 # Variable c_int
XCLASS_BOOL = 4096 # Variable c_int
JOB_NOTIFY_FIELD_DRIVER_NAME = 8 # Variable c_int
ERROR_SUCCESS_REBOOT_REQUIRED = 3010 # Variable c_long
SYSPAL_ERROR = 0 # Variable c_int
DMCOLLATE_FALSE = 0 # Variable c_int
IMAGE_REL_ARM_SECREL = 15 # Variable c_int
SC_MOUSEMENU = 61584 # Variable c_int
JOB_NOTIFY_FIELD_PARAMETERS = 7 # Variable c_int
CONTEXT_DEBUG_REGISTERS = 65552 # Variable c_long
TAPE_DRIVE_TAPE_CAPACITY = 256 # Variable c_int
VT_HRESULT = 25
RPC_X_INVALID_ES_ACTION = 1827 # Variable c_long
PERF_COUNTER_BASE = 196608 # Variable c_int
COLOR_WINDOWTEXT = 8 # Variable c_int
HDI_BITMAP = 16 # Variable c_int
IMAGE_SYM_TYPE_DOUBLE = 7 # Variable c_int
CONVERT10_E_STG_DIB_TO_BITMAP = -2147221050 # Variable c_long
PORT_TYPE_NET_ATTACHED = 8 # Variable c_int
ERROR_TRUSTED_DOMAIN_FAILURE = 1788 # Variable c_long
MOUSEEVENTF_MOVE = 1 # Variable c_int
WGL_SWAP_OVERLAY1 = 2 # Variable c_int
CBEIF_TEXT = 1 # Variable c_int
WS_EX_WINDOWEDGE = 256 # Variable c_long
INET_E_CANNOT_CONNECT = -2146697212 # Variable c_long
MMSYSERR_NODRIVER = 6 # Variable c_int
ILCF_SWAP = 1 # Variable c_int
ERROR_METAFILE_NOT_SUPPORTED = 2003 # Variable c_long
SPLDS_PRINT_SEPARATOR_FILE = 'printSeparatorFile' # Variable POINTER(c_wchar)
psh4 = 1027 # Variable c_int
BST_PUSHED = 4 # Variable c_int
TVM_SETITEMW = 4415 # Variable c_int
NORM_IGNOREWIDTH = 131072 # Variable c_int
cmb3 = 1138 # Variable c_int
ICC_BAR_CLASSES = 4 # Variable c_int
LVM_EDITLABELA = 4119 # Variable c_int
ERROR_CLASS_DOES_NOT_EXIST = 1411 # Variable c_long
VS_FF_SPECIALBUILD = 32 # Variable c_long
SW_RESTORE = 9 # Variable c_int
VFT2_UNKNOWN = 0 # Variable c_long
RPC_S_BINDING_INCOMPLETE = 1819 # Variable c_long
SUBLANG_SPANISH_COSTA_RICA = 5 # Variable c_int
MSGF_COMMCTRL_TOOLBARCUST = 16899 # Variable c_int
CRYPT_E_BAD_ENCODE = -2146885630 # Variable c_long
C1_ALPHA = 256 # Variable c_int
_CONTROL = 32 # Variable c_int
URLPOLICY_JAVA_HIGH = 65536 # Variable c_int
CBEM_SETITEMW = 1036 # Variable c_int
TBM_CLEARTICS = 1033 # Variable c_int
FILE_APPEND_DATA = 4 # Variable c_int
SECTION_QUERY = 1 # Variable c_int
FR_NOUPDOWN = 1024 # Variable c_int
SUBLANG_SPANISH_PARAGUAY = 15 # Variable c_int
ERROR_NOT_ENOUGH_SERVER_MEMORY = 1130 # Variable c_long
IMAGE_REL_PPC_BRTAKEN = 512 # Variable c_int
TBM_SETSELEND = 1036 # Variable c_int
SHERB_NOSOUND = 4 # Variable c_int
MB_DEFAULT_DESKTOP_ONLY = 131072 # Variable c_long
PRINTER_ATTRIBUTE_SHARED = 8 # Variable c_int
HHT_TORIGHT = 1024 # Variable c_int
ENABLE_PROCESSED_INPUT = 1 # Variable c_int
DSTINVERT = 5570569 # Variable c_ulong
DMLERR_POKEACKTIMEOUT = 16395 # Variable c_int
PRINTER_CHANGE_ADD_JOB = 256 # Variable c_int
SS_BLACKFRAME = 7 # Variable c_long
DFCS_MENUCHECK = 1 # Variable c_int
TB_THUMBPOSITION = 4 # Variable c_int
MCI_STATUS_POSITION = 2 # Variable c_long
FONTDLGORD = 1542 # Variable c_int
JOY_BUTTON6 = 32 # Variable c_long
WIN_CERT_REVISION_2_0 = 512 # Variable c_int
PS_DASH = 1 # Variable c_int
TRUST_E_SUBJECT_FORM_UNKNOWN = -2146762749 # Variable c_long
SBT_OWNERDRAW = 4096 # Variable c_int
SORT_GERMAN_PHONE_BOOK = 1 # Variable c_int
DMPAPER_10X11 = 45 # Variable c_int
SE_DACL_PRESENT = 4 # Variable c_int
MCI_OPEN_TYPE = 8192 # Variable c_long
BF_MONO = 32768 # Variable c_int
TVM_HITTEST = 4369 # Variable c_int
ERROR_NO_NETWORK = 1222 # Variable c_long
SHERB_NOPROGRESSUI = 2 # Variable c_int
LANG_SLOVAK = 27 # Variable c_int
FS_THAI = 65536 # Variable c_long
SRVINFO_F_COSERVERINFO = 0
MMIOERR_BASE = 256 # Variable c_int
CBF_SKIP_UNREGISTRATIONS = 1048576 # Variable c_int
cmb5 = 1140 # Variable c_int
LANG_PUNJABI = 70 # Variable c_int
SECURITY_LOGON_IDS_RID = 5 # Variable c_long
ERROR_INVALID_LIBRARY = 4301 # Variable c_long
SO_DISCOPT = 28675 # Variable c_int
IMAGE_COMDAT_SELECT_LARGEST = 6 # Variable c_int
MCIERR_CANNOT_LOAD_DRIVER = 266 # Variable c_int
ERROR_BAD_PROVIDER = 1204 # Variable c_long
TBM_SETTHUMBLENGTH = 1051 # Variable c_int
SPI_SETDRAGFULLWINDOWS = 37 # Variable c_int
BS_FLAT = 32768 # Variable c_long
TBDDRET_TREATPRESSED = 2 # Variable c_int
XST_INIT1 = 3 # Variable c_int
WC_NATIVEFONTCTLA = 'NativeFontCtl' # Variable POINTER(c_char)
PROCESS_CREATE_PROCESS = 128 # Variable c_int
CFS_POINT = 2 # Variable c_int
PERSIST_E_SIZEINDEFINITE = -2146762742 # Variable c_long
RPC_S_INVALID_BOUND = 1734 # Variable c_long
URLACTION_INFODELIVERY_NO_EDITING_CHANNELS = 7425 # Variable c_int
DDE_FNOTPROCESSED = 0 # Variable c_int
RPC_C_MQ_PERMANENT = 1 # Variable c_int
LVIR_ICON = 1 # Variable c_int
EMARCH_ENC_I17_IMM7B_INST_WORD_X = 3 # Variable c_int
SPAPI_E_NO_SUCH_DEVINST = -2146500085 # Variable c_long
SE_PRIVILEGE_ENABLED_BY_DEFAULT = 1 # Variable c_long
CERT_E_PATHLENCONST = -2146762492 # Variable c_long
VOS__WINDOWS16 = 1 # Variable c_long
SND_MEMORY = 4 # Variable c_int
VFT_STATIC_LIB = 7 # Variable c_long
XTYP_UNREGISTER = 32978 # Variable c_int
SB_ENDSCROLL = 8 # Variable c_int
CDRF_NOTIFYPOSTERASE = 64 # Variable c_int
MCI_VD_SEEK_REVERSE = 65536 # Variable c_long
TAPE_DRIVE_LOCK_UNLK_IMMED = 2147483776 # Variable c_uint
TAPE_DRIVE_ABS_BLK_IMMED = 2147491840 # Variable c_uint
INPLACE_S_TRUNCATED = 262560 # Variable c_long
INADDR_LOOPBACK = 2130706433 # Variable c_int
STN_CLICKED = 0 # Variable c_int
TB_THUMBTRACK = 5 # Variable c_int
RPC_C_STATS_CALLS_IN = 0 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_LAST = 4106 # Variable c_long
LC_MARKER = 4 # Variable c_int
DOMAIN_GROUP_RID_CERT_ADMINS = 517 # Variable c_long
DOF_DOCUMENT = 32770 # Variable c_int
TT_ENABLED = 2 # Variable c_int
CO_E_NOT_SUPPORTED = -2147467231 # Variable c_long
DFCS_SCROLLCOMBOBOX = 5 # Variable c_int
WS_EX_PALETTEWINDOW = 392 # Variable c_long
VIF_DIFFCODEPG = 16 # Variable c_long
EMR_SETCOLORSPACE = 100 # Variable c_int
IGP_SELECT = 24 # Variable c_int
IDLFLAG_NONE = 0 # Variable c_int
RPC_S_CANT_CREATE_ENDPOINT = 1720 # Variable c_long
MCI_OVLY_STATUS_STRETCH = 16386 # Variable c_long
EM_GETLIMITTEXT = 213 # Variable c_int
PP_DISPLAYERRORS = 1 # Variable c_int
ERROR_INVALID_TARGET_HANDLE = 114 # Variable c_long
FOF_FILESONLY = 128 # Variable c_int
SIZEOF_RFPO_DATA = 16 # Variable c_int
JOB_CONTROL_SENT_TO_PRINTER = 6 # Variable c_int
SPI_SETCURSORS = 87 # Variable c_int
IMAGE_SCN_MEM_NOT_PAGED = 134217728 # Variable c_int
HSHELL_WINDOWACTIVATED = 4 # Variable c_int
JOB_NOTIFY_FIELD_TOTAL_BYTES = 22 # Variable c_int

tagSTDMSHLFLAGS = c_int # enum
SECTION_ALL_ACCESS = 983071 # Variable c_long
MCIERR_WAVE_OUTPUTSUNSUITABLE = 326 # Variable c_int
CTRY_ZIMBABWE = 263 # Variable c_int
ERROR_HOST_NODE_NOT_AVAILABLE = 5005 # Variable c_long
UNIVERSAL_NAME_INFO_LEVEL = 1 # Variable c_int
ERROR_DISK_CORRUPT = 1393 # Variable c_long
COLOR_3DDKSHADOW = 21 # Variable c_int
BM_GETSTATE = 242 # Variable c_int
CAL_SMONTHNAME1 = 21 # Variable c_int
DMLERR_POSTMSG_FAILED = 16396 # Variable c_int
TAPE_ERASE_SHORT = 0 # Variable c_long
ERROR_NO_MORE_DEVICES = 1248 # Variable c_long
ERROR_NOTIFY_ENUM_DIR = 1022 # Variable c_long
MK_E_INVALIDEXTENSION = -2147221018 # Variable c_long
IME_CMODE_SYMBOL = 1024 # Variable c_int
SUBLANG_SPANISH_GUATEMALA = 4 # Variable c_int

_ACL_INFORMATION_CLASS = c_int # enum
AclRevisionInformation = 1
AclSizeInformation = 2
ERROR_AUTODATASEG_EXCEEDS_64k = 199 # Variable c_long
TVM_GETBKCOLOR = 4383 # Variable c_int

tagEXTCONN = c_int # enum
EXTCONN_STRONG = 1
EXTCONN_WEAK = 2
EXTCONN_CALLABLE = 4
PRINTER_CHANGE_DELETE_PRINT_PROCESSOR = 67108864 # Variable c_int
TVN_ENDLABELEDITW = 4294966836 # Variable c_uint
PRINTER_ERROR_JAM = 2 # Variable c_int
GMEM_DISCARDED = 16384 # Variable c_int
CAPSLOCK_ON = 128 # Variable c_int
MSGF_MAX = 8 # Variable c_int
XST_ADVDATASENT = 15 # Variable c_int
STDOLE_LCID = 0 # Variable c_int
SPI_SETFASTTASKSWITCH = 36 # Variable c_int
SIZE_MINIMIZED = 1 # Variable c_int
S_LEGATO = 1 # Variable c_int
TBIF_IMAGE = 1 # Variable c_int
DTM_GETMCCOLOR = 4103 # Variable c_int
COLORONCOLOR = 3 # Variable c_int
SIF_TRACKPOS = 16 # Variable c_int
MM_WOM_CLOSE = 956 # Variable c_int
EMR_ARCTO = 55 # Variable c_int
TB_CUSTOMIZE = 1051 # Variable c_int
stc3 = 1090 # Variable c_int
DRV_USER = 16384 # Variable c_int
WSAECONNRESET = 10054 # Variable c_int
SO_DISCDATALEN = 28678 # Variable c_int
DL_BEGINDRAG = 1157 # Variable c_int
DATA_E_FIRST = 2147746096 # Variable c_ulong
GL_ID_PRIVATE_FIRST = 32768 # Variable c_int
RC_DEVBITS = 32768 # Variable c_int
CONTEXT_FLOATING_POINT = 65544 # Variable c_long
stc26 = 1113 # Variable c_int
STG_E_INSUFFICIENTMEMORY = -2147287032 # Variable c_long
GDICOMMENT_ENDGROUP = 3 # Variable c_int
EC_QUERYWAITING = 2 # Variable c_int
SOCKET_ERROR = -1 # Variable c_int
DESKTOP_READOBJECTS = 1 # Variable c_long
ERROR_INVALID_THREAD_ID = 1444 # Variable c_long
TBN_HOTITEMCHANGE = 4294966583 # Variable c_uint
_UPPER = 1 # Variable c_int
GL_ID_CHOOSECANDIDATE = 40 # Variable c_int
OLEOBJ_E_NOVERBS = -2147221120 # Variable c_long
CTRY_PARAGUAY = 595 # Variable c_int

tagSTGMOVE = c_int # enum
STGMOVE_MOVE = 0
STGMOVE_COPY = 1
STGMOVE_SHALLOWCOPY = 2
SM_CXSIZE = 30 # Variable c_int
CAL_SABBREVMONTHNAME5 = 38 # Variable c_int
IMAGE_FILE_MACHINE_I386 = 332 # Variable c_int
CF_UNICODETEXT = 13 # Variable c_int
TVE_COLLAPSE = 1 # Variable c_int
IMAGE_DEBUG_TYPE_OMAP_FROM_SRC = 8 # Variable c_int
CTRY_OMAN = 968 # Variable c_int
MIXERCONTROL_CONTROLTYPE_PAN = 1073872897 # Variable c_long
CTRY_JORDAN = 962 # Variable c_int
HHT_TOLEFT = 2048 # Variable c_int
SPI_SETDESKPATTERN = 21 # Variable c_int
CP_WINANSI = 1004 # Variable c_int
RPCFLG_ASYNCHRONOUS = 1073741824 # Variable c_ulong
EMR_DELETEOBJECT = 40 # Variable c_int
WHDR_PREPARED = 2 # Variable c_int
TB_CHANGEBITMAP = 1067 # Variable c_int
IMAGE_REL_SH3_SECTION = 14 # Variable c_int
IMAGE_REL_BASED_ABSOLUTE = 0 # Variable c_int
DTM_SETMCCOLOR = 4102 # Variable c_int
CFERR_CHOOSEFONTCODES = 8192 # Variable c_int
MCI_ANIM_WINDOW_DEFAULT = 0 # Variable c_long
SPI_SETPOWEROFFACTIVE = 86 # Variable c_int
NEXTBAND = 3 # Variable c_int
CF_SCREENFONTS = 1 # Variable c_int
EMR_EXTTEXTOUTA = 83 # Variable c_int
edt11 = 1162 # Variable c_int
TBN_DROPDOWN = 4294966586 # Variable c_uint
SM_CXVSCROLL = 2 # Variable c_int
RPC_C_PROFILE_MATCH_BY_MBR = 3 # Variable c_int
OLE_E_BLANK = -2147221497 # Variable c_long
DESKTOPVERTRES = 117 # Variable c_int
CAL_SMONTHNAME13 = 33 # Variable c_int
SND_ALIAS_SYSTEMASTERISK = 10835 # Variable c_ulong
IMAGE_SUBSYSTEM_WINDOWS_CUI = 3 # Variable c_int
INADDR_ANY = 0 # Variable c_ulong
TRANSPORT_TYPE_DG = 2 # Variable c_int

tagMSHLFLAGS = c_int # enum
MSHLFLAGS_NORMAL = 0
MSHLFLAGS_TABLESTRONG = 1
MSHLFLAGS_TABLEWEAK = 2
MSHLFLAGS_NOPING = 4
UI_CAP_ROT90 = 2 # Variable c_int
TBSTYLE_TOOLTIPS = 256 # Variable c_int
LOCALE_SDATE = 29 # Variable c_int
JOY_BUTTON26 = 33554432 # Variable c_long
LVM_GETBKIMAGEW = 4235 # Variable c_int
RPC_S_NOT_RPC_ERROR = 1823 # Variable c_long
ERROR_THREAD_1_INACTIVE = 210 # Variable c_long
GREEK_CHARSET = 161 # Variable c_int
JOY_RETURNV = 32 # Variable c_long
MCI_SEEK_TO_START = 256 # Variable c_long
HCBT_CLICKSKIPPED = 6 # Variable c_int
LVS_EX_UNDERLINEHOT = 2048 # Variable c_int
MDMVOLFLAG_HIGH = 4 # Variable c_int

_tagOIBDG_FLAGS = c_int # enum
OIBDG_APARTMENTTHREADED = 256
BACKUP_OBJECT_ID = 7 # Variable c_int
TA_BASELINE = 24 # Variable c_int
PST_FAX = 33 # Variable c_ulong
ERROR_CURRENT_DIRECTORY = 16 # Variable c_long
EM_GETLINECOUNT = 186 # Variable c_int
TVIS_OVERLAYMASK = 3840 # Variable c_int
PROGRESS_CLASSW = 'msctls_progress32' # Variable POINTER(c_wchar)
DEFAULT_PITCH = 0 # Variable c_int
ITALIC_FONTTYPE = 512 # Variable c_int
HSHELL_WINDOWDESTROYED = 2 # Variable c_int
CLRBREAK = 9 # Variable c_int
SECTION_MAP_EXECUTE = 8 # Variable c_int
DMTT_SUBDEV = 3 # Variable c_int
ELF_VERSION = 0 # Variable c_int
APPCLASS_MONITOR = 1 # Variable c_long
QDI_STRETCHDIB = 8 # Variable c_int
MCIERR_UNSUPPORTED_FUNCTION = 274 # Variable c_int
CO_E_INIT_SCM_FILE_MAPPING_EXISTS = -2147467249 # Variable c_long
PRINTER_STATUS_PENDING_DELETION = 4 # Variable c_int
MCI_WAVE_STATUS_BITSPERSAMPLE = 16390 # Variable c_long
ERROR_SUCCESS_RESTART_REQUIRED = 3011 # Variable c_long
RT_VERSION = 22 # Variable POINTER(c_wchar)
ACS_TIMER = 8 # Variable c_int
CB_SETITEMHEIGHT = 339 # Variable c_int
WS_EX_CONTROLPARENT = 65536 # Variable c_long
PERF_COUNTER_MULTI_TIMER_INV = 591463680 # Variable c_int
PGK_SHIFT = 1 # Variable c_int
TTM_NEWTOOLRECTA = 1030 # Variable c_int
RPC_C_MQ_USE_EXISTING_SECURITY = 4 # Variable c_int
SYSPAL_NOSTATIC256 = 3 # Variable c_int
SERVICE_CONFIG_DESCRIPTION = 1 # Variable c_int
MCI_OVLY_WINDOW_DEFAULT = 0 # Variable c_long
IMAGE_REL_MIPS_ABSOLUTE = 0 # Variable c_int
TOKEN_ADJUST_PRIVILEGES = 32 # Variable c_int
CTLCOLOR_MAX = 7 # Variable c_int
_DIGIT = 4 # Variable c_int
HKCOMB_S = 2 # Variable c_int
BINDSTRING_ACCEPT_ENCODINGS = 11
URLZONEREG_HKCU = 2
LANG_LITHUANIAN = 39 # Variable c_int
HC_SYSMODALON = 4 # Variable c_int
HCBT_MOVESIZE = 0 # Variable c_int
DCX_INTERSECTUPDATE = 512 # Variable c_long
URLACTION_INFODELIVERY_CURR_MAX = 7430 # Variable c_int
IMAGE_REL_I386_ABSOLUTE = 0 # Variable c_int
CONNECT_REFCOUNT = 64 # Variable c_int
LANG_ALBANIAN = 28 # Variable c_int
RB_GETTOOLTIPS = 1041 # Variable c_int
LOCALE_INEGCURR = 28 # Variable c_int
TB_GETHOTITEM = 1095 # Variable c_int
SC_MANAGER_ENUMERATE_SERVICE = 4 # Variable c_int
MCI_SEQ_DIV_SMPTE_24 = 1217 # Variable c_int
PRINTER_STATUS_WARMING_UP = 65536 # Variable c_int
RPC_C_NS_SYNTAX_DEFAULT = 0 # Variable c_int
LCMAP_SIMPLIFIED_CHINESE = 33554432 # Variable c_int
WGL_SWAP_OVERLAY3 = 8 # Variable c_int
CDIS_DISABLED = 4 # Variable c_int
TRANSPORT_TYPE_CN = 1 # Variable c_int
ERROR_NO_SUCH_SITE = 1249 # Variable c_long
MCS_WEEKNUMBERS = 4 # Variable c_int
NEWOBJECTOPENORD = 1548 # Variable c_int
ICC_USEREX_CLASSES = 512 # Variable c_int
TME_LEAVE = 2 # Variable c_int
LOCALE_IDEFAULTMACCODEPAGE = 4113 # Variable c_int
SECURITY_SERVICE_RID = 6 # Variable c_long
BACKGROUND_GREEN = 32 # Variable c_int
ERROR_DEVICE_IN_USE = 2404 # Variable c_long
RPC_S_WAITONTIMER = -2147417834 # Variable c_long
DT_WORDBREAK = 16 # Variable c_int
WH_SYSMSGFILTER = 6 # Variable c_int
RESOURCEDISPLAYTYPE_SERVER = 2 # Variable c_int
ERROR_WORKING_SET_QUOTA = 1453 # Variable c_long
HS_CROSS = 4 # Variable c_int
IPPORT_EXECSERVER = 512 # Variable c_int
HCBT_SETFOCUS = 9 # Variable c_int
CHANGEKIND_SETDOCUMENTATION = 3
VIFF_DONTDELETEOLD = 2 # Variable c_int
SPAPI_E_DEVICE_INTERFACE_ACTIVE = -2146500069 # Variable c_long
FILE_EXECUTE = 32 # Variable c_int
IMAGE_SYM_TYPE_MOE = 11 # Variable c_int
ERROR_DLL_INIT_FAILED = 1114 # Variable c_long
EMARCH_ENC_I17_IMM41a_SIZE_X = 10 # Variable c_int
ST_ADVISE = 2 # Variable c_int
CBEN_DRAGBEGINW = 4294966487 # Variable c_uint
RC_BITBLT = 1 # Variable c_int
PWR_SUSPENDREQUEST = 1 # Variable c_int
PSM_SETWIZBUTTONS = 1136 # Variable c_int
KEY_CREATE_LINK = 32 # Variable c_int
KLF_REPLACELANG = 16 # Variable c_int
HWND_TOPMOST = 4294967295 # Variable POINTER(HWND__)
FR_HIDEWHOLEWORD = 65536 # Variable c_int
MCI_SAVE_FILE = 256 # Variable c_long
ERROR = 0 # Variable c_int
RPC_E_SERVERFAULT = -2147417851 # Variable c_long
MCI_WAVE_SET_AVGBYTESPERSEC = 524288 # Variable c_long
VOS__WINDOWS32 = 4 # Variable c_long
TB_HIDEBUTTON = 1028 # Variable c_int
LANG_LATVIAN = 38 # Variable c_int
HDI_TEXT = 2 # Variable c_int
TVIS_EXPANDEDONCE = 64 # Variable c_int
VT_PTR = 26

PROXY_PHASE = c_int # enum
PROXY_GETBUFFER = 1
PROXY_MARSHAL = 2
PROXY_SENDRECEIVE = 3
PROXY_UNMARSHAL = 4
ERROR_OPLOCK_NOT_GRANTED = 300 # Variable c_long
CAL_SABBREVMONTHNAME2 = 35 # Variable c_int
IDC_WAIT = 206100 # Variable POINTER(c_wchar)
CACHE_S_SAMECACHE = 262513 # Variable c_long
ERROR_INVALID_CLEANER = 4310 # Variable c_long
RC_FLOODFILL = 4096 # Variable c_int
WSB_PROP_CXHTHUMB = 16 # Variable c_long
IME_ITHOTKEY_UISTYLE_TOGGLE = 514 # Variable c_int
ERROR_ACCOUNT_EXPIRED = 1793 # Variable c_long
IMAGE_REL_BASED_MIPS_JMPADDR16 = 9 # Variable c_int
DLGC_STATIC = 256 # Variable c_int
PAN_LETT_OBLIQUE_WEIGHTED = 10 # Variable c_int
RB_SETTEXTCOLOR = 1045 # Variable c_int
GDT_VALID = 0 # Variable c_int
ASPECTX = 40 # Variable c_int
GMEM_INVALID_HANDLE = 32768 # Variable c_int
LB_SETHORIZONTALEXTENT = 404 # Variable c_int
IP_MULTICAST_IF = 2 # Variable c_int
SPLDS_VERSION_NUMBER = 'versionNumber' # Variable POINTER(c_wchar)
METRICS_USEDEFAULT = -1 # Variable c_int
LB_SETTOPINDEX = 407 # Variable c_int
WNNC_NET_COGENT = 1114112 # Variable c_int
SUBLANG_ARABIC_SYRIA = 10 # Variable c_int
SRVINFO_F_COSERVERINFO2 = 1
SHOW_OPENNOACTIVATE = 4 # Variable c_int
PBT_APMQUERYSTANDBY = 1 # Variable c_int
WINSTA_CREATEDESKTOP = 8 # Variable c_long
VARIANT_NOUSEROVERRIDE = 4 # Variable c_int
DS_CENTER = 2048 # Variable c_long
CO_E_ACESINWRONGORDER = -2147220969 # Variable c_long
MIXERCONTROL_CT_CLASS_SWITCH = 536870912 # Variable c_long
STGM_CONVERT = 131072 # Variable c_long
RPC_S_OUT_OF_RESOURCES = 1721 # Variable c_long
OLE_E_FIRST = -2147221504 # Variable c_long
SUBLANG_ARABIC_LIBYA = 4 # Variable c_int
MOUSEEVENTF_RIGHTUP = 16 # Variable c_int
JOY_BUTTON20 = 524288 # Variable c_long
SPLDS_PRINT_STAPLING_SUPPORTED = 'printStaplingSupported' # Variable POINTER(c_wchar)
MCI_PLAY = 2054 # Variable c_int
DL_CURSORSET = 0 # Variable c_int
AF_IMPLINK = 3 # Variable c_int
ERROR_NO_MEDIA_IN_DRIVE = 1112 # Variable c_long
GCPCLASS_LATINNUMBER = 5 # Variable c_int
PAGE_NOACCESS = 1 # Variable c_int
TBNF_TEXT = 2 # Variable c_int
SEC_WINNT_AUTH_IDENTITY_ANSI = 1 # Variable c_int
IMAGE_NT_OPTIONAL_HDR32_MAGIC = 267 # Variable c_int
SOMAXCONN = 5 # Variable c_int
QDI_SETDIBITS = 1 # Variable c_int
ASPECT_FILTERING = 1 # Variable c_int
VTBIT_I1 = 65536 # Variable c_int
META_INTERSECTCLIPRECT = 1046 # Variable c_int
JOB_OBJECT_SECURITY_FILTER_TOKENS = 8 # Variable c_int
CTRY_BELGIUM = 32 # Variable c_int
LPD_SWAP_EXCHANGE = 512 # Variable c_int
MAP_PRECOMPOSED = 32 # Variable c_int
TRUNCATE_EXISTING = 5 # Variable c_int
PSD_NOWARNING = 128 # Variable c_int
PERF_ELAPSED_TIME = 807666944 # Variable c_int
MCI_VD_GETDEVCAPS_CAN_REVERSE = 16386 # Variable c_long
MCI_VD_GETDEVCAPS_CAV = 131072 # Variable c_long
IPPORT_EFSSERVER = 520 # Variable c_int
CFSTR_MIME_HTML = 'text/html' # Variable POINTER(c_wchar)
GCS_COMPCLAUSE = 32 # Variable c_int
EMR_OFFSETCLIPRGN = 26 # Variable c_int
TB_GETRECT = 1075 # Variable c_int
META_CREATEPENINDIRECT = 762 # Variable c_int
CERT_E_PURPOSE = -2146762490 # Variable c_long
FW_SEMIBOLD = 600 # Variable c_int
TCM_SETEXTENDEDSTYLE = 4916 # Variable c_int
NRC_ACTSES = 15 # Variable c_int
LBN_KILLFOCUS = 5 # Variable c_int
stc11 = 1098 # Variable c_int
ERROR_SEEK = 25 # Variable c_long
PORT_STATUS_POWER_SAVE = 12 # Variable c_int
WS_CHILD = 1073741824 # Variable c_long
CTRY_ROMANIA = 40 # Variable c_int
MDMVOL_MEDIUM = 1 # Variable c_int
ERROR_BAD_THREADID_ADDR = 159 # Variable c_long
RBN_DELETEDBAND = 4294966458 # Variable c_uint
SE_GROUP_ENABLED = 4 # Variable c_long
BF_ADJUST = 8192 # Variable c_int
FORM_USER = 0 # Variable c_int
PSP_HIDEHEADER = 2048 # Variable c_int
LOCALE_SISO639LANGNAME = 89 # Variable c_int
PRINTER_ERROR_INFORMATION = 2147483648 # Variable c_uint
SCS_CHANGEATTR = 18 # Variable c_int
SUBLANG_CHINESE_MACAU = 5 # Variable c_int
NRC_MAXAPPS = 54 # Variable c_int
EVENTLOG_PAIRED_EVENT_INACTIVE = 16 # Variable c_int
SPI_SETMOUSE = 4 # Variable c_int
HDF_CENTER = 2 # Variable c_int
LANG_HEBREW = 13 # Variable c_int
NFS_BUTTON = 8 # Variable c_int
MAXLONG = 2147483647 # Variable c_int
TAPE_DRIVE_END_OF_DATA = 2147549184 # Variable c_uint
DLGC_WANTCHARS = 128 # Variable c_int
GMEM_LOCKCOUNT = 255 # Variable c_int
NTM_MULTIPLEMASTER = 262144 # Variable c_int
SELECT_CAP_SENTENCE = 2 # Variable c_int
SM_CXBORDER = 5 # Variable c_int
DRV_REMOVE = 10 # Variable c_int
IPPORT_SUPDUP = 95 # Variable c_int
ERROR_QUORUMLOG_OPEN_FAILED = 5028 # Variable c_long
CONNECT_PROMPT = 16 # Variable c_int
WS_VISIBLE = 268435456 # Variable c_long
EMR_SETMAPPERFLAGS = 16 # Variable c_int
MMIOM_OPEN = 3 # Variable c_int
BM_CLICK = 245 # Variable c_int
rad7 = 1062 # Variable c_int
ERROR_DYNLINK_FROM_INVALID_RING = 196 # Variable c_long
LVM_GETSTRINGWIDTHW = 4183 # Variable c_int
TAPE_DRIVE_SPACE_IMMEDIATE = 2155872256 # Variable c_uint
MIXERCONTROL_CONTROLTYPE_QSOUNDPAN = 1073872898 # Variable c_long
WNNC_NET_DCE = 1638400 # Variable c_int
LOCALE_SNATIVELANGNAME = 4 # Variable c_int
HDI_WIDTH = 1 # Variable c_int
TVHT_ONITEMLABEL = 4 # Variable c_int
GMR_VISIBLE = 0 # Variable c_int
DC_MAXEXTENT = 5 # Variable c_int
CF_NOFACESEL = 524288 # Variable c_long
OFN_HIDEREADONLY = 4 # Variable c_int
PC_PATHS = 512 # Variable c_int
PAN_MIDLINE_STANDARD_TRIMMED = 2 # Variable c_int
LOCALE_IPOSSEPBYSPACE = 85 # Variable c_int
MCI_SPIN = 2060 # Variable c_int
PANOSE_COUNT = 10 # Variable c_int
SERVICE_ACCEPT_SHUTDOWN = 4 # Variable c_int
MCI_OVLY_WHERE_DESTINATION = 262144 # Variable c_long
MEM_RESERVE = 8192 # Variable c_int
ODS_CHECKED = 8 # Variable c_int
ERROR_EVENTLOG_FILE_CHANGED = 1503 # Variable c_long
JOYCAPS_POVCTS = 64 # Variable c_int
TVM_GETTOOLTIPS = 4377 # Variable c_int
TIME_NOMINUTESORSECONDS = 1 # Variable c_int
LVM_FINDITEMW = 4179 # Variable c_int
PERF_NO_INSTANCES = -1 # Variable c_int
RDW_INVALIDATE = 1 # Variable c_int
TTF_CENTERTIP = 2 # Variable c_int
IME_KHOTKEY_ENGLISH = 82 # Variable c_int
MB_OK = 0 # Variable c_long
NI_CLOSECANDIDATE = 17 # Variable c_int
DMICM_SATURATE = 1 # Variable c_int
HDN_TRACKA = 4294966988 # Variable c_uint
JOB_NOTIFY_FIELD_PORT_NAME = 2 # Variable c_int
EMARCH_ENC_I17_IC_SIZE_X = 1 # Variable c_int
COLOR_INACTIVECAPTIONTEXT = 19 # Variable c_int
FADF_HAVEVARTYPE = 128 # Variable c_int
LBS_OWNERDRAWVARIABLE = 32 # Variable c_long
LVM_SETIMAGELIST = 4099 # Variable c_int
HHT_ONHEADER = 2 # Variable c_int
ERROR_EXTENDED_ERROR = 1208 # Variable c_long
ERROR_SERVICE_EXISTS = 1073 # Variable c_long
EM_SCROLL = 181 # Variable c_int
LVBKIF_STYLE_TILE = 16 # Variable c_int
TME_HOVER = 1 # Variable c_int
STATUS_FLOAT_DENORMAL_OPERAND = 3221225613 # Variable c_ulong
MCS_MULTISELECT = 2 # Variable c_int
CBM_INIT = 4 # Variable c_long
ICM_QUERYPROFILE = 3 # Variable c_int
IME_SMODE_AUTOMATIC = 4 # Variable c_int
FILE_READ_ATTRIBUTES = 128 # Variable c_int
stc27 = 1114 # Variable c_int
ILC_COLOR4 = 4 # Variable c_int
UPDFCACHE_NODATACACHE = 1 # Variable c_int
ABM_NEW = 0 # Variable c_int
NCBADDNAME = 48 # Variable c_int
LOCALE_IPOSSYMPRECEDES = 84 # Variable c_int

tagPENDINGMSG = c_int # enum
PENDINGMSG_CANCELCALL = 0
PENDINGMSG_WAITNOPROCESS = 1
PENDINGMSG_WAITDEFPROCESS = 2
RPC_S_CALL_FAILED = 1726 # Variable c_long
RPC_X_BYTE_COUNT_TOO_SMALL = 1782 # Variable c_long
PSH_WATERMARK = 32768 # Variable c_int
IMAGE_REL_MIPS_REFHALF = 1 # Variable c_int
CTRL_LOGOFF_EVENT = 5 # Variable c_int
UDM_SETRANGE32 = 1135 # Variable c_int
WNNC_NET_CLEARCASE = 1441792 # Variable c_int
PAN_MIDLINE_INDEX = 8 # Variable c_int
CAL_SABBREVMONTHNAME4 = 37 # Variable c_int
WHDR_ENDLOOP = 8 # Variable c_int
NO_PRIORITY = 0 # Variable c_int
PAN_WEIGHT_VERY_LIGHT = 2 # Variable c_int
LVSIL_SMALL = 1 # Variable c_int
AF_BAN = 21 # Variable c_int
TTTOOLINFOA_V1_SIZE = 40 # Variable c_uint
MF_HILITE = 128 # Variable c_long
OLEOBJ_S_INVALIDVERB = 262528 # Variable c_long
LVM_DELETEALLITEMS = 4105 # Variable c_int
DTR_CONTROL_HANDSHAKE = 2 # Variable c_int
GCPCLASS_NEUTRAL = 3 # Variable c_int
SND_ALIAS = 65536 # Variable c_long
RPC_C_STATS_CALLS_OUT = 1 # Variable c_int
NUMCOLORS = 24 # Variable c_int
PERF_COUNTER_QUEUELEN_TYPE = 4523008 # Variable c_int
IMAGE_SYM_CLASS_AUTOMATIC = 1 # Variable c_int
TAPE_DRIVE_WRITE_MARK_IMMED = 2415919104 # Variable c_uint
RPC_S_GROUP_MEMBER_NOT_FOUND = 1898 # Variable c_long
ERROR_BAD_ENVIRONMENT = 10 # Variable c_long
ERROR_INVALID_EA_NAME = 254 # Variable c_long
PROCESSOR_SHx_SH4 = 104 # Variable c_int
MOUSEEVENTF_LEFTDOWN = 2 # Variable c_int
EM_GETRECT = 178 # Variable c_int
FACILITY_STORAGE = 3 # Variable c_int
EN_VSCROLL = 1538 # Variable c_int
ERROR_ENVVAR_NOT_FOUND = 203 # Variable c_long
psh9 = 1032 # Variable c_int
TYPE_E_CANTCREATETMPFILE = -2147316573 # Variable c_long
CAL_SMONTHNAME12 = 32 # Variable c_int
PRINTER_FONTTYPE = 16384 # Variable c_int
LVM_SETICONSPACING = 4149 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_HEADPHONES = 5 # Variable c_long
ERROR_DIRECTORY = 267 # Variable c_long
TVI_SORT = 4294901763 # Variable POINTER(_TREEITEM)
DDD_REMOVE_DEFINITION = 2 # Variable c_int
HCBT_QS = 2 # Variable c_int
SHGNLI_PIDL = 1 # Variable c_int
COMPRESSION_FORMAT_NONE = 0 # Variable c_int
NM_RELEASEDCAPTURE = 4294967280 # Variable c_uint
MOUSEEVENTF_WHEEL = 2048 # Variable c_int
rad3 = 1058 # Variable c_int
IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP = 1024 # Variable c_int
S_WHITEVOICE = 7 # Variable c_int
CDN_TYPECHANGE = 4294966689 # Variable c_uint
TOKEN_ASSIGN_PRIMARY = 1 # Variable c_int
TCM_SETITEMA = 4870 # Variable c_int
PSM_SETCURSEL = 1125 # Variable c_int
CLASSFACTORY_S_FIRST = 262416 # Variable c_long
TB_SETBUTTONWIDTH = 1083 # Variable c_int
FF_MODERN = 48 # Variable c_int
SP_OUTOFMEMORY = -5 # Variable c_int
TBSTYLE_WRAPABLE = 512 # Variable c_int
SHGFI_USEFILEATTRIBUTES = 16 # Variable c_int
STG_E_CANTSAVE = -2147286781 # Variable c_long
WNCON_NOTROUTED = 2 # Variable c_int
ERROR_CLIPBOARD_NOT_OPEN = 1418 # Variable c_long
MMSYSERR_NODRIVERCB = 20 # Variable c_int
LVIS_ACTIVATING = 32 # Variable c_int
LVHT_TORIGHT = 32 # Variable c_int
RPC_E_THREAD_NOT_INIT = -2147417841 # Variable c_long
SB_ISSIMPLE = 1038 # Variable c_int
TKIND_DISPATCH = 4
RUSSIAN_CHARSET = 204 # Variable c_int
NRC_CMDCAN = 11 # Variable c_int
DRVCNF_CANCEL = 0 # Variable c_int
PERF_RAW_BASE = 1073939459 # Variable c_int
SC_TASKLIST = 61744 # Variable c_int
CP_REGION = 2 # Variable c_int
VTBIT_UI2 = 262144 # Variable c_int
psh1 = 1024 # Variable c_int
PSM_REBOOTSYSTEM = 1130 # Variable c_int
ERROR_SEM_USER_LIMIT = 106 # Variable c_long
EMR_RECTANGLE = 43 # Variable c_int
COLOR_BACKGROUND = 1 # Variable c_int
TTM_SETTOOLINFOA = 1033 # Variable c_int
ST_ISSELF = 256 # Variable c_int
NLS_IME_CONVERSION = 8388608 # Variable c_int
GCP_DBCS = 1 # Variable c_int
SND_APPLICATION = 128 # Variable c_int
RBN_GETOBJECT = 4294966464 # Variable c_uint
DWL_DLGPROC = 4 # Variable c_int
STGM_CREATE = 4096 # Variable c_long
RC_GDI20_STATE = 32 # Variable c_int
PST_LAT = 257 # Variable c_ulong
CS_HREDRAW = 2 # Variable c_int
WSAEOPNOTSUPP = 10045 # Variable c_int
WSANO_RECOVERY = 11003 # Variable c_int
MCI_OPEN_SHAREABLE = 256 # Variable c_long
SM_CYVSCROLL = 20 # Variable c_int
RESOURCEDISPLAYTYPE_NETWORK = 6 # Variable c_int
RPC_C_MGMT_IS_SERVER_LISTEN = 3 # Variable c_int
INET_E_OBJECT_NOT_FOUND = -2146697210 # Variable c_long
HELP_SETINDEX = 5 # Variable c_long
LOCALE_SINTLSYMBOL = 21 # Variable c_int
TAPE_DRIVE_LOG_BLK_IMMED = 2147516416 # Variable c_uint
WNNC_NET_VINES = 262144 # Variable c_int
PFD_NEED_SYSTEM_PALETTE = 256 # Variable c_int
CBEIF_DI_SETITEM = 268435456 # Variable c_int
DISP_E_DIVBYZERO = -2147352558 # Variable c_long
IMAGE_REL_SH3_SECREL = 15 # Variable c_int
PARAMFLAG_FOPT = 16 # Variable c_int
ctlLast = 1279 # Variable c_int
SPI_SETMOUSETRAILS = 93 # Variable c_int
CTRY_NORWAY = 47 # Variable c_int
PSBTN_FINISH = 2 # Variable c_int
FW_HEAVY = 900 # Variable c_int
TRUST_E_TIME_STAMP = -2146869243 # Variable c_long
CFSTR_MIME_X_WAV = 'audio/x-wav' # Variable POINTER(c_wchar)
LVN_ODCACHEHINT = 4294967183 # Variable c_uint
MKF_MODIFIERS = 64 # Variable c_int
STGM_SHARE_DENY_NONE = 64 # Variable c_long
LPD_SHARE_STENCIL = 128 # Variable c_int
IMAGE_REL_ALPHA_REFLO = 11 # Variable c_int
DLGC_WANTALLKEYS = 4 # Variable c_int
CLIP_TO_PATH = 4097 # Variable c_int
MCIERR_NEW_REQUIRES_ALIAS = 299 # Variable c_int
DFCS_MENUARROWRIGHT = 4 # Variable c_int
ICC_TREEVIEW_CLASSES = 2 # Variable c_int
URLACTION_JAVA_CURR_MAX = 7168 # Variable c_int
PBM_SETPOS = 1026 # Variable c_int
SPI_SETFONTSMOOTHING = 75 # Variable c_int
DRV_FREE = 6 # Variable c_int
IS_TEXT_UNICODE_NOT_UNICODE_MASK = 3840 # Variable c_int
PRINTER_STATUS_PAGE_PUNT = 524288 # Variable c_int
CP_SYMBOL = 42 # Variable c_int
STG_E_MEDIUMFULL = -2147286928 # Variable c_long
SERVICE_RUNNING = 4 # Variable c_int
PROCESSOR_ARM820 = 2080 # Variable c_int
SECURITY_NT_NON_UNIQUE = 21 # Variable c_long
LZERROR_GLOBLOCK = -6 # Variable c_int
PSH_PROPTITLE = 1 # Variable c_int
NCBADDGRNAME = 54 # Variable c_int
TAPE_DRIVE_PADDING = 262144 # Variable c_int
MCI_VD_STATUS_DISC_SIZE = 16390 # Variable c_long
IPPORT_DAYTIME = 13 # Variable c_int
XMLELEMTYPE_DTD = 4
WSB_PROP_PALETTE = 2048 # Variable c_long
IPPROTO_ICMP = 1 # Variable c_int
TEXTCAPS = 34 # Variable c_int
PRINTACTION_NETINSTALLLINK = 3 # Variable c_int
IMAGE_REL_PPC_ADDR64 = 1 # Variable c_int
ERROR_INVALID_SIGNAL_NUMBER = 209 # Variable c_long
MCI_ESCAPE = 2053 # Variable c_int
LVBKIF_SOURCE_HBITMAP = 1 # Variable c_int
HTBOTTOMRIGHT = 17 # Variable c_int
ERROR_PASSWORD_MUST_CHANGE = 1907 # Variable c_long
PRINTER_ERROR_SEVERE = 536870912 # Variable c_int
TRANSFORM_CTM = 4107 # Variable c_int
edt14 = 1165 # Variable c_int
ERROR_INVALID_TIME = 1901 # Variable c_long
stc18 = 1105 # Variable c_int
PASSTHROUGH = 19 # Variable c_int
PGF_CALCWIDTH = 1 # Variable c_int
ICC_WIN95_CLASSES = 255 # Variable c_int
C3_DIACRITIC = 2 # Variable c_int
NRC_NAMERR = 23 # Variable c_int
RB_GETBANDBORDERS = 1058 # Variable c_int
PID_DICTIONARY = 0 # Variable c_int
HELP_MULTIKEY = 513 # Variable c_long
LOCALE_RETURN_NUMBER = 536870912 # Variable c_int
LVCF_WIDTH = 2 # Variable c_int
WNFMT_INENUM = 16 # Variable c_int
S_OK = 0 # Variable c_long
VT_SAFEARRAY = 27
BAUD_150 = 8 # Variable c_ulong
CB_ADDSTRING = 323 # Variable c_int
CS_E_LAST = 2147746152 # Variable c_ulong
ERROR_ALREADY_EXISTS = 183 # Variable c_long
NUMPRS_TRAILING_WHITE = 2 # Variable c_int
SE_GROUP_ENABLED_BY_DEFAULT = 2 # Variable c_long
MMSYSERR_INVALHANDLE = 5 # Variable c_int
TVM_INSERTITEMA = 4352 # Variable c_int
SE_SHUTDOWN_NAME = 'SeShutdownPrivilege' # Variable POINTER(c_wchar)
CF_OEMTEXT = 7 # Variable c_int
SUBLANG_SWEDISH = 1 # Variable c_int
SE_REMOTE_SHUTDOWN_NAME = 'SeRemoteShutdownPrivilege' # Variable POINTER(c_wchar)
JOB_CONTROL_LAST_PAGE_EJECTED = 7 # Variable c_int
PWR_OK = 1 # Variable c_int
MH_KEEP = 2 # Variable c_int
EMR_REALIZEPALETTE = 52 # Variable c_int
IMAGE_REL_MIPS_SECTION = 10 # Variable c_int
CCS_NOMOVEX = 130 # Variable c_long
ERROR_ILLEGAL_ELEMENT_ADDRESS = 1162 # Variable c_long
STOCK_LAST = 17 # Variable c_int
MOUSEEVENTF_MIDDLEUP = 64 # Variable c_int
PIDSI_APPNAME = 18 # Variable c_long
EM_SCROLLCARET = 183 # Variable c_int
DPD_DELETE_ALL_FILES = 4 # Variable c_int
PCF_SETXCHAR = 32 # Variable c_ulong
PAN_CONTRAST_MEDIUM = 6 # Variable c_int
CERT_E_UNTRUSTEDTESTROOT = -2146762483 # Variable c_long
LB_ERRSPACE = -2 # Variable c_int
PSINJECT_PAGES = 21 # Variable c_int
VFT2_FONT_RASTER = 1 # Variable c_long
ERROR_WMI_INVALID_MOF = 4210 # Variable c_long
IN_CLASSC_NSHIFT = 8 # Variable c_int
DST_ICON = 3 # Variable c_int
CO_E_FAILEDTOGETSECCTX = -2147220991 # Variable c_long
R2_XORPEN = 7 # Variable c_int
MCI_WAVE_STATUS_SAMPLESPERSEC = 16387 # Variable c_long
CP_INSTALLED = 1 # Variable c_int
CDERR_REGISTERMSGFAIL = 12 # Variable c_int
ERROR_PORT_UNREACHABLE = 1234 # Variable c_long
RDH_RECTANGLES = 1 # Variable c_int
PID_LOCALE = 2147483648 # Variable c_uint
LVCFMT_CENTER = 2 # Variable c_int
SET_BOUNDS = 4109 # Variable c_int
CDDS_POSTPAINT = 2 # Variable c_int
ODS_SELECTED = 1 # Variable c_int
HKCOMB_A = 8 # Variable c_int
MDITILE_HORIZONTAL = 1 # Variable c_int
DMLERR_ADVACKTIMEOUT = 16384 # Variable c_int
LMEM_NODISCARD = 32 # Variable c_int
MCI_FORMAT_FRAMES = 3 # Variable c_int
BLTALIGNMENT = 119 # Variable c_int
SERVICE_ERROR_SEVERE = 2 # Variable c_int
PSP_RTLREADING = 16 # Variable c_int
ERROR_FILE_OFFLINE = 4350 # Variable c_long
RPC_S_UNKNOWN_AUTHN_TYPE = 1741 # Variable c_long
PERF_DISPLAY_NOSHOW = 1073741824 # Variable c_int
WNNC_NET_DECORB = 2097152 # Variable c_int
KL_NAMELENGTH = 9 # Variable c_int
SWP_NOSENDCHANGING = 1024 # Variable c_int
STG_E_NOMOREFILES = -2147287022 # Variable c_long
WC_TREEVIEWW = 'SysTreeView32' # Variable POINTER(c_wchar)
MIXERCONTROL_CONTROLTYPE_PEAKMETER = 268566529 # Variable c_long
S_PERIODVOICE = 3 # Variable c_int
VFT_VXD = 5 # Variable c_long
PGF_NORMAL = 1 # Variable c_int
VT_STREAMED_OBJECT = 68
RC_DI_BITMAP = 128 # Variable c_int
VALID_INHERIT_FLAGS = 31 # Variable c_int
EC_USEFONTINFO = 65535 # Variable c_int
DFCS_MENUBULLET = 2 # Variable c_int
RT_MESSAGETABLE = 17 # Variable POINTER(c_wchar)
ILC_COLOR8 = 8 # Variable c_int
CB_DIR = 325 # Variable c_int
TCIS_HIGHLIGHTED = 2 # Variable c_int
WSAHOST_NOT_FOUND = 11001 # Variable c_int
VFT2_DRV_SOUND = 9 # Variable c_long
CBR_128000 = 128000 # Variable c_int
PF_PPC_MOVEMEM_64BIT_OK = 4 # Variable c_int
DMBIN_ONLYONE = 1 # Variable c_int
ERROR_SERVICE_DEPENDENCY_FAIL = 1068 # Variable c_long
MCN_FIRST = 4294966546 # Variable c_uint
RESOURCEDISPLAYTYPE_GENERIC = 0 # Variable c_int
PSIDENT_PSCENTRIC = 1 # Variable c_int
BS_LEFTTEXT = 32 # Variable c_long
TCI_SRCCODEPAGE = 2 # Variable c_int
CO_E_NOMATCHINGSIDFOUND = -2147220980 # Variable c_long
ERROR_FILE_ENCRYPTED = 6002 # Variable c_long
GETTECHNOLOGY = 20 # Variable c_int
MIXERCONTROL_CONTROLTYPE_BASS = 1342373890 # Variable c_long
ABSOLUTE = 1 # Variable c_int
URLPOLICY_NOTIFY_ON_ALLOW = 16 # Variable c_int

__MIDL_IBindStatusCallback_0002 = c_int # enum
NTE_BAD_KEY = -2146893821 # Variable c_long
XTYPF_NODATA = 4 # Variable c_int
HELP_SETWINPOS = 515 # Variable c_long
BSF_POSTMESSAGE = 16 # Variable c_int
LVM_GETTEXTBKCOLOR = 4133 # Variable c_int
SIOCGLOWAT = 1074033411 # Variable c_long
IMN_CLOSECANDIDATE = 4 # Variable c_int
OF_SHARE_COMPAT = 0 # Variable c_int
HANGEUL_CHARSET = 129 # Variable c_int
SECTION_MAP_WRITE = 2 # Variable c_int
EM_GETFIRSTVISIBLELINE = 206 # Variable c_int
HELP_CONTEXTPOPUP = 8 # Variable c_long
stc32 = 1119 # Variable c_int
LANG_VIETNAMESE = 42 # Variable c_int
WSB_PROP_WINSTYLE = 1024 # Variable c_long
ERROR_TOO_MANY_SECRETS = 1381 # Variable c_long
SPLREG_BEEP_ENABLED = 'BeepEnabled' # Variable POINTER(c_wchar)
LB_GETTOPINDEX = 398 # Variable c_int
XST_NULL = 0 # Variable c_int
SPI_GETSERIALKEYS = 62 # Variable c_int
GW_OWNER = 4 # Variable c_int
MCI_OVLY_WINDOW_HWND = 65536 # Variable c_long
WMSZ_BOTTOMRIGHT = 8 # Variable c_int
CS_INSERTCHAR = 8192 # Variable c_int
NM_KEYDOWN = 4294967281 # Variable c_uint
IMAGE_FILE_MACHINE_SH3E = 420 # Variable c_int
LVN_DELETEITEM = 4294967193 # Variable c_uint
TBSTYLE_EX_DRAWDDARROWS = 1 # Variable c_int
RBBS_FIXEDBMP = 32 # Variable c_int
CLIPBRD_E_CANT_OPEN = -2147221040 # Variable c_long
PSM_CHANGED = 1128 # Variable c_int
CF_GDIOBJFIRST = 768 # Variable c_int
IMAGE_REL_ALPHA_ABSOLUTE = 0 # Variable c_int
CAL_SSHORTDATE = 5 # Variable c_int
SE_ERR_ASSOCINCOMPLETE = 27 # Variable c_int
ABM_SETAUTOHIDEBAR = 8 # Variable c_int
SBARS_SIZEGRIP = 256 # Variable c_int
RPC_C_AUTHN_MSN = 17 # Variable c_int
TVS_TRACKSELECT = 512 # Variable c_int
PSM_RESTARTWINDOWS = 1129 # Variable c_int
PAN_FAMILY_TEXT_DISPLAY = 2 # Variable c_int
CP_UTF8 = 65001 # Variable c_int
TC_SA_INTEGER = 128 # Variable c_int
AF_DATAKIT = 9 # Variable c_int
WSAEDQUOT = 10069 # Variable c_int
RPC_E_CHANGED_MODE = -2147417850 # Variable c_long
PSINJECT_DOCNEEDEDRES = 19 # Variable c_int
STARTF_RUNFULLSCREEN = 32 # Variable c_int
PAN_BENT_ARMS_HORZ = 7 # Variable c_int
SZDDESYS_ITEM_SYSITEMS = 'SysItems' # Variable POINTER(c_wchar)
KLF_SETFORPROCESS = 256 # Variable c_int
LVM_SETCOLUMNW = 4192 # Variable c_int
IMAGE_REL_PPC_NEG = 256 # Variable c_int
CBN_SELENDCANCEL = 10 # Variable c_int
ERROR_INVALID_PRINTER_STATE = 1906 # Variable c_long
MIXERCONTROL_CONTROLTYPE_SINGLESELECT = 1879113728 # Variable c_long
ERROR_SETMARK_DETECTED = 1103 # Variable c_long
DEVICE_DEFAULT_FONT = 14 # Variable c_int

tagSERVERCALL = c_int # enum
SERVERCALL_ISHANDLED = 0
SERVERCALL_REJECTED = 1
SERVERCALL_RETRYLATER = 2
PGM_GETBORDER = 5127 # Variable c_int
DT_RASCAMERA = 3 # Variable c_int
SETCHARSET = 772 # Variable c_int
stc28 = 1115 # Variable c_int
DFCS_ADJUSTRECT = 8192 # Variable c_int
CDIS_GRAYED = 2 # Variable c_int
SUBLANG_SPANISH_PERU = 10 # Variable c_int
PAN_XHEIGHT_DUCKING_LARGE = 7 # Variable c_int
JOB_OBJECT_LIMIT_ACTIVE_PROCESS = 8 # Variable c_int
OF_WRITE = 1 # Variable c_int
SB_CTL = 2 # Variable c_int
MCI_SET_DOOR_OPEN = 256 # Variable c_long
TB_GETBITMAP = 1068 # Variable c_int

_TOKEN_INFORMATION_CLASS = c_int # enum
TokenUser = 1
TokenGroups = 2
TokenPrivileges = 3
TokenOwner = 4
TokenPrimaryGroup = 5
TokenDefaultDacl = 6
TokenSource = 7
TokenType = 8
TokenImpersonationLevel = 9
TokenStatistics = 10
TokenRestrictedSids = 11
TokenSessionId = 12
REBARCLASSNAMEW = 'ReBarWindow32' # Variable POINTER(c_wchar)
ERROR_BUSY_DRIVE = 142 # Variable c_long
SOFTDIST_FLAG_USAGE_EMAIL = 1 # Variable c_int
SERVICE_CHANGE_CONFIG = 2 # Variable c_int
SE_DEBUG_NAME = 'SeDebugPrivilege' # Variable POINTER(c_wchar)
META_DIBSTRETCHBLT = 2881 # Variable c_int
PRINTER_ENUM_ICON1 = 65536 # Variable c_int
EMR_SETWORLDTRANSFORM = 35 # Variable c_int
LVN_BEGINRDRAG = 4294967185 # Variable c_uint
CAL_SABBREVMONTHNAME7 = 40 # Variable c_int
OLEIVERB_PRIMARY = 0 # Variable c_long
MIXER_GETLINEINFOF_SOURCE = 1 # Variable c_long
EMR_STRETCHDIBITS = 81 # Variable c_int
BM_GETCHECK = 240 # Variable c_int
SO_TYPE = 4104 # Variable c_int
MCI_GETDEVCAPS_CAN_RECORD = 1 # Variable c_long
PRINTER_ALL_ACCESS = 983052 # Variable c_long
DATE_USE_ALT_CALENDAR = 4 # Variable c_int
CTRY_BELIZE = 501 # Variable c_int
IMAGE_FILE_MACHINE_MIPSFPU = 870 # Variable c_int
MK_E_SYNTAX = -2147221020 # Variable c_long
HDS_HORZ = 0 # Variable c_int
OFN_ENABLETEMPLATEHANDLE = 128 # Variable c_int
LB_ERR = -1 # Variable c_int
JOY_CAL_READXONLY = 1048576 # Variable c_long
FILEOPENORD = 1536 # Variable c_int
TA_TOP = 0 # Variable c_int
IMAGE_REL_MIPS_PAIR = 37 # Variable c_int
TTM_TRACKACTIVATE = 1041 # Variable c_int
CF_PALETTE = 9 # Variable c_int
IMAGE_ROM_OPTIONAL_HDR_MAGIC = 263 # Variable c_int
ACCESS_DENIED_OBJECT_ACE_TYPE = 6 # Variable c_int
SE_OWNER_DEFAULTED = 1 # Variable c_int
TTM_GETDELAYTIME = 1045 # Variable c_int
STANDARD_RIGHTS_WRITE = 131072 # Variable c_long
RB_INSERTBANDA = 1025 # Variable c_int
TB_SAVERESTOREW = 1100 # Variable c_int
E_OUTOFMEMORY = -2147024882 # Variable c_long
GM_ADVANCED = 2 # Variable c_int
TTDT_AUTOMATIC = 0 # Variable c_int
CO_E_INIT_SCM_MUTEX_EXISTS = -2147467250 # Variable c_long
URLPOLICY_ALLOW = 0 # Variable c_int
LC_INTERIORS = 128 # Variable c_int
TCS_BUTTONS = 256 # Variable c_int
SPI_GETSOUNDSENTRY = 64 # Variable c_int
STG_E_TOOMANYOPENFILES = -2147287036 # Variable c_long
RPC_C_MQ_RECOVERABLE = 1 # Variable c_int
RPC_C_MQ_JOURNAL_ALWAYS = 2 # Variable c_int
ENUMPAPERBINS = 31 # Variable c_int
CRYPT_E_INVALID_MSG_TYPE = -2146889724 # Variable c_long
cmb14 = 1149 # Variable c_int
SPIF_SENDWININICHANGE = 2 # Variable c_int
DUPLICATE_CLOSE_SOURCE = 1 # Variable c_int
MMSYSERR_ERROR = 1 # Variable c_int
DTN_WMKEYDOWNA = 4294966539 # Variable c_uint
IMAGE_REL_ARM_BRANCH24 = 3 # Variable c_int
CAL_SMONTHNAME10 = 30 # Variable c_int
SPLDS_PRINT_NOTIFY = 'printNotify' # Variable POINTER(c_wchar)
JOB_NOTIFY_FIELD_USER_NAME = 3 # Variable c_int
CREATE_NEW_CONSOLE = 16 # Variable c_int
CTRY_SWITZERLAND = 41 # Variable c_int
MF_GRAYED = 1 # Variable c_long
CTRY_BELARUS = 375 # Variable c_int
MCI_ANIM_OPEN_WS = 65536 # Variable c_long
URLPOLICY_JAVA_PROHIBIT = 0 # Variable c_int
TOOLBARCLASSNAMEW = 'ToolbarWindow32' # Variable POINTER(c_wchar)
PSM_GETCURRENTPAGEHWND = 1142 # Variable c_int
IMAGE_FILE_BYTES_REVERSED_LO = 128 # Variable c_int
RPC_C_USE_INTERNET_PORT = 1 # Variable c_int

__MIDL_ICodeInstall_0001 = c_int # enum
CIP_DISK_FULL = 0
CIP_ACCESS_DENIED = 1
CIP_NEWER_VERSION_EXISTS = 2
CIP_NAME_CONFLICT = 4
CIP_TRUST_VERIFICATION_COMPONENT_MISSING = 5
CIP_EXE_SELF_REGISTERATION_TIMEOUT = 6
CIP_UNSAFE_TO_ABORT = 7
CIP_NEED_REBOOT = 8
OLE_E_NOTRUNNING = -2147221499 # Variable c_long
TKIND_COCLASS = 5
PT_MOVETO = 6 # Variable c_int
CDERR_LOCKRESFAILURE = 8 # Variable c_int
EMBDHLP_INPROC_HANDLER = 0 # Variable c_long
STATUS_INTEGER_DIVIDE_BY_ZERO = 3221225620 # Variable c_ulong
MAXGETHOSTSTRUCT = 1024 # Variable c_int
LANG_CZECH = 5 # Variable c_int
CC_PREVENTFULLOPEN = 4 # Variable c_int
IMAGE_ARCHIVE_PAD = '\n' # Variable POINTER(c_char)
SE_ERR_NOASSOC = 31 # Variable c_int
EMR_SETICMMODE = 98 # Variable c_int
IME_KHOTKEY_HANJACONVERT = 81 # Variable c_int
IMAGE_DEBUG_TYPE_OMAP_TO_SRC = 7 # Variable c_int
WC_SEPCHARS = 32 # Variable c_int
GDICOMMENT_IDENTIFIER = 1128875079 # Variable c_int
NTE_NO_MEMORY = -2146893810 # Variable c_long
DMPAPER_USER = 256 # Variable c_int
E_INVALIDARG = -2147024809 # Variable c_long
FD_OOB = 4 # Variable c_int
STG_S_CONVERTED = 197120 # Variable c_long
CRYPT_E_NO_KEY_PROPERTY = -2146885621 # Variable c_long
R2_NOTXORPEN = 10 # Variable c_int
BINDSTRING_HEADERS = 1
SE_GROUP_LOGON_ID = 3221225472 # Variable c_ulong
RPC_S_CALL_FAILED_DNE = 1727 # Variable c_long
TVN_ITEMEXPANDEDA = 4294966890 # Variable c_uint
PRINTER_NOTIFY_FIELD_LOCATION = 6 # Variable c_int
PAGE_READWRITE = 4 # Variable c_int
LVS_LIST = 3 # Variable c_int
CAL_SDAYNAME4 = 10 # Variable c_int
PD_PRINTTOFILE = 32 # Variable c_int
MAXLOGICALLOGNAMESIZE = 256 # Variable c_int
LVS_SORTASCENDING = 16 # Variable c_int
CFSTR_MIME_X_PNG = 'image/x-png' # Variable POINTER(c_wchar)
HCBT_SYSCOMMAND = 8 # Variable c_int
CDDS_ITEMPOSTPAINT = 65538 # Variable c_int
LVS_EX_TWOCLICKACTIVATE = 128 # Variable c_int
PGS_VERT = 0 # Variable c_int
KF_UP = 32768 # Variable c_int
CAL_SDAYNAME7 = 13 # Variable c_int
DATABITS_8 = 8 # Variable c_ushort
MB_HELP = 16384 # Variable c_long
IMAGE_SCN_MEM_FARDATA = 32768 # Variable c_int
NUMPRS_PARENS = 128 # Variable c_int
LVA_SNAPTOGRID = 5 # Variable c_int
EMBDHLP_DELAYCREATE = 65536 # Variable c_long
FW_BOLD = 700 # Variable c_int
WGL_SWAP_OVERLAY5 = 32 # Variable c_int
CAL_SDAYNAME6 = 12 # Variable c_int
DMPAPER_LETTER_EXTRA_TRANSVERSE = 56 # Variable c_int
TB_ENABLEBUTTON = 1025 # Variable c_int
MCI_OVLY_WHERE_SOURCE = 131072 # Variable c_long
MCI_BREAK_KEY = 256 # Variable c_long
MF_CONV = 1073741824 # Variable c_int
ERROR_NOT_AUTHENTICATED = 1244 # Variable c_long
BS_PUSHBUTTON = 0 # Variable c_long
TOKEN_EXECUTE = 131072 # Variable c_long
VIF_FILEINUSE = 128 # Variable c_long
TCN_KEYDOWN = 4294966746 # Variable c_uint
STG_E_EXTANTMARSHALLINGS = -2147286776 # Variable c_long
ERROR_DHCP_ADDRESS_CONFLICT = 4100 # Variable c_long
ICC_PROGRESS_CLASS = 32 # Variable c_int
FILE_ADD_SUBDIRECTORY = 4 # Variable c_int
SUBLANG_SPANISH_BOLIVIA = 16 # Variable c_int
RPC_S_MAX_CALLS_TOO_SMALL = 1742 # Variable c_long
SND_ALIAS_SYSTEMEXCLAMATION = 8531 # Variable c_ulong
PBM_SETRANGE = 1025 # Variable c_int
IME_REGWORD_STYLE_EUDC = 1 # Variable c_int
IMAGE_RESOURCE_DATA_IS_DIRECTORY = 2147483648 # Variable c_uint
FW_THIN = 100 # Variable c_int
TME_QUERY = 1073741824 # Variable c_int
MCIERR_UNRECOGNIZED_COMMAND = 261 # Variable c_int
EMARCH_ENC_I17_IMM9D_INST_WORD_X = 3 # Variable c_int
SB_SETTIPTEXTW = 1041 # Variable c_int
EMR_SETPALETTEENTRIES = 50 # Variable c_int
LANG_MARATHI = 78 # Variable c_int
ILC_COLOR = 0 # Variable c_int
BLACKONWHITE = 1 # Variable c_int
ERROR_CAN_NOT_COMPLETE = 1003 # Variable c_long
STG_E_DOCFILECORRUPT = -2147286775 # Variable c_long
MAXERRORLENGTH = 256 # Variable c_int
CAT_E_LAST = 2147746145 # Variable c_ulong
WC_PAGESCROLLERW = 'SysPager' # Variable POINTER(c_wchar)
edt15 = 1166 # Variable c_int
BACKUP_LINK = 5 # Variable c_int
AC_LINE_UNKNOWN = 255 # Variable c_int
PORT_STATUS_TYPE_INFO = 3 # Variable c_int
TTF_SUBCLASS = 16 # Variable c_int
COMMON_LVB_TRAILING_BYTE = 512 # Variable c_int
DDE_FREQUESTED = 4096 # Variable c_int
ERROR_INVALID_OPLOCK_PROTOCOL = 301 # Variable c_long
UDM_SETACCEL = 1131 # Variable c_int
EM_REPLACESEL = 194 # Variable c_int
EM_SETTEXTEX = 1121 # Variable c_int
ST_DEFAULT = 0
ST_KEEPUNDO = 1
ST_SELECTION = 2
ST_NEWCHARS = 4
TB_LOADIMAGES = 1074 # Variable c_int
DFC_MENU = 2 # Variable c_int
TVINSERTSTRUCTA_V1_SIZE = 48 # Variable c_uint
VT_CARRAY = 28
ACCESS_ALLOWED_COMPOUND_ACE_TYPE = 4 # Variable c_int
DT_EDITCONTROL = 8192 # Variable c_int
CBENF_ESCAPE = 3 # Variable c_int
INPLACE_E_LAST = 2147746223 # Variable c_ulong
JOB_OBJECT_SECURITY_NO_ADMIN = 1 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE = 4099 # Variable c_long
TCM_SETIMAGELIST = 4867 # Variable c_int
GCPGLYPH_LINKAFTER = 16384 # Variable c_int
NOTSRCERASE = 1114278 # Variable c_ulong
IMAGE_SCN_ALIGN_4BYTES = 3145728 # Variable c_int
ESB_DISABLE_UP = 1 # Variable c_int
CERTSRV_E_BAD_REQUESTSUBJECT = -2146877439 # Variable c_long
FW_LIGHT = 300 # Variable c_int
TB_TOP = 6 # Variable c_int
EV_DSR = 16 # Variable c_int
EPT_S_CANT_PERFORM_OP = 1752 # Variable c_long
PRINTER_CHANGE_SET_PRINTER = 2 # Variable c_int
COLOR_SCROLLBAR = 0 # Variable c_int
ERROR_LOGON_SESSION_EXISTS = 1363 # Variable c_long
CCS_NOMOVEY = 2 # Variable c_long
ERROR_NO_USER_KEYS = 6006 # Variable c_long
TB_ISBUTTONHIDDEN = 1036 # Variable c_int
CTRY_TRINIDAD_Y_TOBAGO = 1 # Variable c_int
HS_DIAGCROSS = 5 # Variable c_int
RBBIM_ID = 256 # Variable c_int
STD_COPY = 1 # Variable c_int
ERROR_EXE_MACHINE_TYPE_MISMATCH = 216 # Variable c_long
PAN_LETT_NORMAL_OFF_CENTER = 7 # Variable c_int
QS_PAINT = 32 # Variable c_int
OUT_DEFAULT_PRECIS = 0 # Variable c_int
TB_SETTOOLTIPS = 1060 # Variable c_int
MB_ABORTRETRYIGNORE = 2 # Variable c_long
JOY_BUTTON9 = 256 # Variable c_long
GDTR_MIN = 1 # Variable c_int
GL_ID_REVERSECONVERSION = 41 # Variable c_int
MH_DELETE = 3 # Variable c_int
LVS_ALIGNTOP = 0 # Variable c_int
SM_CMOUSEBUTTONS = 43 # Variable c_int
DESKTOPHORZRES = 118 # Variable c_int
ERROR_LOCAL_USER_SESSION_KEY = 1303 # Variable c_long
VFT_UNKNOWN = 0 # Variable c_long
LVHT_ONITEMSTATEICON = 8 # Variable c_int
C2_LEFTTORIGHT = 1 # Variable c_int
NI_FINALIZECONVERSIONRESULT = 20 # Variable c_int
JOB_ACCESS_ADMINISTER = 16 # Variable c_int
CRYPT_E_NOT_DECRYPTED = -2146889718 # Variable c_long
CFSEPCHAR = '+' # Variable c_char
FOF_CONFIRMMOUSE = 2 # Variable c_int
LVM_SCROLL = 4116 # Variable c_int
CDDS_PREPAINT = 1 # Variable c_int
JOB_NOTIFY_FIELD_TIME = 19 # Variable c_int
IMAGE_REL_ALPHA_GPRELHI = 23 # Variable c_int
CONVERT10_S_LAST = 262607 # Variable c_long
ERROR_INVALID_SCROLLBAR_RANGE = 1448 # Variable c_long
THREAD_BASE_PRIORITY_IDLE = -15 # Variable c_int
ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS = 1926 # Variable c_long
TB_SETHOTIMAGELIST = 1076 # Variable c_int
FACILITY_INTERNET = 12 # Variable c_int
PROCESS_DUP_HANDLE = 64 # Variable c_int
CRYPT_E_VERIFY_USAGE_OFFLINE = -2146885591 # Variable c_long
TIME_ZONE_ID_INVALID = 4294967295 # Variable c_ulong
URLPOLICY_CHANNEL_SOFTDIST_PRECACHE = 131072 # Variable c_int
VT_BSTR_BLOB = 4095
MCI_ANIM_PLAY_SPEED = 65536 # Variable c_long
GCPCLASS_PREBOUNDLTR = 128 # Variable c_int
MSGF_COMMCTRL_SIZEHEADER = 16897 # Variable c_int
MCN_SELECT = 4294966550 # Variable c_uint
CRYPT_E_STREAM_INSUFFICIENT_DATA = -2146889711 # Variable c_long
TB_INSERTBUTTONA = 1045 # Variable c_int
PGF_DEPRESSED = 4 # Variable c_int
NCBTRACE = 121 # Variable c_int
WAIT_ABANDONED_0 = 128 # Variable c_ulong
VAR_CALENDAR_HIJRI = 8 # Variable c_ulong
DMPAPER_RESERVED_48 = 48 # Variable c_int
OF_PROMPT = 8192 # Variable c_int
MCIERR_SEQ_PORT_MAPNODEVICE = 339 # Variable c_int
CRYPT_E_FILERESIZED = -2146885595 # Variable c_long
RDW_NOERASE = 32 # Variable c_int
MIXERCONTROL_CT_UNITS_PERCENT = 327680 # Variable c_long
TB_GETHOTIMAGELIST = 1077 # Variable c_int
SPAPI_E_BAD_SERVICE_INSTALLSECT = -2146500073 # Variable c_long
RPC_S_UNKNOWN_MGR_TYPE = 1716 # Variable c_long
TAPE_UNLOAD = 1 # Variable c_long
ERROR_INVALID_EDIT_HEIGHT = 1424 # Variable c_long
TCM_SETITEMEXTRA = 4878 # Variable c_int
IME_HOTKEY_DSWITCH_LAST = 287 # Variable c_int
MFS_GRAYED = 3 # Variable c_long
CONNECT_NEED_DRIVE = 32 # Variable c_int
CAL_SCALNAME = 2 # Variable c_int
ERROR_UNKNOWN_COMPONENT = 1607 # Variable c_long
PIDSI_REVNUMBER = 9 # Variable c_long
EM_GETMARGINS = 212 # Variable c_int
MF_RIGHTJUSTIFY = 16384 # Variable c_long
NUMPRS_EXPONENT = 2048 # Variable c_int
SB_GETTEXTW = 1037 # Variable c_int
SERVICE_QUERY_CONFIG = 1 # Variable c_int
psh5 = 1028 # Variable c_int
DRV_POWER = 15 # Variable c_int
ERROR_EA_LIST_INCONSISTENT = 255 # Variable c_long
INET_E_USE_DEFAULT_SETTING = -2146697198 # Variable c_long
TB_GETBUTTONSIZE = 1082 # Variable c_int
ERROR_NOLOGON_SERVER_TRUST_ACCOUNT = 1809 # Variable c_long
ExeNamePath = 0
MCI_OVLY_PUT_DESTINATION = 262144 # Variable c_long
GCL_CBCLSEXTRA = -20 # Variable c_int
RPC_C_OPT_MQ_AUTHN_LEVEL = 6 # Variable c_int
TAPE_DRIVE_LOGICAL_BLK = 2147500032 # Variable c_uint
PBT_APMQUERYSUSPEND = 0 # Variable c_int
SP_ERROR = -1 # Variable c_int
PRINTER_CHANGE_ALL = 2004353023 # Variable c_int
SM_SHOWSOUNDS = 70 # Variable c_int
ERROR_INSTALL_USEREXIT = 1602 # Variable c_long
PRINTER_CONTROL_SET_STATUS = 4 # Variable c_int
TCIF_PARAM = 8 # Variable c_int
SPAPI_E_INVALID_HWPROFILE = -2146500080 # Variable c_long
ERROR_REVISION_MISMATCH = 1306 # Variable c_long
ERROR_OBJECT_ALREADY_EXISTS = 5010 # Variable c_long
MMSYSERR_BADDB = 14 # Variable c_int
STATUS_WAIT_0 = 0 # Variable c_ulong
SPLREG_SCHEDULER_THREAD_PRIORITY_DEFAULT = 'SchedulerThreadPriorityDefault' # Variable POINTER(c_wchar)
TBN_CUSTHELP = 4294966587 # Variable c_uint
SM_CYCURSOR = 14 # Variable c_int
RT_RCDATA = 16 # Variable POINTER(c_wchar)
MM_MIM_LONGDATA = 964 # Variable c_int
TBS_FIXEDLENGTH = 64 # Variable c_int
E_POINTER = -2147467261 # Variable c_long
WMSZ_TOPRIGHT = 5 # Variable c_int
ERROR_INVALID_MONITOR_HANDLE = 1461 # Variable c_long
JOY_BUTTON4 = 8 # Variable c_int
RPC_C_AUTHZ_NAME = 1 # Variable c_int
PS_JOIN_MITER = 8192 # Variable c_int
HEAP_DISABLE_COALESCE_ON_FREE = 128 # Variable c_int
TBN_GETDISPINFOW = 4294966579 # Variable c_uint
TYPE_E_IOERROR = -2147316574 # Variable c_long
FINDMSGSTRINGW = 'commdlg_FindReplace' # Variable POINTER(c_wchar)
DMPAPER_ENV_11 = 21 # Variable c_int
MCI_ANIM_GETDEVCAPS_CAN_STRETCH = 16391 # Variable c_long
OLE_E_ADVF = -2147221503 # Variable c_long
JOB_OBJECT_UILIMIT_READCLIPBOARD = 2 # Variable c_int
PD_ENABLEPRINTTEMPLATEHANDLE = 65536 # Variable c_int
VARFLAG_FBINDABLE = 4
SERVICE_CONTROL_NETBINDADD = 7 # Variable c_int
CTRY_GERMANY = 49 # Variable c_int
LANG_BULGARIAN = 2 # Variable c_int
XST_INIT2 = 4 # Variable c_int
SM_CYICONSPACING = 39 # Variable c_int
OF_SHARE_DENY_NONE = 64 # Variable c_int
ERROR_LOGIN_TIME_RESTRICTION = 1239 # Variable c_long
BACKGROUND_RED = 64 # Variable c_int
ERROR_INVALID_PRIMARY_GROUP = 1308 # Variable c_long
SUBLANG_GERMAN_LIECHTENSTEIN = 5 # Variable c_int
URLACTION_AUTHENTICATE_CLIENT = 6657 # Variable c_int
IMAGE_SYM_TYPE_ENUM = 10 # Variable c_int
MCI_WAVE_PCM = 1152 # Variable c_int
ERROR_WMI_GUID_NOT_FOUND = 4200 # Variable c_long
TVM_SETINDENT = 4359 # Variable c_int
LOCALE_SGROUPING = 16 # Variable c_int
SCS_POSIX_BINARY = 4 # Variable c_int
CB_SETITEMDATA = 337 # Variable c_int
CAL_SMONTHNAME5 = 25 # Variable c_int
REG_NOTIFY_CHANGE_NAME = 1 # Variable c_long
META_SELECTCLIPREGION = 300 # Variable c_int
CERT_E_ROLE = -2146762493 # Variable c_long
stc29 = 1116 # Variable c_int
RESOURCEUSAGE_SIBLING = 8 # Variable c_int
chx6 = 1045 # Variable c_int
ERROR_IO_PENDING = 997 # Variable c_long
TTM_ENUMTOOLSW = 1082 # Variable c_int
DESKTOP_WRITEOBJECTS = 128 # Variable c_long
MCI_INFO_COPYRIGHT = 8192 # Variable c_long
PSINJECT_EOF = 15 # Variable c_int
PROCESSOR_INTEL_386 = 386 # Variable c_int
TLS_MINIMUM_AVAILABLE = 64 # Variable c_int
JOY_RETURNBUTTONS = 128 # Variable c_long
LVS_ALIGNMASK = 3072 # Variable c_int
IMAGE_REL_SH3_DIRECT32 = 2 # Variable c_int
CO_E_SETSERLHNDLFAILED = -2147220976 # Variable c_long
SE_SELF_RELATIVE = 32768 # Variable c_int
PERF_DISPLAY_SECONDS = 805306368 # Variable c_int
FO_RENAME = 4 # Variable c_int
CAL_SABBREVMONTHNAME6 = 39 # Variable c_int
JOB_NOTIFY_FIELD_STATUS_STRING = 11 # Variable c_int
META_OFFSETCLIPRGN = 544 # Variable c_int
ERROR_NO_RECOVERY_PROGRAM = 1082 # Variable c_long
EV_CTS = 8 # Variable c_int
CBF_FAIL_REQUESTS = 131072 # Variable c_int
SPAPI_E_DI_DONT_INSTALL = -2146500053 # Variable c_long
BINDSTRING_LANGUAGE = 4
MIXERCONTROL_CT_SC_METER_POLLED = 0 # Variable c_long
MCHT_NOWHERE = 0 # Variable c_int
TCM_SETMINTABWIDTH = 4913 # Variable c_int
CFSTR_MIME_X_BITMAP = 'image/x-xbitmap' # Variable POINTER(c_wchar)
VAR_PERINSTANCE = 0
PO_PORTCHANGE = 32 # Variable c_int
HDM_GETITEMRECT = 4615 # Variable c_int
IMAGE_SIZEOF_ARCHIVE_MEMBER_HDR = 60 # Variable c_int
REG_OPTION_OPEN_LINK = 8 # Variable c_long
ERROR_INVALID_MESSAGENAME = 1217 # Variable c_long
PERF_DETAIL_EXPERT = 300 # Variable c_int
REBARBANDINFOW_V3_SIZE = 56 # Variable c_uint
PARITY_SPACE = 4096 # Variable c_ushort
TTM_SETTIPTEXTCOLOR = 1044 # Variable c_int
SO_MAXDG = 28681 # Variable c_int
PROCESSOR_SHx_SH3 = 103 # Variable c_int
LVM_GETVIEWRECT = 4130 # Variable c_int
WC_DEFAULTCHAR = 64 # Variable c_int
MMSYSERR_BADDEVICEID = 2 # Variable c_int
ERROR_CANTOPEN = 1011 # Variable c_long
CO_E_NOMATCHINGNAMEFOUND = -2147220978 # Variable c_long
PRINTER_CHANGE_DELETE_FORM = 262144 # Variable c_int
PSINJECT_DOCSUPPLIEDRES = 20 # Variable c_int
ERROR_LICENSE_QUOTA_EXCEEDED = 1395 # Variable c_long
NAME_FLAGS_MASK = 135 # Variable c_int
SPI_GETDEFAULTINPUTLANG = 89 # Variable c_int
SPI_SETKEYBOARDDELAY = 23 # Variable c_int
HEAP_REALLOC_IN_PLACE_ONLY = 16 # Variable c_int
JOB_OBJECT_TERMINATE_AT_END_OF_JOB = 0 # Variable c_int
PAN_PROPORTION_INDEX = 3 # Variable c_int
PCF_RTSCTS = 2 # Variable c_ulong
CBEN_FIRST = 4294966496 # Variable c_uint
SUBLANG_FRENCH_SWISS = 4 # Variable c_int
LB_SETCOUNT = 423 # Variable c_int
TBSTYLE_CHECK = 2 # Variable c_int
DMCOLOR_MONOCHROME = 1 # Variable c_int
ERROR_REDIR_PAUSED = 72 # Variable c_long
BST_FOCUS = 8 # Variable c_int
WC_DIALOG = 206704 # Variable POINTER(c_wchar)
IME_ITHOTKEY_RESEND_RESULTSTR = 512 # Variable c_int
JobObjectEndOfJobTimeInformation = 6
MSGF_MENU = 2 # Variable c_int
ATTR_INPUT_ERROR = 4 # Variable c_int
MCI_OVLY_WHERE_FRAME = 524288 # Variable c_long
DM_FORMNAME = 65536 # Variable c_long
RPC_E_SERVER_CANTMARSHAL_DATA = -2147418099 # Variable c_long
WS_EX_RIGHTSCROLLBAR = 0 # Variable c_long
MMIO_FINDPROC = 262144 # Variable c_int
WINDOW_BUFFER_SIZE_EVENT = 4 # Variable c_int
FR_SHOWHELP = 128 # Variable c_int
LCS_sRGB = 1934772034 # Variable c_int
MCI_TRACK = 16 # Variable c_long
HTNOWHERE = 0 # Variable c_int
BINDSTRING_UA_COLOR = 8
CB_SETHORIZONTALEXTENT = 350 # Variable c_int
SECURITY_DYNAMIC_TRACKING = 1 # Variable c_int
DIGSIG_E_CRYPTO = -2146762744 # Variable c_long
ES_NOHIDESEL = 256 # Variable c_long
GCP_KASHIDA = 1024 # Variable c_int
GETCOLORTABLE = 5 # Variable c_int
WS_CLIPSIBLINGS = 67108864 # Variable c_long
SPI_SETLANGTOGGLE = 91 # Variable c_int
R2_COPYPEN = 13 # Variable c_int
DISPLAY_DEVICE_PRIMARY_DEVICE = 4 # Variable c_int
WS_TABSTOP = 65536 # Variable c_long
TKIND_ALIAS = 6
IDABORT = 3 # Variable c_int
GDICOMMENT_WINDOWS_METAFILE = 2147483649 # Variable c_uint
WHITENESS = 16711778 # Variable c_ulong
PROCESSOR_ARCHITECTURE_IA64 = 6 # Variable c_int
SBS_HORZ = 0 # Variable c_long
GDICOMMENT_MULTIFORMATS = 1073741828 # Variable c_int
IME_THOTKEY_IME_NONIME_TOGGLE = 112 # Variable c_int
DS_3DLOOK = 4 # Variable c_long
DCBA_FACEDOWNCENTER = 257 # Variable c_int
SZDDESYS_ITEM_HELP = 'Help' # Variable POINTER(c_wchar)
ACTFLG_Assigned = 16 # Variable c_int
IMC_GETCANDIDATEPOS = 7 # Variable c_int
IMAGE_REL_ALPHA_SECTION = 14 # Variable c_int
ENUM_ALL_CALENDARS = 4294967295 # Variable c_uint
RPC_E_OUT_OF_RESOURCES = -2147417855 # Variable c_long
TTM_ACTIVATE = 1025 # Variable c_int
PORT_STATUS_PAPER_OUT = 3 # Variable c_int
TRUST_E_ACTION_UNKNOWN = -2146762750 # Variable c_long
SERVICE_ERROR_IGNORE = 0 # Variable c_int
EMARCH_ENC_I17_IMM5C_INST_WORD_POS_X = 13 # Variable c_int
PCF_PARITY_CHECK = 8 # Variable c_ulong
DOMAIN_ALIAS_RID_GUESTS = 546 # Variable c_long
TBS_LEFT = 4 # Variable c_int
MCIERR_BAD_INTEGER = 270 # Variable c_int
PIPE_READMODE_MESSAGE = 2 # Variable c_int
URLPOLICY_MASK_PERMISSIONS = 15 # Variable c_int
LVIF_DI_SETITEM = 4096 # Variable c_int
DACL_SECURITY_INFORMATION = 4 # Variable c_long
RBBIM_COLORS = 2 # Variable c_int
DS_NOFAILCREATE = 16 # Variable c_long
PD_RETURNDEFAULT = 1024 # Variable c_int
ERROR_UNKNOWN_PORT = 1796 # Variable c_long
OFN_ALLOWMULTISELECT = 512 # Variable c_int
PRINTER_CHANGE_TIMEOUT = 2147483648 # Variable c_uint
STOPBITS_20 = 4 # Variable c_ushort
BS_DIBPATTERNPT = 6 # Variable c_int
CE_MODE = 32768 # Variable c_int
PRINTER_NOTIFY_FIELD_PAGES_PRINTED = 23 # Variable c_int
RBBS_CHILDEDGE = 4 # Variable c_int
ERROR_BAD_DEV_TYPE = 66 # Variable c_long
SM_RESERVED1 = 24 # Variable c_int
SERVICE_INTERACTIVE_PROCESS = 256 # Variable c_int
FACILITY_RPC = 1 # Variable c_int
KEY_CREATE_SUB_KEY = 4 # Variable c_int
TVGN_LASTVISIBLE = 10 # Variable c_int
SPLDS_DRIVER_KEY = 'DsDriver' # Variable POINTER(c_wchar)
GENERIC_EXECUTE = 536870912 # Variable c_long
TBM_GETTHUMBRECT = 1049 # Variable c_int
ODA_SELECT = 2 # Variable c_int
PSD_ENABLEPAGESETUPTEMPLATEHANDLE = 131072 # Variable c_int
ERROR_NO_SUCH_DOMAIN = 1355 # Variable c_long
IDH_CANCEL = 28444 # Variable c_int
GETEXTENTTABLE = 257 # Variable c_int
MCM_SETRANGE = 4114 # Variable c_int
RPC_S_STRING_TOO_LONG = 1743 # Variable c_long
TT_PRIM_LINE = 1 # Variable c_int
MCI_FORMAT_HMS = 1 # Variable c_int
CALLBACK_EVENT = 327680 # Variable c_long
ERROR_SEM_IS_SET = 102 # Variable c_long
LCMAP_LINGUISTIC_CASING = 16777216 # Variable c_int
AF_LAT = 14 # Variable c_int
IMPLTYPEFLAG_FDEFAULT = 1 # Variable c_int
NRC_NAMCONF = 25 # Variable c_int
IMC_SETCOMPOSITIONFONT = 10 # Variable c_int
ERROR_CANNOT_FIND_WND_CLASS = 1407 # Variable c_long
CS_E_NOT_DELETABLE = -2147221147 # Variable c_long
IN_CLASSB_HOST = 65535 # Variable c_int
TCS_HOTTRACK = 64 # Variable c_int
MCIERR_WAVE_SETOUTPUTUNSUITABLE = 327 # Variable c_int
ERROR_ACTIVE_CONNECTIONS = 2402 # Variable c_long
CREATE_UNICODE_ENVIRONMENT = 1024 # Variable c_int
TCN_GETOBJECT = 4294966743 # Variable c_uint
BS_BITMAP = 128 # Variable c_long
REGDB_S_LAST = 262495 # Variable c_long
LBS_USETABSTOPS = 128 # Variable c_long
NTE_SYS_ERR = -2146893791 # Variable c_long
CF_NOVECTORFONTS = 2048 # Variable c_long
MCS_NOTODAY = 16 # Variable c_int
N_TSHIFT = 2 # Variable c_int
TRUST_E_SYSTEM_ERROR = -2146869247 # Variable c_long
AC_SRC_NO_ALPHA = 2 # Variable c_int
HELP_CONTEXTMENU = 10 # Variable c_int
edt16 = 1167 # Variable c_int
RPC_E_CLIENT_CANTMARSHAL_DATA = -2147418101 # Variable c_long
DRV_INSTALL = 9 # Variable c_int
IMAGE_FILE_32BIT_MACHINE = 256 # Variable c_int
MSGF_DDEMGR = 32769 # Variable c_int
IMAGE_SYM_DTYPE_ARRAY = 3 # Variable c_int
MCI_OVLY_WHERE_VIDEO = 1048576 # Variable c_long
TAPE_DRIVE_WRITE_PROTECT = 4096 # Variable c_int
PORT_STATUS_DOOR_OPEN = 7 # Variable c_int
FOREGROUND_GREEN = 2 # Variable c_int
ERROR_RETRY = 1237 # Variable c_long
CRYPT_E_NO_REVOCATION_DLL = -2146885615 # Variable c_long
ACCESS_MAX_MS_ACE_TYPE = 8 # Variable c_int
VT_USERDEFINED = 29
SKF_HOTKEYACTIVE = 4 # Variable c_int
MIXERCONTROL_CONTROLF_MULTIPLE = 2 # Variable c_long
DMPAPER_ENV_B5 = 34 # Variable c_int
MSG_PARTIAL = 32768 # Variable c_int
CO_E_FAILEDTOQUERYCLIENTBLANKET = -2147220987 # Variable c_long
STGFMT_DOCFILE = 5 # Variable c_int
INPLACE_E_FIRST = 2147746208 # Variable c_ulong
GCS_RESULTREADCLAUSE = 1024 # Variable c_int
TAPE_SPACE_SEQUENTIAL_FMKS = 7 # Variable c_long
TVM_SORTCHILDREN = 4371 # Variable c_int
TYPE_E_CIRCULARTYPE = -2147312508 # Variable c_long
RB_GETTEXTCOLOR = 1046 # Variable c_int
HDN_ITEMCLICKA = 4294966994 # Variable c_uint
MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE = 4102 # Variable c_long
TBNF_DI_SETITEM = 268435456 # Variable c_int
SEC_RESERVE = 67108864 # Variable c_int
DMBIN_SMALLFMT = 9 # Variable c_int
URLPOLICY_CHANNEL_SOFTDIST_PROHIBIT = 65536 # Variable c_int
psh6 = 1029 # Variable c_int
CTRY_SERBIA = 381 # Variable c_int
JOY_BUTTON16 = 32768 # Variable c_long
OEM_FIXED_FONT = 10 # Variable c_int
CPS_REVERT = 3 # Variable c_int
PGN_FIRST = 4294966396 # Variable c_uint
DT_TABSTOP = 128 # Variable c_int
ERROR_EOM_OVERFLOW = 1129 # Variable c_long
MINCHAR = 128 # Variable c_int
CCERR_CHOOSECOLORCODES = 20480 # Variable c_int
JOB_OBJECT_LIMIT_JOB_TIME = 4 # Variable c_int
ERROR_FAILED_SERVICE_CONTROLLER_CONNECT = 1063 # Variable c_long
TVS_SHOWSELALWAYS = 32 # Variable c_int
FS_BALTIC = 128 # Variable c_long
DCX_EXCLUDERGN = 64 # Variable c_long
SUBLANG_ITALIAN = 1 # Variable c_int
DRIVE_REMOVABLE = 2 # Variable c_int
HKEY_CURRENT_USER = 2147483649 # Variable POINTER(HKEY__)
SSWF_WINDOW = 2 # Variable c_int
MCI_ANIM_OPEN_PARENT = 131072 # Variable c_long
PIPE_READMODE_BYTE = 0 # Variable c_int
IDC_SIZEWE = 206404 # Variable POINTER(c_wchar)
IMAGE_REL_PPC_ADDR32 = 2 # Variable c_int
XTYP_CONNECT = 4194 # Variable c_int
IE_NOPEN = -3 # Variable c_int
PF_AMD3D_INSTRUCTIONS_AVAILABLE = 7 # Variable c_int
TAPE_SPACE_SEQUENTIAL_SMKS = 9 # Variable c_long
LVM_GETITEMTEXTA = 4141 # Variable c_int
CTRY_PORTUGAL = 351 # Variable c_int
META_OFFSETWINDOWORG = 527 # Variable c_int
SERVICE_STOP_PENDING = 3 # Variable c_int
SB_GETTIPTEXTW = 1043 # Variable c_int
SUBLANG_ARABIC_EGYPT = 3 # Variable c_int
TRUST_E_CERT_SIGNATURE = -2146869244 # Variable c_long
DMPAPER_ENV_DL = 27 # Variable c_int
RPC_C_AUTHN_LEVEL_CONNECT = 2 # Variable c_int
EM_SETMODIFY = 185 # Variable c_int
CB_SHOWDROPDOWN = 335 # Variable c_int
SRCCOPY = 13369376 # Variable c_ulong
SERVICE_CONTROL_NETBINDENABLE = 9 # Variable c_int
STATUS_FLOAT_MULTIPLE_FAULTS = 3221226164 # Variable c_ulong
CAL_SABBREVMONTHNAME10 = 43 # Variable c_int
LVN_SETDISPINFOA = 4294967145 # Variable c_uint
WSADESCRIPTION_LEN = 256 # Variable c_int
DTN_FIRST = 4294966536 # Variable c_uint
ENABLE_WINDOW_INPUT = 8 # Variable c_int
WSAENOPROTOOPT = 10042 # Variable c_int
MK_S_US = 262630 # Variable c_long
HCF_HOTKEYAVAILABLE = 64 # Variable c_int
TCS_SINGLELINE = 0 # Variable c_int
TAPE_DRIVE_HIGH_FEATURES = 2147483648 # Variable c_uint
HDN_ITEMCHANGINGW = 4294966976 # Variable c_uint
IMAGE_REL_SH3_PCREL8_LONG = 10 # Variable c_int
DM_PRINTQUALITY = 1024 # Variable c_long
MOUSEEVENTF_RIGHTDOWN = 8 # Variable c_int
DFCS_SCROLLUP = 0 # Variable c_int
ERROR_NOT_QUORUM_CAPABLE = 5021 # Variable c_long
MCIERR_FILENAME_REQUIRED = 304 # Variable c_int
IDLFLAG_FOUT = 2 # Variable c_int
S_SERMACT = -3 # Variable c_int
SUBLANG_SERBIAN_LATIN = 2 # Variable c_int
SO_OPENTYPE = 28680 # Variable c_int
RPC_BUFFER_ASYNC = 32768 # Variable c_int
ICM_REGISTERICMATCHER = 5 # Variable c_int
SO_CONNDATALEN = 28676 # Variable c_int
MIXER_GETCONTROLDETAILSF_LISTTEXT = 1 # Variable c_long
UPDOWN_CLASSW = 'msctls_updown32' # Variable POINTER(c_wchar)
DIGSIG_E_EXTENSIBILITY = -2146762745 # Variable c_long
CS_DELETE_TRANSFORM = 3 # Variable c_long
CC_WIDE = 16 # Variable c_int
MIDIERR_BADOPENMODE = 70 # Variable c_int
IMAGE_SYM_UNDEFINED = 0 # Variable c_short
PFD_NEED_PALETTE = 128 # Variable c_int
TIME_CALLBACK_EVENT_SET = 16 # Variable c_int
LANG_HINDI = 57 # Variable c_int
LOCALE_SMONTHNAME12 = 67 # Variable c_int
DOMAIN_GROUP_RID_COMPUTERS = 515 # Variable c_long
ERROR_FILE_NOT_ENCRYPTED = 6007 # Variable c_long
HDF_BITMAP = 8192 # Variable c_int
IMAGE_SCN_MEM_DISCARDABLE = 33554432 # Variable c_int
TTM_ADDTOOLA = 1028 # Variable c_int
CO_E_FAILEDTOIMPERSONATE = -2147220992 # Variable c_long
ACCESS_MAX_MS_V2_ACE_TYPE = 3 # Variable c_int
MCI_ANIM_WINDOW_HWND = 65536 # Variable c_long
NULL = 0 # Variable c_int
BS_GROUPBOX = 7 # Variable c_long
DMLERR_DLL_USAGE = 16388 # Variable c_int
TVN_SELCHANGEDW = 4294966845 # Variable c_uint
PGM_GETPOS = 5129 # Variable c_int
EMR_MAX = 104 # Variable c_int
RPC_S_UNSUPPORTED_AUTHN_LEVEL = 1821 # Variable c_long
DllNamePath = 1
TT_AVAILABLE = 1 # Variable c_int
EWX_REBOOT = 2 # Variable c_int
SM_ARRANGE = 56 # Variable c_int
ERROR_NO_USER_SESSION_KEY = 1394 # Variable c_long
PAN_NO_FIT = 1 # Variable c_int
DM_SCALE = 16 # Variable c_long
SPI_ICONHORIZONTALSPACING = 13 # Variable c_int
DM_PANNINGWIDTH = 134217728 # Variable c_long
EM_LINEFROMCHAR = 201 # Variable c_int
WDT_REMOTE_CALL = 1383359575 # Variable c_int
SE_PRIVILEGE_USED_FOR_ACCESS = 2147483648 # Variable c_ulong
RBNM_LPARAM = 4 # Variable c_int
STG_E_NOTCURRENT = -2147286783 # Variable c_long
SUBLANG_DEFAULT = 1 # Variable c_int
NTE_PROV_TYPE_NOT_DEF = -2146893801 # Variable c_long
LC_POLYMARKER = 8 # Variable c_int
SE_INCREASE_QUOTA_NAME = 'SeIncreaseQuotaPrivilege' # Variable POINTER(c_wchar)
GENERIC_WRITE = 1073741824 # Variable c_long
FOF_NOCOPYSECURITYATTRIBS = 2048 # Variable c_int
PRINTER_WRITE = 131080 # Variable c_long
BAUD_134_5 = 4 # Variable c_ulong
ERROR_UNKNOWN_PRINTPROCESSOR = 1798 # Variable c_long
CTLCOLOR_DLG = 4 # Variable c_int
GCP_NEUTRALOVERRIDE = 33554432 # Variable c_long
MDM_COMPRESSION = 1 # Variable c_int
NUMPRS_DECIMAL = 256 # Variable c_int
chx2 = 1041 # Variable c_int
MIXERCONTROL_CONTROLTYPE_BUTTON = 553713664 # Variable c_long
CLRDTR = 6 # Variable c_int

_AUDIT_EVENT_TYPE = c_int # enum
AuditEventObjectAccess = 0
AuditEventDirectoryServiceAccess = 1
DMRES_DRAFT = -1 # Variable c_int
CO_S_NOTALLINTERFACES = 524306 # Variable c_long
MCI_SEQ_DIV_SMPTE_25 = 1218 # Variable c_int
IMAGE_REL_BASED_LOW = 2 # Variable c_int
WC_DISCARDNS = 16 # Variable c_int
LVSIL_STATE = 2 # Variable c_int
CF_PRIVATEFIRST = 512 # Variable c_int
STG_E_WRITEFAULT = -2147287011 # Variable c_long
ERROR_DESTROY_OBJECT_OF_OTHER_THREAD = 1435 # Variable c_long
MIXERCONTROL_CONTROLTYPE_MONO = 536936451 # Variable c_long
ICC_LISTVIEW_CLASSES = 1 # Variable c_int
ERROR_TOO_MANY_SEM_REQUESTS = 103 # Variable c_long
IMAGE_FILE_MACHINE_THUMB = 450 # Variable c_int
FALSE = 0 # Variable c_int
PAN_CONTRAST_MEDIUM_HIGH = 7 # Variable c_int
S_SERDSR = -15 # Variable c_int
HHT_ONDIVIDER = 4 # Variable c_int
PRINTER_CONTROL_RESUME = 2 # Variable c_int
PROCESS_ALL_ACCESS = 2035711 # Variable c_long
ICC_ANIMATE_CLASS = 128 # Variable c_int
EMARCH_ENC_I17_IMM41a_VAL_POS_X = 22 # Variable c_int
AF_NETBIOS = 17 # Variable c_int
STATUS_INVALID_DISPOSITION = 3221225510 # Variable c_ulong
ERROR_PROCESS_ABORTED = 1067 # Variable c_long
TBM_SETRANGEMAX = 1032 # Variable c_int
SSTF_CHARS = 1 # Variable c_int
MCI_INFO_PRODUCT = 256 # Variable c_long
CLSCTX_SERVER = 5 # Variable c_int
LOCALE_NOUSEROVERRIDE = 2147483648 # Variable c_uint
URLACTION_DOWNLOAD_SIGNED_ACTIVEX = 4097 # Variable c_int
S_SERDPT = -12 # Variable c_int
DTS_APPCANPARSE = 16 # Variable c_int
CDERR_NOHOOK = 11 # Variable c_int
PARAMFLAG_FHASDEFAULT = 32 # Variable c_int
DISPLAY_DEVICE_MIRRORING_DRIVER = 8 # Variable c_int
JOB_NOTIFY_FIELD_POSITION = 15 # Variable c_int
STD_ERROR_HANDLE = 4294967284 # Variable c_ulong
DMPAPER_9X11 = 44 # Variable c_int
CLIPCAPS = 36 # Variable c_int
CRYPT_E_RECIPIENT_NOT_FOUND = -2146889717 # Variable c_long
MM_JOY1ZMOVE = 930 # Variable c_int
FNERR_FILENAMECODES = 12288 # Variable c_int
NLS_DBCSCHAR = 65536 # Variable c_int
TBSTATE_CHECKED = 1 # Variable c_int
PC_NONE = 0 # Variable c_int
WSANOTINITIALISED = 10093 # Variable c_int
FORMAT_MESSAGE_ALLOCATE_BUFFER = 256 # Variable c_int
TVINSERTSTRUCTW_V1_SIZE = 48 # Variable c_uint
SPI_GETMENUDROPALIGNMENT = 27 # Variable c_int
SUBLANG_AZERI_LATIN = 1 # Variable c_int
PRINTER_ENUM_CONNECTIONS = 4 # Variable c_int
VAR_STATIC = 1
PRINTER_NOTIFY_FIELD_TOTAL_BYTES = 24 # Variable c_int
TVE_COLLAPSERESET = 32768 # Variable c_int
META_CREATEREGION = 1791 # Variable c_int
LOCALE_SMONTHNAME8 = 63 # Variable c_int
MIXER_GETLINECONTROLSF_ONEBYID = 1 # Variable c_long
HTLEFT = 10 # Variable c_int
MCM_HITTEST = 4110 # Variable c_int
EVENTLOG_BACKWARDS_READ = 8 # Variable c_int
RBS_VARHEIGHT = 512 # Variable c_int
TTM_GETTOOLINFOW = 1077 # Variable c_int
MMIOERR_NETWORKERROR = 270 # Variable c_int
TOKEN_ALL_ACCESS = 983551 # Variable c_long
LOCALE_SABBREVDAYNAME7 = 55 # Variable c_int
ERROR_NO_TRACKING_SERVICE = 1172 # Variable c_long
TB_LINEUP = 0 # Variable c_int
BAUD_38400 = 16384 # Variable c_ulong
VFF_FILEINUSE = 2 # Variable c_int
UPDFCACHE_IFBLANKORONSAVECACHE = 18 # Variable c_int
LVM_GETTOOLTIPS = 4174 # Variable c_int
ERROR_LOCKED = 212 # Variable c_long
IMAGE_REL_IA64_IMM22 = 2 # Variable c_int
IMAGE_SCN_LNK_INFO = 512 # Variable c_int
DTN_USERSTRINGA = 4294966538 # Variable c_uint
DV_E_CLIPFORMAT = -2147221398 # Variable c_long
SHERB_NOCONFIRMATION = 1 # Variable c_int
SERVICE_DEMAND_START = 3 # Variable c_int
ERROR_RESOURCE_NOT_FOUND = 5007 # Variable c_long
SYSTEM_ALARM_OBJECT_ACE_TYPE = 8 # Variable c_int
LOCALE_IDEFAULTCODEPAGE = 11 # Variable c_int
PHYSICALWIDTH = 110 # Variable c_int
DMPAPER_ENV_C5 = 28 # Variable c_int
RPC_E_SERVERCALL_RETRYLATER = -2147417846 # Variable c_long
SM_CXCURSOR = 13 # Variable c_int
CDERR_NOTEMPLATE = 3 # Variable c_int
COLOR_BTNFACE = 15 # Variable c_int
VFT_DRV = 3 # Variable c_long
PERF_TYPE_NUMBER = 0 # Variable c_int
ERROR_EA_TABLE_FULL = 277 # Variable c_long
LANG_MALAYALAM = 76 # Variable c_int
RB_SETBARINFO = 1028 # Variable c_int
RPC_S_INCOMPLETE_NAME = 1755 # Variable c_long
IMC_SETCANDIDATEPOS = 8 # Variable c_int
STATUS_INTEGER_OVERFLOW = 3221225621 # Variable c_ulong
STG_E_FILENOTFOUND = -2147287038 # Variable c_long
HDM_ORDERTOINDEX = 4623 # Variable c_int
GL_ID_READINGCONFLICT = 35 # Variable c_int
JOB_NOTIFY_TYPE = 1 # Variable c_int
MCIERR_DUPLICATE_ALIAS = 289 # Variable c_int
PGM_SETPOS = 5128 # Variable c_int
EMR_EXTFLOODFILL = 53 # Variable c_int

tagSTREAM_SEEK = c_int # enum
STREAM_SEEK_SET = 0
STREAM_SEEK_CUR = 1
STREAM_SEEK_END = 2
RPC_S_ALREADY_REGISTERED = 1711 # Variable c_long
MCI_VD_ESCAPE_STRING = 256 # Variable c_long
ENDDOC = 11 # Variable c_int
VTBIT_UI1 = 131072 # Variable c_int
SEC_WINNT_AUTH_IDENTITY_UNICODE = 2 # Variable c_int
TVM_ENDEDITLABELNOW = 4374 # Variable c_int
SC_SCREENSAVE = 61760 # Variable c_int
E_HANDLE = -2147024890 # Variable c_long
TB_SETBUTTONINFOA = 1090 # Variable c_int
RPC_S_CALLPENDING = -2147417835 # Variable c_long
INET_E_SECURITY_PROBLEM = -2146697202 # Variable c_long
ERROR_REGISTRY_IO_FAILED = 1016 # Variable c_long
TYPE_E_QUALIFIEDNAMEDISALLOWED = -2147319768 # Variable c_long
ERROR_ICM_NOT_ENABLED = 2308 # Variable c_long
SUBLANG_FRENCH_MONACO = 6 # Variable c_int
MMSYSERR_NOMEM = 7 # Variable c_int
IMAGE_REL_MIPS_REFWORD = 2 # Variable c_int
CO_E_TRUSTEEDOESNTMATCHCLIENT = -2147220988 # Variable c_long
LANG_SYSTEM_DEFAULT = 2048 # Variable c_int
DI_DEFAULTSIZE = 8 # Variable c_int
CS_E_INVALID_VERSION = -2147221145 # Variable c_long
SELECTPAPERSOURCE = 18 # Variable c_int
DMPAPER_ENV_C3 = 29 # Variable c_int
PSP_USEHICON = 2 # Variable c_int
TAPE_DRIVE_EJECT_MEDIA = 16777216 # Variable c_int
WSAEALREADY = 10037 # Variable c_int
JOB_STATUS_ERROR = 2 # Variable c_int
TCM_HIGHLIGHTITEM = 4915 # Variable c_int
CO_E_INIT_TLS_CHANNEL_CONTROL = -2147467252 # Variable c_long
SBN_LAST = 4294966397 # Variable c_uint
ERROR_INTERNAL_DB_CORRUPTION = 1358 # Variable c_long
PSBTN_MAX = 6 # Variable c_int
IMAGE_REL_PPC_TYPEMASK = 255 # Variable c_int
SM_RESERVED4 = 27 # Variable c_int
CBEN_ENDEDITW = 4294966490 # Variable c_uint
SW_SHOWNORMAL = 1 # Variable c_int
DRAGDROP_E_ALREADYREGISTERED = -2147221247 # Variable c_long
PAN_MIDLINE_CONSTANT_POINTED = 9 # Variable c_int
IS_TEXT_UNICODE_UNICODE_MASK = 15 # Variable c_int
SM_CYMAXIMIZED = 62 # Variable c_int
SW_INVALIDATE = 2 # Variable c_int
IME_ESC_SYNC_HOTKEY = 4103 # Variable c_int
LANG_GUJARATI = 71 # Variable c_int
WGL_SWAP_OVERLAY7 = 128 # Variable c_int
REG_LEGAL_OPTION = 15 # Variable c_long
POSTSCRIPT_DATA = 37 # Variable c_int
ERROR_TRUST_FAILURE = 1790 # Variable c_long
SECURITY_BUILTIN_DOMAIN_RID = 32 # Variable c_long
SPAPI_E_NO_SUCH_INTERFACE_CLASS = -2146500066 # Variable c_long
MK_E_NOSTORAGE = -2147221011 # Variable c_long
REG_NOTIFY_CHANGE_LAST_SET = 4 # Variable c_long
INHERIT_ONLY_ACE = 8 # Variable c_int
PAN_SERIF_PERP_SANS = 13 # Variable c_int
DRIVE_UNKNOWN = 0 # Variable c_int
RPC_X_SS_CHAR_TRANS_OPEN_FAIL = 1773 # Variable c_long
IME_ESC_SET_EUDC_DICTIONARY = 4100 # Variable c_int
RPC_S_BINDING_HAS_NO_AUTH = 1746 # Variable c_long
PERF_QUERY_OBJECTS = -2147483648 # Variable c_long
IS_TEXT_UNICODE_ASCII16 = 1 # Variable c_int
IME_JHOTKEY_CLOSE_OPEN = 48 # Variable c_int
WSAVERNOTSUPPORTED = 10092 # Variable c_int
JOY_BUTTON17 = 65536 # Variable c_long
ACCESS_MIN_MS_OBJECT_ACE_TYPE = 5 # Variable c_int
NTE_BAD_KEY_STATE = -2146893813 # Variable c_long
MCM_SETMONTHDELTA = 4116 # Variable c_int

_SC_ACTION_TYPE = c_int # enum
TIME_CALLBACK_EVENT_PULSE = 32 # Variable c_int
VIEW_E_FIRST = 2147746112 # Variable c_ulong
ERROR_INVALID_DATATYPE = 1804 # Variable c_long
PAN_BENT_ARMS_SINGLE_SERIF = 10 # Variable c_int
EMR_SAVEDC = 33 # Variable c_int
ATTR_TARGET_NOTCONVERTED = 3 # Variable c_int
PBM_DELTAPOS = 1027 # Variable c_int
RPC_E_CONNECTION_TERMINATED = -2147418106 # Variable c_long
MCIERR_SEQ_PORT_MISCERROR = 340 # Variable c_int
BS_HATCHED = 2 # Variable c_int
CTRY_NEW_ZEALAND = 64 # Variable c_int
SM_PENWINDOWS = 41 # Variable c_int
PAN_STRAIGHT_ARMS_SINGLE_SERIF = 5 # Variable c_int
RPC_S_OBJECT_NOT_FOUND = 1710 # Variable c_long
QS_TIMER = 16 # Variable c_int
TVIS_USERMASK = 61440 # Variable c_int
TVGN_DROPHILITE = 8 # Variable c_int
FROM_LEFT_1ST_BUTTON_PRESSED = 1 # Variable c_int
TBM_GETRANGEMIN = 1025 # Variable c_int
ABM_SETPOS = 3 # Variable c_int
CRYPT_E_NO_REVOCATION_CHECK = -2146885614 # Variable c_long
R2_MASKPEN = 9 # Variable c_int
VS_FF_PRERELEASE = 2 # Variable c_long
TCIF_STATE = 16 # Variable c_int
NUMPRS_STD = 8191 # Variable c_int
CBN_SETFOCUS = 3 # Variable c_int
DT_END_ELLIPSIS = 32768 # Variable c_int

tagOLECONTF = c_int # enum
OLECONTF_EMBEDDINGS = 1
OLECONTF_LINKS = 2
OLECONTF_OTHERS = 4
OLECONTF_ONLYUSER = 8
OLECONTF_ONLYIFRUNNING = 16
VT_LPSTR = 30
RPC_C_LISTEN_MAX_CALLS_DEFAULT = 1234 # Variable c_int
META_POLYPOLYGON = 1336 # Variable c_int
OFN_ENABLEHOOK = 32 # Variable c_int
STANDARD_RIGHTS_READ = 131072 # Variable c_long
MIXERCONTROL_CT_UNITS_DECIBELS = 262144 # Variable c_long
EMR_CHORD = 46 # Variable c_int
IMAGE_SIZEOF_SHORT_NAME = 8 # Variable c_int
TBIF_STATE = 4 # Variable c_int
TAPE_DRIVE_ERASE_IMMEDIATE = 128 # Variable c_int
ERROR_RESOURCE_PROPERTIES_STORED = 5024 # Variable c_long
DM_TTOPTION = 16384 # Variable c_long
STATUS_FLOAT_STACK_CHECK = 3221225618 # Variable c_ulong
ERROR_LOCK_FAILED = 167 # Variable c_long
EMR_SETBKCOLOR = 25 # Variable c_int
SETABORTPROC = 9 # Variable c_int
SPAPI_E_NO_DRIVER_SELECTED = -2146500093 # Variable c_long
CS_VREDRAW = 1 # Variable c_int
ACTFLG_RunLocally = 1 # Variable c_int
WC_IPADDRESSA = 'SysIPAddress32' # Variable POINTER(c_char)
KF_ALTDOWN = 8192 # Variable c_int
STATUS_BREAKPOINT = 2147483651 # Variable c_ulong
ico3 = 1086 # Variable c_int
SPI_SETPOWEROFFTIMEOUT = 82 # Variable c_int
_ALPHA = 259 # Variable c_int
PROCESSOR_ARM720 = 1824 # Variable c_int
ERROR_SPECIAL_USER = 1373 # Variable c_long
IME_ESC_HANJA_MODE = 4104 # Variable c_int
MCI_CDA_TRACK_AUDIO = 1088 # Variable c_int
GMR_DAYSTATE = 1 # Variable c_int
ERROR_WMI_DP_FAILED = 4209 # Variable c_long
ERROR_BAD_TOKEN_TYPE = 1349 # Variable c_long
DT_PLOTTER = 0 # Variable c_int
FILE_CURRENT = 1 # Variable c_int
NIF_ICON = 2 # Variable c_int
NRC_DUPNAME = 13 # Variable c_int
FACILITY_CONTROL = 10 # Variable c_int
IMPLTYPEFLAG_FRESTRICTED = 4 # Variable c_int
EMR_CREATEPALETTE = 49 # Variable c_int
MMSYSERR_DELETEERROR = 18 # Variable c_int
PAN_SERIF_OBTUSE_SQUARE_COVE = 5 # Variable c_int
DFCS_BUTTONRADIOMASK = 2 # Variable c_int
BS_OWNERDRAW = 11 # Variable c_long
SHAREVISTRINGW = 'commdlg_ShareViolation' # Variable POINTER(c_wchar)
TYPE_E_SIZETOOBIG = -2147317563 # Variable c_long
RPC_X_SS_HANDLES_MISMATCH = 1778 # Variable c_long
BLACK_PEN = 7 # Variable c_int
BS_RIGHT = 512 # Variable c_long
WAVE_FORMAT_1S16 = 8 # Variable c_int
PGN_LAST = 4294966346 # Variable c_uint
EC_RIGHTMARGIN = 2 # Variable c_int
IMAGE_REL_BASED_SECTION = 6 # Variable c_int
MCI_VD_STATUS_FORWARD = 16387 # Variable c_long
IPPORT_WHOSERVER = 513 # Variable c_int
ERROR_CANNOT_MAKE = 82 # Variable c_long
LVNI_TOLEFT = 1024 # Variable c_int
EMR_EOF = 14 # Variable c_int
TVIS_BOLD = 16 # Variable c_int
WIN_CERT_TYPE_PKCS_SIGNED_DATA = 2 # Variable c_int
NTE_BAD_KEYSET = -2146893802 # Variable c_long
TVHT_ONITEMBUTTON = 16 # Variable c_int
ERROR_NONPAGED_SYSTEM_RESOURCES = 1451 # Variable c_long
MIXERLINE_COMPONENTTYPE_SRC_ANALOG = 4106 # Variable c_long
ZAFLAGS_NO_UI = 32
SETCOPYCOUNT = 17 # Variable c_int
DMDITHER_LINEART = 4 # Variable c_int
LB_GETCOUNT = 395 # Variable c_int
LCS_GM_ABS_COLORIMETRIC = 8 # Variable c_long
ERROR_CLIPPING_NOT_SUPPORTED = 2005 # Variable c_long
PSINJECT_MAX = 27 # Variable c_int
MCIERR_FILE_WRITE = 349 # Variable c_int
TBCDRF_NOEDGES = 65536 # Variable c_int
HELP_CONTENTS = 3 # Variable c_long
FILE_SUPPORTS_REMOTE_STORAGE = 256 # Variable c_int
MIIM_DATA = 32 # Variable c_int
PCF_XONXOFF = 16 # Variable c_ulong
PSM_PRESSBUTTON = 1137 # Variable c_int
BS_CHECKBOX = 2 # Variable c_long
IMC_GETSTATUSWINDOWPOS = 15 # Variable c_int
ERROR_CANNOT_DETECT_DRIVER_FAILURE = 1080 # Variable c_long
MEM_RESET = 524288 # Variable c_int
ERROR_DOMAIN_CONTROLLER_EXISTS = 1250 # Variable c_long
SUBLANG_SPANISH_COLOMBIA = 9 # Variable c_int
EV_RXCHAR = 1 # Variable c_int
PSM_REMOVEPAGE = 1126 # Variable c_int
URLZONEREG_DEFAULT = 0
TB_GETANCHORHIGHLIGHT = 1098 # Variable c_int
SUBLANG_ENGLISH_AUS = 3 # Variable c_int
LPD_SWAP_COPY = 1024 # Variable c_int
NTE_BAD_TYPE = -2146893814 # Variable c_long
LANG_KOREAN = 18 # Variable c_int
DATADIR_GET = 1
PRINTER_ATTRIBUTE_NETWORK = 16 # Variable c_int
CRYPT_E_PENDING_CLOSE = -2146885617 # Variable c_long
MIXERR_INVALVALUE = 1026 # Variable c_int
ICM_DONE_OUTSIDEDC = 4 # Variable c_int
BS_LEFT = 256 # Variable c_long
FKF_HOTKEYACTIVE = 4 # Variable c_int
psh7 = 1030 # Variable c_int
HOTKEY_CLASSW = 'msctls_hotkey32' # Variable POINTER(c_wchar)
PC_NOCOLLAPSE = 4 # Variable c_int
NT351_INTERFACE_SIZE = 64 # Variable c_int
STARTF_USECOUNTCHARS = 8 # Variable c_int
IPPORT_TIMESERVER = 37 # Variable c_int
ETO_IGNORELANGUAGE = 4096 # Variable c_int
BS_ICON = 64 # Variable c_long
SC_ARRANGE = 61712 # Variable c_int
TCM_GETIMAGELIST = 4866 # Variable c_int
LBN_ERRSPACE = -2 # Variable c_int
RP_LOGON = 1 # Variable c_int
ACCESS_PROPERTY_GUID = 2 # Variable c_int
CLIPBRD_S_LAST = 262623 # Variable c_long
TBIMHT_AFTER = 1 # Variable c_int
REG_RESOURCE_REQUIREMENTS_LIST = 10 # Variable c_int
SEE_MASK_INVOKEIDLIST = 12 # Variable c_int
CAL_SDAYNAME2 = 8 # Variable c_int
DOMAIN_USER_RID_KRBTGT = 502 # Variable c_long
BAUD_56K = 32768 # Variable c_ulong
TTM_GETTEXTA = 1035 # Variable c_int
SM_RESERVED3 = 26 # Variable c_int
LVM_SETWORKAREAS = 4161 # Variable c_int
CB_FINDSTRING = 332 # Variable c_int
UPDFCACHE_ONSTOPCACHE = 4 # Variable c_int
COMPRESSION_FORMAT_LZNT1 = 2 # Variable c_int
MCI_DEVTYPE_CD_AUDIO = 516 # Variable c_int
TVGN_NEXT = 1 # Variable c_int
DMMEDIA_GLOSSY = 3 # Variable c_int
DIB_PAL_COLORS = 1 # Variable c_int
LB_FINDSTRING = 399 # Variable c_int
FLOODFILLBORDER = 0 # Variable c_int
chx3 = 1042 # Variable c_int
SE_SACL_DEFAULTED = 32 # Variable c_int
BSF_QUERY = 1 # Variable c_int
TIMEOUT_ASYNC = 4294967295 # Variable c_uint
MCI_OVLY_GETDEVCAPS_CAN_STRETCH = 16385 # Variable c_long
LB_MSGMAX = 432 # Variable c_int
EMR_GLSRECORD = 102 # Variable c_int
TAPE_DRIVE_INITIATOR = 4 # Variable c_int
ERROR_INVALID_ICON_HANDLE = 1414 # Variable c_long
PFD_SUPPORT_DIRECTDRAW = 8192 # Variable c_int
IDB_VIEW_LARGE_COLOR = 5 # Variable c_int
ERROR_BAD_INHERITANCE_ACL = 1340 # Variable c_long

_CM_ERROR_CONTROL_TYPE = c_int # enum
SevereError = 2
CriticalError = 3
IMAGE_REL_BASED_DIR64 = 10 # Variable c_int
DTN_FORMATW = 4294966553 # Variable c_uint
OLE_E_PROMPTSAVECANCELLED = -2147221492 # Variable c_long
NMPWAIT_NOWAIT = 1 # Variable c_int
RBN_LAST = 4294966437 # Variable c_uint
CF_TIFF = 6 # Variable c_int
DCBA_FACEUPNONE = 0 # Variable c_int
WVR_ALIGNLEFT = 32 # Variable c_int
LVM_GETSELECTEDCOUNT = 4146 # Variable c_int
PROPSETFLAG_ANSI = 2 # Variable c_int
PD_FORCE_SWITCH = 65536
IME_CMODE_EUDC = 512 # Variable c_int
ERROR_REMOTE_STORAGE_NOT_ACTIVE = 4351 # Variable c_long
SM_CYKANJIWINDOW = 18 # Variable c_int
IMAGE_FILE_MACHINE_ARM = 448 # Variable c_int
JOB_OBJECT_LIMIT_PROCESS_TIME = 2 # Variable c_int
PD_RETURNIC = 512 # Variable c_int
DISP_E_BADPARAMCOUNT = -2147352562 # Variable c_long
STG_E_OLDFORMAT = -2147286780 # Variable c_long
TC_SA_CONTIN = 256 # Variable c_int
PI_PARSE_URL = 1
FILE_ATTRIBUTE_COMPRESSED = 2048 # Variable c_int
GWL_WNDPROC = -4 # Variable c_int
IMAGE_FILE_NET_RUN_FROM_SWAP = 2048 # Variable c_int
PERF_COUNTER_TEXT = 2816 # Variable c_int
NTE_PROV_TYPE_NO_MATCH = -2146893797 # Variable c_long
PD_USEDEVMODECOPIES = 262144 # Variable c_int
MCN_GETDAYSTATE = 4294966549 # Variable c_uint
CB_GETHORIZONTALEXTENT = 349 # Variable c_int
TVE_TOGGLE = 3 # Variable c_int
PSWIZB_BACK = 1 # Variable c_int
UDS_ALIGNRIGHT = 4 # Variable c_int
MCIERR_FILE_READ = 348 # Variable c_int
TBS_NOTHUMB = 128 # Variable c_int
LVM_GETEXTENDEDLISTVIEWSTYLE = 4151 # Variable c_int
SZ_ASYNC_CALLEE = 'AsyncCallee' # Variable POINTER(c_wchar)
WAVE_FORMAT_4S16 = 2048 # Variable c_int
CC_WIDESTYLED = 64 # Variable c_int
NULLREGION = 1 # Variable c_int
STRETCHBLT = 2048 # Variable c_int
IMAGE_SCN_ALIGN_1024BYTES = 11534336 # Variable c_int
DC_COLLATE = 22 # Variable c_int
TBCDRF_NOOFFSET = 262144 # Variable c_int
FR_NOWHOLEWORD = 4096 # Variable c_int
JOB_NOTIFY_FIELD_BYTES_PRINTED = 23 # Variable c_int
LANG_UZBEK = 67 # Variable c_int
C1_DIGIT = 4 # Variable c_int
MIXERCONTROL_CONTROLF_DISABLED = 2147483648 # Variable c_ulong
FLOODFILLSURFACE = 1 # Variable c_int
MMIO_INSTALLPROC = 65536 # Variable c_int
IME_CAND_MEANING = 3 # Variable c_int
VAR_CONST = 2
TYPE_E_DUPLICATEID = -2147317562 # Variable c_long
WSAEACCES = 10013 # Variable c_int
HCF_CONFIRMHOTKEY = 8 # Variable c_int
COLOR_HIGHLIGHTTEXT = 14 # Variable c_int
IMAGE_SIZEOF_NT_OPTIONAL64_HEADER = 240 # Variable c_int
MMIO_CREATE = 4096 # Variable c_int
ABM_ACTIVATE = 6 # Variable c_int
CACHE_S_FIRST = 262512 # Variable c_long
DC_COPIES = 18 # Variable c_int
FADF_BSTR = 256 # Variable c_int
IDC_SIZENESW = 206403 # Variable POINTER(c_wchar)

tagBIND_FLAGS = c_int # enum
TTM_UPDATETIPTEXTA = 1036 # Variable c_int
TAPE_DRIVE_REVERSE_POSITION = 2151677952 # Variable c_uint
MCI_GETDEVCAPS_HAS_VIDEO = 3 # Variable c_long
TCIF_RTLREADING = 4 # Variable c_int
UPDFCACHE_NORMALCACHE = 8 # Variable c_int
COLOR_INFOBK = 24 # Variable c_int
DISP_E_NONAMEDARGS = -2147352569 # Variable c_long
SERVICE_PAUSE_PENDING = 6 # Variable c_int
CDN_FOLDERCHANGE = 4294966693 # Variable c_uint
IPPROTO_IGMP = 2 # Variable c_int
ERROR_WMI_ALREADY_ENABLED = 4206 # Variable c_long
SPLDS_PRINT_MEDIA_READY = 'printMediaReady' # Variable POINTER(c_wchar)
ERROR_STATIC_INIT = 4002 # Variable c_long
META_MOVETO = 532 # Variable c_int
SB_THUMBPOSITION = 4 # Variable c_int
PSN_HELP = 4294967091 # Variable c_uint
PSH_DEFAULT = 0 # Variable c_int
COLOR_APPWORKSPACE = 12 # Variable c_int
IDC_NO = 206408 # Variable POINTER(c_wchar)
WNNC_NET_SYMFONET = 1376256 # Variable c_int
FW_EXTRABOLD = 800 # Variable c_int
XTYP_POKE = 16528 # Variable c_int
PRINTER_ENUM_FAVORITE = 4 # Variable c_int
KLF_REORDER = 8 # Variable c_int
HDITEMW_V1_SIZE = 28 # Variable c_uint
DMPAPER_ENV_C6 = 31 # Variable c_int
PAN_MIDLINE_HIGH_TRIMMED = 5 # Variable c_int
IMAGE_REL_I386_REL32 = 20 # Variable c_int
CD_LBSELCHANGE = 0 # Variable c_int
PBM_GETRANGE = 1031 # Variable c_int
ERROR_FILE_EXISTS = 80 # Variable c_long
STARTF_USESHOWWINDOW = 1 # Variable c_int
ERROR_SPECIAL_ACCOUNT = 1371 # Variable c_long
ERROR_QUORUM_RESOURCE = 5020 # Variable c_long
HORZSIZE = 4 # Variable c_int
MCHT_NEXT = 16777216 # Variable c_int
CF_ENABLEHOOK = 8 # Variable c_long
TB_MARKBUTTON = 1030 # Variable c_int
MCI_WAVE_STATUS_FORMATTAG = 16385 # Variable c_long
STGM_SHARE_DENY_READ = 48 # Variable c_long
MK_S_MONIKERALREADYREGISTERED = 262631 # Variable c_long
ERROR_DLL_NOT_FOUND = 1157 # Variable c_long
SM_CXFRAME = 32 # Variable c_int
MCIERR_NO_IDENTITY = 350 # Variable c_int
TKIND_MAX = 8
SZDDESYS_ITEM_TOPICS = 'Topics' # Variable POINTER(c_wchar)
CLIPBRD_E_BAD_DATA = -2147221037 # Variable c_long
PRINTER_NOTIFY_FIELD_CJOBS = 20 # Variable c_int
ERROR_TOO_MANY_SIDS = 1389 # Variable c_long
CO_E_REMOTE_COMMUNICATION_FAILURE = -2147467235 # Variable c_long
MM_MIM_LONGERROR = 966 # Variable c_int
MCI_WAVE_SET_BLOCKALIGN = 1048576 # Variable c_long
EM_LINESCROLL = 182 # Variable c_int
BKMODE_LAST = 2 # Variable c_int
MMIOERR_OUTOFMEMORY = 258 # Variable c_int
FS_TURKISH = 16 # Variable c_long
ERROR_TOO_MANY_MUXWAITERS = 152 # Variable c_long
FR_NOMATCHCASE = 2048 # Variable c_int
SKF_TWOKEYSOFF = 256 # Variable c_int
CTRY_JAPAN = 81 # Variable c_int
TRUST_E_NO_SIGNER_CERT = -2146869246 # Variable c_long
STGFMT_NATIVE = 1 # Variable c_int
SPLDS_PRINT_SHARE_NAME = 'printShareName' # Variable POINTER(c_wchar)
MIXERLINE_COMPONENTTYPE_SRC_FIRST = 4096 # Variable c_long
WAVERR_UNPREPARED = 34 # Variable c_int
ALL_TRANSPORTS = 'M\x00\x00\x00' # Variable POINTER(c_char)
RESOURCEUSAGE_CONTAINER = 2 # Variable c_int
WNNC_NET_APPLETALK = 1245184 # Variable c_int
WB_RIGHT = 1 # Variable c_int
LCMAP_HIRAGANA = 1048576 # Variable c_int
MOD_SHIFT = 4 # Variable c_int
ES_PASSWORD = 32 # Variable c_long
SSF_AVAILABLE = 2 # Variable c_int
HCF_INDICATOR = 32 # Variable c_int
S_QUEUEEMPTY = 0 # Variable c_int
RPC_C_OPT_BINDING_NONCAUSAL = 9 # Variable c_int
HDF_RIGHT = 1 # Variable c_int
BF_DIAGONAL_ENDBOTTOMLEFT = 25 # Variable c_int
PSNRET_INVALID = 1 # Variable c_int
GETPENWIDTH = 16 # Variable c_int
PRF_CHILDREN = 16 # Variable c_long
DSPRINT_UNPUBLISH = 3 # Variable c_int
DRIVE_CDROM = 5 # Variable c_int
WT_EXECUTEINLONGTHREAD = 16 # Variable c_int
TTM_ENUMTOOLSA = 1038 # Variable c_int
TAPE_DRIVE_SET_CMP_BOP_ONLY = 67108864 # Variable c_int
NTE_BAD_DATA = -2146893819 # Variable c_long
NRC_IFBUSY = 33 # Variable c_int
FILE_VOLUME_QUOTAS = 32 # Variable c_int
DISP_E_EXCEPTION = -2147352567 # Variable c_long
DFC_CAPTION = 1 # Variable c_int
DCBA_FACEUPCENTER = 1 # Variable c_int
NTE_BAD_PROVIDER = -2146893805 # Variable c_long
META_LINETO = 531 # Variable c_int
PENDINGTYPE_TOPLEVEL = 1
ERROR_SOURCE_ELEMENT_EMPTY = 1160 # Variable c_long
TBSTYLE_LIST = 4096 # Variable c_int
STG_E_ABNORMALAPIEXIT = -2147286790 # Variable c_long
SPI_SETDOUBLECLKWIDTH = 29 # Variable c_int
DRAGLISTMSGSTRING = 'commctrl_DragListMsg' # Variable POINTER(c_wchar)
ERROR_IS_JOIN_PATH = 147 # Variable c_long
SWP_NOMOVE = 2 # Variable c_int
TTM_DELTOOLW = 1075 # Variable c_int
AC_DST_NO_ALPHA = 32 # Variable c_int
LOCALE_SDAYNAME4 = 45 # Variable c_int
CACHE_E_LAST = 2147746175 # Variable c_ulong
TC_OP_STROKE = 2 # Variable c_int
FRAME_TRAP = 1 # Variable c_int
SM_CXHSCROLL = 21 # Variable c_int
EMR_CREATEMONOBRUSH = 93 # Variable c_int
THREAD_IMPERSONATE = 256 # Variable c_int
FILE_ACTION_MODIFIED = 3 # Variable c_int
IMAGE_SIZEOF_BASE_RELOCATION = 8 # Variable c_int
TBSTATE_HIDDEN = 8 # Variable c_int
CHANGEKIND_GENERAL = 4
LOCALE_SDAYNAME5 = 46 # Variable c_int
MMIO_EMPTYBUF = 16 # Variable c_int
HC_NOREMOVE = 3 # Variable c_int
META_SETPIXEL = 1055 # Variable c_int
PIDSI_EDITTIME = 10 # Variable c_long
IMAGE_SCN_TYPE_NO_PAD = 8 # Variable c_int
PFD_SWAP_LAYER_BUFFERS = 2048 # Variable c_int
MIXERCONTROL_CONTROLTYPE_SIGNED = 805437440 # Variable c_long
PD_PRINTSETUP = 64 # Variable c_int
SO_SNDTIMEO = 4101 # Variable c_int
CLIP_TT_ALWAYS = 32 # Variable c_int
DMPAPER_LETTER_TRANSVERSE = 54 # Variable c_int
LVM_SETHOTCURSOR = 4158 # Variable c_int
DT_CENTER = 1 # Variable c_int
COLORRES = 108 # Variable c_int
CWP_SKIPTRANSPARENT = 4 # Variable c_int
SECURITY_PRINCIPAL_SELF_RID = 10 # Variable c_long
CFSTR_MIME_X_AIFF = 'audio/x-aiff' # Variable POINTER(c_wchar)
BANDINFO = 24 # Variable c_int
JOB_STATUS_PRINTING = 16 # Variable c_int
HTTOPLEFT = 13 # Variable c_int
VIF_TEMPFILE = 1 # Variable c_long
DFC_SCROLL = 3 # Variable c_int
PROCESS_HEAP_ENTRY_MOVEABLE = 16 # Variable c_int
GCPCLASS_HEBREW = 2 # Variable c_int
MAXHALF_PTR = 32767 # Variable c_int
MIXER_GETCONTROLDETAILSF_VALUE = 0 # Variable c_long
FROM_LEFT_4TH_BUTTON_PRESSED = 16 # Variable c_int
VT_LPWSTR = 31
SEE_MASK_CLASSKEY = 3 # Variable c_int
PRINTER_NOTIFY_FIELD_PRINT_PROCESSOR = 9 # Variable c_int
CF_BOTH = 3 # Variable c_int
NM_RETURN = 4294967292 # Variable c_uint
PID_FIRST_USABLE = 2 # Variable c_int
WAVERR_SYNC = 35 # Variable c_int
META_SELECTOBJECT = 301 # Variable c_int
CO_E_RUNAS_LOGON_FAILURE = -2147467238 # Variable c_long
PSP_USETITLE = 8 # Variable c_int
HTTRANSPARENT = -1 # Variable c_int
VARIANT_LOCALBOOL = 16 # Variable c_int
DST_PREFIXTEXT = 2 # Variable c_int
LVS_EX_GRIDLINES = 1 # Variable c_int
FILE_SHARE_WRITE = 2 # Variable c_int
DROPEFFECT_SCROLL = 2147483648 # Variable c_uint
WIZ_BODYX = 92 # Variable c_int
CO_E_ALREADYINITIALIZED = -2147221007 # Variable c_long
SP_DATABITS = 4 # Variable c_ulong
MIXERLINE_COMPONENTTYPE_DST_MONITOR = 3 # Variable c_long
LMEM_ZEROINIT = 64 # Variable c_int
CBR_600 = 600 # Variable c_int
PERF_QUERY_COSTLY = -2147483646 # Variable c_long
URLACTION_SHELL_MAX = 6655 # Variable c_int
WSABASEERR = 10000 # Variable c_int
CO_E_INIT_SCM_EXEC_FAILURE = -2147467247 # Variable c_long
QDI_GETDIBITS = 2 # Variable c_int
ERROR_INVALID_TRANSFORM = 2310 # Variable c_long
PS_TYPE_MASK = 983040 # Variable c_int
URLPOLICY_LOG_ON_DISALLOW = 128 # Variable c_int
COMMPROP_INITIALIZED = 3879531822 # Variable c_ulong
SZDDESYS_ITEM_RTNMSG = 'ReturnMessage' # Variable POINTER(c_wchar)
JOY_BUTTON21 = 1048576 # Variable c_long
TYPE_E_TYPEMISMATCH = -2147316576 # Variable c_long
FILE_CREATE_PIPE_INSTANCE = 4 # Variable c_int
LOAD_DLL_DEBUG_EVENT = 6 # Variable c_int
PAN_SERIF_THIN = 7 # Variable c_int
GET_TAPE_MEDIA_INFORMATION = 0 # Variable c_int
CO_E_INIT_CLASS_CACHE = -2147467255 # Variable c_long
PSBTN_CANCEL = 5 # Variable c_int
PERF_COUNTER_COUNTER = 272696320 # Variable c_int
GGL_PRIVATE = 4 # Variable c_int
PDERR_LOADDRVFAILURE = 4100 # Variable c_int
SW_PARENTOPENING = 3 # Variable c_int
PERF_NO_UNIQUE_ID = -1 # Variable c_int
LOCALE_IDEFAULTANSICODEPAGE = 4100 # Variable c_int
SB_PAGEUP = 2 # Variable c_int
ERROR_INVALID_PRIORITY = 1800 # Variable c_long
MIDI_CACHE_ALL = 1 # Variable c_int
CONNECT_INTERACTIVE = 8 # Variable c_int
MIXER_LONG_NAME_CHARS = 64 # Variable c_int
WNNC_NET_DISTINCT = 2293760 # Variable c_int
WSAECONNABORTED = 10053 # Variable c_int
RPC_E_SERVER_DIED = -2147418105 # Variable c_long
PSM_QUERYSIBLINGS = 1132 # Variable c_int
TKF_TOGGLEKEYSON = 1 # Variable c_int
MCN_LAST = 4294966537 # Variable c_uint
EM_SETMARGINS = 211 # Variable c_int
STD_OUTPUT_HANDLE = 4294967285 # Variable c_ulong
CFSTR_MIME_QUICKTIME = 'video/quicktime' # Variable POINTER(c_wchar)
RB_SETBANDINFOW = 1035 # Variable c_int
NFR_UNICODE = 2 # Variable c_int
PIPE_NOWAIT = 1 # Variable c_int
DESCKIND_TYPECOMP = 3
ERROR_NOT_OWNER = 288 # Variable c_long
TBN_GETDISPINFOA = 4294966580 # Variable c_uint
TTDT_INITIAL = 3 # Variable c_int
OUT_TT_PRECIS = 4 # Variable c_int
TB_GETSTYLE = 1081 # Variable c_int
EMR_CLOSEFIGURE = 61 # Variable c_int
FADF_VARIANT = 2048 # Variable c_int
CO_E_RELOAD_DLL = -2147467230 # Variable c_long
ICC_DATE_CLASSES = 256 # Variable c_int
RBBS_VARIABLEHEIGHT = 64 # Variable c_int
CREATE_ALWAYS = 2 # Variable c_int
NCBDGRECVBC = 35 # Variable c_int
OF_EXIST = 16384 # Variable c_int
CF_WYSIWYG = 32768 # Variable c_long
META_ROUNDRECT = 1564 # Variable c_int
CTRY_EGYPT = 20 # Variable c_int
DCTT_DOWNLOAD_OUTLINE = 8 # Variable c_long
ICM_QUERY = 3 # Variable c_int
DIGSIG_E_DECODE = -2146762746 # Variable c_long
EMR_STROKEPATH = 64 # Variable c_int
RC_PALETTE = 256 # Variable c_int
URLZONEREG_HKLM = 1
LBS_NOTIFY = 1 # Variable c_long
JOB_OBJECT_POST_AT_END_OF_JOB = 1 # Variable c_int
WS_OVERLAPPED = 0 # Variable c_long
SZM_CREATE = 0
SPAPI_E_NO_SUCH_DEVICE_INTERFACE = -2146500059 # Variable c_long
BATTERY_FLAG_HIGH = 1 # Variable c_int
RPC_X_PIPE_CLOSED = 1916 # Variable c_long
PIDSI_LASTPRINTED = 11 # Variable c_long
SUBLANG_SPANISH_ECUADOR = 12 # Variable c_int
IMAGE_FILE_UP_SYSTEM_ONLY = 16384 # Variable c_int
TVM_EXPAND = 4354 # Variable c_int
ERROR_INVALID_CURSOR_HANDLE = 1402 # Variable c_long
WVR_REDRAW = 768 # Variable c_int
ERROR_LOGON_SESSION_COLLISION = 1366 # Variable c_long
CDS_GLOBAL = 8 # Variable c_int
ERROR_MEDIA_UNAVAILABLE = 4308 # Variable c_long
ERROR_ADDRESS_NOT_ASSOCIATED = 1228 # Variable c_long
CFSTR_MIME_JPEG = 'image/jpeg' # Variable POINTER(c_wchar)
DT_PATH_ELLIPSIS = 16384 # Variable c_int
CabFilePath = 3
ERROR_PRINTER_DRIVER_ALREADY_INSTALLED = 1795 # Variable c_long
MEM_FREE = 65536 # Variable c_int
DL_DROPPED = 1159 # Variable c_int
AF_MAX = 22 # Variable c_int
VIF_MISMATCH = 2 # Variable c_long
JOB_STATUS_PRINTED = 128 # Variable c_int
MCIERR_WAVE_OUTPUTSINUSE = 320 # Variable c_int
FILE_VOLUME_IS_COMPRESSED = 32768 # Variable c_int
SBT_RTLREADING = 1024 # Variable c_int
MAILSLOT_WAIT_FOREVER = 4294967295 # Variable c_ulong
CLIPBRD_E_CANT_SET = -2147221038 # Variable c_long
RPC_E_INVALID_DATA = -2147418097 # Variable c_long
CBS_UPPERCASE = 8192 # Variable c_long
JOY_CAL_READXYONLY = 131072 # Variable c_long
CO_E_CANTDETERMINECLASS = -2147221006 # Variable c_long
LVN_ODFINDITEMA = 4294967144 # Variable c_uint
SPAPI_E_DEVINST_ALREADY_EXISTS = -2146500089 # Variable c_long

_URLZONEREG = c_int # enum
ROT_COMPARE_MAX = 2048 # Variable c_int
EMR_ARC = 45 # Variable c_int
PRINTACTION_SERVERPROPERTIES = 7 # Variable c_int
TIME_MIDI = 16 # Variable c_int
GMEM_DDESHARE = 8192 # Variable c_int
IPPORT_TFTP = 69 # Variable c_int
CAL_SDAYNAME3 = 9 # Variable c_int

tagDATADIR = c_int # enum
PBM_SETRANGE32 = 1030 # Variable c_int
PAN_WEIGHT_DEMI = 7 # Variable c_int
TARGET_IS_NT50_OR_LATER = 0 # Variable c_int

tagADVF = c_int # enum
ADVF_DATAONSTOP = 64
ADVFCACHE_NOHANDLER = 8
ADVFCACHE_FORCEBUILTIN = 16
PERF_AVERAGE_BULK = 1073874176 # Variable c_int
NM_RCLICK = 4294967291 # Variable c_uint
ERROR_CHILD_WINDOW_MENU = 1436 # Variable c_long
NTE_BAD_PUBLIC_KEY = -2146893803 # Variable c_long
BDR_OUTER = 3 # Variable c_int
SPLDS_PRINTER_LOCATIONS = 'printerLocations' # Variable POINTER(c_wchar)
DUPLICATE_DEREG = 7 # Variable c_int
DLGC_WANTTAB = 2 # Variable c_int
VTBIT_UI4 = 524288 # Variable c_int
CFS_FORCE_POSITION = 32 # Variable c_int
WINSTA_ENUMERATE = 256 # Variable c_long
LANG_ORIYA = 72 # Variable c_int
INET_E_INVALID_REQUEST = -2146697204 # Variable c_long
SERVICE_INTERROGATE = 128 # Variable c_int
MCI_SYSINFO_QUANTITY = 256 # Variable c_long
SYSPAL_NOSTATIC = 2 # Variable c_int
IMAGE_REL_PPC_REFHI = 16 # Variable c_int
SPLDS_BYTES_PER_MINUTE = 'bytesPerMinute' # Variable POINTER(c_wchar)
TVS_NOTOOLTIPS = 128 # Variable c_int
ERROR_NOT_LOGON_PROCESS = 1362 # Variable c_long
LCS_CALIBRATED_RGB = 0 # Variable c_long
DMICMMETHOD_DRIVER = 3 # Variable c_int
FILE_NOTIFY_CHANGE_ATTRIBUTES = 4 # Variable c_int
STG_S_BLOCK = 197121 # Variable c_long
PIDSI_SUBJECT = 3 # Variable c_long
WS_EX_TOOLWINDOW = 128 # Variable c_long
PI_FILTER_MODE = 2
MCI_SEQ_OFFSET = 1216 # Variable c_int
HC_GETNEXT = 1 # Variable c_int
FILE_ATTRIBUTE_ENCRYPTED = 64 # Variable c_int
stc10 = 1097 # Variable c_int
DWL_MSGRESULT = 0 # Variable c_int
RB_GETBARHEIGHT = 1051 # Variable c_int
LOAD_LIBRARY_AS_DATAFILE = 2 # Variable c_int
IMAGE_SYM_CLASS_MEMBER_OF_STRUCT = 8 # Variable c_int
SP_SERIALCOMM = 1 # Variable c_ulong
IPPORT_BIFFUDP = 512 # Variable c_int
CS_SAVEBITS = 2048 # Variable c_int
SETBREAK = 8 # Variable c_int
IMAGE_DEBUG_TYPE_BORLAND = 9 # Variable c_int
RPC_S_CALL_CANCELLED = 1818 # Variable c_long
CE_IOE = 1024 # Variable c_int
MCS_DAYSTATE = 1 # Variable c_int
MIXERCONTROL_CT_SC_SWITCH_BUTTON = 16777216 # Variable c_long
TPM_LEFTALIGN = 0 # Variable c_long
GCS_COMPATTR = 16 # Variable c_int
CAL_SABBREVMONTHNAME3 = 36 # Variable c_int
AF_ECMA = 8 # Variable c_int
MIXERCONTROL_CT_SC_SWITCH_BOOLEAN = 0 # Variable c_long
LANG_BASQUE = 45 # Variable c_int
SPI_GETKEYBOARDSPEED = 10 # Variable c_int
CCM_GETCOLORSCHEME = 8195 # Variable c_int
CBEN_GETDISPINFOW = 4294966489 # Variable c_uint
OF_CREATE = 4096 # Variable c_int
SM_CXDOUBLECLK = 36 # Variable c_int
VAR_DISPATCH = 3
TA_CENTER = 6 # Variable c_int
chx13 = 1052 # Variable c_int
ERROR_GROUP_NOT_FOUND = 5013 # Variable c_long
TVN_BEGINLABELEDITA = 4294966886 # Variable c_uint
LOCALE_SMONTHNAME4 = 59 # Variable c_int
RPC_S_UUID_LOCAL_ONLY = 1824 # Variable c_long
MOUSE_MOVED = 1 # Variable c_int
HORZRES = 8 # Variable c_int
PSINJECT_BEGINDEFAULTS = 3 # Variable c_int
EMR_POLYBEZIER16 = 85 # Variable c_int
LOCALE_SPOSITIVESIGN = 80 # Variable c_int
PBT_APMSTANDBY = 5 # Variable c_int
C1_PUNCT = 16 # Variable c_int

tagOLEVERBATTRIB = c_int # enum
OLEVERBATTRIB_NEVERDIRTIES = 1
OLEVERBATTRIB_ONCONTAINERMENU = 2
DMPAPER_LEGAL_EXTRA = 51 # Variable c_int
MCI_ANIM_GETDEVCAPS_PALETTES = 16390 # Variable c_long
STATFLAG_NONAME = 1
PRINTER_STATUS_PAPER_JAM = 8 # Variable c_int
MCI_STATUS_START = 512 # Variable c_long
NF_REQUERY = 4 # Variable c_int
MEM_E_INVALID_LINK = -2146959344 # Variable c_long
DATETIMEPICK_CLASSA = 'SysDateTimePick32' # Variable POINTER(c_char)
SM_MOUSEPRESENT = 19 # Variable c_int
SHGFI_PIDL = 8 # Variable c_int
PSD_DEFAULTMINMARGINS = 0 # Variable c_int

tagCLSCTX = c_int # enum
CLSCTX_INPROC_SERVER = 1
CLSCTX_INPROC_HANDLER = 2
CLSCTX_LOCAL_SERVER = 4
CLSCTX_INPROC_SERVER16 = 8
CLSCTX_REMOTE_SERVER = 16
CLSCTX_INPROC_HANDLER16 = 32
CLSCTX_INPROC_SERVERX86 = 64
CLSCTX_INPROC_HANDLERX86 = 128
CLSCTX_ESERVER_HANDLER = 256
ERROR_INVALID_GROUP_ATTRIBUTES = 1345 # Variable c_long
ERROR_SWAPERROR = 999 # Variable c_long
LVN_HOTTRACK = 4294967175 # Variable c_uint
TRUST_E_FAIL = -2146762485 # Variable c_long
TA_BOTTOM = 8 # Variable c_int
DMLERR_SERVER_DIED = 16398 # Variable c_int
SB_RIGHT = 7 # Variable c_int
LZERROR_BADVALUE = -7 # Variable c_int
WS_EX_CONTEXTHELP = 1024 # Variable c_long
MKSYS_ANTIMONIKER = 3
IME_HOTKEY_PRIVATE_LAST = 543 # Variable c_int
IME_CAND_CODE = 2 # Variable c_int
ERROR_NO_LOG_SPACE = 1019 # Variable c_long
JOY_RETURNX = 1 # Variable c_long
PID_CODEPAGE = 1 # Variable c_int
FD_READ = 1 # Variable c_int
SM_CYMENUSIZE = 55 # Variable c_int
CF_ENABLETEMPLATE = 16 # Variable c_long
MCIERR_SEQ_TIMER = 341 # Variable c_int
POSTSCRIPT_IGNORE = 38 # Variable c_int
TCHT_ONITEM = 6 # Variable c_int
LOCALE_SABBREVDAYNAME4 = 52 # Variable c_int
CB_SETTOPINDEX = 348 # Variable c_int
LVM_GETCOUNTPERPAGE = 4136 # Variable c_int
CCS_TOP = 1 # Variable c_long
DT_BOTTOM = 8 # Variable c_int
OFN_SHAREAWARE = 16384 # Variable c_int
RDW_NOCHILDREN = 64 # Variable c_int
HHT_ONDIVOPEN = 8 # Variable c_int
TRUST_E_SUBJECT_NOT_TRUSTED = -2146762748 # Variable c_long
REG_NOTIFY_CHANGE_SECURITY = 8 # Variable c_long
GCP_CLASSIN = 524288 # Variable c_long
DDE_FACK = 32768 # Variable c_int
FKF_INDICATOR = 32 # Variable c_int
MCIERR_MUST_USE_SHAREABLE = 291 # Variable c_int
BATTERY_FLAG_CRITICAL = 4 # Variable c_int
IMAGE_REL_I386_DIR32 = 6 # Variable c_int
STDOLE_MAJORVERNUM = 1 # Variable c_int
DISPATCH_PROPERTYPUT = 4 # Variable c_int
MCI_FORMAT_BYTES = 8 # Variable c_int
PRINTER_ATTRIBUTE_RAW_ONLY = 4096 # Variable c_int
TAPE_DRIVE_WRITE_FILEMARKS = 2181038080 # Variable c_uint
JOY_POVBACKWARD = 18000 # Variable c_int
MMSYSERR_WRITEERROR = 17 # Variable c_int
LR_COLOR = 2 # Variable c_int
DMPAPER_10X14 = 16 # Variable c_int
LCMAP_LOWERCASE = 256 # Variable c_int
AF_APPLETALK = 16 # Variable c_int
WAVECAPS_SYNC = 16 # Variable c_int
URLACTION_INFODELIVERY_NO_CHANNEL_LOGGING = 7430 # Variable c_int
LB_DELETESTRING = 386 # Variable c_int
WH_GETMESSAGE = 3 # Variable c_int
SORT_JAPANESE_UNICODE = 1 # Variable c_int
IME_ESC_QUERY_SUPPORT = 3 # Variable c_int
EXT_DEVICE_CAPS = 4099 # Variable c_int
MCIERR_NOTIFY_ON_AUTO_OPEN = 300 # Variable c_int
MK_E_MUSTBOTHERUSER = -2147221013 # Variable c_long
ERROR_BAD_IMPERSONATION_LEVEL = 1346 # Variable c_long
IMAGE_REL_IA64_ABSOLUTE = 0 # Variable c_int
HDM_SETITEMA = 4612 # Variable c_int
DMLERR_BUSY = 16385 # Variable c_int
HDF_RTLREADING = 4 # Variable c_int
CERTSRV_E_BAD_REQUESTSTATUS = -2146877437 # Variable c_long
BootLoad = 0
SPAPI_E_INVALID_DEVINST_NAME = -2146500091 # Variable c_long
DMPAPER_A3_EXTRA_TRANSVERSE = 68 # Variable c_int
MCI_STATUS_LENGTH = 1 # Variable c_long
PSD_ENABLEPAGESETUPTEMPLATE = 32768 # Variable c_int
TAPE_DRIVE_FIXED = 1 # Variable c_int
SPI_GETICONTITLELOGFONT = 31 # Variable c_int
SUBLANG_ENGLISH_NZ = 5 # Variable c_int
LVHT_ONITEMICON = 2 # Variable c_int
SP_APPABORT = -2 # Variable c_int
MCI_SEQ_STATUS_OFFSET = 16393 # Variable c_long
WS_SYSMENU = 524288 # Variable c_long
DM_COLOR = 2048 # Variable c_long
TIMERR_NOERROR = 0 # Variable c_int
LOCKFILE_EXCLUSIVE_LOCK = 2 # Variable c_int
TYPE_E_UNKNOWNLCID = -2147319762 # Variable c_long
XCLASS_FLAGS = 16384 # Variable c_int
WS_EX_TOPMOST = 8 # Variable c_long
RGN_XOR = 3 # Variable c_int
IMAGE_SCN_ALIGN_4096BYTES = 13631488 # Variable c_int
RPC_E_CANTCALLOUT_INASYNCCALL = -2147418108 # Variable c_long
RPC_S_NO_INTERFACES = 1817 # Variable c_long
PRINTER_NOTIFY_TYPE = 0 # Variable c_int
LVS_TYPEMASK = 3 # Variable c_int
SEE_MASK_ASYNCOK = 1048576 # Variable c_int
PRINTER_STATUS_BUSY = 512 # Variable c_int
LVM_GETSELECTIONMARK = 4162 # Variable c_int
LVA_ALIGNLEFT = 1 # Variable c_int
CTRY_COLOMBIA = 57 # Variable c_int
PDERR_DEFAULTDIFFERENT = 4108 # Variable c_int
MMSYSERR_KEYNOTFOUND = 15 # Variable c_int
TAPE_DRIVE_SET_ECC = 2147483904 # Variable c_uint
MIXERCONTROL_CONTROLTYPE_BOOLEAN = 536936448 # Variable c_long
RPC_S_INVALID_ASYNC_CALL = 1915 # Variable c_long
MIDIERR_STILLPLAYING = 65 # Variable c_int
LR_CREATEDIBSECTION = 8192 # Variable c_int
PROCESSOR_ARCHITECTURE_SHX = 4 # Variable c_int
MCI_SET_AUDIO_RIGHT = 2 # Variable c_long
ERROR_INVALID_GW_COMMAND = 1443 # Variable c_long
XST_UNADVSENT = 12 # Variable c_int
ISC_SHOWUICANDIDATEWINDOW = 1 # Variable c_int
SPI_GETFONTSMOOTHING = 74 # Variable c_int
LVS_OWNERDRAWFIXED = 1024 # Variable c_int
XMLELEMTYPE_PI = 5
CTRY_ISRAEL = 972 # Variable c_int
SUBLANG_SPANISH_MEXICAN = 2 # Variable c_int
PROPSETHDR_OSVERSION_UNKNOWN = 4294967295 # Variable c_uint
DMLERR_FIRST = 16384 # Variable c_int
WAVE_FORMAT_4M16 = 1024 # Variable c_int
PSINJECT_ENDSETUP = 8 # Variable c_int
MIDIPROP_TEMPO = 2 # Variable c_long
TB_SETCMDID = 1066 # Variable c_int
CSTR_GREATER_THAN = 3 # Variable c_int
IMAGE_FILE_MACHINE_R10000 = 360 # Variable c_int
CB_GETITEMDATA = 336 # Variable c_int
SID_RECOMMENDED_SUB_AUTHORITIES = 1 # Variable c_int
TAPE_DRIVE_RELATIVE_BLKS = 2147614720 # Variable c_uint
ERROR_PAGED_SYSTEM_RESOURCES = 1452 # Variable c_long
TRUST_E_PROVIDER_UNKNOWN = -2146762751 # Variable c_long
ERROR_INVALID_SPI_VALUE = 1439 # Variable c_long
SHGFI_ATTR_SPECIFIED = 131072 # Variable c_int
DISPLAY_DEVICE_MULTI_DRIVER = 2 # Variable c_int
TAPE_LOGICAL_POSITION = 1 # Variable c_long
DOMAIN_ALIAS_RID_ACCOUNT_OPS = 548 # Variable c_long
ERROR_INVALID_ACCEL_HANDLE = 1403 # Variable c_long
PRINTER_ERROR_OUTOFTONER = 4 # Variable c_int
VFT2_FONT_VECTOR = 2 # Variable c_long
VT_RECORD = 36
JOB_POSITION_UNSPECIFIED = 0 # Variable c_int
URLPOLICY_CHANNEL_SOFTDIST_AUTOINSTALL = 196608 # Variable c_int
CE_FRAME = 8 # Variable c_int
LBSELCHSTRINGW = 'commdlg_LBSelChangedNotify' # Variable POINTER(c_wchar)
rct1 = 1080 # Variable c_int
PARAMFLAG_FOUT = 2 # Variable c_int
INET_E_REDIRECT_TO_DIR = -2146697195 # Variable c_long
IMAGE_REL_ALPHA_GPREL32 = 3 # Variable c_int
COLOR_BTNSHADOW = 16 # Variable c_int
VTBIT_I4 = 8 # Variable c_int
FILE_READ_DATA = 1 # Variable c_int

tagDESCKIND = c_int # enum
DESCKIND_FUNCDESC = 1
DESCKIND_VARDESC = 2
DESCKIND_IMPLICITAPPOBJ = 4
DESCKIND_MAX = 5
PF_FLOATING_POINT_EMULATED = 1 # Variable c_int
MCI_ANIM_GETDEVCAPS_SLOW_RATE = 16387 # Variable c_long
SSF_INDICATOR = 4 # Variable c_int
SECURITY_LOGON_IDS_RID_COUNT = 3 # Variable c_long
SUBLANG_CHINESE_SIMPLIFIED = 2 # Variable c_int
CP_THREAD_ACP = 3 # Variable c_int
CC_SOLIDCOLOR = 128 # Variable c_int
GMDI_GOINTOPOPUPS = 2 # Variable c_long
CO_E_START_SERVICE_FAILURE = -2147467236 # Variable c_long
SO_DISCDATA = 28674 # Variable c_int
TF_REUSE_SOCKET = 2 # Variable c_int
STATUS_ACCESS_VIOLATION = 3221225477 # Variable c_ulong
TVIS_STATEIMAGEMASK = 61440 # Variable c_int
CLR_DEFAULT = 4278190080 # Variable c_ulong
JOB_OBJECT_UILIMIT_GLOBALATOMS = 32 # Variable c_int
NTM_ITALIC = 1 # Variable c_long
MCI_CDA_TRACK_OTHER = 1089 # Variable c_int
SPI_GETGRIDGRANULARITY = 18 # Variable c_int
RPC_S_INVALID_ENDPOINT_FORMAT = 1706 # Variable c_long
NDR_LITTLE_ENDIAN = 16 # Variable c_ulong
RESOURCETYPE_RESERVED = 8 # Variable c_int
NUMPRS_TRAILING_PLUS = 8 # Variable c_int
MCIERR_INTERNAL = 277 # Variable c_int
HELP_TCARD_OTHER_CALLER = 17 # Variable c_int
TB_GETITEMRECT = 1053 # Variable c_int
CTRY_SOUTH_KOREA = 82 # Variable c_int
NM_SETFOCUS = 4294967289 # Variable c_uint
RDW_ERASENOW = 512 # Variable c_int
FS_HEBREW = 32 # Variable c_long
BSM_APPLICATIONS = 8 # Variable c_int
PF_XMMI_INSTRUCTIONS_AVAILABLE = 6 # Variable c_int
CBEN_GETDISPINFOA = 4294966496 # Variable c_uint
HDN_BEGINDRAG = 4294966986 # Variable c_uint
DNS_FILTEROFF = 8 # Variable c_int
VIF_CANNOTRENAME = 8192 # Variable c_long
PAN_STROKE_GRADUAL_DIAG = 2 # Variable c_int
MCHT_CALENDARBK = 131072 # Variable c_int
MCIERR_WAVE_SETINPUTINUSE = 323 # Variable c_int
WC_COMPOSITECHECK = 512 # Variable c_int
RPC_C_MGMT_STOP_SERVER_LISTEN = 4 # Variable c_int
TVM_SETINSERTMARK = 4378 # Variable c_int
NTE_BAD_HASH = -2146893822 # Variable c_long
CT_CTYPE1 = 1 # Variable c_int
BALTIC_CHARSET = 186 # Variable c_int
HC_SYSMODALOFF = 5 # Variable c_int
TME_CANCEL = 2147483648 # Variable c_uint
SHGFI_EXETYPE = 8192 # Variable c_int
CERTSRV_E_NO_REQUEST = -2146877438 # Variable c_long
LVM_DELETEITEM = 4104 # Variable c_int
LVHITTESTINFO_V1_SIZE = 16 # Variable c_uint
HDN_ITEMDBLCLICKW = 4294966973 # Variable c_uint
ERROR_SERVICE_DOES_NOT_EXIST = 1060 # Variable c_long
PBT_APMQUERYSTANDBYFAILED = 3 # Variable c_int
ACM_OPENA = 1124 # Variable c_int
MMIO_ALLOCBUF = 65536 # Variable c_int
ERROR_INVALID_MEDIA_POOL = 4302 # Variable c_long
WAVE_FORMAT_PCM = 1 # Variable c_int
HWND_BROADCAST = 65535 # Variable POINTER(HWND__)
PROCESSOR_ARCHITECTURE_ALPHA64 = 7 # Variable c_int
NCBDGSENDBC = 34 # Variable c_int

tagLIBFLAGS = c_int # enum
LIBFLAG_FRESTRICTED = 1
LIBFLAG_FCONTROL = 2
HTERROR = -2 # Variable c_int
RBBIM_BACKGROUND = 128 # Variable c_int
MF_OWNERDRAW = 256 # Variable c_long
RASTERCAPS = 38 # Variable c_int
TBM_GETNUMTICS = 1040 # Variable c_int
FS_SYMBOL = 2147483648 # Variable c_ulong
MCIERR_WAVE_SETINPUTUNSUITABLE = 329 # Variable c_int
FACILITY_WIN32 = 7 # Variable c_int
STATUS_PRIVILEGED_INSTRUCTION = 3221225622 # Variable c_ulong
IME_ESC_GET_EUDC_DICTIONARY = 4099 # Variable c_int
TBNF_IMAGE = 1 # Variable c_int
WSAENAMETOOLONG = 10063 # Variable c_int
TB_BUTTONCOUNT = 1048 # Variable c_int
FR_DIALOGTERM = 64 # Variable c_int
SPAPI_E_NO_CLASS_DRIVER_LIST = -2146500072 # Variable c_long
SIZE_OF_80387_REGISTERS = 80 # Variable c_int
IDANI_CAPTION = 3 # Variable c_int
IME_CMODE_CHARCODE = 32 # Variable c_int
SZM_DELETE = 1
LVM_GETITEMCOUNT = 4100 # Variable c_int
MCI_MODE_PAUSE = 529 # Variable c_int
VREFRESH = 116 # Variable c_int
MCM_SETCURSEL = 4098 # Variable c_int
PST_UNSPECIFIED = 0 # Variable c_ulong
CB_GETDROPPEDSTATE = 343 # Variable c_int
SPLDS_PRINT_ORIENTATIONS_SUPPORTED = 'printOrientationsSupported' # Variable POINTER(c_wchar)
SECURITY_VALID_SQOS_FLAGS = 2031616 # Variable c_int
EMR_INTERSECTCLIPRECT = 30 # Variable c_int
WT_EXECUTEINUITHREAD = 2 # Variable c_int
SWP_NOSIZE = 1 # Variable c_int
BS_VCENTER = 3072 # Variable c_long
MMIO_CREATERIFF = 32 # Variable c_int
PSH_USEHICON = 2 # Variable c_int
SORT_GEORGIAN_TRADITIONAL = 0 # Variable c_int
InfFilePath = 4
WSAELOOP = 10062 # Variable c_int
LANG_TAMIL = 73 # Variable c_int
ERROR_DS_CANT_ON_RDN = 1935 # Variable c_long
ERROR_NOT_SUBSTED = 137 # Variable c_long
MM_MOM_DONE = 969 # Variable c_int
ERROR_INVALID_STACKSEG = 189 # Variable c_long
MSGF_COMMCTRL_BEGINDRAG = 16896 # Variable c_int
MCI_OVLY_GETDEVCAPS_MAX_WINDOWS = 16387 # Variable c_long
CDIS_HOT = 64 # Variable c_int
SM_SLOWMACHINE = 73 # Variable c_int
JOYSTICKID2 = 1 # Variable c_int
DV_E_STGMEDIUM = -2147221402 # Variable c_long
CO_E_SERVER_STOPPING = -2146959352 # Variable c_long
CRYPT_E_INVALID_X500_STRING = -2146885597 # Variable c_long
CLIPBRD_E_LAST = 2147746271 # Variable c_ulong
SPAPI_E_BAD_SECTION_NAME_LINE = -2146500607 # Variable c_long
SERVICE_RECOGNIZER_DRIVER = 8 # Variable c_int
WDT_INPROC_CALL = 1215587415 # Variable c_int
LVHT_NOWHERE = 1 # Variable c_int
SYSTEM_FLAG_DISKLESS_CLIENT = 2 # Variable c_int
SPI_SETMINIMIZEDMETRICS = 44 # Variable c_int
RB_SETPARENT = 1031 # Variable c_int
ERROR_NETNAME_DELETED = 64 # Variable c_long
MF_DELETE = 512 # Variable c_long
chx1 = 1040 # Variable c_int
PRINTER_NOTIFY_FIELD_PORT_NAME = 3 # Variable c_int
WH_CALLWNDPROC = 4 # Variable c_int
XST_UNADVACKRCVD = 14 # Variable c_int
FILE_GENERIC_READ = 1179785 # Variable c_long
SHIFT_PRESSED = 16 # Variable c_int
IMAGE_REL_PPC_IFGLUE = 13 # Variable c_int
LOCALE_SLONGDATE = 32 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC = 4101 # Variable c_long
CFSTR_MIME_X_REALAUDIO = 'audio/x-pn-realaudio' # Variable POINTER(c_wchar)
EMR_SETVIEWPORTORGEX = 12 # Variable c_int
PD_SHOWHELP = 2048 # Variable c_int
ERROR_NOT_SUPPORTED = 50 # Variable c_long
FILE_CASE_PRESERVED_NAMES = 2 # Variable c_int
JOB_OBJECT_UILIMIT_EXITWINDOWS = 128 # Variable c_int
LOCALE_IOPTIONALCALENDAR = 4107 # Variable c_int
chx7 = 1046 # Variable c_int
TRACKBAR_CLASSA = 'msctls_trackbar32' # Variable POINTER(c_char)
IE_OPEN = -2 # Variable c_int
CBEN_DELETEITEM = 4294966494 # Variable c_uint
WMSZ_BOTTOMLEFT = 7 # Variable c_int
WAVERR_LASTERROR = 35 # Variable c_int
SCALINGFACTORY = 115 # Variable c_int
DMPAPER_ENV_ITALY = 36 # Variable c_int
PRINTER_EXECUTE = 131080 # Variable c_long
PI_FORCE_ASYNC = 4
MIIM_TYPE = 16 # Variable c_int
WH_MAX = 12 # Variable c_int
MCI_FORMAT_SMPTE_30DROP = 7 # Variable c_int
REMOTE_NAME_INFO_LEVEL = 2 # Variable c_int
XST_POKEACKRCVD = 8 # Variable c_int
ERROR_BAD_DRIVER_LEVEL = 119 # Variable c_long
E_NOTIMPL = -2147467263 # Variable c_long
QS_ALLINPUT = 255 # Variable c_int
IPPORT_FINGER = 79 # Variable c_int
TYPE_E_CANTLOADLIBRARY = -2147312566 # Variable c_long
CDERR_NOHINSTANCE = 4 # Variable c_int
REFERENCE_WHITE_MAX = 10000 # Variable c_ushort
TVS_LINESATROOT = 4 # Variable c_int
INHERITED_ACE = 16 # Variable c_int
MIXERLINE_TARGETTYPE_AUX = 5 # Variable c_int
TYSPEC_TYPELIB = 5
IMAGE_SYM_TYPE_CHAR = 2 # Variable c_int
VIEW_PARENTFOLDER = 8 # Variable c_int
SRCPAINT = 15597702 # Variable c_ulong
TCS_MULTISELECT = 4 # Variable c_int
NM_KILLFOCUS = 4294967288 # Variable c_uint
IPPORT_NAMESERVER = 42 # Variable c_int
RBHT_CLIENT = 3 # Variable c_int
IGNORE = 0 # Variable c_int
ERROR_SEEK_ON_DEVICE = 132 # Variable c_long
TBIF_SIZE = 64 # Variable c_int
PSBTN_APPLYNOW = 4 # Variable c_int
TC_IA_ABLE = 1024 # Variable c_int
RPC_S_SERVER_UNAVAILABLE = 1722 # Variable c_long
ERROR_CANT_RESOLVE_FILENAME = 1921 # Variable c_long
MIDIERR_DONT_CONTINUE = 71 # Variable c_int
RPCFLG_LOCAL_CALL = 268435456 # Variable c_ulong
META_FLOODFILL = 1049 # Variable c_int
ERROR_ACCOUNT_LOCKED_OUT = 1909 # Variable c_long
CLASS_E_CLASSNOTAVAILABLE = -2147221231 # Variable c_long
BM_SETSTYLE = 244 # Variable c_int
WS_OVERLAPPEDWINDOW = 13565952 # Variable c_long
JOB_NOTIFY_FIELD_START_TIME = 17 # Variable c_int

__MIDL_IInternetZoneManager_0002 = c_int # enum
ZAFLAGS_CUSTOM_EDIT = 1
ZAFLAGS_ADD_SITES = 2
ZAFLAGS_REQUIRE_VERIFICATION = 4
ZAFLAGS_INCLUDE_PROXY_OVERRIDE = 8
ZAFLAGS_INCLUDE_INTRANET_SITES = 16
ZAFLAGS_SUPPORTS_VERIFICATION = 64
ZAFLAGS_UNC_AS_INTRANET = 128
IPPORT_DISCARD = 9 # Variable c_int
ERROR_CANNOT_COPY = 266 # Variable c_long
REGDB_E_LAST = 2147746143 # Variable c_ulong
SUBLANG_ENGLISH_US = 1 # Variable c_int
PST_PARALLELPORT = 2 # Variable c_ulong
CBEM_SETEXSTYLE = 1032 # Variable c_int
PFD_DRAW_TO_WINDOW = 4 # Variable c_int
BN_UNHILITE = 3 # Variable c_int
MEM_E_INVALID_ROOT = -2146959351 # Variable c_long
RPC_S_ADDRESS_ERROR = 1768 # Variable c_long
PAGE_EXECUTE = 16 # Variable c_int
HELPINFO_MENUITEM = 2 # Variable c_int
MCI_NOTIFY_SUCCESSFUL = 1 # Variable c_int
MH_CREATE = 1 # Variable c_int
MAXINT_PTR = 2147483647 # Variable c_long

VARENUM = c_int # enum
VT_NULL = 1
VT_I2 = 2
VT_I4 = 3
VT_R4 = 4
VT_R8 = 5
VT_CY = 6
VT_BSTR = 8
VT_DISPATCH = 9
VT_ERROR = 10
VT_BOOL = 11
VT_VARIANT = 12
VT_UNKNOWN = 13
VT_DECIMAL = 14
VT_I1 = 16
VT_UI1 = 17
VT_VOID = 24
VT_FILETIME = 64
VT_BLOB = 65
VT_STREAM = 66
VT_STORAGE = 67
VT_STORED_OBJECT = 69
VT_BLOB_OBJECT = 70
VT_CF = 71
VT_CLSID = 72
VT_VECTOR = 4096
VT_ARRAY = 8192
VT_BYREF = 16384
VT_RESERVED = 32768
VT_ILLEGAL = 65535
VT_ILLEGALMASKED = 4095
VT_TYPEMASK = 4095
ERROR_DEVICE_NOT_AVAILABLE = 4319 # Variable c_long
XST_ADVACKRCVD = 13 # Variable c_int
SPI_SETKEYBOARDSPEED = 11 # Variable c_int
DMPAPER_A4_TRANSVERSE = 55 # Variable c_int
PRINTER_STATUS_ERROR = 2 # Variable c_int
HIDE_WINDOW = 0 # Variable c_int
CC_ENABLEHOOK = 16 # Variable c_int
CONVERT10_E_FIRST = 2147746240 # Variable c_ulong
ERROR_USER_MAPPED_FILE = 1224 # Variable c_long
lst14 = 1133 # Variable c_int
LVN_FIRST = 4294967196 # Variable c_uint
IGP_PROPERTY = 4 # Variable c_int
SECURITY_LOCAL_SYSTEM_RID = 18 # Variable c_long
FF_DONTCARE = 0 # Variable c_int
RB_GETBANDCOUNT = 1036 # Variable c_int
LOCALE_INEGSEPBYSPACE = 87 # Variable c_int
IDI_ASTERISK = 206102 # Variable POINTER(c_wchar)
NONANTIALIASED_QUALITY = 3 # Variable c_int
TVM_GETITEMRECT = 4356 # Variable c_int
CTRL_CLOSE_EVENT = 2 # Variable c_int
PGK_MENU = 4 # Variable c_int
WHITEONBLACK = 2 # Variable c_int
PAN_BENT_ARMS_VERT = 9 # Variable c_int
RESETDEV = 7 # Variable c_int
FOF_SIMPLEPROGRESS = 256 # Variable c_int
OF_READ = 0 # Variable c_int
IDH_HELP = 28445 # Variable c_int
MM_WOM_OPEN = 955 # Variable c_int
SERVICE_STATE_ALL = 3 # Variable c_int
CO_E_MSI_ERROR = -2147467229 # Variable c_long
EASTEUROPE_CHARSET = 238 # Variable c_int
SBM_GETRANGE = 227 # Variable c_int
ERROR_BUSY = 170 # Variable c_long
STG_LAYOUT_INTERLEAVED = 1 # Variable c_long
TCS_FOCUSONBUTTONDOWN = 4096 # Variable c_int
TBN_QUERYDELETE = 4294966589 # Variable c_uint
VOS_OS232 = 196608 # Variable c_long
CLIPBRD_E_FIRST = 2147746256 # Variable c_ulong
ACTFLG_RunOnce = 2 # Variable c_int
DMLERR_LOW_MEMORY = 16391 # Variable c_int
EMR_POLYBEZIERTO16 = 88 # Variable c_int
CTRY_INDIA = 91 # Variable c_int
IO_COMPLETION_MODIFY_STATE = 2 # Variable c_int
CONNDLG_CONN_POINT = 2 # Variable c_int
SB_SIMPLE = 1033 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_TELEPHONE = 6 # Variable c_long
VAR_LOCALBOOL = 16 # Variable c_ulong
IMAGE_SCN_MEM_WRITE = 2147483648 # Variable c_uint
FILE_WRITE_ATTRIBUTES = 256 # Variable c_int
CERT_E_EXPIRED = -2146762495 # Variable c_long
GCS_RESULTREADSTR = 512 # Variable c_int
PAN_XHEIGHT_INDEX = 9 # Variable c_int
RPC_S_NOT_ALL_OBJS_UNEXPORTED = 1758 # Variable c_long
LVM_SETITEMPOSITION = 4111 # Variable c_int
CUDR_NOCLOSEGAPS = 4 # Variable c_int
TB_ADDSTRINGW = 1101 # Variable c_int
ERROR_SIGNAL_PENDING = 162 # Variable c_long
SPLDS_UNC_NAME = 'uNCName' # Variable POINTER(c_wchar)
IMAGE_SYM_CLASS_UNION_TAG = 12 # Variable c_int
PGN_CALCSIZE = 4294966394 # Variable c_uint
TCIF_IMAGE = 2 # Variable c_int
DFCS_SCROLLLEFT = 2 # Variable c_int
PSIDENT_GDICENTRIC = 0 # Variable c_int
JOYERR_BASE = 160 # Variable c_int
MDM_CCITT_OVERRIDE = 64 # Variable c_int
PRINTER_ENUM_NETWORK = 64 # Variable c_int
NRC_SABORT = 24 # Variable c_int
MF_APPEND = 256 # Variable c_long
SS_ICON = 3 # Variable c_long
MIXERLINE_COMPONENTTYPE_SRC_LINE = 4098 # Variable c_long
PRINTER_STATUS_POWER_SAVE = 16777216 # Variable c_int
MEM_DECOMMIT = 16384 # Variable c_int
WC_LISTVIEWA = 'SysListView32' # Variable POINTER(c_char)
SC_GROUP_IDENTIFIERW = '+' # Variable c_wchar
ODS_FOCUS = 16 # Variable c_int
ERROR_INVALID_FIELD = 1616 # Variable c_long
FNOINVERT = 2 # Variable c_int
RPC_C_STATS_PKTS_OUT = 3 # Variable c_int
IN_CLASSB_NET = 4294901760 # Variable c_uint
LANG_SANSKRIT = 79 # Variable c_int
PSD_DISABLEPAGEPAINTING = 524288 # Variable c_int
SUBLANG_SPANISH_EL_SALVADOR = 17 # Variable c_int
SPAPI_E_INVALID_CLASS_INSTALLER = -2146500083 # Variable c_long
MCI_ANIM_RECT = 65536 # Variable c_long
CBR_14400 = 14400 # Variable c_int
ERROR_NOT_REGISTRY_FILE = 1017 # Variable c_long
ERROR_INVALID_PASSWORD = 86 # Variable c_long
ERROR_INSTALL_SERVICE = 1601 # Variable c_long
ERROR_INSTALL_SOURCE_ABSENT = 1612 # Variable c_long
BATTERY_FLAG_UNKNOWN = 255 # Variable c_int
S_SERDVL = -9 # Variable c_int
DISPID_PROPERTYPUT = -3 # Variable c_int
TB_SETSTATE = 1041 # Variable c_int
RESOURCETYPE_PRINT = 2 # Variable c_int
DMPAPER_JAPANESE_POSTCARD = 43 # Variable c_int
IMAGE_COMDAT_SELECT_ASSOCIATIVE = 5 # Variable c_int
LB_GETCARETINDEX = 415 # Variable c_int
ICC_HOTKEY_CLASS = 64 # Variable c_int
CC_PIE = 2 # Variable c_int
TBSTYLE_NOPREFIX = 32 # Variable c_int
MCIERR_DEVICE_LOCKED = 288 # Variable c_int
IMAGE_COMDAT_SELECT_EXACT_MATCH = 4 # Variable c_int
MEM_PRIVATE = 131072 # Variable c_int
SE_DACL_PROTECTED = 4096 # Variable c_int
IMAGE_REL_I386_SECREL = 11 # Variable c_int
PBM_STEPIT = 1029 # Variable c_int
CDM_SETDEFEXT = 1130 # Variable c_int
SB_PAGERIGHT = 3 # Variable c_int
WEOF = '\uffff' # Variable c_wchar
MOD_CONTROL = 2 # Variable c_int
CREATE_NEW_PROCESS_GROUP = 512 # Variable c_int
STG_E_LOCKVIOLATION = -2147287007 # Variable c_long
IOC_INOUT = 3221225472 # Variable c_uint
TAPE_DRIVE_TENSION_IMMED = 2147483712 # Variable c_uint
QDI_DIBTOSCREEN = 4 # Variable c_int
MCI_SAVE = 2067 # Variable c_int
CURVECAPS = 28 # Variable c_int
IMN_PRIVATE = 14 # Variable c_int
NCBSENDNA = 113 # Variable c_int
VFT_APP = 1 # Variable c_long
ERROR_BAD_DEVICE = 1200 # Variable c_long
COLOR_MENUTEXT = 7 # Variable c_int
TA_RTLREADING = 256 # Variable c_int
CERTSRV_E_PROPERTY_EMPTY = -2146877436 # Variable c_long
ETO_GLYPH_INDEX = 16 # Variable c_int

tagSYSKIND = c_int # enum
SYS_WIN16 = 0
SYS_WIN32 = 1
SYS_MAC = 2
STATUS_ABANDONED_WAIT_0 = 128 # Variable c_ulong
RPC_E_CLIENT_DIED = -2147418104 # Variable c_long
NLS_ALPHANUMERIC = 0 # Variable c_int
PERF_OBJECT_TIMER = 2097152 # Variable c_int
CTRL_C_EVENT = 0 # Variable c_int
WS_MAXIMIZE = 16777216 # Variable c_long
cmb15 = 1150 # Variable c_int
LOGON32_LOGON_NETWORK = 3 # Variable c_int
JOY_BUTTON19 = 262144 # Variable c_long
STDOLE2_LCID = 0 # Variable c_int
VAR_VALIDDATE = 4 # Variable c_ulong
HKEY_DYN_DATA = 2147483654 # Variable POINTER(HKEY__)
CO_E_INIT_SCM_MAP_VIEW_OF_FILE = -2147467248 # Variable c_long
PD_DISABLEPRINTTOFILE = 524288 # Variable c_int
JOY_POVCENTERED = 65535 # Variable c_ushort
OBJ_EXTPEN = 11 # Variable c_int
COLOR_WINDOW = 5 # Variable c_int
SPAPI_E_DUPLICATE_FOUND = -2146500094 # Variable c_long
GHND = 66 # Variable c_int
STATUS_INVALID_HANDLE = 3221225480 # Variable c_ulong
SM_CXMAXIMIZED = 61 # Variable c_int
LB_GETITEMRECT = 408 # Variable c_int
CONNECT_RESERVED = 4278190080 # Variable c_uint
VER_PLATFORM_WIN32_NT = 2 # Variable c_int
PRINTER_STATUS_INITIALIZING = 32768 # Variable c_int
SM_SWAPBUTTON = 23 # Variable c_int
CERT_E_VALIDITYPERIODNESTING = -2146762494 # Variable c_long
FSB_REGULAR_MODE = 0 # Variable c_int
ANSI_FIXED_FONT = 11 # Variable c_int
MCI_WAVE_SET_SAMPLESPERSEC = 262144 # Variable c_long
THREAD_PRIORITY_NORMAL = 0 # Variable c_int
WINSTA_ACCESSGLOBALATOMS = 32 # Variable c_long
MCI_DEVTYPE_DAT = 517 # Variable c_int
PRINTER_STATUS_NOT_AVAILABLE = 4096 # Variable c_int
ERROR_SUCCESS = 0 # Variable c_long
MK_E_NOINVERSE = -2147221012 # Variable c_long
PD_HIDEPRINTTOFILE = 1048576 # Variable c_int
RPC_C_OPT_MQ_TIME_TO_REACH_QUEUE = 7 # Variable c_int
UNICODE = 1 # Variable c_int
GDI_ERROR = 4294967295 # Variable c_ulong
LVS_EX_ONECLICKACTIVATE = 64 # Variable c_int
SS_RIGHT = 2 # Variable c_long
CCHFORMNAME = 32 # Variable c_int

__MIDL_IInternetSecurityManager_0001 = c_int # enum
PUAF_DEFAULT = 0
PUAF_NOUI = 1
PUAF_ISFILE = 2
PUAF_WARN_IF_DENIED = 4
PUAF_FORCEUI_FOREGROUND = 8
PUAF_CHECK_TIFS = 16
ERROR_DEPENDENT_SERVICES_RUNNING = 1051 # Variable c_long
STATUS_ILLEGAL_INSTRUCTION = 3221225501 # Variable c_ulong
ERROR_UNKNOWN_PRINTER_DRIVER = 1797 # Variable c_long
PAN_STROKE_RAPID_VERT = 6 # Variable c_int
LMEM_FIXED = 0 # Variable c_int
TCM_REMOVEIMAGE = 4906 # Variable c_int
PM_REMOVE = 1 # Variable c_int
FONTMAPPER_MAX = 10 # Variable c_int
RIGHT_ALT_PRESSED = 1 # Variable c_int
DrwFilePath = 5
RPC_E_INVALIDMETHOD = -2147417849 # Variable c_long
ATTR_TARGET_CONVERTED = 1 # Variable c_int
UPDFCACHE_ONSAVECACHE = 2 # Variable c_int
SORT_HUNGARIAN_DEFAULT = 0 # Variable c_int
FIONBIO = 2147772030 # Variable c_ulong
TRUST_E_BASIC_CONSTRAINTS = -2146869223 # Variable c_long
stc19 = 1106 # Variable c_int
PAN_SERIF_TRIANGLE = 10 # Variable c_int
RB_ENDDRAG = 1049 # Variable c_int
ERROR_PASSWORD_EXPIRED = 1330 # Variable c_long
EV_TXEMPTY = 4 # Variable c_int
MENU_EVENT = 8 # Variable c_int
PROCESS_CREATE_THREAD = 2 # Variable c_int
ERROR_SHARING_PAUSED = 70 # Variable c_long
SE_UNSOLICITED_INPUT_NAME = 'SeUnsolicitedInputPrivilege' # Variable POINTER(c_wchar)
TBSTYLE_FLAT = 2048 # Variable c_int

__MIDL_IInternetSecurityManager_0002 = c_int # enum
JOB_NOTIFY_FIELD_UNTIL_TIME = 18 # Variable c_int
E_ABORT = -2147467260 # Variable c_long
TB_MOVEBUTTON = 1106 # Variable c_int
SM_CYICON = 12 # Variable c_int
LVM_FINDITEMA = 4109 # Variable c_int
FINDMSGSTRINGA = 'commdlg_FindReplace' # Variable POINTER(c_char)
DMICM_USER = 256 # Variable c_int
FileSystemType = 2
PSH_USEHPLWATERMARK = 131072 # Variable c_int
ERROR_SECRET_TOO_LONG = 1382 # Variable c_long
REFERENCE_BLACK_MAX = 4000 # Variable c_ushort
SERVICE_CONTROL_PARAMCHANGE = 6 # Variable c_int
TAPE_DRIVE_ERASE_SHORT = 16 # Variable c_int
EMR_EXTTEXTOUTW = 84 # Variable c_int
LOCALE_SNATIVEDIGITS = 19 # Variable c_int
ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE = 5033 # Variable c_long
TVIF_CHILDREN = 64 # Variable c_int
DMBIN_CASSETTE = 14 # Variable c_int
VS_VERSION_INFO = 1 # Variable c_int
RPC_S_NO_ENTRY_NAME = 1735 # Variable c_long
CDDS_ITEMPREERASE = 65539 # Variable c_int
IDH_MISSING_CONTEXT = 28441 # Variable c_int
INET_E_REDIRECTING = -2146697196 # Variable c_long
IO_REPARSE_TAG_VALID_VALUES = 3758161919 # Variable c_uint
LPD_SHARE_DEPTH = 64 # Variable c_int
SCALINGFACTORX = 114 # Variable c_int
PSBTN_HELP = 6 # Variable c_int
DMDISPLAYFLAGS_TEXTMODE = 4 # Variable c_int
RGN_AND = 1 # Variable c_int
PI_USE_WORKERTHREAD = 8
WS_MINIMIZEBOX = 131072 # Variable c_long
CREATE_NO_WINDOW = 134217728 # Variable c_int
CTRY_UNITED_KINGDOM = 44 # Variable c_int
IMAGE_SYM_ABSOLUTE = -1 # Variable c_short
SWP_NOACTIVATE = 16 # Variable c_int
BF_BOTTOM = 8 # Variable c_int
NRC_INVADDRESS = 57 # Variable c_int
SC_PREVWINDOW = 61520 # Variable c_int
DMPAPER_A4SMALL = 10 # Variable c_int
LBS_MULTIPLESEL = 8 # Variable c_long
SOFTDIST_ADSTATE_NONE = 0 # Variable c_int
LTGRAY_BRUSH = 1 # Variable c_int
DMPAPER_B4 = 12 # Variable c_int
ENDSESSION_LOGOFF = 2147483648 # Variable c_uint
COMPRESSION_FORMAT_DEFAULT = 1 # Variable c_int
PSBTN_BACK = 0 # Variable c_int
ERROR_TOO_MANY_SESS = 69 # Variable c_long
IDC_ARROW = 206098 # Variable POINTER(c_wchar)
IMAGE_REL_PPC_ADDR32NB = 10 # Variable c_int
IME_CMODE_ALPHANUMERIC = 0 # Variable c_int
LOCALE_SMONTHNAME6 = 61 # Variable c_int
CAL_SMONTHNAME7 = 27 # Variable c_int
ERROR_REC_NON_EXISTENT = 4005 # Variable c_long
EMR_CREATECOLORSPACE = 99 # Variable c_int
PRINTER_ATTRIBUTE_ENABLE_BIDI = 2048 # Variable c_int
VOS__PM16 = 2 # Variable c_long
RPC_C_CANCEL_INFINITE_TIMEOUT = -1 # Variable c_int
RDW_NOFRAME = 2048 # Variable c_int
STG_E_TERMINATED = -2147286526 # Variable c_long
RBN_FIRST = 4294966465 # Variable c_uint
NTE_EXISTS = -2146893809 # Variable c_long
RB_SIZETORECT = 1047 # Variable c_int
ERROR_NOT_ENOUGH_QUOTA = 1816 # Variable c_long
DTM_GETMONTHCAL = 4104 # Variable c_int
STD_FILENEW = 6 # Variable c_int
FO_COPY = 2 # Variable c_int
MK_E_LAST = 2147746287 # Variable c_ulong
HDS_FULLDRAG = 128 # Variable c_int
MMSYSERR_VALNOTFOUND = 19 # Variable c_int
GGO_BITMAP = 1 # Variable c_int
ABE_RIGHT = 2 # Variable c_int
PAN_WEIGHT_LIGHT = 3 # Variable c_int
CB_DELETESTRING = 324 # Variable c_int

LATENCY_TIME = c_int # enum
LT_DONT_CARE = 0
LT_LOWEST_LATENCY = 1
TBN_TOOLBARCHANGE = 4294966588 # Variable c_uint
CTRY_AUSTRIA = 43 # Variable c_int
LOCALE_SMONDECIMALSEP = 22 # Variable c_int
CREATE_SHARED_WOW_VDM = 4096 # Variable c_int
IMAGE_SYM_TYPE_FLOAT = 6 # Variable c_int
IMAGE_OS2_SIGNATURE_LE = 17740 # Variable c_int
COLOR_ACTIVEBORDER = 10 # Variable c_int
R2_NOT = 6 # Variable c_int
MFCOMMENT = 15 # Variable c_int
CTRY_BRUNEI_DARUSSALAM = 673 # Variable c_int
WNNC_NET_MSNET = 65536 # Variable c_int
ERROR_NO_MATCH = 1169 # Variable c_long
IMAGE_RESOURCE_NAME_IS_STRING = 2147483648 # Variable c_uint
TTM_SETTOOLINFOW = 1078 # Variable c_int
SBM_ENABLE_ARROWS = 228 # Variable c_int
SM_CMETRICS = 76 # Variable c_int
EMR_LINETO = 54 # Variable c_int
EN_UPDATE = 1024 # Variable c_int
EM_LIMITTEXT = 197 # Variable c_int
MCI_VD_STEP_REVERSE = 131072 # Variable c_long
PSH_WIZARD97 = 8192 # Variable c_int
JOY_BUTTON4CHG = 2048 # Variable c_int
PFD_STEREO = 2 # Variable c_int
PO_REN_PORT = 52 # Variable c_int
TBSTYLE_DROPDOWN = 8 # Variable c_int
FLI_GLYPHS = 262144 # Variable c_long
ERROR_RPL_NOT_ALLOWED = 4006 # Variable c_long
PRINTER_NOTIFY_FIELD_STATUS = 18 # Variable c_int
LV_MAX_WORKAREAS = 16 # Variable c_int
PAN_LETT_OBLIQUE_ROUNDED = 13 # Variable c_int
TTM_NEWTOOLRECTW = 1076 # Variable c_int
SM_CXSMICON = 49 # Variable c_int
CBENF_KILLFOCUS = 1 # Variable c_int
LB_SETITEMHEIGHT = 416 # Variable c_int
NTE_BAD_ALGID = -2146893816 # Variable c_long
UDS_ARROWKEYS = 32 # Variable c_int
PROCESSOR_ALPHA_21064 = 21064 # Variable c_int
MSGF_MESSAGEBOX = 1 # Variable c_int
CRYPT_E_ATTRIBUTES_MISSING = -2146889713 # Variable c_long
LBS_EXTENDEDSEL = 2048 # Variable c_long
DRAFTMODE = 7 # Variable c_int
RPC_IF_AUTOLISTEN = 1 # Variable c_int
CTRY_ARGENTINA = 54 # Variable c_int
NRC_INUSE = 22 # Variable c_int
ERROR_SPL_NO_STARTDOC = 3003 # Variable c_long
CCS_BOTTOM = 3 # Variable c_long
SUBLANG_SPANISH_ARGENTINA = 11 # Variable c_int
RBBIM_LPARAM = 1024 # Variable c_int
ST_BLOCKNEXT = 128 # Variable c_int
PSINJECT_PSADOBE = 1 # Variable c_int
ERROR_GEN_FAILURE = 31 # Variable c_long
DC_MODEL = 24 # Variable c_int
LMEM_NOCOMPACT = 16 # Variable c_int
HALFTONE = 4 # Variable c_int
TIMER_MODIFY_STATE = 2 # Variable c_int
SS_ENDELLIPSIS = 16384 # Variable c_long
STG_E_SEEKERROR = -2147287015 # Variable c_long
ANSI_CHARSET = 0 # Variable c_int
PRINTER_ATTRIBUTE_QUEUED = 1 # Variable c_int
STG_E_INVALIDFUNCTION = -2147287039 # Variable c_long
SPI_SETLOWPOWERACTIVE = 85 # Variable c_int
TKF_AVAILABLE = 2 # Variable c_int
LR_DEFAULTCOLOR = 0 # Variable c_int
RB_MINIMIZEBAND = 1054 # Variable c_int
PGF_GRAYED = 2 # Variable c_int
MMIO_DEFAULTBUFFER = 8192 # Variable c_int
PIPE_ACCESS_INBOUND = 1 # Variable c_int
LVN_ENDLABELEDITW = 4294967120 # Variable c_uint
DOMAIN_GROUP_RID_ADMINS = 512 # Variable c_long
ERROR_NESTING_NOT_ALLOWED = 215 # Variable c_long
RTS_CONTROL_HANDSHAKE = 2 # Variable c_int
ACCESS_MIN_MS_ACE_TYPE = 0 # Variable c_int
BINDSTRING_ACCEPT_MIMES = 2
LVS_OWNERDATA = 4096 # Variable c_int
URLACTION_SHELL_FILE_DOWNLOAD = 6147 # Variable c_int
COLOR_WINDOWFRAME = 6 # Variable c_int
APPCMD_MASK = 4080 # Variable c_long
LZERROR_UNKNOWNALG = -8 # Variable c_int
FOURCC_MEM = 541934925 # Variable c_ulong
CERT_E_ISSUERCHAINING = -2146762489 # Variable c_long
OFN_EXPLORER = 524288 # Variable c_int
TBSTYLE_CUSTOMERASE = 8192 # Variable c_int
IE_HARDWARE = -10 # Variable c_int
MDMSPKRFLAG_DIAL = 2 # Variable c_int
GCP_MAXEXTENT = 1048576 # Variable c_long
VIF_SHARINGVIOLATION = 1024 # Variable c_long
PAN_SERIF_COVE = 2 # Variable c_int
VIEW_SORTDATE = 6 # Variable c_int
LOCALE_SLANGUAGE = 2 # Variable c_int
WIN_CERT_REVISION_1_0 = 256 # Variable c_int
URLACTION_SCRIPT_MIN = 5120 # Variable c_int
LVHT_BELOW = 16 # Variable c_int
CO_E_OLE1DDE_DISABLED = -2147467242 # Variable c_long
LVM_GETTEXTCOLOR = 4131 # Variable c_int
IME_ESC_SEQUENCE_TO_INTERNAL = 4097 # Variable c_int
LANG_ITALIAN = 16 # Variable c_int
IDB_STD_SMALL_COLOR = 0 # Variable c_int
IMAGE_COMDAT_SELECT_NODUPLICATES = 1 # Variable c_int
C3_HALFWIDTH = 64 # Variable c_int
LVM_SETCALLBACKMASK = 4107 # Variable c_int
DRV_EXITSESSION = 11 # Variable c_int
TTM_GETTIPTEXTCOLOR = 1047 # Variable c_int
HDN_ITEMCHANGEDW = 4294966975 # Variable c_uint
NRC_NOCALL = 20 # Variable c_int
WC_NATIVEFONTCTLW = 'NativeFontCtl' # Variable POINTER(c_wchar)
TCN_SELCHANGE = 4294966745 # Variable c_uint
ERROR_WRONG_PASSWORD = 1323 # Variable c_long
NCBDELNAME = 49 # Variable c_int
QS_INPUT = 7 # Variable c_int
RBN_AUTOSIZE = 4294966462 # Variable c_uint
SM_MIDEASTENABLED = 74 # Variable c_int
PERF_DELTA_BASE = 8388608 # Variable c_int
S_PERIOD2048 = 2 # Variable c_int
ARW_STARTRIGHT = 1 # Variable c_long
RPC_C_PROTSEQ_MAX_REQS_DEFAULT = 10 # Variable c_int
WS_EX_APPWINDOW = 262144 # Variable c_long
SCS_SETRECONVERTSTRING = 65536 # Variable c_int
MIDI_UNCACHE = 4 # Variable c_int
JOY_BUTTON11 = 1024 # Variable c_long
ERROR_IS_JOIN_TARGET = 133 # Variable c_long
WH_MIN = -1 # Variable c_int
ACCESS_MAX_MS_V4_ACE_TYPE = 8 # Variable c_int
TCHT_ONITEMLABEL = 4 # Variable c_int
LR_LOADMAP3DCOLORS = 4096 # Variable c_int
PROP_MED_CXDLG = 227 # Variable c_int
GDICOMMENT_BEGINGROUP = 2 # Variable c_int
IDH_GENERIC_HELP_BUTTON = 28442 # Variable c_int
NRC_INCOMP = 6 # Variable c_int
ERROR_NOT_CONTAINER = 1207 # Variable c_long
SIMULATED_FONTTYPE = 32768 # Variable c_int
DLGC_HASSETSEL = 8 # Variable c_int
JOY_CAL_READRONLY = 33554432 # Variable c_long
SPLDS_DRIVER_VERSION = 'driverVersion' # Variable POINTER(c_wchar)
CP_NONE = 0 # Variable c_int
MCI_LOAD_FILE = 256 # Variable c_long
PSINJECT_SHOWPAGE = 12 # Variable c_int
TAPE_INITIATOR_PARTITIONS = 2 # Variable c_long
HKM_SETHOTKEY = 1025 # Variable c_int
EMR_POLYDRAW16 = 92 # Variable c_int
PS_ENDCAP_SQUARE = 256 # Variable c_int
XTYP_REGISTER = 32930 # Variable c_int
LOCALE_ITIMEMARKPOSN = 4101 # Variable c_int
META_SETTEXTALIGN = 302 # Variable c_int
GCL_REVERSECONVERSION = 2 # Variable c_int
_OUT_TO_DEFAULT = 0 # Variable c_int
SW_HIDE = 0 # Variable c_int
CTRY_LEBANON = 961 # Variable c_int
PAN_PROP_OLD_STYLE = 2 # Variable c_int
PAN_STROKEVARIATION_INDEX = 5 # Variable c_int
LPD_DOUBLEBUFFER = 1 # Variable c_int
SPI_GETWORKAREA = 48 # Variable c_int
WSB_PROP_CYHSCROLL = 4 # Variable c_long
RPC_S_NO_PRINC_NAME = 1822 # Variable c_long
SERVER_READ = 131074 # Variable c_long
SB_GETTEXTLENGTHA = 1027 # Variable c_int
PFD_DOUBLEBUFFER_DONTCARE = 1073741824 # Variable c_int
LPTx = 128 # Variable c_int
ERROR_NETLOGON_NOT_STARTED = 1792 # Variable c_long
CCS_VERT = 128 # Variable c_long
SPI_GETFILTERKEYS = 50 # Variable c_int
WGL_SWAP_UNDERLAY10 = 33554432 # Variable c_int

tagDCOM_CALL_STATE = c_int # enum
CTRY_VENEZUELA = 58 # Variable c_int
ERROR_RESOURCE_NOT_PRESENT = 4316 # Variable c_long
PSN_QUERYCANCEL = 4294967087 # Variable c_uint
IMAGE_SCN_ALIGN_8BYTES = 4194304 # Variable c_int
IMAGE_COMDAT_SELECT_ANY = 2 # Variable c_int
CREATE_FOR_IMPORT = 1 # Variable c_int
ERROR_PATH_NOT_FOUND = 3 # Variable c_long
PORT_STATUS_NO_TONER = 6 # Variable c_int
ERROR_MR_MID_NOT_FOUND = 317 # Variable c_long
UPDOWN_CLASSA = 'msctls_updown32' # Variable POINTER(c_char)
PFD_GENERIC_ACCELERATED = 4096 # Variable c_int
IS_TEXT_UNICODE_ILLEGAL_CHARS = 256 # Variable c_int
IMAGE_FILE_SYSTEM = 4096 # Variable c_int
MIXER_SHORT_NAME_CHARS = 16 # Variable c_int
CO_E_CANT_REMOTE = -2147467245 # Variable c_long
BF_BOTTOMRIGHT = 12 # Variable c_int
RBS_AUTOSIZE = 8192 # Variable c_int
CTRY_HUNGARY = 36 # Variable c_int
GGL_STRING = 3 # Variable c_int
DNS_FILTERON = 4 # Variable c_int
IMAGE_REL_IA64_SECREL64I = 13 # Variable c_int
ERROR_UNKNOWN_FEATURE = 1606 # Variable c_long
LCID_SUPPORTED = 2 # Variable c_int
TIME_NOTIMEMARKER = 4 # Variable c_int
DISPLAY_DEVICE_VGA_COMPATIBLE = 16 # Variable c_int
WGL_SWAP_UNDERLAY3 = 262144 # Variable c_int
PAN_XHEIGHT_DUCKING_SMALL = 5 # Variable c_int
VOS_OS216_PM16 = 131074 # Variable c_long
MCI_FORMAT_SMPTE_30 = 6 # Variable c_int
EMR_PIXELFORMAT = 104 # Variable c_int
TVC_UNKNOWN = 0 # Variable c_int
SBS_LEFTALIGN = 2 # Variable c_long
SetupNamePath = 6
MEM_RELEASE = 32768 # Variable c_int
WAIT_OBJECT_0 = 0 # Variable c_ulong
IOC_VOID = 536870912 # Variable c_int
PARAMFLAG_FIN = 1 # Variable c_int
MCI_ANIM_WINDOW_TEXT = 524288 # Variable c_long
STG_TOEND = 4294967295 # Variable c_ulong
SW_MINIMIZE = 6 # Variable c_int
CACHE_E_FIRST = 2147746160 # Variable c_ulong
IMAGE_REL_ARM_SECTION = 14 # Variable c_int
CE_TXFULL = 256 # Variable c_int
FILE_ATTRIBUTE_REPARSE_POINT = 1024 # Variable c_int
MOUSE_EVENT = 2 # Variable c_int
_WIN32_IE = 1024 # Variable c_int
CBS_OWNERDRAWFIXED = 16 # Variable c_long
LVSICF_NOINVALIDATEALL = 1 # Variable c_int
ERROR_ALREADY_WAITING = 1904 # Variable c_long
APPLICATION_ERROR_MASK = 536870912 # Variable c_int
CO_E_RELEASED = -2147220993 # Variable c_long
GCL_HMODULE = -16 # Variable c_int
SS_BLACKRECT = 4 # Variable c_long
SE_SACL_PRESENT = 16 # Variable c_int
TCS_FOCUSNEVER = 32768 # Variable c_int
DISPID_CONSTRUCTOR = -6 # Variable c_int
SPI_GETLOWPOWERTIMEOUT = 79 # Variable c_int
MIXER_OBJECTF_HMIDIOUT = 2952790016 # Variable c_ulong
LVS_EX_HEADERDRAGDROP = 16 # Variable c_int
ERROR_INVALID_HANDLE = 6 # Variable c_long
PERF_COUNTER_LARGE_RAWCOUNT = 65792 # Variable c_int
RPC_C_IMP_LEVEL_IMPERSONATE = 3 # Variable c_int
NCBRECVANY = 22 # Variable c_int
RPC_C_MQ_AUTHN_LEVEL_NONE = 0 # Variable c_int
HDN_ITEMCHANGEDA = 4294966995 # Variable c_uint
ERROR_INVALID_VERIFY_SWITCH = 118 # Variable c_long
RPC_E_SERVER_DIED_DNE = -2147418094 # Variable c_long
cbNDRContext = 20 # Variable c_int
ACCESS_MAX_MS_V3_ACE_TYPE = 4 # Variable c_int
PURGE_TXCLEAR = 4 # Variable c_int
WS_MAXIMIZEBOX = 65536 # Variable c_long
TAPE_DRIVE_EOT_WZ_SIZE = 8192 # Variable c_int
PWR_FAIL = -1 # Variable c_int
WC_TABCONTROLW = 'SysTabControl32' # Variable POINTER(c_wchar)
NTE_BAD_SIGNATURE = -2146893818 # Variable c_long
LPTR = 64 # Variable c_int
HCBT_MINMAX = 1 # Variable c_int
ESB_DISABLE_RIGHT = 2 # Variable c_int
VIF_BUFFTOOSMALL = 262144 # Variable c_long
CFSTR_MIME_PDF = 'application/pdf' # Variable POINTER(c_wchar)
NTM_BOLD = 32 # Variable c_long
PI_MIMEVERIFICATION = 16
MCHT_PREV = 33554432 # Variable c_int
PAN_LETT_OBLIQUE_CONTACT = 9 # Variable c_int
LOCK_WRITE = 1
RBBIM_SIZE = 64 # Variable c_int
EMR_CREATEBRUSHINDIRECT = 39 # Variable c_int
PFD_SUPPORT_GDI = 16 # Variable c_int
MDM_FORCED_EC = 4 # Variable c_int
RDW_UPDATENOW = 256 # Variable c_int
LR_COPYRETURNORG = 4 # Variable c_int
GMEM_SHARE = 8192 # Variable c_int
RPC_S_INVALID_RPC_PROTSEQ = 1704 # Variable c_long
MAXIMUM_REPARSE_DATA_BUFFER_SIZE = 16384 # Variable c_int
MCI_SEEK = 2055 # Variable c_int
PID_MODIFY_TIME = 2147483649 # Variable c_uint
PROCESSOR_ARCHITECTURE_ALPHA = 2 # Variable c_int
LVM_SETSELECTIONMARK = 4163 # Variable c_int
MIXER_SETCONTROLDETAILSF_CUSTOM = 1 # Variable c_long
EMARCH_ENC_I17_IC_INST_WORD_POS_X = 12 # Variable c_int
SB_SETPARTS = 1028 # Variable c_int
CBN_SELCHANGE = 1 # Variable c_int
OLEOBJ_E_INVALIDVERB = -2147221119 # Variable c_long
CFSTR_MIME_AVI = 'video/avi' # Variable POINTER(c_wchar)
URLACTION_ACTIVEX_MAX = 5119 # Variable c_int
PD_ENABLESETUPHOOK = 8192 # Variable c_int
CBN_KILLFOCUS = 4 # Variable c_int

tagINVOKEKIND = c_int # enum
INVOKE_FUNC = 1
INVOKE_PROPERTYGET = 2
INVOKE_PROPERTYPUT = 4
INVOKE_PROPERTYPUTREF = 8
FILE_FLAG_RANDOM_ACCESS = 268435456 # Variable c_int
XST_REQSENT = 5 # Variable c_int
LOCALE_SMONTHNAME7 = 62 # Variable c_int
MKSYS_URLMONIKER = 6 # Variable c_int
TBSTYLE_REGISTERDROP = 16384 # Variable c_int
MCI_BREAK = 2065 # Variable c_int
MMIOERR_UNBUFFERED = 266 # Variable c_int
ERROR_SERVICE_NOT_FOUND = 1243 # Variable c_long
TB_PAGEDOWN = 3 # Variable c_int
ERROR_SERVER_DISABLED = 1341 # Variable c_long
rad13 = 1068 # Variable c_int
RBNM_STYLE = 2 # Variable c_int
DFCS_BUTTONRADIOIMAGE = 1 # Variable c_int
IMAGE_REL_IA64_PCREL21M = 7 # Variable c_int
NM_HOVER = 4294967283 # Variable c_uint
SPI_SETLOWPOWERTIMEOUT = 81 # Variable c_int
LANG_SERBIAN = 26 # Variable c_int
BATTERY_PERCENTAGE_UNKNOWN = 255 # Variable c_int
QS_MOUSE = 6 # Variable c_int
CRYPT_E_UNEXPECTED_ENCODING = -2146889723 # Variable c_long
ERROR_WINS_INTERNAL = 4000 # Variable c_long
MCHT_CALENDARDATENEXT = 16908289 # Variable c_int
INPLACE_S_FIRST = 262560 # Variable c_long
DL_COPYCURSOR = 2 # Variable c_int
MDM_FLOWCONTROL_HARD = 16 # Variable c_int
SND_LOOP = 8 # Variable c_int
SE_ERR_DLLNOTFOUND = 32 # Variable c_int
LVFI_PARTIAL = 8 # Variable c_int
DISP_E_PARAMNOTOPTIONAL = -2147352561 # Variable c_long
CDRF_SKIPDEFAULT = 4 # Variable c_int
GCS_COMPSTR = 8 # Variable c_int
CB_ERR = -1 # Variable c_int
IMAGE_DEBUG_TYPE_UNKNOWN = 0 # Variable c_int
HCBT_ACTIVATE = 5 # Variable c_int
IPPORT_TELNET = 23 # Variable c_int
SPAPI_E_NO_ASSOCIATED_CLASS = -2146500096 # Variable c_long
LB_SETSEL = 389 # Variable c_int
ERROR_WINDOW_OF_OTHER_THREAD = 1408 # Variable c_long
PSD_DISABLEPRINTER = 32 # Variable c_int
TB_COMMANDTOINDEX = 1049 # Variable c_int
PROCESS_SET_INFORMATION = 512 # Variable c_int
TYPE_E_INVALIDID = -2147317553 # Variable c_long
MMIOERR_TOOMANYOPENFILES = 271 # Variable c_int
IMAGE_SYM_DTYPE_NULL = 0 # Variable c_int
META_POLYGON = 804 # Variable c_int
LC_STYLED = 32 # Variable c_int
NRC_NOSAPS = 55 # Variable c_int
IGP_CONVERSION = 8 # Variable c_int
JOY_CAL_READ5 = 4194304 # Variable c_long
SM_CYSIZE = 31 # Variable c_int
PRINTER_CHANGE_PRINTER_DRIVER = 1879048192 # Variable c_int
E_ACCESSDENIED = -2147024891 # Variable c_long
SPLDS_PRINT_BIN_NAMES = 'printBinNames' # Variable POINTER(c_wchar)
CONNECT_DEFERRED = 1024 # Variable c_int
CRYPT_E_NO_TRUSTED_SIGNER = -2146885589 # Variable c_long
IMAGE_REL_BASED_HIGH3ADJ = 11 # Variable c_int
DTM_SETMCFONT = 4105 # Variable c_int
IS_TEXT_UNICODE_NOT_ASCII_MASK = 61440 # Variable c_int
MKF_CONFIRMHOTKEY = 8 # Variable c_int
WS_EX_CLIENTEDGE = 512 # Variable c_long
LANG_SWAHILI = 65 # Variable c_int
FNERR_INVALIDFILENAME = 12290 # Variable c_int
MM_MIXM_LINE_CHANGE = 976 # Variable c_int
RPC_C_OPT_MQ_DELIVERY = 1 # Variable c_int
EV_RING = 256 # Variable c_int
STATUS_DATATYPE_MISALIGNMENT = 2147483650 # Variable c_ulong
CO_E_EXCEEDSYSACLLIMIT = -2147220970 # Variable c_long
BACKUP_PROPERTY_DATA = 6 # Variable c_int
LVM_GETITEMSPACING = 4147 # Variable c_int
TTN_GETDISPINFOA = 4294966776 # Variable c_uint
SND_ALIAS_SYSTEMWELCOME = 22355 # Variable c_ulong
EMR_SETMAPMODE = 17 # Variable c_int
LANG_ENGLISH = 9 # Variable c_int
WGL_SWAP_OVERLAY12 = 4096 # Variable c_int
OFN_CREATEPROMPT = 8192 # Variable c_int
PAN_ANY = 0 # Variable c_int
HICF_MOUSE = 1 # Variable c_int
ERROR_RELOC_CHAIN_XEEDS_SEGLIM = 201 # Variable c_long
TAPE_LOAD = 0 # Variable c_long
MIXERR_INVALLINE = 1024 # Variable c_int
MCI_SEQ_SET_SLAVE = 262144 # Variable c_long
WSAEPFNOSUPPORT = 10046 # Variable c_int
PGM_RECALCSIZE = 5122 # Variable c_int
URLOSTRM_USECACHEDCOPY_ONLY = 1 # Variable c_int
REG_OPTION_BACKUP_RESTORE = 4 # Variable c_long
MCI_SEQ_SET_PORT = 131072 # Variable c_long
MIXERCONTROL_CT_CLASS_CUSTOM = 0 # Variable c_long
frm4 = 1079 # Variable c_int
HANGUP_COMPLETE = 5 # Variable c_int
RB_HITTEST = 1032 # Variable c_int
_LEADBYTE = 32768 # Variable c_int
CTRY_CHILE = 56 # Variable c_int
RC_BITMAP64 = 8 # Variable c_int
TBM_GETTHUMBLENGTH = 1052 # Variable c_int
CDDS_PREERASE = 3 # Variable c_int
IMAGE_REL_IA64_PCREL21F = 8 # Variable c_int
CF_NOSCRIPTSEL = 8388608 # Variable c_long
PRINTER_CHANGE_DELETE_JOB = 1024 # Variable c_int
TBN_LAST = 4294966576 # Variable c_uint
IDC_SIZENS = 206405 # Variable POINTER(c_wchar)
LB_ITEMFROMPOINT = 425 # Variable c_int
SET_TAPE_DRIVE_INFORMATION = 1 # Variable c_int
CBN_SELENDOK = 9 # Variable c_int
RPC_C_AUTHN_DEC_PUBLIC = 4 # Variable c_int
CFSTR_MIME_X_EMF = 'image/x-emf' # Variable POINTER(c_wchar)
MARSHALINTERFACE_MIN = 500 # Variable c_int
IMAGE_SYM_CLASS_FAR_EXTERNAL = 68 # Variable c_int
EVENTLOG_END_ALL_PAIRED_EVENTS = 4 # Variable c_int
RESOURCEDISPLAYTYPE_GROUP = 5 # Variable c_int
IMAGE_SYM_TYPE_WORD = 13 # Variable c_int
PAN_SERIF_BONE = 8 # Variable c_int
LVM_GETISEARCHSTRINGA = 4148 # Variable c_int
TVN_KEYDOWN = 4294966884 # Variable c_uint
TCI_SRCCHARSET = 1 # Variable c_int
MCHT_CALENDARDAY = 131074 # Variable c_int
PAGE_WRITECOMBINE = 1024 # Variable c_int
WH_CBT = 5 # Variable c_int
ERROR_RESOURCE_ONLINE = 5019 # Variable c_long
SPI_GETNONCLIENTMETRICS = 41 # Variable c_int
USER_MARSHAL_FC_BYTE = 1 # Variable c_int
SUBLANG_LITHUANIAN = 1 # Variable c_int
DCBA_FACEDOWNLEFT = 258 # Variable c_int
TB_ISBUTTONCHECKED = 1034 # Variable c_int
MCI_DEVTYPE_OVERLAY = 515 # Variable c_int
DS_CONTROL = 1024 # Variable c_long
MM_TEXT = 1 # Variable c_int
CACHE_S_LAST = 262527 # Variable c_long
CTRY_PRCHINA = 86 # Variable c_int
SCS_CAP_SETRECONVERTSTRING = 4 # Variable c_int
DRV_ENABLE = 2 # Variable c_int
OPEN_EXISTING = 3 # Variable c_int
SW_SCROLLCHILDREN = 1 # Variable c_int
ERROR_LIBRARY_OFFLINE = 4305 # Variable c_long
IMAGE_REL_IA64_LTOFF22 = 10 # Variable c_int
PRINTER_CHANGE_DELETE_PRINTER = 4 # Variable c_int
ARW_DOWN = 4 # Variable c_long
DRAGDROP_E_NOTREGISTERED = -2147221248 # Variable c_long
NRC_BADDR = 7 # Variable c_int
CONTEXT_INTEGER = 65538 # Variable c_long
EM_POSFROMCHAR = 214 # Variable c_int
PF_ALPHA_BYTE_INSTRUCTIONS = 5 # Variable c_int
MCM_FIRST = 4096 # Variable c_int
TB_SETINSERTMARKCOLOR = 1112 # Variable c_int
ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT = 1807 # Variable c_long
ERROR_CHILD_NOT_COMPLETE = 129 # Variable c_long
RBBIM_CHILD = 16 # Variable c_int
grp1 = 1072 # Variable c_int
FILE_BEGIN = 0 # Variable c_int
LVM_ENSUREVISIBLE = 4115 # Variable c_int
CS_NOMOVECARET = 16384 # Variable c_int
NTM_OPENTYPE = 131072 # Variable c_int
rad14 = 1069 # Variable c_int
QS_KEY = 1 # Variable c_int
SUBLANG_MALAY_BRUNEI_DARUSSALAM = 2 # Variable c_int
IMN_CLOSESTATUSWINDOW = 1 # Variable c_int
DFCS_CAPTIONMAX = 2 # Variable c_int
SPAPI_E_SECTION_NAME_TOO_LONG = -2146500606 # Variable c_long
MIDICAPS_LRVOLUME = 2 # Variable c_int
DDL_READWRITE = 0 # Variable c_int
LVN_KEYDOWN = 4294967141 # Variable c_uint
TB_MAPACCELERATORW = 1114 # Variable c_int
C3_NONSPACING = 1 # Variable c_int
PD_COLLATE = 16 # Variable c_int
TPM_LEFTBUTTON = 0 # Variable c_long
URLPOLICY_LOG_ON_ALLOW = 64 # Variable c_int
MCI_WAIT = 2 # Variable c_long
IME_CMODE_NATIVE = 1 # Variable c_int
SIOCATMARK = 1074033415 # Variable c_long
RT_ANIICON = 34 # Variable POINTER(c_wchar)
TCS_FORCELABELLEFT = 32 # Variable c_int
SEE_MASK_CLASSNAME = 1 # Variable c_int
CO_E_INIT_SHARED_ALLOCATOR = -2147467257 # Variable c_long
DMPAPER_LETTER = 1 # Variable c_int
CC_ROUNDRECT = 256 # Variable c_int
ES_CENTER = 1 # Variable c_long
DMBIN_UPPER = 1 # Variable c_int
S_PERIOD1024 = 1 # Variable c_int
LVN_ITEMACTIVATE = 4294967182 # Variable c_uint
LANG_AZERI = 44 # Variable c_int
I_IMAGECALLBACK = -1 # Variable c_int
DNS_REGISTER = 1 # Variable c_int
EWX_SHUTDOWN = 1 # Variable c_int
EMR_POLYGON16 = 86 # Variable c_int
IMAGE_SIZEOF_ROM_OPTIONAL_HEADER = 56 # Variable c_int
PARAMFLAG_FLCID = 4 # Variable c_int
PERF_100NSEC_MULTI_TIMER_INV = 592512256 # Variable c_int
TBIF_LPARAM = 16 # Variable c_int
SPLDS_PRINT_MAX_RESOLUTION_SUPPORTED = 'printMaxResolutionSupported' # Variable POINTER(c_wchar)
RPC_X_WRONG_PIPE_ORDER = 1831 # Variable c_long
ERROR_DATABASE_FULL = 4314 # Variable c_long
MMIO_EXIST = 16384 # Variable c_int
DDL_ARCHIVE = 32 # Variable c_int
FSB_ENCARTA_MODE = 1 # Variable c_int
TBIF_COMMAND = 32 # Variable c_int
GL_ID_INPUTSYMBOL = 39 # Variable c_int
SWP_NOCOPYBITS = 256 # Variable c_int
CTRY_NICARAGUA = 505 # Variable c_int
SE_MACHINE_ACCOUNT_NAME = 'SeMachineAccountPrivilege' # Variable POINTER(c_wchar)
ERROR_TOO_MANY_LUIDS_REQUESTED = 1333 # Variable c_long
CCM_SETBKCOLOR = 8193 # Variable c_int
ERROR_NOT_ENOUGH_MEMORY = 8 # Variable c_long
IME_KHOTKEY_SHAPE_TOGGLE = 80 # Variable c_int
FAILED_ACCESS_ACE_FLAG = 128 # Variable c_int
MCN_SELCHANGE = 4294966547 # Variable c_uint
RPC_E_DISCONNECTED = -2147417848 # Variable c_long
PRSPEC_PROPID = 1 # Variable c_int
SERVICE_ACCEPT_STOP = 1 # Variable c_int
TVN_SETDISPINFOA = 4294966892 # Variable c_uint
TTM_GETMAXTIPWIDTH = 1049 # Variable c_int
KF_MENUMODE = 4096 # Variable c_int
IP_DONTFRAGMENT = 9 # Variable c_int
MIDIERR_NODEVICE = 68 # Variable c_int
SERVICE_PAUSED = 7 # Variable c_int
CB_GETLBTEXTLEN = 329 # Variable c_int
TTDT_RESHOW = 1 # Variable c_int
CO_E_OBJNOTREG = -2147220997 # Variable c_long
DISPID_COLLECT = -8 # Variable c_int

tagAPPINFOTYPE = c_int # enum
APPINFO_PUBLISHED = 1
APPINFO_SHELLNEW = 2
APPINFO_INSERTABLE = 3
ACE_OBJECT_TYPE_PRESENT = 1 # Variable c_int
DFC_BUTTON = 4 # Variable c_int
COMMON_LVB_LEADING_BYTE = 256 # Variable c_int
WAVECAPS_VOLUME = 4 # Variable c_int

tagCHANGEKIND = c_int # enum
CHANGEKIND_ADDMEMBER = 0
CHANGEKIND_INVALIDATE = 5
CHANGEKIND_CHANGEFAILED = 6
IMAGE_SUBSYSTEM_NATIVE_WINDOWS = 8 # Variable c_int
DRV_QUERYCONFIGURE = 8 # Variable c_int
WS_EX_TRANSPARENT = 32 # Variable c_long
PSP_DEFAULT = 0 # Variable c_int
CRYPT_E_FILE_ERROR = -2146885629 # Variable c_long
DD_DEFSCROLLINTERVAL = 50 # Variable c_int
ERROR_DS_INCORRECT_ROLE_OWNER = 1931 # Variable c_long
HKL_NEXT = 1 # Variable c_int
LANG_KANNADA = 75 # Variable c_int
RESOURCEDISPLAYTYPE_NDSCONTAINER = 11 # Variable c_int
RB_MOVEBAND = 1063 # Variable c_int
ERROR_DATABASE_DOES_NOT_EXIST = 1065 # Variable c_long
CBR_19200 = 19200 # Variable c_int
NRC_NOWILD = 21 # Variable c_int
SHGFI_SELECTED = 65536 # Variable c_int
OPENTYPE_FONTTYPE = 65536 # Variable c_int
JOB_OBJECT_TERMINATE = 8 # Variable c_int
PORT_STATUS_TYPE_WARNING = 2 # Variable c_int
CONNECT_REDIRECT = 128 # Variable c_int
EM_SETSEL = 177 # Variable c_int
PERF_TIMER_100NS = 1048576 # Variable c_int
DMPAPER_CSHEET = 24 # Variable c_int
VIEW_SORTSIZE = 5 # Variable c_int
IMAGE_SCN_MEM_NOT_CACHED = 67108864 # Variable c_int
MIDI_IO_STATUS = 32 # Variable c_long
LOCALE_SDECIMAL = 14 # Variable c_int
ERROR_DATABASE_FAILURE = 4313 # Variable c_long
WIN_CERT_TYPE_RESERVED_1 = 3 # Variable c_int
ERROR_OPEN_FAILED = 110 # Variable c_long
TVI_LAST = 4294901762 # Variable POINTER(_TREEITEM)
ERROR_INVALID_BLOCK_LENGTH = 1106 # Variable c_long
KLF_UNLOADPREVIOUS = 4 # Variable c_int
HICF_ACCELERATOR = 4 # Variable c_int
RC_GDI20_OUTPUT = 16 # Variable c_int
SPAPI_E_NO_INF = -2146500086 # Variable c_long
STG_E_INUSE = -2147286784 # Variable c_long
ERROR_INVALID_LB_MESSAGE = 1432 # Variable c_long
MCI_OVLY_RECT = 65536 # Variable c_long
HCF_HOTKEYSOUND = 16 # Variable c_int
ERROR_SINGLE_INSTANCE_APP = 1152 # Variable c_long
DLGC_WANTARROWS = 1 # Variable c_int
PAN_SERIF_ROUNDED = 15 # Variable c_int
LVA_DEFAULT = 0 # Variable c_int
XMLELEMTYPE_OTHER = 6
DC_FILEDEPENDENCIES = 14 # Variable c_int
DMDITHER_NONE = 1 # Variable c_int
SPAPI_E_ERROR_NOT_INSTALLED = -2146496512 # Variable c_long
ICM_ADDPROFILE = 1 # Variable c_int
CF_NOSTYLESEL = 1048576 # Variable c_long
STATUS_SEGMENT_NOTIFICATION = 1073741829 # Variable c_ulong
LVM_CREATEDRAGIMAGE = 4129 # Variable c_int
PERF_SIZE_DWORD = 0 # Variable c_int
PERF_100NSEC_TIMER = 542180608 # Variable c_int
SLE_WARNING = 3 # Variable c_int
LOCK_EXCLUSIVE = 2
HTCLOSE = 20 # Variable c_int
IMAGE_SEPARATE_DEBUG_MISMATCH = 32768 # Variable c_int
IMAGE_REL_SH3_DIRECT16 = 1 # Variable c_int
DM_REPOSITION = 1026 # Variable c_int
PD_ENABLESETUPTEMPLATEHANDLE = 131072 # Variable c_int
IMAGE_WEAK_EXTERN_SEARCH_LIBRARY = 2 # Variable c_int
CLASSFACTORY_E_FIRST = 2147746064 # Variable c_ulong
CONVERT10_E_OLESTREAM_BITMAP_TO_DIB = -2147221053 # Variable c_long
PSINJECT_BEGINPROLOG = 5 # Variable c_int
TCS_EX_FLATSEPARATORS = 1 # Variable c_int
SE_PRIVILEGE_ENABLED = 2 # Variable c_long
ERROR_ADDRESS_ALREADY_ASSOCIATED = 1227 # Variable c_long
SKF_INDICATOR = 32 # Variable c_int
ERROR_CALL_NOT_IMPLEMENTED = 120 # Variable c_long
SS_ENHMETAFILE = 15 # Variable c_long
MAILSLOT_NO_MESSAGE = 4294967295 # Variable c_ulong
CBF_SKIP_REGISTRATIONS = 524288 # Variable c_int
SIZE_MAXIMIZED = 2 # Variable c_int
TC_GP_TRAP = 2 # Variable c_int
CBF_SKIP_CONNECT_CONFIRMS = 262144 # Variable c_int
MM_MIM_DATA = 963 # Variable c_int
IA64_STUB_NOT_AVAILABLE = 4294967295 # Variable c_void_p
NM_CUSTOMDRAW = 4294967284 # Variable c_uint
SS_LEFTNOWORDWRAP = 12 # Variable c_long
CTRY_DENMARK = 45 # Variable c_int
SPI_GETDRAGFULLWINDOWS = 38 # Variable c_int
NULL_PEN = 8 # Variable c_int
NLS_VALID_LOCALE_MASK = 1048575 # Variable c_int
DOMAIN_ALIAS_RID_SYSTEM_OPS = 549 # Variable c_long
DMPAPER_A5 = 11 # Variable c_int
ACL_REVISION_DS = 4 # Variable c_int
DFCS_FLAT = 16384 # Variable c_int
RPC_E_TIMEOUT = -2147417825 # Variable c_long
JOYSTICKID1 = 0 # Variable c_int
MCI_SET = 2061 # Variable c_int
CWP_SKIPDISABLED = 2 # Variable c_int
GCPCLASS_NUMERICSEPARATOR = 8 # Variable c_int
SHGFI_ATTRIBUTES = 2048 # Variable c_int
LVM_SETBKCOLOR = 4097 # Variable c_int
TPM_NONOTIFY = 128 # Variable c_long
TB_GETBUTTONTEXTA = 1069 # Variable c_int
COMMON_LVB_REVERSE_VIDEO = 16384 # Variable c_int
ERROR_LOGON_FAILURE = 1326 # Variable c_long
PAN_STROKE_GRADUAL_HORZ = 5 # Variable c_int
TC_SCROLLBLT = 65536 # Variable c_int
JOB_OBJECT_UILIMIT_DISPLAYSETTINGS = 16 # Variable c_int
SB_GETPARTS = 1030 # Variable c_int
MCI_COPY = 2130 # Variable c_int
PAN_MIDLINE_HIGH_SERIFED = 7 # Variable c_int
TVS_CHECKBOXES = 256 # Variable c_int
DMLERR_DLL_NOT_INITIALIZED = 16387 # Variable c_int
PSINJECT_ENDPAGECOMMENTS = 9 # Variable c_int
LCS_WINDOWS_COLOR_SPACE = 1466527264 # Variable c_int
NRC_TOOMANY = 34 # Variable c_int
CB_GETDROPPEDCONTROLRECT = 338 # Variable c_int
TVGN_PREVIOUS = 2 # Variable c_int
PGM_SETBORDER = 5126 # Variable c_int
SP_HANDSHAKING = 16 # Variable c_ulong
WIZ_CXBMP = 80 # Variable c_int
SPLDS_PRINT_START_TIME = 'printStartTime' # Variable POINTER(c_wchar)
DT_VCENTER = 4 # Variable c_int
MCI_MODE_NOT_READY = 524 # Variable c_int
FILE_WRITE_DATA = 2 # Variable c_int
PSINJECT_VMSAVE = 17 # Variable c_int
MCIERR_MISSING_COMMAND_STRING = 267 # Variable c_int
IMAGE_SCN_ALIGN_64BYTES = 7340032 # Variable c_int
SE_GROUP_DEFAULTED = 2 # Variable c_int
STG_S_MONITORING = 197123 # Variable c_long
ACCESS_DENIED_ACE_TYPE = 1 # Variable c_int
ERROR_SERVICE_NO_THREAD = 1054 # Variable c_long
MCIERR_WAVE_INPUTUNSPECIFIED = 325 # Variable c_int
DM_SETDEFID = 1025 # Variable c_int
IGP_SETCOMPSTR = 20 # Variable c_int
CBEM_INSERTITEMW = 1035 # Variable c_int
TTM_GETTIPBKCOLOR = 1046 # Variable c_int
ERROR_MEDIA_INCOMPATIBLE = 4315 # Variable c_long
STG_E_DISKISWRITEPROTECTED = -2147287021 # Variable c_long
JOY_CAL_READ6 = 8388608 # Variable c_long
CB_OKAY = 0 # Variable c_int
ERROR_HOTKEY_ALREADY_REGISTERED = 1409 # Variable c_long
PSN_RESET = 4294967093 # Variable c_uint
REGDB_E_READREGDB = -2147221168 # Variable c_long
HTBOTTOMLEFT = 16 # Variable c_int
SB_GETICON = 1044 # Variable c_int
psh12 = 1035 # Variable c_int
DOWNLOADHEADER = 4111 # Variable c_int
MCIERR_ILLEGAL_FOR_AUTO_OPEN = 303 # Variable c_int
IMAGE_SIZEOF_LINENUMBER = 6 # Variable c_int
WMSZ_TOP = 3 # Variable c_int
DMPAPER_TABLOID_EXTRA = 52 # Variable c_int
FD_WRITE = 2 # Variable c_int
CDN_SELCHANGE = 4294966694 # Variable c_uint
MCI_GETDEVCAPS = 2059 # Variable c_int
TB_PRESSBUTTON = 1027 # Variable c_int
PROCESSOR_HITACHI_SH3E = 10004 # Variable c_int
PRINTACTION_OPENNETPRN = 5 # Variable c_int
PORT_STATUS_PAPER_PROBLEM = 5 # Variable c_int
REPLACEDLGORD = 1541 # Variable c_int
JOB_NOTIFY_FIELD_TOTAL_PAGES = 20 # Variable c_int
MCI_VD_SPIN_DOWN = 131072 # Variable c_long
CO_E_DLLNOTFOUND = -2147221000 # Variable c_long
REG_EXPAND_SZ = 2 # Variable c_int
MIXERCONTROL_CONTROLF_UNIFORM = 1 # Variable c_long
PS_JOIN_BEVEL = 4096 # Variable c_int
MCI_DEVTYPE_WAVEFORM_AUDIO = 522 # Variable c_int
IPPORT_TTYLINK = 87 # Variable c_int
TIME_ZONE_ID_DAYLIGHT = 2 # Variable c_int
WGL_SWAP_UNDERLAY13 = 268435456 # Variable c_int
VFT2_DRV_NETWORK = 6 # Variable c_long
CTRY_AUSTRALIA = 61 # Variable c_int
MCI_FORMAT_MILLISECONDS = 0 # Variable c_int
TOOLBARCLASSNAMEA = 'ToolbarWindow32' # Variable POINTER(c_char)
EMR_SELECTOBJECT = 37 # Variable c_int
LVBKIF_STYLE_NORMAL = 0 # Variable c_int
GWL_STYLE = -16 # Variable c_int
PERF_SAMPLE_BASE = 1073939457 # Variable c_int
POLYFILL_LAST = 2 # Variable c_int
TVS_DISABLEDRAGDROP = 16 # Variable c_int
BAUD_600 = 32 # Variable c_ulong
SE_RESTORE_NAME = 'SeRestorePrivilege' # Variable POINTER(c_wchar)
PORT_STATUS_USER_INTERVENTION = 8 # Variable c_int
MEVT_TEMPO = 1 # Variable c_ubyte
FR_REPLACE = 16 # Variable c_int
ERROR_PRINTQ_FULL = 61 # Variable c_long
rad15 = 1070 # Variable c_int
HINSTANCE_ERROR = 32 # Variable c_int
FKF_HOTKEYSOUND = 16 # Variable c_int
TB_CHECKBUTTON = 1026 # Variable c_int
LC_WIDESTYLED = 64 # Variable c_int
HTTOPRIGHT = 14 # Variable c_int
KEYEVENTF_KEYUP = 2 # Variable c_int
SERVICE_ACCEPT_PARAMCHANGE = 8 # Variable c_int
WH_JOURNALPLAYBACK = 1 # Variable c_int
PRINTER_NOTIFY_FIELD_OBJECT_GUID = 26 # Variable c_int
PAN_STRAIGHT_ARMS_HORZ = 2 # Variable c_int
TB_SETMAXTEXTROWS = 1084 # Variable c_int
RB_GETROWCOUNT = 1037 # Variable c_int
SW_SHOW = 5 # Variable c_int
MIXERCONTROL_CONTROLTYPE_VOLUME = 1342373889 # Variable c_long
MONO_FONT = 8 # Variable c_int
WB_LEFT = 0 # Variable c_int
MIDIERR_BASE = 64 # Variable c_int
LOCALE_USER_DEFAULT = 1024 # Variable c_ulong
SM_CYCAPTION = 4 # Variable c_int
DATE_SHORTDATE = 1 # Variable c_int
ES_NUMBER = 8192 # Variable c_long
TECHNOLOGY = 2 # Variable c_int
PERF_SAMPLE_COUNTER = 4260864 # Variable c_int
CBEN_ENDEDITA = 4294966491 # Variable c_uint
PRINTER_ACCESS_ADMINISTER = 4 # Variable c_int
ABN_STATECHANGE = 0 # Variable c_int
RC_DIBTODEV = 512 # Variable c_int
ERROR_PIPE_NOT_CONNECTED = 233 # Variable c_long
ERROR_INVALID_SERVICE_CONTROL = 1052 # Variable c_long
SM_CXDRAG = 68 # Variable c_int
DM_ICMMETHOD = 8388608 # Variable c_long
HKCOMB_NONE = 1 # Variable c_int
RBBIM_IMAGE = 8 # Variable c_int
VFT2_DRV_COMM = 10 # Variable c_long
PROVIDER_KEEPS_VALUE_LENGTH = 1 # Variable c_int
PRINTACTION_TESTPAGE = 4 # Variable c_int
CRYPT_E_UNEXPECTED_MSG_TYPE = -2146885622 # Variable c_long
EMR_SETMITERLIMIT = 58 # Variable c_int
BI_RGB = 0 # Variable c_long
LVM_SETBKIMAGEA = 4164 # Variable c_int
LOCALE_IFIRSTDAYOFWEEK = 4108 # Variable c_int
NFR_ANSI = 1 # Variable c_int
OFN_NOVALIDATE = 256 # Variable c_int
PORT_STATUS_PAPER_JAM = 2 # Variable c_int
WSB_PROP_CXVSCROLL = 8 # Variable c_long
TVM_SELECTITEM = 4363 # Variable c_int
HICF_DUPACCEL = 8 # Variable c_int
TVS_HASBUTTONS = 1 # Variable c_int
EMR_POLYGON = 3 # Variable c_int
ARW_STARTTOP = 2 # Variable c_long
HTOBJECT = 19 # Variable c_int
_MAX_FNAME = 256 # Variable c_int
GCPCLASS_LATIN = 1 # Variable c_int
MOD_ON_KEYUP = 2048 # Variable c_int
SHGFI_SMALLICON = 1 # Variable c_int
TYPE_E_DLLFUNCTIONNOTFOUND = -2147319761 # Variable c_long
chx5 = 1044 # Variable c_int
SERVICE_CONTROL_PAUSE = 2 # Variable c_int
URLACTION_SHELL_VERB = 6148 # Variable c_int
TA_MASK = 287 # Variable c_int
OLE_E_INVALIDRECT = -2147221491 # Variable c_long
SE_ERR_DDEBUSY = 30 # Variable c_int
WA_ACTIVE = 1 # Variable c_int
JOYERR_NOERROR = 0 # Variable c_int
WC_IPADDRESSW = 'SysIPAddress32' # Variable POINTER(c_wchar)
LVA_ALIGNTOP = 2 # Variable c_int
CDM_GETSPEC = 1124 # Variable c_int
MCI_ANIM_REALIZE_NORM = 65536 # Variable c_long
RPC_S_NO_CALL_ACTIVE = 1725 # Variable c_long
MF_SEPARATOR = 2048 # Variable c_long
RPC_S_ZERO_DIVIDE = 1767 # Variable c_long
IME_REGWORD_STYLE_USER_LAST = 4294967295 # Variable c_uint
RPC_C_PROTECT_LEVEL_DEFAULT = 0 # Variable c_int
IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT = 13 # Variable c_int
HIST_ADDTOFAVORITES = 3 # Variable c_int
MAX_LEADBYTES = 12 # Variable c_int
CONTEXT_i486 = 65536 # Variable c_int
PRINTER_CHANGE_DELETE_PORT = 4194304 # Variable c_int
MIDIERR_LASTERROR = 71 # Variable c_int
CDIS_FOCUS = 16 # Variable c_int
BS_RADIOBUTTON = 4 # Variable c_long
EDGE_ETCHED = 6 # Variable c_int
TTN_GETDISPINFOW = 4294966766 # Variable c_uint
SHGFI_OPENICON = 2 # Variable c_int
MCI_ANIM_STATUS_HPAL = 16388 # Variable c_long
KF_DLGMODE = 2048 # Variable c_int
LVM_SUBITEMHITTEST = 4153 # Variable c_int
ERROR_SUBST_TO_JOIN = 141 # Variable c_long
SHAREVISTRINGA = 'commdlg_ShareViolation' # Variable POINTER(c_char)
FACILITY_DISPATCH = 2 # Variable c_int
IMAGE_REL_PPC_SECRELLO = 19 # Variable c_int
ERROR_FULLSCREEN_MODE = 1007 # Variable c_long
SS_WHITERECT = 6 # Variable c_long
SECURITY_EFFECTIVE_ONLY = 524288 # Variable c_int
CS_ENABLE = 1 # Variable c_long
MCI_VD_PLAY_REVERSE = 65536 # Variable c_long
CB_ERRSPACE = -2 # Variable c_int
IE_MEMORY = -4 # Variable c_int
rad8 = 1063 # Variable c_int
CERT_E_CRITICAL = -2146762491 # Variable c_long
WH_MOUSE = 7 # Variable c_int
HCF_AVAILABLE = 2 # Variable c_int
IO_REPARSE_TAG_HSM = 3221225476 # Variable c_uint
XST_ADVDATAACKRCVD = 16 # Variable c_int
ERROR_INVALID_ACL = 1336 # Variable c_long
ERROR_INVALID_COLORSPACE = 2307 # Variable c_long
ERROR_FILE_NOT_FOUND = 2 # Variable c_long
CBEIF_LPARAM = 32 # Variable c_int
LVN_SETDISPINFOW = 4294967118 # Variable c_uint
ERROR_SERVICE_SPECIFIC_ERROR = 1066 # Variable c_long
NORMAL_PRIORITY_CLASS = 32 # Variable c_int
grp4 = 1075 # Variable c_int
CO_E_FAILEDTOOPENTHREADTOKEN = -2147220990 # Variable c_long
MK_E_NOOBJECT = -2147221019 # Variable c_long
HTMAXBUTTON = 9 # Variable c_int
LANG_FAEROESE = 56 # Variable c_int
ERROR_INVALID_ID_AUTHORITY = 1343 # Variable c_long
SEMAPHORE_MODIFY_STATE = 2 # Variable c_int
LVHT_ONITEM = 14 # Variable c_int
IMAGE_REL_ALPHA_SECRELLO = 17 # Variable c_int
URLACTION_INFODELIVERY_MIN = 7424 # Variable c_int
ERROR_NO_MORE_SEARCH_HANDLES = 113 # Variable c_long
SMTO_NORMAL = 0 # Variable c_int
MCI_STATUS_CURRENT_TRACK = 8 # Variable c_long
IMAGE_REL_PPC_ADDR24 = 3 # Variable c_int
REG_OPTION_VOLATILE = 1 # Variable c_long
ENUM_REGISTRY_SETTINGS = 4294967294 # Variable c_ulong
MAXIMUM_ALLOWED = 33554432 # Variable c_long
JOB_OBJECT_LIMIT_WORKINGSET = 1 # Variable c_int
HS_VERTICAL = 1 # Variable c_int
CONVERT10_E_LAST = 2147746255 # Variable c_ulong
TIME_TICKS = 32 # Variable c_int
ERROR_ITERATED_DATA_EXCEEDS_64k = 194 # Variable c_long
VARCMP_LT = 0 # Variable c_int
DC_MANUFACTURER = 23 # Variable c_int
STD_FIND = 12 # Variable c_int
WC_LISTVIEWW = 'SysListView32' # Variable POINTER(c_wchar)
IDC_CROSS = 206101 # Variable POINTER(c_wchar)
LANG_ESTONIAN = 37 # Variable c_int
SEM_NOOPENFILEERRORBOX = 32768 # Variable c_int
C3_SYMBOL = 8 # Variable c_int
BAUD_115200 = 131072 # Variable c_ulong
ACCESS_DS_OBJECT_TYPE_NAME_W = 'Directory Service Object' # Variable POINTER(c_wchar)
DMRES_MEDIUM = -3 # Variable c_int
ERROR_WRONG_DISK = 34 # Variable c_long
NUMPRS_THOUSANDS = 512 # Variable c_int
FACILITY_MEDIASERVER = 13 # Variable c_int
ctlFirst = 1024 # Variable c_int
TVN_SETDISPINFOW = 4294966843 # Variable c_uint
ERROR_TOO_MANY_LINKS = 1142 # Variable c_long
MMIO_DENYNONE = 64 # Variable c_int
ANIMATE_CLASSW = 'SysAnimate32' # Variable POINTER(c_wchar)
LANG_GERMAN = 7 # Variable c_int
MINSHORT = 32768 # Variable c_int
PC_SCANLINE = 8 # Variable c_int
PDERR_INITFAILURE = 4102 # Variable c_int
CO_E_LAUNCH_PERMSSION_DENIED = -2147467237 # Variable c_long
SEC_VLM = 33554432 # Variable c_int
PAN_PROP_EXPANDED = 5 # Variable c_int
PIPE_UNLIMITED_INSTANCES = 255 # Variable c_int
DNS_UNREGISTER = 2 # Variable c_int
DATABITS_5 = 1 # Variable c_ushort
LVS_NOSCROLL = 8192 # Variable c_int
rct2 = 1081 # Variable c_int
SM_RESERVED2 = 25 # Variable c_int
IMPLINK_HIGHEXPER = 158 # Variable c_int
SCS_PIF_BINARY = 3 # Variable c_int
IDI_HAND = 206099 # Variable POINTER(c_wchar)
STD_REPLACE = 13 # Variable c_int
SIF_POS = 4 # Variable c_int
DMBIN_MANUAL = 4 # Variable c_int
CF_TEXT = 1 # Variable c_int
IMAGE_REL_BASED_REL32 = 7 # Variable c_int
NDR_BIG_ENDIAN = 0 # Variable c_ulong
ERROR_SERVICE_NOT_ACTIVE = 1062 # Variable c_long
VIF_OUTOFSPACE = 256 # Variable c_long
FILE_FLAG_OVERLAPPED = 1073741824 # Variable c_int
IME_CMODE_LANGUAGE = 3 # Variable c_int
MCI_OVLY_INFO_TEXT = 65536 # Variable c_long
META_CHORD = 2096 # Variable c_int
DFCS_CAPTIONHELP = 4 # Variable c_int
IMAGE_SCN_NO_DEFER_SPEC_EXC = 16384 # Variable c_int
LB_SELECTSTRING = 396 # Variable c_int
CTRY_MACEDONIA = 389 # Variable c_int
CWP_SKIPINVISIBLE = 1 # Variable c_int
MIXER_OBJECTF_MIDIIN = 1073741824 # Variable c_long
PERF_SAMPLE_FRACTION = 549585920 # Variable c_int
LVN_COLUMNCLICK = 4294967188 # Variable c_uint
SIOCGHIWAT = 1074033409 # Variable c_long
RPC_S_NO_CONTEXT_AVAILABLE = 1765 # Variable c_long
LVFI_NEARESTXY = 64 # Variable c_int

_CLASSPATHTYPE = c_int # enum
TBTS_BOTTOM = 2 # Variable c_int
PAN_PROP_MODERN = 3 # Variable c_int
REG_LEGAL_CHANGE_FILTER = 15 # Variable c_long
VER_PLATFORM_WIN32_WINDOWS = 1 # Variable c_int
ERROR_NO_TOKEN = 1008 # Variable c_long
DIGSIG_E_ENCODE = -2146762747 # Variable c_long
SO_CONNOPTLEN = 28677 # Variable c_int
TAPE_DRIVE_SELECT = 2 # Variable c_int
OLE_S_FIRST = 262144 # Variable c_long
SE_SYSTEM_PROFILE_NAME = 'SeSystemProfilePrivilege' # Variable POINTER(c_wchar)
WA_CLICKACTIVE = 2 # Variable c_int
TRUE = 1 # Variable c_int
LVS_AUTOARRANGE = 256 # Variable c_int
MOUSETRAILS = 39 # Variable c_int
SPLDS_PRINT_KEEP_PRINTED_JOBS = 'printKeepPrintedJobs' # Variable POINTER(c_wchar)
PSP_USEHEADERTITLE = 4096 # Variable c_int
ERROR_GENERIC_NOT_MAPPED = 1360 # Variable c_long
FILE_FLAG_WRITE_THROUGH = 2147483648 # Variable c_uint
DO_PRINTFILE = 1414419024 # Variable c_long
PI_DATAPROGRESS = 64
MCHT_CALENDARDATEPREV = 33685505 # Variable c_int
SS_SIMPLE = 11 # Variable c_long
FOF_RENAMEONCOLLISION = 8 # Variable c_int
UD_MINVAL = -32767 # Variable c_int
OFN_PATHMUSTEXIST = 2048 # Variable c_int
RPC_IF_OLE = 2 # Variable c_int
DDE_FBUSY = 16384 # Variable c_int
CE_BREAK = 16 # Variable c_int
ERROR_INVALID_DOMAINNAME = 1212 # Variable c_long
MIDIERR_INVALIDSETUP = 69 # Variable c_int
DFCS_CAPTIONCLOSE = 0 # Variable c_int
SPAPI_E_INVALID_FILTER_DRIVER = -2146500052 # Variable c_long
IMPORT_OBJECT_HDR_SIG2 = 65535 # Variable c_int
HDN_DIVIDERDBLCLICKW = 4294966971 # Variable c_uint
CLR_NONE = 4294967295 # Variable c_ulong
TKF_CONFIRMHOTKEY = 8 # Variable c_int
FILE_END = 2 # Variable c_int
JOY_USEDEADZONE = 2048 # Variable c_long
CBS_HASSTRINGS = 512 # Variable c_long
ERROR_NOT_DOS_DISK = 26 # Variable c_long
CONVERT10_E_STG_NO_STD_STREAM = -2147221051 # Variable c_long
ERROR_END_OF_MEDIA = 1100 # Variable c_long
MSG_OOB = 1 # Variable c_int
IMAGE_SCN_ALIGN_128BYTES = 8388608 # Variable c_int
IMAGE_REL_MIPS_SECRELLO = 12 # Variable c_int
URLACTION_JAVA_MAX = 7423 # Variable c_int
S_WHITE2048 = 6 # Variable c_int
SS_CENTERIMAGE = 512 # Variable c_long
CDIS_MARKED = 128 # Variable c_int
LOCALE_SMONTHNAME1 = 56 # Variable c_int
ENHMETA_SIGNATURE = 1179469088 # Variable c_int
XTYP_SHIFT = 4 # Variable c_int
LVSCW_AUTOSIZE = -1 # Variable c_int
TBS_NOTICKS = 16 # Variable c_int
OFN_EXTENSIONDIFFERENT = 1024 # Variable c_int
ERROR_MEDIA_NOT_AVAILABLE = 4318 # Variable c_long
IMAGE_OS2_SIGNATURE = 17742 # Variable c_int
TYPE_E_AMBIGUOUSNAME = -2147319764 # Variable c_long
MCHT_TITLEBTNNEXT = 16842755 # Variable c_int
LB_SETLOCALE = 421 # Variable c_int
MCI_ANIM_UPDATE_HDC = 131072 # Variable c_long
ILLUMINANT_D75 = 7 # Variable c_int
SET_POLY_MODE = 4104 # Variable c_int
DV_E_STATDATA = -2147221401 # Variable c_long
MCIERR_INVALID_DEVICE_NAME = 263 # Variable c_int
SE_LOCK_MEMORY_NAME = 'SeLockMemoryPrivilege' # Variable POINTER(c_wchar)
GCP_SYMSWAPOFF = 8388608 # Variable c_long
CO_E_FAILEDTOOPENPROCESSTOKEN = -2147220967 # Variable c_long
NTE_NOT_FOUND = -2146893807 # Variable c_long
ERROR_NETWORK_UNREACHABLE = 1231 # Variable c_long
RPC_C_PARM_BUFFER_LENGTH = 2 # Variable c_int
FS_GREEK = 8 # Variable c_long
FACILITY_MSMQ = 14 # Variable c_int
ASYNCH = 128 # Variable c_int
PBT_APMOEMEVENT = 11 # Variable c_int
REG_NOTIFY_CHANGE_ATTRIBUTES = 2 # Variable c_long
WGL_SWAP_OVERLAY15 = 32768 # Variable c_int
USER_MARSHAL_FC_CHAR = 2 # Variable c_int
INVALID_SOCKET = 4294967295 # Variable c_uint
SEF_SACL_AUTO_INHERIT = 2 # Variable c_int
PST_RS423 = 4 # Variable c_ulong
CTRL_SHUTDOWN_EVENT = 6 # Variable c_int
CDS_RESET = 1073741824 # Variable c_int
PORT_TYPE_REDIRECTED = 4 # Variable c_int
LVN_ODFINDITEMW = 4294967117 # Variable c_uint
TVN_DELETEITEMA = 4294966887 # Variable c_uint
CO_E_CREATEPROCESS_FAILURE = -2147467240 # Variable c_long
CE_DNS = 2048 # Variable c_int
TAPE_DRIVE_LOAD_UNLOAD = 2147483649 # Variable c_uint
MB_NOFOCUS = 32768 # Variable c_long
DRIVE_REMOTE = 4 # Variable c_int
S_ALLTHRESHOLD = 2 # Variable c_int
MEM_MAPPED = 262144 # Variable c_int
VIEW_LARGEICONS = 0 # Variable c_int
HELPMSGSTRINGA = 'commdlg_help' # Variable POINTER(c_char)
RPC_S_COMM_FAILURE = 1820 # Variable c_long
ERROR_TAG_NOT_FOUND = 2302 # Variable c_long
IMN_SETCONVERSIONMODE = 6 # Variable c_int
CLIP_STROKE_PRECIS = 2 # Variable c_int
GGO_BEZIER = 3 # Variable c_int
EMBDHLP_CREATENOW = 0 # Variable c_long
TMPF_FIXED_PITCH = 1 # Variable c_int
STATUS_PENDING = 259 # Variable c_ulong
DMTT_DOWNLOAD_OUTLINE = 4 # Variable c_int
ICON_BIG = 1 # Variable c_int
SHOW_ICONWINDOW = 2 # Variable c_int
BACKGROUND_INTENSITY = 128 # Variable c_int
THREAD_PRIORITY_ABOVE_NORMAL = 1 # Variable c_int

tagBINDSTRING = c_int # enum
BINDSTRING_EXTRA_URL = 3
BINDSTRING_USERNAME = 5
BINDSTRING_OS = 9
BINDSTRING_USER_AGENT = 10
BINDSTRING_POST_COOKIE = 12
BINDSTRING_URL = 14
CFSTR_MIME_RICHTEXT = 'text/richtext' # Variable POINTER(c_wchar)
ERROR_DRIVE_MEDIA_MISMATCH = 4303 # Variable c_long
DBG_CONTROL_BREAK = 1073807368 # Variable c_ulong
SPI_SETSOUNDSENTRY = 65 # Variable c_int
PAN_CULTURE_LATIN = 0 # Variable c_int
CO_E_FAILEDTOSETDACL = -2147220986 # Variable c_long
LOCALE_SABBREVMONTHNAME10 = 77 # Variable c_int
REGDB_E_WRITEREGDB = -2147221167 # Variable c_long
PROCESSOR_ARCHITECTURE_MIPS = 1 # Variable c_int
RGB_GAMMA_MIN = 1344 # Variable c_ushort
BN_PAINT = 1 # Variable c_int
ERROR_MAPPED_ALIGNMENT = 1132 # Variable c_long
IDRETRY = 4 # Variable c_int
LB_DIR = 397 # Variable c_int
LVM_GETBKCOLOR = 4096 # Variable c_int
CB_GETLOCALE = 346 # Variable c_int
VARCMP_GT = 2 # Variable c_int
ICC_UPDOWN_CLASS = 16 # Variable c_int
PRINTER_CHANGE_WRITE_JOB = 2048 # Variable c_int
OUT_CHARACTER_PRECIS = 2 # Variable c_int
WVR_HREDRAW = 256 # Variable c_int
STDOLE2_MAJORVERNUM = 2 # Variable c_int
EM_UNDO = 199 # Variable c_int
CTRY_BAHRAIN = 973 # Variable c_int
TAPE_DRIVE_TENSION = 2147483650 # Variable c_uint
BAUD_110 = 2 # Variable c_ulong
CF_DSPENHMETAFILE = 142 # Variable c_int
ERROR_FAIL_I24 = 83 # Variable c_long
RPC_C_AUTHN_LEVEL_CALL = 3 # Variable c_int
CBEM_GETEXSTYLE = 1033 # Variable c_int
INPLACE_E_NOTOOLSPACE = -2147221087 # Variable c_long
PGM_FIRST = 5120 # Variable c_int
MCHT_TITLEBK = 65536 # Variable c_int
USER_MARSHAL_FC_FLOAT = 10 # Variable c_int
frm1 = 1076 # Variable c_int
CFERR_NOFONTS = 8193 # Variable c_int
VOS__PM32 = 3 # Variable c_long
SW_SHOWDEFAULT = 10 # Variable c_int
rad16 = 1071 # Variable c_int
CDN_FIRST = 4294966695 # Variable c_uint
EMR_DELETECOLORSPACE = 101 # Variable c_int
PSINJECT_PAGENUMBER = 25 # Variable c_int
CE_OOP = 4096 # Variable c_int
RPC_C_OPT_MQ_AUTHN_SERVICE = 5 # Variable c_int
MCI_GETDEVCAPS_CAN_PLAY = 8 # Variable c_long
VIF_OUTOFMEMORY = 32768 # Variable c_long
SUBLANG_KOREAN = 1 # Variable c_int
DMLERR_NO_ERROR = 0 # Variable c_int
MKF_REPLACENUMBERS = 128 # Variable c_int
MUTANT_QUERY_STATE = 1 # Variable c_int
DMPAPER_RESERVED_49 = 49 # Variable c_int
TTM_TRACKPOSITION = 1042 # Variable c_int
SETDIBSCALING = 32 # Variable c_int
GCP_NUMERICSLOCAL = 134217728 # Variable c_long
PDERR_PRINTERCODES = 4096 # Variable c_int

__MIDL_IBindStatusCallback_0004 = c_int # enum
BSCF_FIRSTDATANOTIFICATION = 1
BSCF_INTERMEDIATEDATANOTIFICATION = 2
BSCF_DATAFULLYAVAILABLE = 8
BSCF_AVAILABLEDATASIZEUNKNOWN = 16
TCHT_NOWHERE = 1 # Variable c_int
MCM_GETMONTHDELTA = 4115 # Variable c_int
STANDARD_RIGHTS_REQUIRED = 983040 # Variable c_long
TB_SETDRAWTEXTFLAGS = 1094 # Variable c_int
THREAD_QUERY_INFORMATION = 64 # Variable c_int
BM_SETCHECK = 241 # Variable c_int
CO_E_INVALIDSID = -2147220982 # Variable c_long
CBF_FAIL_SELFCONNECTIONS = 4096 # Variable c_int
CTRY_MALAYSIA = 60 # Variable c_int
ERROR_QUORUM_RESOURCE_ONLINE_FAILED = 5027 # Variable c_long
LVCF_IMAGE = 16 # Variable c_int
PROP_LG_CYDLG = 218 # Variable c_int
ERROR_SESSION_CREDENTIAL_CONFLICT = 1219 # Variable c_long
ARW_HIDE = 8 # Variable c_long
KEY_ALL_ACCESS = 983103 # Variable c_long
STATUS_FLOAT_OVERFLOW = 3221225617 # Variable c_ulong
REGISTERED = 4 # Variable c_int
PSP_PREMATURE = 1024 # Variable c_int
EXIT_THREAD_DEBUG_EVENT = 4 # Variable c_int
RPC_X_NO_MORE_ENTRIES = 1772 # Variable c_long
TAPE_UNLOCK = 4 # Variable c_long
LOCALE_IDIGITS = 17 # Variable c_int
IMAGE_SYM_DTYPE_FUNCTION = 2 # Variable c_int
CO_E_FAILEDTOGENUUID = -2147220973 # Variable c_long
IMEVER_0310 = 196618 # Variable c_int
FILE_SUPPORTS_ENCRYPTION = 131072 # Variable c_int
SUBLANG_FRENCH_BELGIAN = 2 # Variable c_int
CFSTR_MIME_X_ART = 'image/x-jg' # Variable POINTER(c_wchar)
SUBLANG_ENGLISH_BELIZE = 10 # Variable c_int
ICM_QUERYMATCH = 7 # Variable c_int
IMAGE_REL_MIPS_REFWORDNB = 34 # Variable c_int
BS_MONOPATTERN = 9 # Variable c_int
ERROR_MEMBER_IN_GROUP = 1320 # Variable c_long
EN_HSCROLL = 1537 # Variable c_int
SPI_SETBEEP = 2 # Variable c_int
ERROR_FULL_BACKUP = 4004 # Variable c_long
PD_SELECTION = 1 # Variable c_int
LVN_ODSTATECHANGED = 4294967181 # Variable c_uint
FS_WANSUNG = 524288 # Variable c_long
AUX_MAPPER = 4294967295 # Variable c_uint
IMAGE_SYM_CLASS_FILE = 103 # Variable c_int
PROCESS_HEAP_REGION = 1 # Variable c_int
FD_SETSIZE = 64 # Variable c_int
RESOURCEDISPLAYTYPE_DIRECTORY = 9 # Variable c_int
PRINTER_CONTROL_PURGE = 3 # Variable c_int
STN_ENABLE = 2 # Variable c_int
PSH_PROPSHEETPAGE = 8 # Variable c_int
IMAGE_SYM_CLASS_STRUCT_TAG = 10 # Variable c_int
S_FALSE = 1 # Variable c_long
LANG_NEPALI = 97 # Variable c_int
FSHIFT = 4 # Variable c_int
TTM_SETMARGIN = 1050 # Variable c_int
EMR_SETARCDIRECTION = 57 # Variable c_int
SB_PAGEDOWN = 3 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_DIGITAL = 1 # Variable c_long
CB_GETDROPPEDWIDTH = 351 # Variable c_int
TBM_SETTICFREQ = 1044 # Variable c_int
MCIERR_FILE_NOT_SAVED = 286 # Variable c_int
PSINJECT_TRAILER = 14 # Variable c_int
FORMAT_MESSAGE_ARGUMENT_ARRAY = 8192 # Variable c_int
EMR_PIE = 47 # Variable c_int
PROPSHEETPAGEA_V1_SIZE = 40 # Variable c_uint
DCX_PARENTCLIP = 32 # Variable c_long
CAL_SABBREVDAYNAME2 = 15 # Variable c_int
SS_SUNKEN = 4096 # Variable c_long
SEM_NOALIGNMENTFAULTEXCEPT = 4 # Variable c_int
SSF_SOUNDSENTRYON = 1 # Variable c_int
AF_NS = 6 # Variable c_int
SBS_SIZEBOX = 8 # Variable c_long
BF_MIDDLE = 2048 # Variable c_int
WNCON_DYNAMIC = 8 # Variable c_int
WGL_SWAP_OVERLAY9 = 512 # Variable c_int
ERROR_ALREADY_RUNNING_LKG = 1074 # Variable c_long
TAPE_TENSION = 2 # Variable c_long
LANG_TELUGU = 74 # Variable c_int
DIALOPTION_QUIET = 128 # Variable c_int
KLF_ACTIVATE = 1 # Variable c_int
MCI_SET_AUDIO = 2048 # Variable c_long
TVM_GETITEMHEIGHT = 4380 # Variable c_int

_SID_NAME_USE = c_int # enum
SidTypeUser = 1
SidTypeGroup = 2
SidTypeDomain = 3
SidTypeAlias = 4
SidTypeWellKnownGroup = 5
SidTypeInvalid = 7
SidTypeUnknown = 8
SidTypeComputer = 9
FADF_HAVEIID = 64 # Variable c_int
ERROR_TOO_MANY_CONTEXT_IDS = 1384 # Variable c_long
DT_SINGLELINE = 32 # Variable c_int
TBM_SETTOOLTIPS = 1053 # Variable c_int
MIN_PRIORITY = 1 # Variable c_int
EWX_FORCE = 4 # Variable c_int
MCI_MODE_SEEK = 528 # Variable c_int
IMAGE_REL_BASED_HIGH = 1 # Variable c_int
SPLDS_URL = 'url' # Variable POINTER(c_wchar)
MB_MISCMASK = 49152 # Variable c_long
STG_E_NOTFILEBASEDSTORAGE = -2147286777 # Variable c_long
STD_CUT = 0 # Variable c_int
DMPAPER_ENV_MONARCH = 37 # Variable c_int
XST_DATARCVD = 6 # Variable c_int
LOCKFILE_FAIL_IMMEDIATELY = 1 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER = 4100 # Variable c_long
EMR_SETVIEWPORTEXTEX = 11 # Variable c_int
ERROR_SERVICE_DEPENDENCY_DELETED = 1075 # Variable c_long
DISC_UPDATE_PROFILE = 1 # Variable c_int
CO_E_SCM_ERROR = -2146959358 # Variable c_long
LVS_NOSORTHEADER = 32768 # Variable c_int
WAVE_FORMAT_4M08 = 256 # Variable c_int
ERROR_PRINTER_HAS_JOBS_QUEUED = 3009 # Variable c_long
PRINTER_NOTIFY_FIELD_TOTAL_PAGES = 22 # Variable c_int
RESOURCEDISPLAYTYPE_TREE = 10 # Variable c_int
EMR_STROKEANDFILLPATH = 63 # Variable c_int
LANG_HUNGARIAN = 14 # Variable c_int
PD_PAGENUMS = 2 # Variable c_int
SE_ERR_OOM = 8 # Variable c_int
ERROR_COMMITMENT_LIMIT = 1455 # Variable c_long
IPPORT_CMDSERVER = 514 # Variable c_int
SM_CXSMSIZE = 52 # Variable c_int
CTRY_VIET_NAM = 84 # Variable c_int
NCBLISTEN = 17 # Variable c_int
SOCK_RAW = 3 # Variable c_int
FW_DONTCARE = 0 # Variable c_int
SUBLANG_ENGLISH_ZIMBABWE = 12 # Variable c_int
IMAGE_REL_ALPHA_REFLONG = 1 # Variable c_int
ERROR_EA_FILE_CORRUPT = 276 # Variable c_long
NTE_BAD_PROV_TYPE = -2146893804 # Variable c_long
GCP_DIACRITIC = 256 # Variable c_int
SUBLANG_ARABIC_ALGERIA = 5 # Variable c_int
VIEW_VIEWMENU = 12 # Variable c_int
WINSTA_ACCESSCLIPBOARD = 4 # Variable c_long
FR_REPLACEALL = 32 # Variable c_int
ICM_UNREGISTERICMATCHER = 6 # Variable c_int
SEMAPHORE_ALL_ACCESS = 2031619 # Variable c_long
ERROR_MORE_WRITES = 1120 # Variable c_long
DOMAIN_ALIAS_RID_REPLICATOR = 552 # Variable c_long
XST_POKESENT = 7 # Variable c_int
VIF_CANNOTCREATE = 2048 # Variable c_long
LVBKIF_STYLE_MASK = 16 # Variable c_int
STD_PASTE = 2 # Variable c_int
MIXERCONTROL_CT_SUBCLASS_MASK = 251658240 # Variable c_long
SEEK_SET = 0 # Variable c_int
FR_HIDEUPDOWN = 16384 # Variable c_int
PST_RS449 = 5 # Variable c_ulong
MCI_SET_AUDIO_LEFT = 1 # Variable c_long
IGP_GETIMEVERSION = 4294967292 # Variable c_ulong
PARAMFLAG_NONE = 0 # Variable c_int
RPC_BUFFER_EXTRA = 16384 # Variable c_int
RPC_E_INVALID_DATAPACKET = -2147418103 # Variable c_long
CDN_INCLUDEITEM = 4294966688 # Variable c_uint
IMAGE_FILE_DEBUG_STRIPPED = 512 # Variable c_int
SM_CXDLGFRAME = 7 # Variable c_int
TKIND_INTERFACE = 3
LANG_CATALAN = 3 # Variable c_int
IMAGE_REL_SH3_DIRECT8_LONG = 5 # Variable c_int
SPLDS_PRINTER_NAME = 'printerName' # Variable POINTER(c_wchar)
FILEOKSTRINGW = 'commdlg_FileNameOK' # Variable POINTER(c_wchar)
PRINTER_CHANGE_FORM = 458752 # Variable c_int
URLACTION_SHELL_MOVE_OR_COPY = 6146 # Variable c_int
DCX_CLIPSIBLINGS = 16 # Variable c_long
TTM_GETTOOLINFOA = 1032 # Variable c_int
SBM_GETSCROLLINFO = 234 # Variable c_int
CHINESEBIG5_CHARSET = 136 # Variable c_int
IMAGE_REL_MIPS_JMPADDR = 3 # Variable c_int
MIIM_CHECKMARKS = 8 # Variable c_int
IMAGE_DEBUG_TYPE_CODEVIEW = 2 # Variable c_int
PSH_USEPSTARTPAGE = 64 # Variable c_int
MEM_4MB_PAGES = 2147483648 # Variable c_uint
GCL_REVERSE_LENGTH = 3 # Variable c_int
FR_HIDEMATCHCASE = 32768 # Variable c_int
WC_COMBOBOXEXW = 'ComboBoxEx32' # Variable POINTER(c_wchar)
PERF_DISPLAY_NO_SUFFIX = 0 # Variable c_int
SPLDS_PRINT_END_TIME = 'printEndTime' # Variable POINTER(c_wchar)
MCI_WAVE_SET_ANYOUTPUT = 134217728 # Variable c_long
MCI_ANIM_GETDEVCAPS_CAN_REVERSE = 16385 # Variable c_long
MCI_SET_TIME_FORMAT = 1024 # Variable c_long
SOCK_SEQPACKET = 5 # Variable c_int
ERROR_BAD_ARGUMENTS = 160 # Variable c_long
SW_PARENTCLOSING = 1 # Variable c_int
MS_RLSD_ON = 128 # Variable c_ulong
WNNC_NET_INTERGRAPH = 1310720 # Variable c_int
WNNC_NET_LANSTEP = 524288 # Variable c_int
ERROR_BAD_DESCRIPTOR_FORMAT = 1361 # Variable c_long
CAL_SABBREVDAYNAME3 = 16 # Variable c_int
HFILE_ERROR = -1 # Variable c_int
PI_SYNCHRONOUS = 128
ABM_GETSTATE = 4 # Variable c_int
WSF_VISIBLE = 1 # Variable c_long
OLERENDER_NONE = 0
HHT_NOWHERE = 1 # Variable c_int
JOY_BUTTON5 = 16 # Variable c_long
MCIERR_OUTOFRANGE = 282 # Variable c_int
MAP_FOLDDIGITS = 128 # Variable c_int
DM_UPDATE = 1 # Variable c_int
SHGFI_SHELLICONSIZE = 4 # Variable c_int
SP_NOTREPORTED = 16384 # Variable c_int
IO_REPARSE_TAG_NSSRECOVER = 2147483654 # Variable c_uint
MCIERR_DEVICE_TYPE_REQUIRED = 287 # Variable c_int
THREAD_DIRECT_IMPERSONATION = 512 # Variable c_int
IS_TEXT_UNICODE_CONTROLS = 4 # Variable c_int
CRYPT_E_SIGNER_NOT_FOUND = -2146889714 # Variable c_long
NDR_EBCDIC_CHAR = 1 # Variable c_ulong
CO_E_ERRORINAPP = -2147221001 # Variable c_long
NTE_OP_OK = 0 # Variable c_int
DATE_LONGDATE = 2 # Variable c_int
LVN_GETINFOTIPA = 4294967139 # Variable c_uint
META_SETVIEWPORTORG = 525 # Variable c_int
XMLELEMTYPE_TEXT = 1
MAX_PATH = 260 # Variable c_int
OFN_READONLY = 1 # Variable c_int
DMRES_HIGH = -4 # Variable c_int
TVM_SETTOOLTIPS = 4376 # Variable c_int
ERROR_SIGNAL_REFUSED = 156 # Variable c_long
CTRY_ALGERIA = 213 # Variable c_int
IMPLTYPEFLAG_FSOURCE = 2 # Variable c_int
IME_SMODE_NONE = 0 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_SPEAKERS = 4 # Variable c_long
LVM_SETTOOLTIPS = 4170 # Variable c_int
LOCALE_SMONTHNAME2 = 57 # Variable c_int
TVM_GETNEXTITEM = 4362 # Variable c_int
MK_E_NEEDGENERIC = -2147221022 # Variable c_long
DT_EXPANDTABS = 64 # Variable c_int
scr8 = 1175 # Variable c_int
SM_CYDRAG = 69 # Variable c_int
SBN_FIRST = 4294966416 # Variable c_uint
ERROR_SECTOR_NOT_FOUND = 27 # Variable c_long
PERF_DISPLAY_PERCENT = 536870912 # Variable c_int
TCP_BSDURGENT = 28672 # Variable c_int
CO_E_SERVER_EXEC_FAILURE = -2146959355 # Variable c_long
LVM_SETCOLUMNORDERARRAY = 4154 # Variable c_int
RPC_C_PROTECT_LEVEL_CALL = 3 # Variable c_int
TWOSTOPBITS = 2 # Variable c_int
TTS_NOPREFIX = 2 # Variable c_int
_NLSCMPERROR = 2147483647 # Variable c_int
SRCERASE = 4457256 # Variable c_ulong
IMAGE_REL_IA64_DIR32 = 4 # Variable c_int
WS_EX_STATICEDGE = 131072 # Variable c_long
MUTANT_ALL_ACCESS = 2031617 # Variable c_long
IMAGE_REL_ALPHA_SECREL = 15 # Variable c_int
SPI_GETLOWPOWERACTIVE = 83 # Variable c_int
SPLDS_ASSET_NUMBER = 'assetNumber' # Variable POINTER(c_wchar)
META_ANIMATEPALETTE = 1078 # Variable c_int
ILC_COLOR16 = 16 # Variable c_int
ERROR_DUPLICATE_SERVICE_NAME = 1078 # Variable c_long
URLACTION_CHANNEL_SOFTDIST_PERMISSIONS = 7685 # Variable c_int
DV_E_TYMED = -2147221399 # Variable c_long
CS_E_PACKAGE_NOTFOUND = -2147221148 # Variable c_long
WGL_SWAP_OVERLAY14 = 16384 # Variable c_int
SPI_SETDESKWALLPAPER = 20 # Variable c_int
LBS_SORT = 2 # Variable c_long
META_RECTANGLE = 1051 # Variable c_int
UPDFCACHE_ALLBUTNODATACACHE = 2147483646 # Variable c_ulong
R2_LAST = 16 # Variable c_int
PST_RS422 = 3 # Variable c_ulong
PFD_TYPE_RGBA = 0 # Variable c_int
PSD_RETURNDEFAULT = 1024 # Variable c_int
HTMINBUTTON = 8 # Variable c_int
LBS_NOINTEGRALHEIGHT = 256 # Variable c_long
ERROR_INVALID_SECURITY_DESCR = 1338 # Variable c_long
HHT_BELOW = 512 # Variable c_int
CUDR_NOPRIMARY = 16 # Variable c_int
ERROR_SHUTDOWN_IN_PROGRESS = 1115 # Variable c_long
TVM_GETINDENT = 4358 # Variable c_int
SECURITY_INTERACTIVE_RID = 4 # Variable c_long
ERROR_BAD_DRIVER = 2001 # Variable c_long
DRV_DISABLE = 5 # Variable c_int
ERROR_TOO_MANY_POSTS = 298 # Variable c_long
IMAGE_ARCHIVE_START_SIZE = 8 # Variable c_int
INET_E_CANNOT_LOAD_DATA = -2146697201 # Variable c_long
FLI_MASK = 4155 # Variable c_int
DISC_NO_FORCE = 64 # Variable c_int
GL_LEVEL_WARNING = 3 # Variable c_int
PRINTACTION_OPEN = 0 # Variable c_int
TC_CR_90 = 8 # Variable c_int
WS_EX_RTLREADING = 8192 # Variable c_long
FS_ARABIC = 64 # Variable c_long
ATTR_CONVERTED = 2 # Variable c_int
BAUD_USER = 268435456 # Variable c_ulong
S_THRESHOLD = 1 # Variable c_int
ERROR_INVALID_ENVIRONMENT = 1805 # Variable c_long
ERROR_INVALID_EA_HANDLE = 278 # Variable c_long
RPC_NCA_FLAGS_IDEMPOTENT = 1 # Variable c_int
WS_CLIPCHILDREN = 33554432 # Variable c_long
TIME_FORCE24HOURFORMAT = 8 # Variable c_int
MCM_GETCOLOR = 4107 # Variable c_int
PAN_SERIF_SQUARE_COVE = 4 # Variable c_int
TIME_BYTES = 4 # Variable c_int
IME_HOTKEY_PRIVATE_FIRST = 512 # Variable c_int
ILC_PALETTE = 2048 # Variable c_int
BS_3STATE = 5 # Variable c_long
PROCESSOR_PPC_603 = 603 # Variable c_int
UDS_HORZ = 64 # Variable c_int
SERVICE_START = 16 # Variable c_int
BSF_NOHANG = 8 # Variable c_int
ERROR_DEPENDENCY_NOT_FOUND = 5002 # Variable c_long
PSH_WIZARDHASFINISH = 16 # Variable c_int
MAP_COMPOSITE = 64 # Variable c_int
URLPOLICY_DISALLOW = 3 # Variable c_int
IMAGE_SYM_CLASS_NULL = 0 # Variable c_int
DC_DUPLEX = 7 # Variable c_int
ERROR_INVALID_OPERATION = 4317 # Variable c_long
DC_ORIENTATION = 17 # Variable c_int
CBR_115200 = 115200 # Variable c_int
ERROR_JOIN_TO_SUBST = 140 # Variable c_long
RPC_E_CLIENT_CANTUNMARSHAL_DATA = -2147418100 # Variable c_long
LANG_KONKANI = 87 # Variable c_int
SM_CYMENUCHECK = 72 # Variable c_int
DISPID_DESTRUCTOR = -7 # Variable c_int
URLACTION_JAVA_PERMISSIONS = 7168 # Variable c_int
TYPEFLAG_FHIDDEN = 16
NTE_DOUBLE_ENCRYPT = -2146893806 # Variable c_long
EXCEPTION_CONTINUE_EXECUTION = -1 # Variable c_int
PGM_SETCHILD = 5121 # Variable c_int
ERROR_BAD_LENGTH = 24 # Variable c_long
RPC_S_NO_PROTSEQS_REGISTERED = 1714 # Variable c_long
CLRRTS = 4 # Variable c_int
CRYPT_E_AUTH_ATTR_MISSING = -2146889722 # Variable c_long
IMAGE_WEAK_EXTERN_SEARCH_NOLIBRARY = 1 # Variable c_int
JOY_BUTTON8 = 128 # Variable c_long
CTRY_SLOVENIA = 386 # Variable c_int
LVIS_STATEIMAGEMASK = 61440 # Variable c_int
SE_CREATE_PERMANENT_NAME = 'SeCreatePermanentPrivilege' # Variable POINTER(c_wchar)
THREAD_TERMINATE = 1 # Variable c_int
TB_INSERTBUTTONW = 1091 # Variable c_int
ERROR_OPEN_FILES = 2401 # Variable c_long
CDIS_CHECKED = 8 # Variable c_int
ERROR_INVALID_SERVICENAME = 1213 # Variable c_long
SPI_GETMOUSETRAILS = 94 # Variable c_int
SPI_GETTOGGLEKEYS = 52 # Variable c_int
RPC_C_STATS_PKTS_IN = 2 # Variable c_int
ATTR_INPUT = 0 # Variable c_int
CBEN_LAST = 4294966466 # Variable c_uint
TB_SETBUTTONINFOW = 1088 # Variable c_int
LVIR_BOUNDS = 0 # Variable c_int
RT_CURSOR = 1 # Variable POINTER(c_wchar)
MCIERR_INVALID_FILE = 296 # Variable c_int
CF_OWNERDISPLAY = 128 # Variable c_int
EMARCH_ENC_I17_IC_INST_WORD_X = 3 # Variable c_int
RBN_BEGINDRAG = 4294966461 # Variable c_uint
SS_USERITEM = 10 # Variable c_long
SUBLANG_SPANISH_DOMINICAN_REPUBLIC = 7 # Variable c_int
NI_CHANGECANDIDATELIST = 19 # Variable c_int
HS_FDIAGONAL = 2 # Variable c_int
EVENTLOG_PAIRED_EVENT_ACTIVE = 8 # Variable c_int
IMC_GETCOMPOSITIONFONT = 9 # Variable c_int
IMAGE_DEBUG_TYPE_RESERVED10 = 10 # Variable c_int
TB_GETOBJECT = 1086 # Variable c_int
VAR_FOURDIGITYEARS = 64 # Variable c_ulong
ERROR_GROUP_EXISTS = 1318 # Variable c_long
FF_SWISS = 32 # Variable c_int
TIME_MS = 1 # Variable c_int
MCI_USER_MESSAGES = 3072 # Variable c_int
PERSIST_E_NOTSELFSIZING = -2146762741 # Variable c_long
TBS_TOOLTIPS = 256 # Variable c_int
SKF_STICKYKEYSON = 1 # Variable c_int
UOI_USER_SID = 4 # Variable c_int
MCIERR_DEVICE_NOT_READY = 276 # Variable c_int
MIXER_OBJECTF_HMIDIIN = 3221225472 # Variable c_ulong
ERROR_DS_GC_NOT_AVAILABLE = 1938 # Variable c_long
stc30 = 1117 # Variable c_int
WAVE_FORMAT_1M16 = 4 # Variable c_int
STATUS_TIMEOUT = 258 # Variable c_ulong
LVCFMT_BITMAP_ON_RIGHT = 4096 # Variable c_int
DMBIN_LARGEFMT = 10 # Variable c_int
ERROR_SAME_DRIVE = 143 # Variable c_long
PRIVILEGE_SET_ALL_NECESSARY = 1 # Variable c_int
WGL_SWAP_OVERLAY6 = 64 # Variable c_int
LOCALE_INEGNUMBER = 4112 # Variable c_int
ERROR_NO_NET_OR_BAD_PATH = 1203 # Variable c_long
PRINTER_ENUM_LOCAL = 2 # Variable c_int
LVITEMA_V1_SIZE = 36 # Variable c_uint
PAN_SERIF_NORMAL_SANS = 11 # Variable c_int
LOCALE_STHOUSAND = 15 # Variable c_int
WINSTA_EXITWINDOWS = 64 # Variable c_long
VIF_WRITEPROT = 64 # Variable c_long
AF_UNIX = 1 # Variable c_int
DI_MASK = 1 # Variable c_int
MCI_VD_PLAY_SLOW = 1048576 # Variable c_long
EV_RLSD = 32 # Variable c_int
LB_GETTEXT = 393 # Variable c_int
EMR_POLYTEXTOUTA = 96 # Variable c_int
MS_NBF = 'MNBF' # Variable POINTER(c_char)
TVS_HASLINES = 2 # Variable c_int
REG_DWORD_BIG_ENDIAN = 5 # Variable c_int
MARKPARITY = 3 # Variable c_int
TIME_ONESHOT = 0 # Variable c_int
IMAGE_DEBUG_MISC_EXENAME = 1 # Variable c_int
IMAGE_REL_I386_REL16 = 2 # Variable c_int
PBS_VERTICAL = 4 # Variable c_int
DOF_MULTIPLE = 32772 # Variable c_int
LOCALE_ICALENDARTYPE = 4105 # Variable c_int
ERROR_LAST_ADMIN = 1322 # Variable c_long
CF_EFFECTS = 256 # Variable c_long
PSH_HEADER = 524288 # Variable c_int
SPLDS_PRINT_COLLATE = 'printCollate' # Variable POINTER(c_wchar)
CM_OUT_OF_GAMUT = 255 # Variable c_int
EDGE_BUMP = 9 # Variable c_int
SUBLANG_GERMAN = 1 # Variable c_int
INET_E_NO_VALID_MEDIA = -2146697206 # Variable c_long
NFS_EDIT = 1 # Variable c_int
CPS_COMPLETE = 1 # Variable c_int
QS_MOUSEBUTTON = 4 # Variable c_int
ETO_OPAQUE = 2 # Variable c_int
MCIERR_NO_INTEGER = 312 # Variable c_int
ERROR_DEPENDENCY_ALREADY_EXISTS = 5003 # Variable c_long
SETCOLORTABLE = 4 # Variable c_int
MIXERLINE_LINEF_ACTIVE = 1 # Variable c_long
DC_ICON = 4 # Variable c_int
ILC_COLORDDB = 254 # Variable c_int
IMAGE_SCN_LNK_REMOVE = 2048 # Variable c_int
HTCAPTION = 2 # Variable c_int
PRINTER_NOTIFY_FIELD_DEVMODE = 7 # Variable c_int
PROCESS_VM_WRITE = 32 # Variable c_int
FOREGROUND_RED = 4 # Variable c_int
MDIS_ALLCHILDSTYLES = 1 # Variable c_int
scr7 = 1174 # Variable c_int
IME_PROP_SPECIAL_UI = 131072 # Variable c_int
HDM_GETITEMCOUNT = 4608 # Variable c_int
RBNM_ID = 1 # Variable c_int
R2_BLACK = 1 # Variable c_int
MH_CLEANUP = 4 # Variable c_int
BSM_ALLDESKTOPS = 16 # Variable c_int
INET_E_REDIRECT_FAILED = -2146697196 # Variable c_long
RPC_X_PIPE_EMPTY = 1918 # Variable c_long
HKM_SETRULES = 1027 # Variable c_int
RB_SETBANDINFOA = 1030 # Variable c_int
TBTS_LEFT = 1 # Variable c_int
SPI_GETMOUSEKEYS = 54 # Variable c_int
MMIOERR_PATHNOTFOUND = 267 # Variable c_int
ERROR_DOMAIN_CONTROLLER_NOT_FOUND = 1908 # Variable c_long
CBF_FAIL_ADVISES = 16384 # Variable c_int
GETTECHNOLGY = 20 # Variable c_int
LVNI_BELOW = 512 # Variable c_int
SPI_SETWORKAREA = 47 # Variable c_int
LVM_GETCOLUMNW = 4191 # Variable c_int
DIFFERENCE = 11 # Variable c_int
TVN_SELCHANGEDA = 4294966894 # Variable c_uint
RPC_E_WRONG_THREAD = -2147417842 # Variable c_long
CTLCOLOR_LISTBOX = 2 # Variable c_int
ERROR_DS_CANT_MOD_OBJ_CLASS = 1936 # Variable c_long
NCBFINDNAME = 120 # Variable c_int
ERROR_EVENTLOG_CANT_START = 1501 # Variable c_long
MAC_CHARSET = 77 # Variable c_int
MMIOERR_CANNOTEXPAND = 264 # Variable c_int
RPC_E_INVALID_OBJECT = -2147417836 # Variable c_long
SWP_FRAMECHANGED = 32 # Variable c_int
SEE_MASK_IDLIST = 4 # Variable c_int
ERROR_WMI_ITEMID_NOT_FOUND = 4202 # Variable c_long
SORT_CHINESE_UNICODE = 1 # Variable c_int
CERT_E_UNTRUSTEDROOT = -2146762487 # Variable c_long
ERROR_NO_SUCH_GROUP = 1319 # Variable c_long
JOB_STATUS_PAPEROUT = 64 # Variable c_int
PSINJECT_ENDPROLOG = 6 # Variable c_int
DTM_SETFORMATW = 4146 # Variable c_int
MCI_DEVTYPE_FIRST_USER = 4096 # Variable c_int
LOCALE_SMONTHNAME10 = 65 # Variable c_int
ERROR_FILE_CORRUPT = 1392 # Variable c_long
DMPAPER_EXECUTIVE = 7 # Variable c_int
RPC_E_CALL_REJECTED = -2147418111 # Variable c_long
CAL_SMONTHNAME8 = 28 # Variable c_int
HDS_BUTTONS = 2 # Variable c_int
ERROR_INVALID_STARTING_CODESEG = 188 # Variable c_long
DM_YRESOLUTION = 8192 # Variable c_long
ERROR_NO_SYSTEM_RESOURCES = 1450 # Variable c_long
CT_CTYPE3 = 4 # Variable c_int
IP_ADD_MEMBERSHIP = 5 # Variable c_int
DATABITS_7 = 4 # Variable c_ushort
SPLDS_LOCATION = 'location' # Variable POINTER(c_wchar)
WGL_SWAP_UNDERLAY12 = 134217728 # Variable c_int
PIDSI_AUTHOR = 4 # Variable c_long
ERROR_INVALID_NETNAME = 1214 # Variable c_long
IPM_ISBLANK = 1129 # Variable c_int
MB_ERR_INVALID_CHARS = 8 # Variable c_int
IO_REPARSE_TAG_RESERVED_ZERO = 0 # Variable c_int
CDIS_DEFAULT = 32 # Variable c_int
EMR_EXTCREATEFONTINDIRECTW = 82 # Variable c_int
RPC_S_UNSUPPORTED_TRANS_SYN = 1730 # Variable c_long
TVS_NOSCROLL = 8192 # Variable c_int
SS_OWNERDRAW = 13 # Variable c_long
SETMITERLIMIT = 23 # Variable c_int
ILLUMINANT_DEVICE_DEFAULT = 0 # Variable c_int
SUBLANG_SPANISH_NICARAGUA = 19 # Variable c_int
DS_SYSMODAL = 2 # Variable c_long
SPLDS_PRINT_MAC_ADDRESS = 'printMACAddress' # Variable POINTER(c_wchar)
SB_TOP = 6 # Variable c_int
ISC_SHOWUIALL = 3221225487 # Variable c_uint
RDW_ERASE = 4 # Variable c_int
WGL_SWAP_UNDERLAY11 = 67108864 # Variable c_int
ERROR_NOT_ALL_ASSIGNED = 1300 # Variable c_long
QS_ALLPOSTMESSAGE = 256 # Variable c_int
RPC_X_WRONG_STUB_VERSION = 1829 # Variable c_long
NCBCALL = 16 # Variable c_int
RPC_S_SEND_INCOMPLETE = 1913 # Variable c_long
TAPE_DRIVE_SET_EOT_WZ_SIZE = 4194304 # Variable c_int
OLE_E_ADVISENOTSUPPORTED = -2147221501 # Variable c_long
SECURITY_IDENTIFICATION = 65536 # Variable c_int
FD_CLOSE = 32 # Variable c_int
LBN_SETFOCUS = 4 # Variable c_int
SM_CYMINSPACING = 48 # Variable c_int
TTM_SETDELAYTIME = 1027 # Variable c_int
MDMSPKR_ON = 2 # Variable c_int
cmb1 = 1136 # Variable c_int
MKF_INDICATOR = 32 # Variable c_int
ERROR_INVALID_MINALLOCSIZE = 195 # Variable c_long
S_SERBDNT = -5 # Variable c_int
TAPE_ABSOLUTE_BLOCK = 1 # Variable c_long
META_RESIZEPALETTE = 313 # Variable c_int
TVIF_SELECTEDIMAGE = 32 # Variable c_int

tagVARKIND = c_int # enum
TIMERR_NOCANDO = 97 # Variable c_int
EMR_GLSBOUNDEDRECORD = 103 # Variable c_int
SC_MANAGER_CONNECT = 1 # Variable c_int
EVENTLOG_SUCCESS = 0 # Variable c_int
XTYP_DISCONNECT = 32962 # Variable c_int
DFCS_BUTTON3STATE = 8 # Variable c_int
IMAGE_SCN_CNT_UNINITIALIZED_DATA = 128 # Variable c_int
PI_APARTMENTTHREADED = 256
JOYERR_PARMS = 165 # Variable c_int
DATA_S_SAMEFORMATETC = 262448 # Variable c_long
OLERENDER_DRAW = 1
KEY_EVENT = 1 # Variable c_int
LVS_TYPESTYLEMASK = 64512 # Variable c_int
MB_ICONHAND = 16 # Variable c_long
LINECAPS = 30 # Variable c_int
CBEIF_INDENT = 16 # Variable c_int
DSS_NORMAL = 0 # Variable c_int
SBS_SIZEBOXBOTTOMRIGHTALIGN = 4 # Variable c_long
SOCK_RDM = 4 # Variable c_int
ROTFLAGS_REGISTRATIONKEEPSALIVE = 1 # Variable c_int
GMEM_MOVEABLE = 2 # Variable c_int
SWP_NOREDRAW = 8 # Variable c_int
PORT_TYPE_READ = 2 # Variable c_int
WINDING = 2 # Variable c_int
WNCON_SLOWLINK = 4 # Variable c_int
MCI_STATUS_MEDIA_PRESENT = 5 # Variable c_long
HDM_FIRST = 4608 # Variable c_int
MCI_FORMAT_TMSF = 10 # Variable c_int
IPPROTO_RAW = 255 # Variable c_int
DMDITHER_FINE = 3 # Variable c_int
SPI_GETMOUSE = 3 # Variable c_int
EN_SETFOCUS = 256 # Variable c_int
WNNC_NET_RDR2SAMPLE = 2424832 # Variable c_int
DC_HASDEFID = 21323 # Variable c_int
TVM_DELETEITEM = 4353 # Variable c_int
_SPACE = 8 # Variable c_int
STGM_SHARE_EXCLUSIVE = 16 # Variable c_long
DM_DEFAULTSOURCE = 512 # Variable c_long
LOCALE_SMONTHNAME3 = 58 # Variable c_int
SUBLANG_SPANISH_MODERN = 3 # Variable c_int
CRYPT_E_BAD_MSG = -2146885619 # Variable c_long
URLMON_OPTION_USERAGENT = 268435457 # Variable c_int
ERROR_BAD_PATHNAME = 161 # Variable c_long
IME_PROP_AT_CARET = 65536 # Variable c_int
SOFTDIST_ADSTATE_AVAILABLE = 1 # Variable c_int
IMN_SETOPENSTATUS = 8 # Variable c_int
CC_ANYCOLOR = 256 # Variable c_int
STARTF_USEFILLATTRIBUTE = 16 # Variable c_int
HS_HORIZONTAL = 0 # Variable c_int
HKCOMB_SC = 16 # Variable c_int
EM_CANUNDO = 198 # Variable c_int
RELATIVE = 2 # Variable c_int
RB_INSERTBANDW = 1034 # Variable c_int
ENABLE_MOUSE_INPUT = 16 # Variable c_int
SPI_SETMENUDROPALIGNMENT = 28 # Variable c_int
LVM_GETEDITCONTROL = 4120 # Variable c_int
DCX_VALIDATE = 2097152 # Variable c_long
CA_LOG_FILTER = 2 # Variable c_int
ACM_OPENW = 1127 # Variable c_int
CREATE_DEFAULT_ERROR_MODE = 67108864 # Variable c_int
ERROR_INVALID_COMBOBOX_MESSAGE = 1422 # Variable c_long
CTRY_SWEDEN = 46 # Variable c_int
QUERYROPSUPPORT = 40 # Variable c_int
TVN_ITEMEXPANDINGW = 4294966842 # Variable c_uint
CS_E_FIRST = 2147746148 # Variable c_ulong
ERROR_INVALID_SID = 1337 # Variable c_long
NMPWAIT_USE_DEFAULT_WAIT = 0 # Variable c_int
cmb12 = 1147 # Variable c_int
WGL_SWAP_OVERLAY13 = 8192 # Variable c_int
VARIANT_FALSE = 0 # Variable c_short
DC_TEXT = 8 # Variable c_int
TYPE_E_WRONGTYPEKIND = -2147319766 # Variable c_long

tagTYPEFLAGS = c_int # enum
TYPEFLAG_FAPPOBJECT = 1
TYPEFLAG_FCANCREATE = 2
TYPEFLAG_FLICENSED = 4
TYPEFLAG_FPREDECLID = 8
TYPEFLAG_FCONTROL = 32
TYPEFLAG_FDUAL = 64
TYPEFLAG_FNONEXTENSIBLE = 128
TYPEFLAG_FOLEAUTOMATION = 256
MaxJobObjectInfoClass = 8
CBS_OEMCONVERT = 128 # Variable c_long
WSAEINTR = 10004 # Variable c_int
IE_BAUDRATE = -12 # Variable c_int
TOKEN_SOURCE_LENGTH = 8 # Variable c_int
BM_SETSTATE = 243 # Variable c_int
RPC_S_NAME_SERVICE_UNAVAILABLE = 1762 # Variable c_long
PROP_LG_CXDLG = 252 # Variable c_int
DMLERR_UNADVACKTIMEOUT = 16400 # Variable c_int
TB_INDETERMINATE = 1029 # Variable c_int

IMPORT_OBJECT_NAME_TYPE = c_int # enum
IMPORT_OBJECT_ORDINAL = 0
IMPORT_OBJECT_NAME = 1
IMPORT_OBJECT_NAME_NO_PREFIX = 2
IMAGE_REL_ALPHA_BRADDR = 7 # Variable c_int
IMAGE_SCN_CNT_CODE = 32 # Variable c_int
CCS_NODIVIDER = 64 # Variable c_long
JOB_NOTIFY_FIELD_DEVMODE = 9 # Variable c_int
ES_UPPERCASE = 8 # Variable c_long
VFT2_DRV_SYSTEM = 7 # Variable c_long
MCI_OVLY_PUT_FRAME = 524288 # Variable c_long
DISP_CHANGE_BADMODE = -2 # Variable c_int
NUMPRS_NEG = 65536 # Variable c_int
SORT_CHINESE_BOPOMOFO = 3 # Variable c_int
SPI_SETDOUBLECLICKTIME = 32 # Variable c_int
OLEOBJ_S_INVALIDHWND = 262530 # Variable c_long
LVIR_LABEL = 2 # Variable c_int
UDS_HOTTRACK = 256 # Variable c_int
TBCD_CHANNEL = 3 # Variable c_int
RIP_EVENT = 9 # Variable c_int
ERROR_NOT_CONNECTED = 2250 # Variable c_long
NRC_OSRESNOTAV = 53 # Variable c_int
SECURITY_IMPERSONATION = 131072 # Variable c_int
MDMVOL_HIGH = 2 # Variable c_int
WNNC_NET_10NET = 327680 # Variable c_int
DMLERR_NO_CONV_ESTABLISHED = 16394 # Variable c_int
GW_HWNDFIRST = 0 # Variable c_int
ERROR_FLOPPY_UNKNOWN_ERROR = 1124 # Variable c_long
WSAENOTEMPTY = 10066 # Variable c_int
CD_LBSELNOITEMS = -1 # Variable c_int
LOCALE_SABBREVMONTHNAME11 = 78 # Variable c_int
TCS_TABS = 0 # Variable c_int
ERROR_UNKNOWN_PROPERTY = 1608 # Variable c_long
IME_ESC_PRIVATE_FIRST = 2048 # Variable c_int
NTE_PERM = -2146893808 # Variable c_long
SIOCSLOWAT = 2147775234 # Variable c_ulong
DMLERR_UNFOUND_QUEUE_ID = 16401 # Variable c_int
SPLDS_PRINT_RATE = 'printRate' # Variable POINTER(c_wchar)
CS_OWNDC = 32 # Variable c_int
JOB_NOTIFY_FIELD_NOTIFY_NAME = 4 # Variable c_int
TAPE_FORMAT = 5 # Variable c_long
AC_LINE_OFFLINE = 0 # Variable c_int
SND_NOSTOP = 16 # Variable c_int
TBN_RESET = 4294966591 # Variable c_uint
LOCK_ONLYONCE = 4
STRICT = 1 # Variable c_int
TLS_OUT_OF_INDEXES = 4294967295 # Variable c_ulong
RPC_S_UNSUPPORTED_TYPE = 1732 # Variable c_long
MCM_GETMONTHRANGE = 4103 # Variable c_int
CONNDLG_HIDE_BOX = 8 # Variable c_int
ERROR_DC_NOT_FOUND = 1425 # Variable c_long
WNFMT_ABBREVIATED = 2 # Variable c_int
IME_SMODE_SINGLECONVERT = 2 # Variable c_int
TYPE1_FONTTYPE = 131072 # Variable c_int
PROCESSOR_HITACHI_SH4 = 10005 # Variable c_int
ERROR_NULL_LM_PASSWORD = 1304 # Variable c_long
DMBIN_MIDDLE = 3 # Variable c_int
IMAGE_REL_ALPHA_INLINE_REFLONG = 9 # Variable c_int
SPI_GETSHOWSOUNDS = 56 # Variable c_int
SS_GRAYRECT = 5 # Variable c_long
REG_OPENED_EXISTING_KEY = 2 # Variable c_long
ERROR_IS_SUBST_TARGET = 149 # Variable c_long
DM_COPIES = 256 # Variable c_long
SM_CXMIN = 28 # Variable c_int
MCHT_CALENDARDATE = 131073 # Variable c_int
rad10 = 1065 # Variable c_int
RESOURCEUSAGE_ALL = 19 # Variable c_int
SCS_OS216_BINARY = 5 # Variable c_int
HELP_COMMAND = 258 # Variable c_long
DMPAPER_DSHEET = 25 # Variable c_int
LVCFMT_COL_HAS_IMAGES = 32768 # Variable c_int
TAPE_SPACE_END_OF_DATA = 4 # Variable c_long
COMPRESSION_ENGINE_STANDARD = 0 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_FIRST = 0 # Variable c_long
WT_EXECUTEINIOTHREAD = 1 # Variable c_int
LVN_ITEMCHANGING = 4294967196 # Variable c_uint
ACCESS_ALLOWED_ACE_TYPE = 0 # Variable c_int
RPC_E_NO_SYNC = -2147417824 # Variable c_long
TCM_GETCURFOCUS = 4911 # Variable c_int
CBEN_INSERTITEM = 4294966495 # Variable c_uint
PAN_SERIF_SQUARE = 6 # Variable c_int
JOY_BUTTON24 = 8388608 # Variable c_long
RPC_S_NO_ENDPOINT_FOUND = 1708 # Variable c_long
IN_CLASSC_HOST = 255 # Variable c_int
GL_ID_NODICTIONARY = 16 # Variable c_int
PROFILE_USER = 268435456 # Variable c_int
EN_CHANGE = 768 # Variable c_int
LVCOLUMNW_V1_SIZE = 24 # Variable c_uint
FILE_READ_EA = 8 # Variable c_int
C3_FULLWIDTH = 128 # Variable c_int
BATTERY_FLAG_CHARGING = 8 # Variable c_int
TTN_FIRST = 4294966776 # Variable c_uint
ESB_DISABLE_LEFT = 1 # Variable c_int
MDM_TONE_DIAL = 256 # Variable c_int
VIEW_NEWFOLDER = 11 # Variable c_int
EMARCH_ENC_I17_IC_VAL_POS_X = 21 # Variable c_int
LOCALE_NEUTRAL = 0 # Variable c_ulong
MMIO_READ = 0 # Variable c_int
SPLDS_PRINT_MIN_Y_EXTENT = 'printMinYExtent' # Variable POINTER(c_wchar)
ERROR_IS_JOINED = 134 # Variable c_long
TC_VA_ABLE = 16384 # Variable c_int
DOMAIN_ALIAS_RID_POWER_USERS = 547 # Variable c_long
MK_SHIFT = 4 # Variable c_int
OLE_E_NOCACHE = -2147221498 # Variable c_long
IME_ITHOTKEY_RECONVERTSTRING = 515 # Variable c_int
MMIO_DENYWRITE = 32 # Variable c_int
DTM_GETSYSTEMTIME = 4097 # Variable c_int
ERROR_GROUP_NOT_ONLINE = 5014 # Variable c_long
IMAGE_SYM_CLASS_SECTION = 104 # Variable c_int
SPAPI_E_NO_CONFIGMGR_SERVICES = -2146500061 # Variable c_long
EVENTLOG_START_PAIRED_EVENT = 1 # Variable c_int
SPAPI_E_NO_COMPAT_DRIVERS = -2146500056 # Variable c_long
PD_NOPAGENUMS = 8 # Variable c_int
SM_CYMENU = 15 # Variable c_int
RPC_S_INTERFACE_NOT_FOUND = 1759 # Variable c_long
SECURITY_ANONYMOUS_LOGON_RID = 7 # Variable c_long
SOFTDIST_FLAG_USAGE_PRECACHE = 2 # Variable c_int
RPC_NCA_FLAGS_BROADCAST = 2 # Variable c_int
PAGE_EXECUTE_WRITECOPY = 128 # Variable c_int
PWR_SUSPENDRESUME = 2 # Variable c_int
DTR_CONTROL_ENABLE = 1 # Variable c_int
JOY_BUTTON10 = 512 # Variable c_long
DT_METAFILE = 5 # Variable c_int
BF_DIAGONAL_ENDTOPLEFT = 19 # Variable c_int
GCP_LIGATE = 32 # Variable c_int
FILE_ATTRIBUTE_READONLY = 1 # Variable c_int
IMAGE_REL_ALPHA_REFQUAD = 2 # Variable c_int
IMAGE_SIZEOF_NT_OPTIONAL32_HEADER = 224 # Variable c_int
HDF_BITMAP_ON_RIGHT = 4096 # Variable c_int
GDT_NONE = 1 # Variable c_int
MEM_TOP_DOWN = 1048576 # Variable c_int
CB_SETLOCALE = 345 # Variable c_int
ARW_UP = 4 # Variable c_long
ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE = 5030 # Variable c_long
NI_SETCANDIDATE_PAGESTART = 22 # Variable c_int
URLACTION_INFODELIVERY_NO_REMOVING_CHANNELS = 7426 # Variable c_int
CREATE_SUSPENDED = 4 # Variable c_int
LPSTR_TEXTCALLBACKW = 4294967295 # Variable POINTER(c_wchar)
LVS_REPORT = 1 # Variable c_int
TB_MAPACCELERATORA = 1102 # Variable c_int
SB_SETTEXTW = 1035 # Variable c_int
WAVE_FORMAT_4S08 = 512 # Variable c_int
EMARCH_ENC_I17_IMM41c_SIZE_X = 23 # Variable c_int
MIXER_OBJECTF_HWAVEIN = 2684354560 # Variable c_ulong
SO_RCVBUF = 4098 # Variable c_int
LVM_GETITEMTEXTW = 4211 # Variable c_int
IMAGE_SCN_MEM_READ = 1073741824 # Variable c_int
HELP_SETCONTENTS = 5 # Variable c_long
IMPLTYPEFLAG_FDEFAULTVTABLE = 8 # Variable c_int
SC_CLOSE = 61536 # Variable c_int
RPC_C_IMP_LEVEL_ANONYMOUS = 1 # Variable c_int
SB_HORZ = 0 # Variable c_int
SUBLANG_SERBIAN_CYRILLIC = 3 # Variable c_int
LVS_EX_REGIONAL = 512 # Variable c_int
RC_STRETCHDIB = 8192 # Variable c_int
ERROR_LABEL_TOO_LONG = 154 # Variable c_long
RT_PLUGPLAY = 25 # Variable POINTER(c_wchar)
TCS_TOOLTIPS = 16384 # Variable c_int
META_STRETCHBLT = 2851 # Variable c_int
ERROR_MEMBER_NOT_IN_ALIAS = 1377 # Variable c_long
STGM_WRITE = 1 # Variable c_long
RTS_CONTROL_DISABLE = 0 # Variable c_int
COLOR_ACTIVECAPTION = 2 # Variable c_int
RPC_S_INVALID_VERS_OPTION = 1756 # Variable c_long
VOS_OS216 = 131072 # Variable c_long
SCS_SETSTR = 9 # Variable c_int
EMARCH_ENC_I17_SIGN_VAL_POS_X = 63 # Variable c_int
LZERROR_BADINHANDLE = -1 # Variable c_int
NLS_IME_DISABLE = 536870912 # Variable c_int
OFN_NOTESTFILECREATE = 65536 # Variable c_int
MIXER_GETLINECONTROLSF_ONEBYTYPE = 2 # Variable c_long
CC_ENABLETEMPLATEHANDLE = 64 # Variable c_int
PCF_RLSD = 4 # Variable c_ulong
STG_E_PROPSETMISMATCHED = -2147286800 # Variable c_long
VIFF_FORCEINSTALL = 1 # Variable c_int
MIXERCONTROL_CT_CLASS_FADER = 1342177280 # Variable c_long
JOB_CONTROL_CANCEL = 3 # Variable c_int
CO_E_INIT_TLS_SET_CHANNEL_CONTROL = -2147467253 # Variable c_long
ERROR_INVALID_LIST_FORMAT = 153 # Variable c_long

tagVARFLAGS = c_int # enum
VARFLAG_FREADONLY = 1
VARFLAG_FSOURCE = 2
VARFLAG_FREQUESTEDIT = 8
VARFLAG_FDISPLAYBIND = 16
GRADIENT_FILL_OP_FLAG = 255 # Variable c_int
DD_DEFSCROLLINSET = 11 # Variable c_int
MCI_OPEN_ALIAS = 1024 # Variable c_long
rct3 = 1082 # Variable c_int
SM_CXFULLSCREEN = 16 # Variable c_int
MM_MOM_POSITIONCB = 970 # Variable c_int
IOCPARM_MASK = 127 # Variable c_int
URLACTION_CHANNEL_SOFTDIST_MAX = 7935 # Variable c_int
CF_DSPMETAFILEPICT = 131 # Variable c_int
PI_CLSIDLOOKUP = 32
DFCS_CAPTIONMIN = 1 # Variable c_int
PFD_STEREO_DONTCARE = 2147483648 # Variable c_uint
NTE_BAD_FLAGS = -2146893815 # Variable c_long
DC_BINADJUST = 19 # Variable c_int
DBG_TERMINATE_PROCESS = 1073807364 # Variable c_ulong
SPLDS_PRINT_MAX_Y_EXTENT = 'printMaxYExtent' # Variable POINTER(c_wchar)
HIST_VIEWTREE = 4 # Variable c_int
IGP_SENTENCE = 12 # Variable c_int
SM_CXICONSPACING = 38 # Variable c_int
APD_STRICT_UPGRADE = 1 # Variable c_int
MIXERCONTROL_CONTROLTYPE_SIGNEDMETER = 268566528 # Variable c_long
TYPE_E_ELEMENTNOTFOUND = -2147319765 # Variable c_long
TBM_GETLINESIZE = 1048 # Variable c_int
READ_CONTROL = 131072 # Variable c_long
STYLE_DESCRIPTION_SIZE = 32 # Variable c_int
PSWIZB_DISABLEDFINISH = 8 # Variable c_int
PRINTER_NOTIFY_FIELD_STATUS_STRING = 19 # Variable c_int
ERROR_DS_NO_ATTRIBUTE_OR_VALUE = 1923 # Variable c_long
MCI_SYSINFO_INSTALLNAME = 2048 # Variable c_long
ERROR_INVALID_WINDOW_HANDLE = 1400 # Variable c_long
MM_MIXM_CONTROL_CHANGE = 977 # Variable c_int
ACN_STOP = 2 # Variable c_int
PSINJECT_ENDSTREAM = 16 # Variable c_int
RBS_REGISTERDROP = 4096 # Variable c_int
CDRF_NOTIFYSUBITEMDRAW = 32 # Variable c_int
TPM_RETURNCMD = 256 # Variable c_long
MCI_RESUME = 2133 # Variable c_int
EMR_SETBRUSHORGEX = 13 # Variable c_int
LANG_DUTCH = 19 # Variable c_int
RPC_C_AUTHN_WINNT = 10 # Variable c_int
WVR_ALIGNBOTTOM = 64 # Variable c_int
DTN_WMKEYDOWNW = 4294966552 # Variable c_uint
TVN_BEGINRDRAGA = 4294966888 # Variable c_uint
WHDR_BEGINLOOP = 4 # Variable c_int
PSM_GETTABCONTROL = 1140 # Variable c_int
RPC_E_CANTCALLOUT_AGAIN = -2147418095 # Variable c_long
IDC_SIZE = 206400 # Variable POINTER(c_wchar)
WSAEADDRNOTAVAIL = 10049 # Variable c_int
ACCESS_DS_OBJECT_TYPE_NAME_A = 'Directory Service Object' # Variable POINTER(c_char)
ST_ISLOCAL = 4 # Variable c_int
PSD_ENABLEPAGEPAINTHOOK = 262144 # Variable c_int
UDS_ALIGNLEFT = 8 # Variable c_int
TAPE_PSEUDO_LOGICAL_BLOCK = 3 # Variable c_long
DROPEFFECT_LINK = 4 # Variable c_int
PF_FLOATING_POINT_PRECISION_ERRATA = 0 # Variable c_int
SBM_SETRANGEREDRAW = 230 # Variable c_int
MIXER_GETCONTROLDETAILSF_QUERYMASK = 15 # Variable c_long
PBT_APMPOWERSTATUSCHANGE = 10 # Variable c_int
MCI_INFO_MEDIA_IDENTITY = 2048 # Variable c_long
CO_E_WRONG_SERVER_IDENTITY = -2147467243 # Variable c_long
PGM_SETBKCOLOR = 5124 # Variable c_int
TCM_GETITEMA = 4869 # Variable c_int
SEE_MASK_NO_CONSOLE = 32768 # Variable c_int
PARAMFLAG_FRETVAL = 8 # Variable c_int
PRINTER_STATUS_PAPER_OUT = 16 # Variable c_int
HTHELP = 21 # Variable c_int
DISP_E_NOTACOLLECTION = -2147352559 # Variable c_long
LOCALE_SNATIVECTRYNAME = 8 # Variable c_int
LC_POLYLINE = 2 # Variable c_int
GGO_NATIVE = 2 # Variable c_int
RPC_INTERFACE_HAS_PIPES = 1 # Variable c_int
STG_E_ACCESSDENIED = -2147287035 # Variable c_long
JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO = 4 # Variable c_int
SETLINEJOIN = 22 # Variable c_int
CS_GLOBALCLASS = 16384 # Variable c_int
LBS_HASSTRINGS = 64 # Variable c_long
SS_GRAYFRAME = 8 # Variable c_long
PI_CLASSINSTALL = 512
DM_ICMINTENT = 16777216 # Variable c_long
PRINTER_NOTIFY_INFO_DISCARDED = 1 # Variable c_int
OLERENDER_FORMAT = 2
JOY_RETURNY = 2 # Variable c_long
JOB_STATUS_PAUSED = 1 # Variable c_int
VARIABLE_PITCH = 2 # Variable c_int
ERROR_NO_VOLUME_LABEL = 125 # Variable c_long
DTM_SETRANGE = 4100 # Variable c_int
MB_SETFOREGROUND = 65536 # Variable c_long
DTS_LONGDATEFORMAT = 4 # Variable c_int
HKM_GETHOTKEY = 1026 # Variable c_int
THREAD_GET_CONTEXT = 8 # Variable c_int
LOCALE_ILZERO = 18 # Variable c_int
CTLCOLOR_SCROLLBAR = 5 # Variable c_int
SECURITY_ENTERPRISE_CONTROLLERS_RID = 9 # Variable c_long
DISPLAY_DEVICE_ATTACHED_TO_DESKTOP = 1 # Variable c_int
GCW_ATOM = -32 # Variable c_int
VIF_CANNOTREADDST = 131072 # Variable c_long
SM_CLEANBOOT = 67 # Variable c_int
IMN_SETSENTENCEMODE = 7 # Variable c_int
DM_ORIENTATION = 1 # Variable c_long
BDR_SUNKENINNER = 8 # Variable c_int
IMAGE_REL_PPC_SECTION = 12 # Variable c_int
INET_E_DATA_NOT_AVAILABLE = -2146697209 # Variable c_long
GGL_INDEX = 2 # Variable c_int
CBS_DROPDOWN = 2 # Variable c_long
RDW_FRAME = 1024 # Variable c_int
stc7 = 1094 # Variable c_int
AF_HYLINK = 15 # Variable c_int
INET_E_ERROR_FIRST = -2146697214 # Variable c_long
BF_RIGHT = 4 # Variable c_int
SEVERITY_SUCCESS = 0 # Variable c_int
PS_SOLID = 0 # Variable c_int
CB_GETEXTENDEDUI = 342 # Variable c_int
WSANO_DATA = 11004 # Variable c_int
PERF_TEXT_ASCII = 65536 # Variable c_int
MCI_ANIM_INFO_TEXT = 65536 # Variable c_long
RPC_C_IMP_LEVEL_DELEGATE = 4 # Variable c_int
MB_DEFBUTTON4 = 768 # Variable c_long
CRYPT_E_REVOKED = -2146885616 # Variable c_long
MOD_RIGHT = 16384 # Variable c_int
HANDLE_FLAG_INHERIT = 1 # Variable c_int
IMAGE_REL_I386_SECTION = 10 # Variable c_int
PRINTACTION_DOCUMENTDEFAULTS = 6 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_DIGITAL = 4097 # Variable c_long
ES_AUTOHSCROLL = 128 # Variable c_long
SM_CYFRAME = 33 # Variable c_int
CBS_LOWERCASE = 16384 # Variable c_long
HTMENU = 5 # Variable c_int
MEVT_SHORTMSG = 0 # Variable c_ubyte
TBSTYLE_BUTTON = 0 # Variable c_int
RPC_S_WRONG_KIND_OF_BINDING = 1701 # Variable c_long
PROCESSOR_MOTOROLA_821 = 821 # Variable c_int
MCI_NOTIFY_ABORTED = 4 # Variable c_int
JOY_BUTTON27 = 67108864 # Variable c_long
HDF_LEFT = 0 # Variable c_int
SC_SIZE = 61440 # Variable c_int
EV_BREAK = 64 # Variable c_int
SO_LINGER = 128 # Variable c_int
DC_SMALLCAP = 2 # Variable c_int
IMAGE_REL_ALPHA_REFLONGNB = 16 # Variable c_int
DMPAPER_ENV_PERSONAL = 38 # Variable c_int
DMDUP_VERTICAL = 2 # Variable c_int
IMEVER_0400 = 262144 # Variable c_int
MCIERR_WAVE_INPUTSINUSE = 322 # Variable c_int
MCI_VD_GETDEVCAPS_CLV = 65536 # Variable c_long
SPI_GETWINDOWSEXTENSION = 92 # Variable c_int
HICF_ARROWKEYS = 2 # Variable c_int
CDERR_GENERALCODES = 0 # Variable c_int
DM_MODIFY = 8 # Variable c_int
SPAPI_E_INVALID_PROPPAGE_PROVIDER = -2146500060 # Variable c_long
CFS_EXCLUDE = 128 # Variable c_int
IME_ESC_GETHELPFILENAME = 4107 # Variable c_int
MCI_SEQ_STATUS_SLAVE = 16391 # Variable c_long
EXTTEXTOUT = 512 # Variable c_int
ERROR_INVALID_COMPUTERNAME = 1210 # Variable c_long
lst9 = 1128 # Variable c_int
SUBLANG_SPANISH_CHILE = 13 # Variable c_int
GWL_ID = -12 # Variable c_int
PGF_SCROLLUP = 1 # Variable c_int
rct4 = 1083 # Variable c_int
NUMPRS_USE_ALL = 4096 # Variable c_int
EM_SETPASSWORDCHAR = 204 # Variable c_int
URLACTION_ACTIVEX_MIN = 4608 # Variable c_int
SPAPI_E_DEVINFO_NOT_REGISTERED = -2146500088 # Variable c_long
MMSYSERR_INVALIDALIAS = 13 # Variable c_int
N_BTSHFT = 4 # Variable c_int
XTYP_ERROR = 32770 # Variable c_int
ES_CONTINUOUS = 2147483648 # Variable c_ulong
SND_ALIAS_START = 0 # Variable c_int
IMAGE_SUBSYSTEM_WINDOWS_GUI = 2 # Variable c_int
HTVSCROLL = 7 # Variable c_int
NM_DBLCLK = 4294967293 # Variable c_uint
GETPAIRKERNTABLE = 258 # Variable c_int
JOB_OBJECT_ALL_ACCESS = 2031647 # Variable c_long
BAUD_128K = 65536 # Variable c_ulong
LVM_GETITEMW = 4171 # Variable c_int
SW_SHOWNOACTIVATE = 4 # Variable c_int
JOB_ALL_ACCESS = 983056 # Variable c_long
TAPE_DRIVE_GET_LOGICAL_BLK = 2097152 # Variable c_int
PAN_LETT_OBLIQUE_OFF_CENTER = 14 # Variable c_int
QUERY_CAN_NAVIGATE = 7
JOY_BUTTON12 = 2048 # Variable c_long
RPC_C_MGMT_INQ_STATS = 2 # Variable c_int
ACCESS_MAX_MS_OBJECT_ACE_TYPE = 8 # Variable c_int
PROP_MED_CYDLG = 215 # Variable c_int
PRINTER_ATTRIBUTE_PUBLISHED = 8192 # Variable c_int
DATA_S_FIRST = 262448 # Variable c_long
TBSTATE_WRAP = 32 # Variable c_int
PROCESSOR_PPC_601 = 601 # Variable c_int
OLEIVERB_INPLACEACTIVATE = -5 # Variable c_long
JOYCAPS_HASPOV = 16 # Variable c_int
THREAD_BASE_PRIORITY_LOWRT = 15 # Variable c_int
GET_TAPE_DRIVE_INFORMATION = 1 # Variable c_int
DMBIN_ENVELOPE = 5 # Variable c_int
PERF_AVERAGE_TIMER = 805438464 # Variable c_int
WSAEISCONN = 10056 # Variable c_int
ERROR_INVALID_FORM_SIZE = 1903 # Variable c_long
JOB_NOTIFY_FIELD_STATUS = 10 # Variable c_int
IMAGE_VXD_SIGNATURE = 17740 # Variable c_int
FILE_SHARE_DELETE = 4 # Variable c_int
TVGN_FIRSTVISIBLE = 5 # Variable c_int
SUBLANG_KASHMIRI_INDIA = 2 # Variable c_int
ERROR_WMI_INVALID_REGINFO = 4211 # Variable c_long
META_ESCAPE = 1574 # Variable c_int
JOY_CAL_READ4 = 524288 # Variable c_long
TCM_SETITEMW = 4925 # Variable c_int
WPF_RESTORETOMAXIMIZED = 2 # Variable c_int
ERROR_RESMON_ONLINE_FAILED = 5018 # Variable c_long
GCPCLASS_PREBOUNDRTL = 64 # Variable c_int
NIF_TIP = 4 # Variable c_int
IMAGE_REL_SH3_SIZEOF_SECTION = 13 # Variable c_int
TAPE_DRIVE_WRITE_SHORT_FMKS = 2214592512 # Variable c_uint
DMPAPER_A4_PLUS = 60 # Variable c_int
MCM_SETFIRSTDAYOFWEEK = 4111 # Variable c_int
LANG_TURKISH = 31 # Variable c_int
PAN_MIDLINE_LOW_TRIMMED = 11 # Variable c_int
ERROR_CONNECTION_ABORTED = 1236 # Variable c_long
PERF_TEXT_UNICODE = 0 # Variable c_int
ERROR_MEDIA_CHANGED = 1110 # Variable c_long
LCMAP_FULLWIDTH = 8388608 # Variable c_int
CO_E_NOTINITIALIZED = -2147221008 # Variable c_long
PRINTER_HTML_VIEW_INSTALL = 2 # Variable c_int
rad11 = 1066 # Variable c_int
COMMON_LVB_GRID_HORIZONTAL = 1024 # Variable c_int
FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 8192 # Variable c_int
CC_CIRCLES = 1 # Variable c_int
WGL_SWAP_UNDERLAY15 = 1073741824 # Variable c_int
RPC_X_BAD_STUB_DATA = 1783 # Variable c_long
STDOLE2_MINORVERNUM = 0 # Variable c_int
SPI_SETFILTERKEYS = 51 # Variable c_int
TB_GETPADDING = 1110 # Variable c_int
MCI_FORMAT_MSF = 2 # Variable c_int
SPI_SETICONTITLEWRAP = 26 # Variable c_int
SC_NEXTWINDOW = 61504 # Variable c_int
JOY_POVRIGHT = 9000 # Variable c_int
CBS_OWNERDRAWVARIABLE = 32 # Variable c_long
MIXERCONTROL_CONTROLTYPE_UNSIGNEDMETER = 268632064 # Variable c_long
MCI_PUT = 2114 # Variable c_int
LVN_ENDLABELEDITA = 4294967190 # Variable c_uint
MCIERR_PARAM_OVERFLOW = 268 # Variable c_int
LCS_DEVICE_RGB = 1 # Variable c_long
GRADIENT_FILL_TRIANGLE = 2 # Variable c_int
HDN_BEGINTRACKW = 4294966970 # Variable c_uint
ERROR_NO_BROWSER_SERVERS_FOUND = 6118 # Variable c_long
LVCF_ORDER = 32 # Variable c_int
PSH_USEICONID = 4 # Variable c_int
IMAGE_SCN_LNK_OTHER = 256 # Variable c_int
TB_GETIMAGELIST = 1073 # Variable c_int
PRINTER_CHANGE_ADD_FORM = 65536 # Variable c_int
LVS_SHOWSELALWAYS = 8 # Variable c_int
UDN_DELTAPOS = 4294966574 # Variable c_uint
IMAGE_SUBSYSTEM_OS2_CUI = 5 # Variable c_int
FNERR_BUFFERTOOSMALL = 12291 # Variable c_int
OEM_CHARSET = 255 # Variable c_int
FILE_FILE_COMPRESSION = 16 # Variable c_int
PRINTER_CHANGE_PRINT_PROCESSOR = 117440512 # Variable c_int
JOY_CAL_READ3 = 262144 # Variable c_long
FILE_NOTIFY_CHANGE_DIR_NAME = 2 # Variable c_int
TIME_ZONE_ID_UNKNOWN = 0 # Variable c_int
PSM_UNCHANGED = 1133 # Variable c_int
TBSTYLE_CHECKGROUP = 6 # Variable c_int
E_NOINTERFACE = -2147467262 # Variable c_long
DLL_PROCESS_DETACH = 0 # Variable c_int
RESTORE_CTM = 4100 # Variable c_int
IN_CLASSB_MAX = 65536 # Variable c_int
TB_SETEXTENDEDSTYLE = 1108 # Variable c_int
RPC_E_SERVER_CANTUNMARSHAL_DATA = -2147418098 # Variable c_long
TIMERR_STRUCT = 129 # Variable c_int
CTRY_HONG_KONG = 852 # Variable c_int
MMIO_FINDLIST = 64 # Variable c_int
ST_INLIST = 64 # Variable c_int
ERROR_NO_TRUST_SAM_ACCOUNT = 1787 # Variable c_long
IME_CMODE_FIXED = 2048 # Variable c_int
MCI_ALL_DEVICE_ID = 4294967295 # Variable c_uint
WSAESTALE = 10070 # Variable c_int
DROPEFFECT_COPY = 1 # Variable c_int
CDS_UPDATEREGISTRY = 1 # Variable c_int
PSH_HASHELP = 512 # Variable c_int
ERROR_CORE_RESOURCE = 5026 # Variable c_long
SM_CYHSCROLL = 3 # Variable c_int
DS_CONTEXTHELP = 8192 # Variable c_long
CDN_INITDONE = 4294966695 # Variable c_uint
MMIO_DELETE = 512 # Variable c_int
ERROR_NO_WILDCARD_CHARACTERS = 1417 # Variable c_long
PRINTER_READ = 131080 # Variable c_long
MK_MBUTTON = 16 # Variable c_int
GWL_EXSTYLE = -20 # Variable c_int
VTBIT_DECIMAL = 16384 # Variable c_int
DV_E_LINDEX = -2147221400 # Variable c_long
TBN_BEGINADJUST = 4294966593 # Variable c_uint
LVM_GETORIGIN = 4137 # Variable c_int
C3_IDEOGRAPH = 256 # Variable c_int
MCI_INFO_FILE = 512 # Variable c_long
OFN_LONGNAMES = 2097152 # Variable c_int
PAN_SERIFSTYLE_INDEX = 1 # Variable c_int
PERF_COUNTER_TIMER = 541132032 # Variable c_int
SND_ALIAS_SYSTEMQUESTION = 16211 # Variable c_ulong
UNICODE_NULL = '\x00' # Variable c_wchar
EM_GETHANDLE = 189 # Variable c_int
IPPORT_NETSTAT = 15 # Variable c_int
EMR_BITBLT = 76 # Variable c_int
MEVT_LONGMSG = 128 # Variable c_ubyte
URLACTION_HTML_SUBMIT_FORMS_FROM = 5634 # Variable c_int
RPC_S_CALL_IN_PROGRESS = 1791 # Variable c_long
CBF_SKIP_ALLNOTIFICATIONS = 3932160 # Variable c_int
PERF_DATA_VERSION = 1 # Variable c_int
TC_OP_CHARACTER = 1 # Variable c_int
DN_DEFAULTPRN = 1 # Variable c_int
PRINTACTION_PROPERTIES = 1 # Variable c_int
SPLDS_PRINT_MAX_X_EXTENT = 'printMaxXExtent' # Variable POINTER(c_wchar)
THREAD_SUSPEND_RESUME = 2 # Variable c_int
ACL_REVISION3 = 3 # Variable c_int
SPAPI_E_INVALID_REG_PROPERTY = -2146500087 # Variable c_long
scr5 = 1172 # Variable c_int
PRINTER_CONTROL_PAUSE = 1 # Variable c_int
ERROR_INVALID_HOOK_FILTER = 1426 # Variable c_long
LANG_FARSI = 41 # Variable c_int
CF_LOCALE = 16 # Variable c_int
MOD_MIDIPORT = 1 # Variable c_int
ABN_POSCHANGED = 1 # Variable c_int
IDI_APPLICATION = 206098 # Variable POINTER(c_wchar)
TAPE_DRIVE_ERASE_BOP_ONLY = 64 # Variable c_int
JOY_BUTTON28 = 134217728 # Variable c_long
SUBLANG_ARABIC_KUWAIT = 13 # Variable c_int
CB_SELECTSTRING = 333 # Variable c_int
S_WHITE512 = 4 # Variable c_int
IMAGE_REL_IA64_SECREL32 = 14 # Variable c_int
GL_ID_TOOMANYSTROKE = 34 # Variable c_int
SPLDS_PRINTER_NAME_ALIASES = 'printerNameAliases' # Variable POINTER(c_wchar)
SHGFI_LARGEICON = 0 # Variable c_int
TCN_SELCHANGING = 4294966744 # Variable c_uint
HICF_RESELECT = 64 # Variable c_int
PSD_MINMARGINS = 1 # Variable c_int
PAN_LETT_NORMAL_WEIGHTED = 3 # Variable c_int
WIZ_BODYCX = 184 # Variable c_int
CLIPBRD_E_CANT_CLOSE = -2147221036 # Variable c_long
TCM_GETCURSEL = 4875 # Variable c_int
DL_CANCELDRAG = 1160 # Variable c_int
WH_HARDWARE = 8 # Variable c_int
LVN_GETDISPINFOA = 4294967146 # Variable c_uint
WB_ISDELIMITER = 2 # Variable c_int
VTBIT_R8 = 32 # Variable c_int
IMAGE_REL_ALPHA_MATCH = 13 # Variable c_int
CC_FULLOPEN = 2 # Variable c_int
TMPF_VECTOR = 2 # Variable c_int
TAPE_DRIVE_SET_REPORT_SMKS = 2147485696 # Variable c_uint
NM_OUTOFMEMORY = 4294967295 # Variable c_uint
ANIMATE_CLASSA = 'SysAnimate32' # Variable POINTER(c_char)
MIDIPROP_TIMEDIV = 1 # Variable c_long
SERKF_SERIALKEYSON = 1 # Variable c_int
stc15 = 1102 # Variable c_int
IMAGE_NT_SIGNATURE = 17744 # Variable c_int
IMAGE_REL_ALPHA_SECRELHI = 18 # Variable c_int

tagTYSPEC = c_int # enum
TYSPEC_MIMETYPE = 2
TYSPEC_PROGID = 3
TYSPEC_IID = 4
EMARCH_ENC_I17_IMM41c_INST_WORD_X = 2 # Variable c_int
_LOWER = 2 # Variable c_int
LOCALE_SABBREVMONTHNAME12 = 79 # Variable c_int
MKSYS_GENERICCOMPOSITE = 1
REGULAR_FONTTYPE = 1024 # Variable c_int
GCP_JUSTIFY = 65536 # Variable c_long
MCI_UPDATE = 2132 # Variable c_int
ES_WANTRETURN = 4096 # Variable c_long
LVS_SINGLESEL = 4 # Variable c_int
TKF_INDICATOR = 32 # Variable c_int
CTRY_SYRIA = 963 # Variable c_int
ISC_SHOWUIGUIDELINE = 1073741824 # Variable c_int
TTM_GETTEXTW = 1080 # Variable c_int
SEEK_END = 2 # Variable c_int
TVM_GETEDITCONTROL = 4367 # Variable c_int
SM_CYDOUBLECLK = 37 # Variable c_int
FILE_TRAVERSE = 32 # Variable c_int
CBEM_GETITEMA = 1028 # Variable c_int
TRUST_E_COUNTER_SIGNER = -2146869245 # Variable c_long
SET_MIRROR_MODE = 4110 # Variable c_int
FACILITY_ITF = 4 # Variable c_int
ES_READONLY = 2048 # Variable c_long
WH_FOREGROUNDIDLE = 11 # Variable c_int
MAXINTATOM = 49152 # Variable c_int
DISP_E_BADINDEX = -2147352565 # Variable c_long
STATUS_FLOAT_INVALID_OPERATION = 3221225616 # Variable c_ulong
TC_EA_DOUBLE = 512 # Variable c_int
SYSTEM_AUDIT_OBJECT_ACE_TYPE = 7 # Variable c_int
DMTT_DOWNLOAD = 2 # Variable c_int
SUBLANG_DUTCH_BELGIAN = 2 # Variable c_int
TBDDRET_DEFAULT = 0 # Variable c_int
PROCESSOR_INTEL_PENTIUM = 586 # Variable c_int
AF_DLI = 13 # Variable c_int
MM_JOY2ZMOVE = 931 # Variable c_int
LVFI_STRING = 2 # Variable c_int
SPAPI_E_INVALID_MACHINENAME = -2146500064 # Variable c_long
TB_SETDISABLEDIMAGELIST = 1078 # Variable c_int
CF_DSPBITMAP = 130 # Variable c_int

_CM_SERVICE_LOAD_TYPE = c_int # enum
SystemLoad = 1
AutoLoad = 2
DemandLoad = 3
DisableLoad = 4
MIXERCONTROL_CONTROLTYPE_EQUALIZER = 1342373892 # Variable c_long
SKF_CONFIRMHOTKEY = 8 # Variable c_int
cmb2 = 1137 # Variable c_int
ERROR_ALLOTTED_SPACE_EXCEEDED = 1344 # Variable c_long
LOCALE_ICURRDIGITS = 25 # Variable c_int
EMR_RESIZEPALETTE = 51 # Variable c_int
OLE_S_STATIC = 262145 # Variable c_long
MSGF_NEXTWINDOW = 6 # Variable c_int
LPD_STEREO = 2 # Variable c_int
CONVERT10_S_FIRST = 262592 # Variable c_long
RBBS_BREAK = 1 # Variable c_int
PRINTER_ENUM_NAME = 8 # Variable c_int
CUDR_NORMAL = 0 # Variable c_int
ERROR_SEVERITY_SUCCESS = 0 # Variable c_int
TVN_BEGINRDRAGW = 4294966839 # Variable c_uint
TVHT_ABOVE = 256 # Variable c_int
PSN_FIRST = 4294967096 # Variable c_uint
GCPCLASS_ARABIC = 2 # Variable c_int
PROOF_QUALITY = 2 # Variable c_int
OLERENDER_ASIS = 3
CS_BYTEALIGNCLIENT = 4096 # Variable c_int
ERROR_NOT_FOUND = 1168 # Variable c_long
REGISTERING = 0 # Variable c_int
MCI_RECORD_INSERT = 256 # Variable c_long
TC_RA_ABLE = 8192 # Variable c_int
TOKEN_QUERY_SOURCE = 16 # Variable c_int
RPC_C_PROFILE_ALL_ELT = 1 # Variable c_int
IMAGE_REL_MIPS_GPREL = 6 # Variable c_int
TPM_BOTTOMALIGN = 32 # Variable c_long
LANG_POLISH = 21 # Variable c_int
PARITY_EVEN = 1024 # Variable c_ushort
CB_MSGMAX = 354 # Variable c_int
DMPAPER_ENV_12 = 22 # Variable c_int
ERROR_WMI_GUID_DISCONNECTED = 4207 # Variable c_long
IME_THOTKEY_SHAPE_TOGGLE = 113 # Variable c_int
ERROR_EVENTLOG_FILE_CORRUPT = 1500 # Variable c_long
LR_LOADTRANSPARENT = 32 # Variable c_int
PERF_NUMBER_HEX = 0 # Variable c_int
DMDITHER_USER = 256 # Variable c_int
HOTKEY_CLASSA = 'msctls_hotkey32' # Variable POINTER(c_char)
TVGN_PREVIOUSVISIBLE = 7 # Variable c_int
MA_NOACTIVATEANDEAT = 4 # Variable c_int
SERVICE_START_PENDING = 2 # Variable c_int
SS_BITMAP = 14 # Variable c_long
LVM_UPDATE = 4138 # Variable c_int
WS_THICKFRAME = 262144 # Variable c_long
JOY_RETURNCENTERED = 1024 # Variable c_long
IMAGE_DIRECTORY_ENTRY_IMPORT = 1 # Variable c_int
EM_SETRECT = 179 # Variable c_int
TBM_SETSELSTART = 1035 # Variable c_int
KEY_SET_VALUE = 2 # Variable c_int
URLPOLICY_JAVA_LOW = 196608 # Variable c_int
SPI_SETMOUSEKEYS = 55 # Variable c_int
STATUS_ILLEGAL_VLM_REFERENCE = 3221226176 # Variable c_ulong
WIZ_CYDLG = 140 # Variable c_int
S_WHITE1024 = 5 # Variable c_int
MB_DEFBUTTON1 = 0 # Variable c_long
XCLASS_NOTIFICATION = 32768 # Variable c_int
ERROR_HOST_NODE_NOT_RESOURCE_OWNER = 5015 # Variable c_long
SUBLANG_ARABIC_BAHRAIN = 15 # Variable c_int
SUBLANG_SYS_DEFAULT = 2 # Variable c_int
TAPE_DRIVE_FORMAT = 2684354560 # Variable c_uint
WSAEWOULDBLOCK = 10035 # Variable c_int
HDM_LAYOUT = 4613 # Variable c_int
FS_JOHAB = 2097152 # Variable c_long
MMIO_READWRITE = 2 # Variable c_int
DISP_E_PARAMNOTFOUND = -2147352572 # Variable c_long
URLPOLICY_AUTHENTICATE_CHALLENGE_RESPONSE = 65536 # Variable c_int
ERROR_WINDOW_NOT_COMBOBOX = 1423 # Variable c_long
lst11 = 1130 # Variable c_int
SUBLANG_CHINESE_HONGKONG = 3 # Variable c_int
LANG_ASSAMESE = 77 # Variable c_int
IN_CLASSB_NSHIFT = 16 # Variable c_int
MIIM_STATE = 1 # Variable c_int
CALLBACK_THREAD = 131072 # Variable c_long
cmb10 = 1145 # Variable c_int
WGL_SWAP_OVERLAY11 = 2048 # Variable c_int
DT_CALCRECT = 1024 # Variable c_int
SUBLANG_ENGLISH_TRINIDAD = 11 # Variable c_int
PSCB_PRECREATE = 2 # Variable c_int
RT_FONTDIR = 7 # Variable POINTER(c_wchar)
MIXER_OBJECTF_WAVEIN = 536870912 # Variable c_long
CBS_SORT = 256 # Variable c_long
MM_STREAM_OPEN = 980 # Variable c_int
ILLUMINANT_D65 = 6 # Variable c_int
IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT = 11 # Variable c_int
PC_POLYGON = 1 # Variable c_int
ERROR_SERVER_NOT_DISABLED = 1342 # Variable c_long
ERROR_BAD_PIPE = 230 # Variable c_long
TAPE_DRIVE_WRITE_LONG_FMKS = 2281701376 # Variable c_uint
LPSTR_TEXTCALLBACKA = 4294967295 # Variable POINTER(c_char)
MCI_WAVE_MAPPER = 1153 # Variable c_int
TAPE_FIXED_PARTITIONS = 0 # Variable c_long
SWP_DEFERERASE = 8192 # Variable c_int
MAX_JOYSTICKOEMVXDNAME = 260 # Variable c_int
SIF_DISABLENOSCROLL = 8 # Variable c_int
LVCFMT_JUSTIFYMASK = 3 # Variable c_int
MMIOERR_SHARINGVIOLATION = 269 # Variable c_int
PERF_COUNTER_DELTA = 4195328 # Variable c_int

xmlelemTYPE = c_int # enum
XMLELEMTYPE_ELEMENT = 0
PAN_WEIGHT_NORD = 11 # Variable c_int
RPC_C_MGMT_INQ_PRINC_NAME = 1 # Variable c_int
PORT_STATUS_TYPE_ERROR = 1 # Variable c_int
HDN_FIRST = 4294966996 # Variable c_uint
C2_EUROPETERMINATOR = 5 # Variable c_int
CBEM_INSERTITEMA = 1025 # Variable c_int
SPLREG_ARCHITECTURE = 'Architecture' # Variable POINTER(c_wchar)
TB_GETINSERTMARK = 1103 # Variable c_int
WNNC_NET_FTP_NFS = 786432 # Variable c_int
HEAP_TAIL_CHECKING_ENABLED = 32 # Variable c_int
ERROR_NO_SHUTDOWN_IN_PROGRESS = 1116 # Variable c_long
IDOK = 1 # Variable c_int

tagMKSYS = c_int # enum
MKSYS_NONE = 0
MKSYS_FILEMONIKER = 2
MKSYS_ITEMMONIKER = 4
LVM_SORTITEMS = 4144 # Variable c_int
IME_ESC_AUTOMATA = 4105 # Variable c_int
IDANI_OPEN = 1 # Variable c_int
HKCOMB_SA = 32 # Variable c_int
PSD_SHOWHELP = 2048 # Variable c_int
PRINTER_ERROR_WARNING = 1073741824 # Variable c_int
ERROR_INVALID_SERVICE_LOCK = 1071 # Variable c_long
RT_ICON = 3 # Variable POINTER(c_wchar)
LVM_GETHOTITEM = 4157 # Variable c_int
PSH_USEPAGELANG = 2097152 # Variable c_int
ERROR_FILENAME_EXCED_RANGE = 206 # Variable c_long
MAXIMUM_WAIT_OBJECTS = 64 # Variable c_int
FILE_ACTION_RENAMED_OLD_NAME = 4 # Variable c_int
LOCALE_SABBREVDAYNAME3 = 51 # Variable c_int
DC_EMF_COMPLIANT = 20 # Variable c_int
PIDSI_WORDCOUNT = 15 # Variable c_long
TURKISH_CHARSET = 162 # Variable c_int
FRERR_FINDREPLACECODES = 16384 # Variable c_int
OLE_E_ENUM_NOMORE = -2147221502 # Variable c_long
EMBDHLP_INPROC_SERVER = 1 # Variable c_long
LOCALE_SDAYNAME6 = 47 # Variable c_int
SUBLANG_SPANISH = 1 # Variable c_int
POSTSCRIPT_INJECTION = 4118 # Variable c_int
NTE_PROVIDER_DLL_FAIL = -2146893795 # Variable c_long
SPI_SETSCREENSAVETIMEOUT = 15 # Variable c_int
BOLD_FONTTYPE = 256 # Variable c_int
MCI_VD_STEP_FRAMES = 65536 # Variable c_long
MF_END = 128 # Variable c_long
LVS_EX_INFOTIP = 1024 # Variable c_int
PC_RECTANGLE = 2 # Variable c_int
PRINTER_ATTRIBUTE_KEEPPRINTEDJOBS = 256 # Variable c_int
IDC_HELP = 206417 # Variable POINTER(c_wchar)
TBIMHT_BACKGROUND = 2 # Variable c_int
LOCALE_IINTLCURRDIGITS = 26 # Variable c_int
GCPCLASS_POSTBOUNDLTR = 32 # Variable c_int
ERROR_WRONG_EFS = 6005 # Variable c_long
DDL_EXCLUSIVE = 32768 # Variable c_int
SPI_GETPOWEROFFTIMEOUT = 80 # Variable c_int
LR_COPYDELETEORG = 8 # Variable c_int
MB_ICONMASK = 240 # Variable c_long
MULTIFILEOPENORD = 1537 # Variable c_int
rad12 = 1067 # Variable c_int
SECURITY_SQOS_PRESENT = 1048576 # Variable c_int
GETVECTORBRUSHSIZE = 27 # Variable c_int
DMPAPER_B_PLUS = 58 # Variable c_int
LBS_NOREDRAW = 4 # Variable c_long
SS_CENTER = 1 # Variable c_long
ASYNC_MODE_COMPATIBILITY = 1 # Variable c_long
SW_SHOWMAXIMIZED = 3 # Variable c_int
ERROR_DEV_NOT_EXIST = 55 # Variable c_long
UOI_FLAGS = 1 # Variable c_int
TVN_BEGINLABELEDITW = 4294966837 # Variable c_uint
ERROR_INVALID_MESSAGE = 1002 # Variable c_long
CALLBACK_NULL = 0 # Variable c_long
LVHT_ABOVE = 8 # Variable c_int
FW_MEDIUM = 500 # Variable c_int
CTRY_FRANCE = 33 # Variable c_int
SPLDS_PRINT_ATTRIBUTES = 'printAttributes' # Variable POINTER(c_wchar)
ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION = 1459 # Variable c_long
C2_EUROPESEPARATOR = 4 # Variable c_int
STG_E_INVALIDNAME = -2147286788 # Variable c_long
DMDITHER_RESERVED9 = 9 # Variable c_int
ERROR_DS_CROSS_DOM_MOVE_ERROR = 1937 # Variable c_long
SO_CONNDATA = 28672 # Variable c_int
PAN_LETTERFORM_INDEX = 7 # Variable c_int
S_PERIOD512 = 0 # Variable c_int
PSN_WIZNEXT = 4294967089 # Variable c_uint
DM_PELSHEIGHT = 1048576 # Variable c_long
RPC_S_ALREADY_LISTENING = 1713 # Variable c_long
IMAGE_SYM_CLASS_FUNCTION = 101 # Variable c_int
SPI_SETGRIDGRANULARITY = 19 # Variable c_int
TTM_GETTOOLCOUNT = 1037 # Variable c_int
MCI_DEVTYPE_SEQUENCER = 523 # Variable c_int
DC_FIELDS = 1 # Variable c_int
UDS_WRAP = 1 # Variable c_int
SYMBOL_CHARSET = 2 # Variable c_int
ERROR_NO_DATA_DETECTED = 1104 # Variable c_long
DV_E_DVTARGETDEVICE = -2147221403 # Variable c_long
ERROR_DS_INVALID_ATTRIBUTE_SYNTAX = 1924 # Variable c_long
EM_GETLINE = 196 # Variable c_int
SECURITY_BATCH_RID = 3 # Variable c_long
JOY_RETURNPOVCTS = 512 # Variable c_long
ERROR_HOTKEY_NOT_REGISTERED = 1419 # Variable c_long
CCM_SETCOLORSCHEME = 8194 # Variable c_int
IP_MULTICAST_LOOP = 4 # Variable c_int
MCIERR_DEVICE_NOT_INSTALLED = 306 # Variable c_int
NDR_INT_REP_MASK = 240 # Variable c_ulong
IMAGE_REL_IA64_GPREL22 = 9 # Variable c_int
C1_SPACE = 8 # Variable c_int
ERROR_OUT_OF_STRUCTURES = 84 # Variable c_long
NTM_TYPE1 = 1048576 # Variable c_int
IME_CMODE_KATAKANA = 2 # Variable c_int
MCI_SEQ_STATUS_PORT = 16387 # Variable c_long
URLACTION_ACTIVEX_TREATASUNTRUSTED = 4613 # Variable c_int
TBS_RIGHT = 0 # Variable c_int
TB_GETROWS = 1064 # Variable c_int
GCS_RESULTCLAUSE = 4096 # Variable c_int
OBJ_BITMAP = 7 # Variable c_int
DOMAIN_ALIAS_RID_PRINT_OPS = 550 # Variable c_long
GCL_CONVERSION = 1 # Variable c_int
LOCALE_USE_CP_ACP = 1073741824 # Variable c_int
VIEW_SMALLICONS = 1 # Variable c_int
SUBLANG_AZERI_CYRILLIC = 2 # Variable c_int
CRYPT_E_SECURITY_SETTINGS = -2146885594 # Variable c_long
STG_E_INVALIDHANDLE = -2147287034 # Variable c_long
USER_MARSHAL_FC_DOUBLE = 12 # Variable c_int
C1_CNTRL = 32 # Variable c_int
GW_HWNDNEXT = 2 # Variable c_int
GW_ENABLEDPOPUP = 6
THAI_CHARSET = 222 # Variable c_int
THREAD_SET_THREAD_TOKEN = 128 # Variable c_int
MCI_ANIM_PUT_DESTINATION = 262144 # Variable c_long
ESB_ENABLE_BOTH = 0 # Variable c_int
PRINTER_STATUS_PAPER_PROBLEM = 64 # Variable c_int
DRV_CONFIGURE = 7 # Variable c_int
PFD_OVERLAY_PLANE = 1 # Variable c_int
SECTION_EXTEND_SIZE = 16 # Variable c_int
CUDR_NORESOLVEPOSITIONS = 2 # Variable c_int
EMR_SETWINDOWEXTEX = 9 # Variable c_int
BAUD_075 = 1 # Variable c_ulong
RESOURCE_GLOBALNET = 2 # Variable c_int
RB_GETRECT = 1033 # Variable c_int
IMAGE_DIRECTORY_ENTRY_ARCHITECTURE = 7 # Variable c_int
MCIERR_SEQ_NOMIDIPRESENT = 343 # Variable c_int
MCI_SEQ_SET_OFFSET = 16777216 # Variable c_long
ELF_CULTURE_LATIN = 0 # Variable c_int
TTM_ADDTOOLW = 1074 # Variable c_int
PSP_HASHELP = 32 # Variable c_int
MM_JOY2BUTTONDOWN = 950 # Variable c_int
TF_WRITE_BEHIND = 4 # Variable c_int
ERROR_SPECIAL_GROUP = 1372 # Variable c_long
stc2 = 1089 # Variable c_int
DDL_DRIVES = 16384 # Variable c_int
PERF_TYPE_TEXT = 2048 # Variable c_int
GCL_CBWNDEXTRA = -18 # Variable c_int
IMAGE_DIRECTORY_ENTRY_GLOBALPTR = 8 # Variable c_int
SSTF_NONE = 0 # Variable c_int
IMAGE_SYM_CLASS_ARGUMENT = 9 # Variable c_int
IDLE_PRIORITY_CLASS = 64 # Variable c_int
stc12 = 1099 # Variable c_int
ERROR_BAD_USERNAME = 2202 # Variable c_long
DISP_CHANGE_SUCCESSFUL = 0 # Variable c_int
IMAGE_REL_ALPHA_REFQ1 = 21 # Variable c_int
LVM_GETCOLUMNWIDTH = 4125 # Variable c_int
ANTIALIASED_QUALITY = 4 # Variable c_int
FILE_FLAG_BACKUP_SEMANTICS = 33554432 # Variable c_int
WAVE_FORMAT_QUERY = 1 # Variable c_int
EMR_POLYPOLYLINE16 = 90 # Variable c_int
ERROR_INVALID_SHARENAME = 1215 # Variable c_long
ESB_DISABLE_BOTH = 3 # Variable c_int
OFN_SHAREFALLTHROUGH = 2 # Variable c_int
PERF_COUNTER_RAWCOUNT = 65536 # Variable c_int
CP_SUPPORTED = 2 # Variable c_int
TBM_GETBUDDY = 1057 # Variable c_int
RB_GETROWHEIGHT = 1038 # Variable c_int
META_CREATEFONTINDIRECT = 763 # Variable c_int
LVM_SETCOLUMNA = 4122 # Variable c_int
MIXERCONTROL_CONTROLTYPE_MICROTIME = 1610809344 # Variable c_long
TBM_GETTOOLTIPS = 1054 # Variable c_int
MOVEFILE_WRITE_THROUGH = 8 # Variable c_int
ERROR_ACCOUNT_RESTRICTION = 1327 # Variable c_long
OUT_STRING_PRECIS = 1 # Variable c_int
SERVICE_CONTINUE_PENDING = 5 # Variable c_int
ERROR_INCORRECT_ADDRESS = 1241 # Variable c_long
ACL_REVISION4 = 4 # Variable c_int
SUBLANG_MALAY_MALAYSIA = 1 # Variable c_int
PORT_TYPE_WRITE = 1 # Variable c_int
PT_LINETO = 2 # Variable c_int
URLACTION_SHELL_MIN = 6144 # Variable c_int
GETSCALINGFACTOR = 14 # Variable c_int
DMBIN_TRACTOR = 8 # Variable c_int
MF_CHANGE = 128 # Variable c_long
IS_TEXT_UNICODE_NULL_BYTES = 4096 # Variable c_int
C2_ARABICNUMBER = 6 # Variable c_int
SW_SHOWNA = 8 # Variable c_int
PROCESSOR_ARCHITECTURE_PPC = 3 # Variable c_int
TVN_BEGINDRAGA = 4294966889 # Variable c_uint
LOCALE_SABBREVMONTHNAME13 = 4111 # Variable c_int
USER_MARSHAL_FC_LONG = 8 # Variable c_int
PS_NULL = 5 # Variable c_int
GW_HWNDPREV = 3 # Variable c_int
TCM_GETITEMRECT = 4874 # Variable c_int
MAXBYTE = 255 # Variable c_int
TVIS_EXPANDPARTIAL = 128 # Variable c_int
SUBLANG_ARABIC_YEMEN = 9 # Variable c_int
LANG_DANISH = 6 # Variable c_int
IMAGE_SYM_CLASS_MEMBER_OF_ENUM = 16 # Variable c_int
TB_SETINSERTMARK = 1104 # Variable c_int
LOCALE_SNEGATIVESIGN = 81 # Variable c_int
ERROR_DEVICE_NOT_CONNECTED = 1167 # Variable c_long
MHDR_DONE = 1 # Variable c_int
MCM_GETMAXTODAYWIDTH = 4117 # Variable c_int
STG_E_UNKNOWN = -2147286787 # Variable c_long
CFSTR_MIME_X_MSVIDEO = 'video/x-msvideo' # Variable POINTER(c_wchar)
MCI_ANIM_STEP_REVERSE = 65536 # Variable c_long
ERROR_TOKEN_ALREADY_IN_USE = 1375 # Variable c_long
NFS_ALL = 16 # Variable c_int
WSB_PROP_CXHSCROLL = 2 # Variable c_long
IDC_SIZENWSE = 206402 # Variable POINTER(c_wchar)
SPLREG_PORT_THREAD_PRIORITY_DEFAULT = 'PortThreadPriorityDefault' # Variable POINTER(c_wchar)
GRADIENT_FILL_RECT_V = 1 # Variable c_int
PROTOCOLFLAG_NO_PICS_CHECK = 1 # Variable c_int
HHT_ABOVE = 256 # Variable c_int
LB_SETTABSTOPS = 402 # Variable c_int
RPC_E_INVALID_HEADER = -2147417839 # Variable c_long
IME_CHOTKEY_SYMBOL_TOGGLE = 18 # Variable c_int
PAN_FAMILY_DECORATIVE = 4 # Variable c_int
SUBLANG_URDU_PAKISTAN = 1 # Variable c_int
SM_CXMAXTRACK = 59 # Variable c_int
ERROR_INVALID_MEMBER = 1388 # Variable c_long
SKF_AUDIBLEFEEDBACK = 64 # Variable c_int
MDM_V23_OVERRIDE = 1024 # Variable c_int
FILE_WRITE_EA = 16 # Variable c_int
TB_REPLACEBITMAP = 1070 # Variable c_int
CTRY_ITALY = 39 # Variable c_int
STATUS_FLOAT_UNDERFLOW = 3221225619 # Variable c_ulong
ERROR_NO_EFS = 6004 # Variable c_long

tagOLECLOSE = c_int # enum
PRINTER_NOTIFY_FIELD_SEPFILE = 8 # Variable c_int
GB2312_CHARSET = 134 # Variable c_int
IMN_SETCOMPOSITIONFONT = 10 # Variable c_int
EVENTLOG_INFORMATION_TYPE = 4 # Variable c_int
DMPAPER_ENV_10 = 20 # Variable c_int
CREATE_FORCEDOS = 8192 # Variable c_int
ERROR_NO_SUCH_ALIAS = 1376 # Variable c_long
WS_DLGFRAME = 4194304 # Variable c_long
FRERR_BUFFERLENGTHZERO = 16385 # Variable c_int
RPC_BUFFER_COMPLETE = 4096 # Variable c_int
MIXERCONTROL_CONTROLTYPE_UNSIGNED = 805502976 # Variable c_long
MARSHAL_S_LAST = 262447 # Variable c_long
VFT2_DRV_INPUTMETHOD = 11 # Variable c_long
USER_MARSHAL_FC_USHORT = 7 # Variable c_int
URLTEMPLATE_PREDEFINED_MIN = 65536
PRINTER_ATTRIBUTE_HIDDEN = 32 # Variable c_int
FILE_NOTIFY_CHANGE_SIZE = 8 # Variable c_int
NM_LAST = 4294967197 # Variable c_uint
BN_DISABLE = 4 # Variable c_int
CC_CHORD = 4 # Variable c_int

tagPENDINGTYPE = c_int # enum
TBM_SETPOS = 1029 # Variable c_int
BS_USERBUTTON = 8 # Variable c_long
ID_PSREBOOTSYSTEM = 3 # Variable c_int
MCIERR_MISSING_PARAMETER = 273 # Variable c_int
MB_DEFBUTTON3 = 512 # Variable c_long
WS_EX_LEFT = 0 # Variable c_long
URLTEMPLATE_LOW = 65536
DTM_GETRANGE = 4099 # Variable c_int
EFS_USE_RECOVERY_KEYS = 1 # Variable c_int
DISP_E_UNKNOWNNAME = -2147352570 # Variable c_long
SIF_PAGE = 2 # Variable c_int
GM_COMPATIBLE = 1 # Variable c_int
COLOR_BTNTEXT = 18 # Variable c_int
SC_MANAGER_LOCK = 8 # Variable c_int
TBSTATE_INDETERMINATE = 16 # Variable c_int
DMLERR_SYS_ERROR = 16399 # Variable c_int
PRINTER_NOTIFY_FIELD_ATTRIBUTES = 13 # Variable c_int
IMAGE_COMDAT_SELECT_NEWEST = 7 # Variable c_int
ERROR_IO_INCOMPLETE = 996 # Variable c_long
LVS_EX_CHECKBOXES = 4 # Variable c_int
TAPE_SETMARKS = 0 # Variable c_long
ERROR_LOGIN_WKSTA_RESTRICTION = 1240 # Variable c_long
MIXERCONTROL_CT_SC_TIME_MILLISECS = 16777216 # Variable c_long
RPC_S_UNKNOWN_AUTHN_SERVICE = 1747 # Variable c_long
PGF_SCROLLLEFT = 4 # Variable c_int
TB_GETBUTTONTEXTW = 1099 # Variable c_int
ERROR_NO_RECOVERY_POLICY = 6003 # Variable c_long
ERROR_INVALID_LOGON_HOURS = 1328 # Variable c_long
IMAGE_FILE_EXECUTABLE_IMAGE = 2 # Variable c_int
CF_USESTYLE = 128 # Variable c_long
lst1 = 1120 # Variable c_int
HEAP_ZERO_MEMORY = 8 # Variable c_int
SND_ALIAS_ID = 1114112 # Variable c_long
PRINTER_NOTIFY_FIELD_COMMENT = 5 # Variable c_int
CRYPT_E_OID_FORMAT = -2146889725 # Variable c_long
COLOR_ADJ_MAX = 100 # Variable c_short
PRSPEC_LPWSTR = 0 # Variable c_int
PDERR_SETUPFAILURE = 4097 # Variable c_int
DCBA_FACEDOWNRIGHT = 259 # Variable c_int
TB_ENDTRACK = 8 # Variable c_int
IS_TEXT_UNICODE_REVERSE_ASCII16 = 16 # Variable c_int
ETO_RTLREADING = 128 # Variable c_int
MCI_ANIM_WINDOW_STATE = 262144 # Variable c_long
SERVICE_WIN32 = 48 # Variable c_int
RPC_C_AUTHN_DCE_PRIVATE = 1 # Variable c_int
RPC_S_INVALID_ASYNC_HANDLE = 1914 # Variable c_long
EMR_POLYBEZIER = 2 # Variable c_int
FS_LATIN2 = 2 # Variable c_long
DTN_LAST = 4294966497 # Variable c_uint
ERROR_FILE_INVALID = 1006 # Variable c_long
CREATE_PROCESS_DEBUG_EVENT = 3 # Variable c_int
LVN_GETDISPINFOW = 4294967119 # Variable c_uint
MDM_FLOWCONTROL_SOFT = 32 # Variable c_int
VIEW_S_ALREADY_FROZEN = 262464 # Variable c_long
C1_LOWER = 2 # Variable c_int
CTRY_ECUADOR = 593 # Variable c_int
IMAGE_REL_IA64_RESERVED_12 = 18 # Variable c_int
SECURITY_CREATOR_GROUP_SERVER_RID = 3 # Variable c_long
TAPE_DRIVE_ABSOLUTE_BLK = 2147487744 # Variable c_uint
DMPAPER_LETTERSMALL = 2 # Variable c_int
CLIP_MASK = 15 # Variable c_int
REFERENCE_BLACK_MIN = 0 # Variable c_ushort
CO_S_LAST = 262655 # Variable c_long
MF_UNCHECKED = 0 # Variable c_long
URLACTION_HTML_FONT_DOWNLOAD = 5636 # Variable c_int
APPCMD_CLIENTONLY = 16 # Variable c_long

tagOLERENDER = c_int # enum
VARIANT_ALPHABOOL = 2 # Variable c_int
MCIERR_NONAPPLICABLE_FUNCTION = 302 # Variable c_int
SND_FILENAME = 131072 # Variable c_long
SPI_SETSCREENREADER = 71 # Variable c_int
DISP_E_UNKNOWNLCID = -2147352564 # Variable c_long
CONVERT10_S_NO_PRESENTATION = 262592 # Variable c_long
DISPATCH_PROPERTYGET = 2 # Variable c_int
EM_GETSEL = 176 # Variable c_int
ERROR_PIPE_BUSY = 231 # Variable c_long
LVM_INSERTCOLUMNA = 4123 # Variable c_int
CDS_NORESET = 268435456 # Variable c_int
JOB_NOTIFY_FIELD_DATATYPE = 5 # Variable c_int
DTS_TIMEFORMAT = 9 # Variable c_int

_JOBOBJECTINFOCLASS = c_int # enum
JobObjectBasicProcessIdList = 3
JobObjectAssociateCompletionPortInformation = 7
ERROR_ARENA_TRASHED = 7 # Variable c_long
CBR_2400 = 2400 # Variable c_int
JOY_BUTTON1CHG = 256 # Variable c_int
DDD_EXACT_MATCH_ON_REMOVE = 4 # Variable c_int
CDS_SETRECT = 536870912 # Variable c_int
LVS_EX_TRACKSELECT = 8 # Variable c_int
SSTF_BORDER = 2 # Variable c_int
IMAGE_NUMBEROF_DIRECTORY_ENTRIES = 16 # Variable c_int
CRYPT_E_MSG_ERROR = -2146889727 # Variable c_long
CRYPT_E_NOT_IN_CTL = -2146885590 # Variable c_long
JOB_OBJECT_SET_SECURITY_ATTRIBUTES = 16 # Variable c_int
LANG_KAZAK = 63 # Variable c_int
SETALLJUSTVALUES = 771 # Variable c_int
SUBLANG_URDU_INDIA = 2 # Variable c_int
CRYPT_E_BAD_LEN = -2146885631 # Variable c_long
JOY_RETURNRAWDATA = 256 # Variable c_long
EMR_INVERTRGN = 73 # Variable c_int
TB_ISBUTTONPRESSED = 1035 # Variable c_int
SPAPI_E_DEVICE_INTERFACE_REMOVED = -2146500068 # Variable c_long
DV_E_DVASPECT = -2147221397 # Variable c_long
VIEW_S_FIRST = 262464 # Variable c_long
COLOR_BTNHIGHLIGHT = 20 # Variable c_int
E_UNEXPECTED = -2147418113 # Variable c_long
PRINTER_NOTIFY_FIELD_PRINTER_NAME = 1 # Variable c_int
LCS_GM_IMAGES = 4 # Variable c_long
EMR_POLYPOLYGON16 = 91 # Variable c_int
PSM_SETTITLEA = 1135 # Variable c_int
RPC_C_AUTHZ_NONE = 0 # Variable c_int
DMPAPER_15X11 = 46 # Variable c_int
ERROR_INVALID_PASSWORDNAME = 1216 # Variable c_long
DMDITHER_RESERVED8 = 8 # Variable c_int
LOCALE_ITIME = 35 # Variable c_int
CB_FINDSTRINGEXACT = 344 # Variable c_int
ERROR_GLOBAL_ONLY_HOOK = 1429 # Variable c_long
ERROR_BADDB = 1009 # Variable c_long
MOUSEEVENTF_ABSOLUTE = 32768 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_LINE = 2 # Variable c_long
ERROR_DS_NOT_INSTALLED = 1251 # Variable c_long
REG_RESOURCE_LIST = 8 # Variable c_int
HDS_HIDDEN = 8 # Variable c_int
_OUT_TO_STDERR = 1 # Variable c_int
TB_PAGEUP = 2 # Variable c_int
RESOURCE_RECENT = 4 # Variable c_int
SCROLLLOCK_ON = 64 # Variable c_int
MCHT_TITLEYEAR = 65538 # Variable c_int
ERROR_WMI_DP_NOT_FOUND = 4204 # Variable c_long
HCF_HOTKEYACTIVE = 4 # Variable c_int
ES_AUTOVSCROLL = 64 # Variable c_long
OLE_E_LAST = -2147221249 # Variable c_long
MCI_WAVE_STATUS_AVGBYTESPERSEC = 16388 # Variable c_long
SPI_SETSTICKYKEYS = 59 # Variable c_int
ERROR_SERVICE_ALREADY_RUNNING = 1056 # Variable c_long
TVM_SETTEXTCOLOR = 4382 # Variable c_int
SW_FORCEMINIMIZE = 11 # Variable c_int
RPC_NCA_FLAGS_DEFAULT = 0 # Variable c_int
MCI_SET_DOOR_CLOSED = 512 # Variable c_long
DI_IMAGE = 2 # Variable c_int
AF_SNA = 11 # Variable c_int
MCI_SET_AUDIO_ALL = 0 # Variable c_long
SOFTDIST_ADSTATE_DOWNLOADED = 2 # Variable c_int
LVM_REDRAWITEMS = 4117 # Variable c_int
TAPE_DRIVE_SET_PADDING = 2147484672 # Variable c_uint
IDI_EXCLAMATION = 206101 # Variable POINTER(c_wchar)
CP_WINUNICODE = 1200 # Variable c_int
CONTEXT_i386 = 65536 # Variable c_int
VS_USER_DEFINED = 100 # Variable c_int
SIZEPALETTE = 104 # Variable c_int
PROPSETFLAG_DEFAULT = 0 # Variable c_int
OWNER_SECURITY_INFORMATION = 1 # Variable c_long
DMORIENT_LANDSCAPE = 2 # Variable c_int
REFERENCE_WHITE_MIN = 6000 # Variable c_ushort
MCI_OVLY_PUT_VIDEO = 1048576 # Variable c_long
ERROR_INVALID_MESSAGEDEST = 1218 # Variable c_long
CDDS_ITEMPREPAINT = 65537 # Variable c_int
LOCALE_IDEFAULTLANGUAGE = 9 # Variable c_int
CONVERT10_E_OLESTREAM_PUT = -2147221055 # Variable c_long
GCS_COMPREADSTR = 1 # Variable c_int
FOF_WANTMAPPINGHANDLE = 32 # Variable c_int
LZERROR_BADOUTHANDLE = -2 # Variable c_int
PRINTER_NOTIFY_FIELD_PRIORITY = 14 # Variable c_int
MCI_SEQ_STATUS_NAME = 16395 # Variable c_long
CO_S_FIRST = 262640 # Variable c_long
LVNI_ALL = 0 # Variable c_int
C3_VOWELMARK = 4 # Variable c_int
STARTF_USESTDHANDLES = 256 # Variable c_int
INPLACE_E_NOTUNDOABLE = -2147221088 # Variable c_long
SET_CLIP_BOX = 4108 # Variable c_int
WNNC_NET_9TILES = 589824 # Variable c_int
CONTAINER_INHERIT_ACE = 2 # Variable c_int
HDI_LPARAM = 8 # Variable c_int
MIXERCONTROL_CONTROLTYPE_MUX = 1879113729 # Variable c_long
SM_CYDLGFRAME = 8 # Variable c_int
scr3 = 1170 # Variable c_int
GM_LAST = 2 # Variable c_int
ERROR_INVALID_DOMAIN_STATE = 1353 # Variable c_long
LMEM_MOVEABLE = 2 # Variable c_int
TB_SETANCHORHIGHLIGHT = 1097 # Variable c_int
MCI_VD_MEDIA_OTHER = 1028 # Variable c_int
CTRY_PUERTO_RICO = 1 # Variable c_int
CS_CLASSDC = 64 # Variable c_int
MCI_ANIM_WHERE_DESTINATION = 262144 # Variable c_long
OFN_OVERWRITEPROMPT = 2 # Variable c_int
TTM_UPDATE = 1053 # Variable c_int
MCIERR_WAVE_INPUTSUNSUITABLE = 328 # Variable c_int
HEAP_MAXIMUM_TAG = 4095 # Variable c_int
TVM_SETITEMHEIGHT = 4379 # Variable c_int
IMAGE_REL_ALPHA_REFQ3 = 19 # Variable c_int
TVM_GETINSERTMARKCOLOR = 4390 # Variable c_int
JOB_STATUS_USER_INTERVENTION = 1024 # Variable c_int
ERROR_ALREADY_INITIALIZED = 1247 # Variable c_long
CTRY_TURKEY = 90 # Variable c_int
LCS_SIGNATURE = 1347637059 # Variable c_int
MCI_SEQ_NONE = 65533 # Variable c_int
LB_GETSEL = 391 # Variable c_int
SE_ERR_ACCESSDENIED = 5 # Variable c_int
IN_CLASSC_NET = 4294967040 # Variable c_uint
RPC_S_UUID_NO_ADDRESS = 1739 # Variable c_long
DLGC_DEFPUSHBUTTON = 16 # Variable c_int
IMAGE_FILE_RELOCS_STRIPPED = 1 # Variable c_int
SPAPI_E_CLASS_MISMATCH = -2146500095 # Variable c_long
MF_CHECKED = 8 # Variable c_long
DC_VERSION = 10 # Variable c_int
LMEM_DISCARDED = 16384 # Variable c_int
TCM_FIRST = 4864 # Variable c_int
PIPE_ACCESS_OUTBOUND = 2 # Variable c_int
RESOURCEDISPLAYTYPE_SHARE = 3 # Variable c_int
WNFMT_MULTILINE = 1 # Variable c_int
ERROR_NO_INHERITANCE = 1391 # Variable c_long
STATUS_USER_APC = 192 # Variable c_ulong
PERF_COUNTER_FRACTION = 131072 # Variable c_int
URLACTION_ACTIVEX_RUN = 4608 # Variable c_int
SPI_GETPOWEROFFACTIVE = 84 # Variable c_int
ERROR_APP_WRONG_OS = 1151 # Variable c_long
LVM_GETIMAGELIST = 4098 # Variable c_int
SO_KEEPALIVE = 8 # Variable c_int
CDM_FIRST = 1124 # Variable c_int
MMSYSERR_INVALFLAG = 10 # Variable c_int
MB_OKCANCEL = 1 # Variable c_long
PSH_STRETCHWATERMARK = 262144 # Variable c_int
IMAGE_REL_IA64_SECTION = 11 # Variable c_int
WS_BORDER = 8388608 # Variable c_long
ERROR_INVALID_ACCESS = 12 # Variable c_long
CTRY_YEMEN = 967 # Variable c_int
TVM_GETISEARCHSTRINGW = 4416 # Variable c_int
DBG_TERMINATE_THREAD = 1073807363 # Variable c_ulong
FILE_DELETE_CHILD = 64 # Variable c_int
MIDICAPS_CACHE = 4 # Variable c_int
IMAGE_SYM_TYPE_BYTE = 12 # Variable c_int
UDS_AUTOBUDDY = 16 # Variable c_int
SB_LEFT = 6 # Variable c_int
SEE_MASK_UNICODE = 16384 # Variable c_int
WMSZ_TOPLEFT = 4 # Variable c_int
HINST_COMMCTRL = 4294967295 # Variable POINTER(HINSTANCE__)
SE_PROF_SINGLE_PROCESS_NAME = 'SeProfileSingleProcessPrivilege' # Variable POINTER(c_wchar)
HEAP_CREATE_ENABLE_TRACING = 131072 # Variable c_int
GRADIENT_FILL_RECT_H = 0 # Variable c_int
PO_DELETE = 19 # Variable c_int
MERGEPAINT = 12255782 # Variable c_ulong
DS_ABSALIGN = 1 # Variable c_long
SPAPI_E_LINE_NOT_FOUND = -2146500350 # Variable c_long
MCI_INFO_NAME = 4096 # Variable c_long
cmb4 = 1139 # Variable c_int
HIST_FAVORITES = 2 # Variable c_int
TC_SF_X_YINDEP = 32 # Variable c_int
JOB_NOTIFY_FIELD_MACHINE_NAME = 1 # Variable c_int
RBN_ENDDRAG = 4294966460 # Variable c_uint
SOFTKEYBOARD_TYPE_T1 = 1 # Variable c_int
MCI_LOAD = 2128 # Variable c_int
TBN_BEGINDRAG = 4294966595 # Variable c_uint
WINSTA_ENUMDESKTOPS = 1 # Variable c_long
JOB_STATUS_OFFLINE = 32 # Variable c_int
PAN_LETT_NORMAL_ROUNDED = 6 # Variable c_int
IMAGE_REL_PPC_PAIR = 18 # Variable c_int
SPI_SETSERIALKEYS = 63 # Variable c_int
SUBLANG_FRENCH_CANADIAN = 3 # Variable c_int
TYPE_E_INCONSISTENTPROPFUNCS = -2147312509 # Variable c_long
ERROR_TOO_MANY_OPEN_FILES = 4 # Variable c_long
COMPLEXREGION = 3 # Variable c_int
DISP_E_TYPEMISMATCH = -2147352571 # Variable c_long
MESSAGE_RESOURCE_UNICODE = 1 # Variable c_int
STARTF_FORCEOFFFEEDBACK = 128 # Variable c_int
URLACTION_CHANNEL_SOFTDIST_MIN = 7680 # Variable c_int
ABM_REMOVE = 1 # Variable c_int
MB_USERICON = 128 # Variable c_long
BF_TOP = 2 # Variable c_int
PERSIST_E_SIZEDEFINITE = -2146762743 # Variable c_long
HDM_CREATEDRAGIMAGE = 4624 # Variable c_int
VFT2_DRV_PRINTER = 1 # Variable c_long
BS_AUTOCHECKBOX = 3 # Variable c_long
ODT_LISTBOX = 2 # Variable c_int
ERROR_RESOURCE_NOT_ONLINE = 5004 # Variable c_long
SEC_COMMIT = 134217728 # Variable c_int
PRINTER_ENUM_ICONMASK = 16711680 # Variable c_int
SB_BOTTOM = 7 # Variable c_int
RPC_S_NO_MORE_BINDINGS = 1806 # Variable c_long
NCBACTION = 119 # Variable c_int
SERVICE_ALL_ACCESS = 983551 # Variable c_long
MM_MCINOTIFY = 953 # Variable c_int
PSM_SETCURSELID = 1138 # Variable c_int

_tagPI_FLAGS = c_int # enum
CAT_E_NODESCRIPTION = -2147221151 # Variable c_long
MB_DEFBUTTON2 = 256 # Variable c_long
IMAGE_SCN_LNK_NRELOC_OVFL = 16777216 # Variable c_int
MF_CALLBACKS = 134217728 # Variable c_int
MFT_RIGHTORDER = 8192 # Variable c_long
CWCSTORAGENAME = 32 # Variable c_int
ERROR_NO_MORE_ITEMS = 259 # Variable c_long
TTM_DELTOOLA = 1029 # Variable c_int
ERROR_MORE_DATA = 234 # Variable c_long
JOB_OBJECT_QUERY = 4 # Variable c_int
IMN_OPENCANDIDATE = 5 # Variable c_int
MHDR_INQUEUE = 4 # Variable c_int
psh13 = 1036 # Variable c_int
MARSHAL_E_FIRST = 2147746080 # Variable c_ulong
TVN_ITEMEXPANDINGA = 4294966891 # Variable c_uint

tagMEMCTX = c_int # enum
MEMCTX_TASK = 1
MEMCTX_SHARED = 2
MEMCTX_MACSYSTEM = 3
MEMCTX_UNKNOWN = -1
MEMCTX_SAME = -2
LVM_APPROXIMATEVIEWRECT = 4160 # Variable c_int
UDS_SETBUDDYINT = 2 # Variable c_int
IME_CMODE_SOFTKBD = 128 # Variable c_int
ERROR_SEM_OWNER_DIED = 105 # Variable c_long
LBS_STANDARD = 10485763 # Variable c_long
TPM_HORIZONTAL = 0 # Variable c_long
LVM_GETHOVERTIME = 4168 # Variable c_int
DROPEFFECT_NONE = 0 # Variable c_int
HDN_TRACKW = 4294966968 # Variable c_uint
FADF_RECORD = 32 # Variable c_int
IMAGE_FILE_MACHINE_UNKNOWN = 0 # Variable c_int
MMIO_RWMODE = 3 # Variable c_int
PCF_DTRDSR = 1 # Variable c_ulong
EMR_POLYLINETO16 = 89 # Variable c_int
EV_EVENT2 = 4096 # Variable c_int
lst2 = 1121 # Variable c_int
MCI_WAVE_GETDEVCAPS_INPUTS = 16385 # Variable c_long
WHITE_BRUSH = 0 # Variable c_int
ESB_DISABLE_DOWN = 2 # Variable c_int
ERROR_SERVICE_REQUEST_TIMEOUT = 1053 # Variable c_long
INPLACE_S_LAST = 262575 # Variable c_long
CO_E_FAILEDTOCREATEFILE = -2147220972 # Variable c_long
PROP_SM_CYDLG = 188 # Variable c_int
SBM_SETSCROLLINFO = 233 # Variable c_int
GL_ID_INPUTCODE = 38 # Variable c_int
CO_E_IIDREG_INCONSISTENT = -2147467232 # Variable c_long
SLE_MINORERROR = 2 # Variable c_int
PC_STYLED = 32 # Variable c_int
STG_E_REVERTED = -2147286782 # Variable c_long
DDD_RAW_TARGET_PATH = 1 # Variable c_int
ERROR_CONNECTION_UNAVAIL = 1201 # Variable c_long
CAL_IYEAROFFSETRANGE = 3 # Variable c_int
OLEIVERB_SHOW = -1 # Variable c_long
NM_SETCURSOR = 4294967279 # Variable c_uint
R2_MASKPENNOT = 5 # Variable c_int
TYPE_E_UNSUPFORMAT = -2147319783 # Variable c_long
SC_MINIMIZE = 61472 # Variable c_int
PRINTER_STATUS_SERVER_UNKNOWN = 8388608 # Variable c_int
MCIERR_MISSING_DEVICE_NAME = 292 # Variable c_int
IMAGE_REL_IA64_RESERVED_13 = 19 # Variable c_int
RESOURCETYPE_UNKNOWN = 4294967295 # Variable c_uint
EMR_POLYPOLYGON = 8 # Variable c_int
MDM_BLIND_DIAL = 512 # Variable c_int
URLACTION_HTML_SUBMIT_FORMS = 5633 # Variable c_int
LB_GETANCHORINDEX = 413 # Variable c_int
ODS_DISABLED = 4 # Variable c_int
PRF_NONCLIENT = 2 # Variable c_long
edt12 = 1163 # Variable c_int
TYMED_ISTORAGE = 8
SND_NOWAIT = 8192 # Variable c_long
TYPE_E_REGISTRYACCESS = -2147319780 # Variable c_long

tagTYPEKIND = c_int # enum
TKIND_UNION = 7
PAN_WEIGHT_INDEX = 2 # Variable c_int
MCI_CD_OFFSET = 1088 # Variable c_int
EMR_SETPIXELV = 15 # Variable c_int
MCIERR_DEVICE_LENGTH = 310 # Variable c_int
DEF_PRIORITY = 1 # Variable c_int
WGL_SWAP_MAIN_PLANE = 1 # Variable c_int
LZERROR_GLOBALLOC = -5 # Variable c_int
DATABITS_6 = 2 # Variable c_ushort
LVM_SETITEMTEXTA = 4142 # Variable c_int
NTE_NO_KEY = -2146893811 # Variable c_long
MCM_GETRANGE = 4113 # Variable c_int
MF_REMOVE = 4096 # Variable c_long
TVHT_TORIGHT = 1024 # Variable c_int
ELF_VENDOR_SIZE = 4 # Variable c_int
MIDIPROP_GET = 1073741824 # Variable c_long
ERROR_REMOTE_STORAGE_MEDIA_ERROR = 4352 # Variable c_long
EMARCH_ENC_I17_IMM5C_SIZE_X = 5 # Variable c_int
MCI_SEQ_SMPTE = 16388 # Variable c_int
JOY_BUTTON1 = 1 # Variable c_int
TBM_GETPTICS = 1038 # Variable c_int
PAN_LETT_NORMAL_SQUARE = 8 # Variable c_int
DISP_CHANGE_BADFLAGS = -4 # Variable c_int
MSG_DONTROUTE = 4 # Variable c_int
SUBLANG_DUTCH = 1 # Variable c_int
HOTKEYF_CONTROL = 2 # Variable c_int
FILE_SHARE_READ = 1 # Variable c_int
LOCALE_IDEFAULTCOUNTRY = 10 # Variable c_int
SEM_FAILCRITICALERRORS = 1 # Variable c_int
VOS_NT = 262144 # Variable c_long
PAGE_NOCACHE = 512 # Variable c_int
SPLREG_MAJOR_VERSION = 'MajorVersion' # Variable POINTER(c_wchar)
STGFMT_STORAGE = 0 # Variable c_int
LB_GETSELITEMS = 401 # Variable c_int
LVM_INSERTCOLUMNW = 4193 # Variable c_int
EMR_POLYLINE = 4 # Variable c_int
TAPE_DRIVE_FORMAT_IMMEDIATE = 3221225472 # Variable c_uint
CONNDLG_USE_MRU = 4 # Variable c_int
IMAGE_SYM_CLASS_REGISTER = 4 # Variable c_int
NTE_KEYSET_NOT_DEF = -2146893799 # Variable c_long
DIALOPTION_BILLING = 64 # Variable c_int
ERROR_INVALID_INDEX = 1413 # Variable c_long
SPAPI_E_CANT_LOAD_CLASS_ICON = -2146500084 # Variable c_long
DOMAIN_USER_RID_ADMIN = 500 # Variable c_long
DSS_DISABLED = 32 # Variable c_int
FORMATDLGORD31 = 1543 # Variable c_int
SPACEPARITY = 4 # Variable c_int
NUMFONTS = 22 # Variable c_int
PIPE_SERVER_END = 1 # Variable c_int
edt13 = 1164 # Variable c_int
JOHAB_CHARSET = 130 # Variable c_int
ERROR_KEY_HAS_CHILDREN = 1020 # Variable c_long
MCI_OVLY_WINDOW_ENABLE_STRETCH = 1048576 # Variable c_long
ABM_QUERYPOS = 2 # Variable c_int
MCI_SEQ_SET_MASTER = 524288 # Variable c_long
DMPAPER_FOLIO = 14 # Variable c_int
IMAGE_DIRECTORY_ENTRY_EXCEPTION = 3 # Variable c_int
lst15 = 1134 # Variable c_int
REGDB_E_FIRST = 2147746128 # Variable c_ulong
RT_ANICURSOR = 33 # Variable POINTER(c_wchar)
SBT_NOBORDERS = 256 # Variable c_int
ERROR_INTERNAL_DB_ERROR = 1383 # Variable c_long
FACILITY_WINDOWS = 8 # Variable c_int
JOY_BUTTON3CHG = 1024 # Variable c_int
LVBKIF_SOURCE_NONE = 0 # Variable c_int
S_STACCATO = 2 # Variable c_int
CLASSFACTORY_E_LAST = 2147746079 # Variable c_ulong
IP_TOS = 8 # Variable c_int
RB_GETDROPTARGET = 8196 # Variable c_int
CDN_FILEOK = 4294966690 # Variable c_uint
HTHSCROLL = 6 # Variable c_int
OLEOBJ_S_LAST = 262543 # Variable c_long
TB_DELETEBUTTON = 1046 # Variable c_int
PERF_COUNTER_BULK_COUNT = 272696576 # Variable c_int
HELP_KEY = 257 # Variable c_long
SERVER_EXECUTE = 131074 # Variable c_long
LVM_SETITEMA = 4102 # Variable c_int
NUMBRUSHES = 16 # Variable c_int
ERROR_INVALID_HOOK_HANDLE = 1404 # Variable c_long
DFCS_BUTTONCHECK = 0 # Variable c_int
JOY_BUTTON18 = 131072 # Variable c_long
CBF_SKIP_DISCONNECTS = 2097152 # Variable c_int
TVHT_ONITEMSTATEICON = 64 # Variable c_int
MEVT_COMMENT = 130 # Variable c_ubyte
TVS_SINGLEEXPAND = 1024 # Variable c_int
WAIT_ABANDONED = 128 # Variable c_ulong
ARW_RIGHT = 0 # Variable c_long
SPAPI_E_NO_ASSOCIATED_SERVICE = -2146500071 # Variable c_long
PRINTER_STATUS_MANUAL_FEED = 32 # Variable c_int
OBJ_ENHMETAFILE = 13 # Variable c_int
IPM_SETADDRESS = 1125 # Variable c_int
IMAGE_DIRECTORY_ENTRY_SECURITY = 4 # Variable c_int
DTS_RIGHTALIGN = 32 # Variable c_int
scr2 = 1169 # Variable c_int
SPAPI_E_WRONG_INF_STYLE = -2146500352 # Variable c_long
QUERY_EXPIRATION_DATE = 1
IME_CONFIG_REGISTERWORD = 2 # Variable c_int
PIDSI_KEYWORDS = 5 # Variable c_long
MCI_GETDEVCAPS_COMPOUND_DEVICE = 6 # Variable c_long
XST_INCOMPLETE = 1 # Variable c_int
EMARCH_ENC_I17_IMM41c_INST_WORD_POS_X = 0 # Variable c_int
MOD_LEFT = 32768 # Variable c_int
MM_DRVM_OPEN = 976 # Variable c_int
PSM_SETTITLEW = 1144 # Variable c_int
TVIF_STATE = 8 # Variable c_int
CLIP_LH_ANGLES = 16 # Variable c_int
SB_GETTEXTLENGTHW = 1036 # Variable c_int
TB_ADDBITMAP = 1043 # Variable c_int
LOCALE_SENGCOUNTRY = 4098 # Variable c_int
IMAGE_REL_ALPHA_REFQ2 = 20 # Variable c_int
DM_DISPLAYFLAGS = 2097152 # Variable c_long
MF_STRING = 0 # Variable c_long
GCS_COMPREADATTR = 2 # Variable c_int
RASTER_FONTTYPE = 1 # Variable c_int
MCI_ANIM_PUT_SOURCE = 131072 # Variable c_long
PAGESETUPDLGORD = 1546 # Variable c_int
DCTT_BITMAP = 1 # Variable c_long
PSN_APPLY = 4294967094 # Variable c_uint
WVR_VREDRAW = 512 # Variable c_int
ICON_SMALL = 0 # Variable c_int
LVM_GETNEXTITEM = 4108 # Variable c_int

tagMSHCTX = c_int # enum
MSHCTX_LOCAL = 0
MSHCTX_NOSHAREDMEM = 1
URLPOLICY_CREDENTIALS_MUST_PROMPT_USER = 65536 # Variable c_int
REG_DWORD = 4 # Variable c_int
NCBRESET = 50 # Variable c_int
SM_CYSCREEN = 1 # Variable c_int
EMR_POLYLINE16 = 87 # Variable c_int
DTN_DROPDOWN = 4294966542 # Variable c_uint
CERTDB_E_JET_ERROR = -2146873344 # Variable c_long
SE_ERR_PNF = 3 # Variable c_int
EMR_CREATEDIBPATTERNBRUSHPT = 94 # Variable c_int
TTN_SHOW = 4294966775 # Variable c_uint
VS_FF_INFOINFERRED = 16 # Variable c_long
DWL_USER = 8 # Variable c_int
CONNDLG_NOT_PERSIST = 32 # Variable c_int
PAN_BENT_ARMS_WEDGE = 8 # Variable c_int
WSAETOOMANYREFS = 10059 # Variable c_int
IMAGE_SCN_SCALE_INDEX = 1 # Variable c_int
PSM_SETFINISHTEXTW = 1145 # Variable c_int
DISPATCH_PROPERTYPUTREF = 8 # Variable c_int
DATABITS_16X = 32 # Variable c_ushort
CBEM_SETIMAGELIST = 1026 # Variable c_int
LVN_BEGINLABELEDITW = 4294967121 # Variable c_uint
STATUS_SINGLE_STEP = 2147483652 # Variable c_ulong
ERROR_RESOURCE_DISABLED = 4309 # Variable c_long
KEY_ENUMERATE_SUB_KEYS = 8 # Variable c_int
WSAESOCKTNOSUPPORT = 10044 # Variable c_int
CF_HDROP = 15 # Variable c_int
IPPORT_RJE = 77 # Variable c_int
RPC_C_MGMT_INQ_IF_IDS = 0 # Variable c_int
SUBLANG_UZBEK_LATIN = 1 # Variable c_int
MOUSEEVENTF_LEFTUP = 4 # Variable c_int
TAPE_DRIVE_TAPE_REMAINING = 512 # Variable c_int
SHOW_OPENWINDOW = 1 # Variable c_int
IMAGE_REL_IA64_IMM14 = 1 # Variable c_int
GMEM_NOT_BANKED = 4096 # Variable c_int
SC_MANAGER_CREATE_SERVICE = 2 # Variable c_int
TTM_GETCURRENTTOOLA = 1039 # Variable c_int
ERROR_TIMEOUT = 1460 # Variable c_long
N_TMASK = 48 # Variable c_int
CFSTR_MIME_X_WMF = 'image/x-wmf' # Variable POINTER(c_wchar)
MCI_SYSINFO_OPEN = 512 # Variable c_long
VIF_CANNOTDELETECUR = 16384 # Variable c_long
DTS_UPDOWN = 1 # Variable c_int
CTRY_EL_SALVADOR = 503 # Variable c_int
HEAP_CREATE_ALIGN_16 = 65536 # Variable c_int
MIXERCONTROL_CONTROLTYPE_BOOLEANMETER = 268500992 # Variable c_long
ERROR_LOGON_TYPE_NOT_GRANTED = 1385 # Variable c_long
ERROR_DEVICE_NOT_PARTITIONED = 1107 # Variable c_long
FILEOKSTRINGA = 'commdlg_FileNameOK' # Variable POINTER(c_char)
ERROR_JOURNAL_HOOK_SET = 1430 # Variable c_long
TCM_GETEXTENDEDSTYLE = 4917 # Variable c_int
SS_WHITEFRAME = 9 # Variable c_long
HELPMSGSTRINGW = 'commdlg_help' # Variable POINTER(c_wchar)
EVENTLOG_SEEK_READ = 2 # Variable c_int
HDM_SETITEMW = 4620 # Variable c_int
SSWF_NONE = 0 # Variable c_int
HDM_GETITEMA = 4611 # Variable c_int
RT_HTML = 35 # Variable POINTER(c_wchar)
MAXUHALF_PTR = 65535 # Variable c_int
cmb7 = 1142 # Variable c_int
WINVER = 1024 # Variable c_int
ABM_GETAUTOHIDEBAR = 7 # Variable c_int
TTM_HITTESTA = 1034 # Variable c_int
IMAGE_DIRECTORY_ENTRY_IAT = 12 # Variable c_int
MCI_SEQ_STATUS_DIVTYPE = 16394 # Variable c_long
IDHELP = 9 # Variable c_int
CB_SETDROPPEDWIDTH = 352 # Variable c_int
IMAGE_REL_I386_DIR16 = 1 # Variable c_int
LVCFMT_LEFT = 0 # Variable c_int
TCS_RIGHT = 2 # Variable c_int
SPIF_UPDATEINIFILE = 1 # Variable c_int
ERROR_INVALID_EXE_SIGNATURE = 191 # Variable c_long
RT_VXD = 32 # Variable POINTER(c_wchar)
MK_RBUTTON = 2 # Variable c_int
DOMAIN_USER_RID_GUEST = 501 # Variable c_long
ENABLEDUPLEX = 28 # Variable c_int
GL_ID_INPUTRADICAL = 37 # Variable c_int
IO_REPARSE_TAG_RESERVED_ONE = 1 # Variable c_int
CTRY_JAMAICA = 1 # Variable c_int

__MIDL_IInternetZoneManager_0001 = c_int # enum
MAX_ZONE_PATH = 260
MAX_ZONE_DESCRIPTION = 200
WSAEMFILE = 10024 # Variable c_int
SPI_GETFASTTASKSWITCH = 35 # Variable c_int
IMAGE_REL_PPC_BRNTAKEN = 1024 # Variable c_int
ODT_COMBOBOX = 3 # Variable c_int
TBM_GETSELSTART = 1041 # Variable c_int

_TOKEN_TYPE = c_int # enum
TokenPrimary = 1
TokenImpersonation = 2
ERROR_INVALID_WORKSTATION = 1329 # Variable c_long
MCM_GETMAXSELCOUNT = 4099 # Variable c_int
FACILITY_SETUPAPI = 15 # Variable c_int
MIXERLINE_TARGETTYPE_UNDEFINED = 0 # Variable c_int
ERROR_PROTOCOL_UNREACHABLE = 1233 # Variable c_long
psh15 = 1038 # Variable c_int
TVN_ENDLABELEDITA = 4294966885 # Variable c_uint
RTL_RESOURCE_TYPE = 1 # Variable c_int
TCM_HITTEST = 4877 # Variable c_int

tagREGCLS = c_int # enum
REGCLS_SINGLEUSE = 0
REGCLS_MULTIPLEUSE = 1
REGCLS_MULTI_SEPARATE = 2
REGCLS_SUSPENDED = 4
LOCALE_SABBREVDAYNAME1 = 49 # Variable c_int
INET_E_CONNECTION_TIMEOUT = -2146697205 # Variable c_long
ABE_BOTTOM = 3 # Variable c_int
REG_SZ = 1 # Variable c_int
WAIT_FAILED = 4294967295 # Variable c_ulong
TBSTATE_ELLIPSES = 64 # Variable c_int

_tagQUERYOPTION = c_int # enum
QUERY_TIME_OF_LAST_CHANGE = 2
QUERY_CONTENT_ENCODING = 3
QUERY_CONTENT_TYPE = 4
QUERY_REFRESH = 5
QUERY_RECOMBINE = 6
RPC_S_PROCNUM_OUT_OF_RANGE = 1745 # Variable c_long
BF_DIAGONAL = 16 # Variable c_int
NCBENUM = 55 # Variable c_int
FILE_ATTRIBUTE_DIRECTORY = 16 # Variable c_int
STATUSCLASSNAMEA = 'msctls_statusbar32' # Variable POINTER(c_char)
DMICM_CONTRAST = 2 # Variable c_int
DIB_RGB_COLORS = 0 # Variable c_int
FKF_FILTERKEYSON = 1 # Variable c_int
META_FRAMEREGION = 1065 # Variable c_int
BATTERY_LIFE_UNKNOWN = 4294967295 # Variable c_uint
NRC_CANCEL = 38 # Variable c_int
HDN_DIVIDERDBLCLICKA = 4294966991 # Variable c_uint
CTRY_DOMINICAN_REPUBLIC = 1 # Variable c_int
FD_CONNECT = 16 # Variable c_int
MCI_OVLY_WINDOW_TEXT = 524288 # Variable c_long
SOFTKEYBOARD_TYPE_C1 = 2 # Variable c_int
IMAGE_ICON = 1 # Variable c_int
FKF_CONFIRMHOTKEY = 8 # Variable c_int
IMAGE_SYM_TYPE_PCODE = 32768 # Variable c_int
MM_MOM_OPEN = 967 # Variable c_int
PIDSI_COMMENTS = 6 # Variable c_long
GGO_GRAY4_BITMAP = 5 # Variable c_int
LVKF_CONTROL = 2 # Variable c_int
URLPOLICY_NOTIFY_ON_DISALLOW = 32 # Variable c_int
CO_E_CONVERSIONFAILED = -2147220981 # Variable c_long
FILE_FLAG_OPEN_REPARSE_POINT = 2097152 # Variable c_int
SE_SYSTEMTIME_NAME = 'SeSystemtimePrivilege' # Variable POINTER(c_wchar)
RPC_C_AUTHN_LEVEL_DEFAULT = 0 # Variable c_int
PFD_UNDERLAY_PLANE = -1 # Variable c_int
HELP_TCARD = 32768 # Variable c_int
SPI_SETICONS = 88 # Variable c_int
SECURITY_CONTEXT_TRACKING = 262144 # Variable c_int
TVIS_CUT = 4 # Variable c_int
PRINTER_STATUS_NO_TONER = 262144 # Variable c_int
EMR_MASKBLT = 78 # Variable c_int
PIDSI_THUMBNAIL = 17 # Variable c_long
grp3 = 1074 # Variable c_int
TIME_PERIODIC = 1 # Variable c_int
HKEY_USERS = 2147483651 # Variable POINTER(HKEY__)
NTE_KEYSET_ENTRY_BAD = -2146893798 # Variable c_long
DL_STOPCURSOR = 1 # Variable c_int
IPPROTO_IP = 0 # Variable c_int
MK_S_REDUCED_TO_SELF = 262626 # Variable c_long
JOB_STATUS_BLOCKED_DEVQ = 512 # Variable c_int
RPC_C_QOS_CAPABILITIES_MUTUAL_AUTH = 1 # Variable c_int
OLEOBJ_E_LAST = 2147746191 # Variable c_ulong
PRINTER_ENUM_REMOTE = 16 # Variable c_int
RPC_E_CALL_COMPLETE = -2147417833 # Variable c_long
IDC_SIZEALL = 206406 # Variable POINTER(c_wchar)
RPC_E_UNSECURE_CALL = -2147417832 # Variable c_long
SPI_SETMOUSEBUTTONSWAP = 33 # Variable c_int
MCI_OVLY_PUT_SOURCE = 131072 # Variable c_long
ES_USER_PRESENT = 4 # Variable c_ulong
IN_CLASSA_HOST = 16777215 # Variable c_int

tagSTATFLAG = c_int # enum
MM_JOY1BUTTONDOWN = 949 # Variable c_int
LVIF_INDENT = 16 # Variable c_int
RPC_E_NO_CONTEXT = -2147417826 # Variable c_long
TAPE_DRIVE_RESERVED_BIT = 2147483648 # Variable c_uint
CLASSFACTORY_S_LAST = 262431 # Variable c_long
TAPE_PSEUDO_LOGICAL_POSITION = 2 # Variable c_long
BS_TOP = 1024 # Variable c_long
MK_E_NO_NORMALIZED = -2146959353 # Variable c_long
PROCESS_HEAP_UNCOMMITTED_RANGE = 2 # Variable c_int
PSP_USEICONID = 4 # Variable c_int
RPC_C_AUTHZ_DCE = 2 # Variable c_int
LOCALE_SCOUNTRY = 6 # Variable c_int
WSB_PROP_HSTYLE = 512 # Variable c_long
PERF_COUNTER_HISTOGRAM_TYPE = 2147483648 # Variable c_uint
INET_E_QUERYOPTION_UNKNOWN = -2146697197 # Variable c_long
CS_DISABLE = 2 # Variable c_long
PGF_HOT = 8 # Variable c_int
ERROR_DS_NO_RIDS_ALLOCATED = 1929 # Variable c_long
ERROR_DS_RIDMGR_INIT_ERROR = 1932 # Variable c_long
SORT_STRINGSORT = 4096 # Variable c_int
IMAGE_SIZEOF_RELOCATION = 10 # Variable c_int
MEVT_F_SHORT = 0 # Variable c_long
TVN_DELETEITEMW = 4294966838 # Variable c_uint
OF_CANCEL = 2048 # Variable c_int
URLACTION_SCRIPT_RUN = 5120 # Variable c_int
STGM_DIRECT = 0 # Variable c_long
IMAGE_SYM_TYPE_STRUCT = 8 # Variable c_int
RPC_S_ENTRY_ALREADY_EXISTS = 1760 # Variable c_long
FIOASYNC = 2147772029 # Variable c_ulong
ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE = 2305 # Variable c_long
MMIO_UNICODEPROC = 16777216 # Variable c_int
ERROR_DRIVE_LOCKED = 108 # Variable c_long
DISP_E_ARRAYISLOCKED = -2147352563 # Variable c_long
PRINTER_CHANGE_SET_FORM = 131072 # Variable c_int
LANG_GREEK = 8 # Variable c_int
SEE_MASK_FLAG_DDEWAIT = 256 # Variable c_int
RPC_NCA_FLAGS_MAYBE = 4 # Variable c_int
UOI_NAME = 2 # Variable c_int
PROCESSOR_ARM_7TDMI = 70001 # Variable c_int
CCM_GETDROPTARGET = 8196 # Variable c_int
RGN_OR = 2 # Variable c_int
PS_USERSTYLE = 7 # Variable c_int
PSBTN_OK = 3 # Variable c_int
CLR_INVALID = 4294967295 # Variable c_uint
NRC_LOCKFAIL = 60 # Variable c_int
COLOR_GRAYTEXT = 17 # Variable c_int
S_SERDST = -16 # Variable c_int
CFSTR_MIME_GIF = 'image/gif' # Variable POINTER(c_wchar)
IME_ITHOTKEY_PREVIOUS_COMPOSITION = 513 # Variable c_int
IPPROTO_GGP = 3 # Variable c_int
SE_ERR_FNF = 2 # Variable c_int
GDT_ERROR = -1 # Variable c_int
PROCESS_HEAP_ENTRY_DDESHARE = 32 # Variable c_int
ERROR_SUBST_TO_SUBST = 139 # Variable c_long
DMRES_LOW = -2 # Variable c_int
CTLCOLOR_BTN = 3 # Variable c_int
PERF_QUERY_GLOBAL = -2147483647 # Variable c_long
PLANES = 14 # Variable c_int
HANGUL_CHARSET = 129 # Variable c_int
GL_LEVEL_FATAL = 1 # Variable c_int
WS_EX_NOPARENTNOTIFY = 4 # Variable c_long
LVCFMT_RIGHT = 1 # Variable c_int
IP_DEFAULT_MULTICAST_TTL = 1 # Variable c_int
MDMSPKR_OFF = 0 # Variable c_int
ABE_LEFT = 0 # Variable c_int
NMTTDISPINFOA_V1_SIZE = 104 # Variable c_uint
SPLREG_MINOR_VERSION = 'MinorVersion' # Variable POINTER(c_wchar)
ERROR_NT_CROSS_ENCRYPTION_REQUIRED = 1386 # Variable c_long
MCI_SEQ_DIV_PPQN = 1216 # Variable c_int
EXIT_SUCCESS = 0 # Variable c_int
TBN_GETOBJECT = 4294966584 # Variable c_uint
ERROR_TRANSFORM_NOT_SUPPORTED = 2004 # Variable c_long
NTM_REGULAR = 64 # Variable c_long
RPC_S_CANNOT_SUPPORT = 1764 # Variable c_long
SPAPI_E_MACHINE_UNAVAILABLE = -2146500062 # Variable c_long
CO_E_FIRST = 2147746288 # Variable c_ulong
VIEW_NETDISCONNECT = 10 # Variable c_int
MIXER_GETLINEINFOF_LINEID = 2 # Variable c_long
MM_MIM_OPEN = 961 # Variable c_int
FORMATDLGORD30 = 1544 # Variable c_int
MM_HIENGLISH = 5 # Variable c_int
WSAEUSERS = 10068 # Variable c_int
SERVICE_CONTROL_NETBINDREMOVE = 8 # Variable c_int
RT_DIALOG = 5 # Variable POINTER(c_wchar)
LOCALE_ICURRENCY = 27 # Variable c_int
IMAGE_REL_MIPS_REFLO = 5 # Variable c_int
IMAGE_FILE_MACHINE_MIPSFPU16 = 1126 # Variable c_int
EMR_POLYDRAW = 56 # Variable c_int
INET_E_USE_DEFAULT_PROTOCOLHANDLER = -2146697199 # Variable c_long
ERROR_INVALID_MODULETYPE = 190 # Variable c_long
NLS_ROMAN = 4194304 # Variable c_int
MCI_VD_PLAY_SCAN = 524288 # Variable c_long
STD_HELP = 11 # Variable c_int
lst16 = 1135 # Variable c_int
PIDSI_TITLE = 2 # Variable c_long
IMAGE_DOS_SIGNATURE = 23117 # Variable c_int
CERT_E_MALFORMED = -2146762488 # Variable c_long
WAVE_FORMAT_1M08 = 1 # Variable c_int
FR_MATCHCASE = 4 # Variable c_int
EMR_POLYTEXTOUTW = 97 # Variable c_int
WAVE_FORMAT_1S08 = 2 # Variable c_int
ERROR_CANT_OPEN_ANONYMOUS = 1347 # Variable c_long
NUMPRS_HEX_OCT = 64 # Variable c_int
TAPE_DRIVE_SETMARKS = 2148532224 # Variable c_uint
UDM_SETBASE = 1133 # Variable c_int
COLOR_CAPTIONTEXT = 9 # Variable c_int
FILE_SUPPORTS_SPARSE_FILES = 64 # Variable c_int
JOY_BUTTON13 = 4096 # Variable c_long
DMPAPER_ESHEET = 26 # Variable c_int
SM_CYEDGE = 46 # Variable c_int
CBES_EX_CASESENSITIVE = 16 # Variable c_int
EMR_STRETCHBLT = 77 # Variable c_int
IMAGE_REL_PPC_IMGLUE = 14 # Variable c_int
PAN_STRAIGHT_ARMS_WEDGE = 3 # Variable c_int
RB_DRAGMOVE = 1050 # Variable c_int
PGB_TOPORLEFT = 0 # Variable c_int
WINSTA_READSCREEN = 512 # Variable c_long
HELP_TCARD_DATA = 16 # Variable c_int
OR_INVALID_OXID = 1910 # Variable c_long
IMAGE_REL_I386_SEG12 = 9 # Variable c_int
TB_GETINSERTMARKCOLOR = 1113 # Variable c_int
FILE_ATTRIBUTE_TEMPORARY = 256 # Variable c_int
MERGECOPY = 12583114 # Variable c_ulong
FCONTROL = 8 # Variable c_int
TVM_GETITEMW = 4414 # Variable c_int
JOB_NOTIFY_FIELD_SECURITY_DESCRIPTOR = 12 # Variable c_int
WGL_SWAP_OVERLAY2 = 4 # Variable c_int
STM_SETIMAGE = 370 # Variable c_int
LOCALE_SABBREVMONTHNAME8 = 75 # Variable c_int
TVS_INFOTIP = 2048 # Variable c_int
SPLDS_PRINT_STATUS = 'printStatus' # Variable POINTER(c_wchar)
MIXERLINE_COMPONENTTYPE_DST_LAST = 8 # Variable c_long
IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG = 10 # Variable c_int
ERROR_OBJECT_IN_LIST = 5011 # Variable c_long
scr1 = 1168 # Variable c_int
TVN_BEGINDRAGW = 4294966840 # Variable c_uint
CDDS_ITEMPOSTERASE = 65540 # Variable c_int
MCM_SETMAXSELCOUNT = 4100 # Variable c_int
PC_WIDESTYLED = 64 # Variable c_int
SOCK_STREAM = 1 # Variable c_int
DI_NORMAL = 3 # Variable c_int
TVIS_DROPHILITED = 8 # Variable c_int
MCI_BREAK_HWND = 512 # Variable c_long
CB_SETEXTENDEDUI = 341 # Variable c_int
EM_SETTABSTOPS = 203 # Variable c_int
DOUBLE_CLICK = 2 # Variable c_int
IMN_SETCOMPOSITIONWINDOW = 11 # Variable c_int
IMAGE_SEPARATE_DEBUG_SIGNATURE = 18756 # Variable c_int
URLACTION_NETWORK_CURR_MAX = 6657 # Variable c_int
PRINTER_NOTIFY_FIELD_SERVER_NAME = 0 # Variable c_int
IMAGE_SCN_MEM_PRELOAD = 524288 # Variable c_int
ES_OEMCONVERT = 1024 # Variable c_long
RC_BANDING = 2 # Variable c_int
FADF_DISPATCH = 1024 # Variable c_int
FILE_TYPE_DISK = 1 # Variable c_int
CE_PTO = 512 # Variable c_int
MMIO_WRITE = 1 # Variable c_int
CTRY_CARIBBEAN = 1 # Variable c_int
CP_UTF7 = 65000 # Variable c_int
EVENTLOG_AUDIT_FAILURE = 16 # Variable c_int
EM_FMTLINES = 200 # Variable c_int
DC_SIZE = 8 # Variable c_int
SEF_DACL_AUTO_INHERIT = 1 # Variable c_int
RES_ICON = 1 # Variable c_int
MAX_LANA = 254 # Variable c_int
TCI_SRCFONTSIG = 3 # Variable c_int
GL_LEVEL_NOGUIDELINE = 0 # Variable c_int
IMAGE_SUBSYSTEM_POSIX_CUI = 7 # Variable c_int
LBSELCHSTRINGA = 'commdlg_LBSelChangedNotify' # Variable POINTER(c_char)
MCIERR_EXTRA_CHARACTERS = 305 # Variable c_int
DMLERR_DATAACKTIMEOUT = 16386 # Variable c_int
EXCEPTION_EXECUTE_HANDLER = 1 # Variable c_int
MIXERCONTROL_CONTROLTYPE_ONOFF = 536936449 # Variable c_long
CO_E_CLASS_CREATE_FAILED = -2146959359 # Variable c_long
IMN_OPENSTATUSWINDOW = 2 # Variable c_int
PSPCB_RELEASE = 1 # Variable c_int
EMR_SELECTCLIPPATH = 67 # Variable c_int
EMR_PLGBLT = 79 # Variable c_int
ILD_TRANSPARENT = 1 # Variable c_int
NTE_SIGNATURE_FILE_BAD = -2146893796 # Variable c_long
STGM_FAILIFTHERE = 0 # Variable c_long
MAXPROPPAGES = 100 # Variable c_int
VFT2_DRV_MOUSE = 5 # Variable c_long
NUMPENS = 18 # Variable c_int
LISTEN_OUTSTANDING = 1 # Variable c_int
LANG_BELARUSIAN = 35 # Variable c_int
TAPE_DRIVE_LOCK_UNLOCK = 2147483652 # Variable c_uint
LVIR_SELECTBOUNDS = 3 # Variable c_int
DTS_SHORTDATEFORMAT = 0 # Variable c_int
CTRY_POLAND = 48 # Variable c_int
MFT_RADIOCHECK = 512 # Variable c_long
DSPRINT_UPDATE = 2 # Variable c_int
PDERR_NODEVICES = 4103 # Variable c_int
RBHT_NOWHERE = 1 # Variable c_int
BS_DIBPATTERN8X8 = 8 # Variable c_int
IPM_GETADDRESS = 1126 # Variable c_int
ERROR_CHILD_MUST_BE_VOLATILE = 1021 # Variable c_long
RPC_C_BINDING_DEFAULT_TIMEOUT = 5 # Variable c_int
IMAGE_REL_ARM_ABSOLUTE = 0 # Variable c_int
LCMAP_HALFWIDTH = 4194304 # Variable c_int
DMMEDIA_TRANSPARENCY = 2 # Variable c_int
ERROR_INVALID_HANDLE_STATE = 1609 # Variable c_long
PCF_SPECIALCHARS = 256 # Variable c_ulong
SUBLANG_ITALIAN_SWISS = 2 # Variable c_int
ARW_TOPRIGHT = 3 # Variable c_long
WSAEBADF = 10009 # Variable c_int
VS_FF_DEBUG = 1 # Variable c_long
SERVER_ACCESS_ADMINISTER = 1 # Variable c_int
MF_USECHECKBITMAPS = 512 # Variable c_long
EXCEPTION_CONTINUE_SEARCH = 0 # Variable c_int
MAXSTRETCHBLTMODE = 4 # Variable c_int
NRC_DUPENV = 48 # Variable c_int
MCIERR_OUT_OF_MEMORY = 264 # Variable c_int
DMDITHER_COARSE = 2 # Variable c_int
HDF_JUSTIFYMASK = 3 # Variable c_int
ERROR_SCREEN_ALREADY_LOCKED = 1440 # Variable c_long
cmb6 = 1141 # Variable c_int
NORM_IGNORECASE = 1 # Variable c_int
LVN_BEGINLABELEDITA = 4294967191 # Variable c_uint
MM_MCISIGNAL = 971 # Variable c_int
OF_REOPEN = 32768 # Variable c_int
VERTRES = 10 # Variable c_int
PFD_SWAP_COPY = 1024 # Variable c_int
DONT_RESOLVE_DLL_REFERENCES = 1 # Variable c_int
PRINTACTION_NETINSTALL = 2 # Variable c_int
PSD_INWININIINTLMEASURE = 0 # Variable c_int
CS_E_NO_CLASSSTORE = -2147221144 # Variable c_long
NCBCANCEL = 53 # Variable c_int
UD_MAXVAL = 32767 # Variable c_int
CTRY_MEXICO = 52 # Variable c_int
RPC_X_SS_CHAR_TRANS_SHORT_FILE = 1774 # Variable c_long
TBDDRET_NODEFAULT = 1 # Variable c_int
SIF_ALL = 23 # Variable c_int
S_SERDDR = -14 # Variable c_int
SHGFI_DISPLAYNAME = 512 # Variable c_int
DMLERR_REENTRANCY = 16397 # Variable c_int
IMPLINK_IP = 155 # Variable c_int
TRANSPORT_TYPE_WMSG = 8 # Variable c_int
QS_ALLEVENTS = 191 # Variable c_int
OFN_ENABLETEMPLATE = 64 # Variable c_int
NM_FIRST = 0 # Variable c_uint
TAPE_DRIVE_VARIABLE_BLOCK = 2048 # Variable c_int
MCI_STOP = 2056 # Variable c_int
CDERR_FINDRESFAILURE = 6 # Variable c_int
PSH_MODELESS = 1024 # Variable c_int
TBCDRF_NOETCHEDEFFECT = 1048576 # Variable c_int
MOVEFILE_DELAY_UNTIL_REBOOT = 4 # Variable c_int
HDM_SETORDERARRAY = 4626 # Variable c_int
PAN_MIDLINE_STANDARD_SERIFED = 4 # Variable c_int
PRINTER_ATTRIBUTE_WORK_OFFLINE = 1024 # Variable c_int
psh16 = 1039 # Variable c_int
SHGFI_ICONLOCATION = 4096 # Variable c_int
SUBLANG_UZBEK_CYRILLIC = 2 # Variable c_int
RPC_C_MQ_JOURNAL_NONE = 0 # Variable c_int
NETINFO_PRINTERRED = 8 # Variable c_int
PRINTDLGORD = 1538 # Variable c_int
SO_REUSEADDR = 4 # Variable c_int
HKEY_PERFORMANCE_DATA = 2147483652 # Variable POINTER(HKEY__)
MCI_OVLY_OPEN_PARENT = 131072 # Variable c_long
MCI_GETDEVCAPS_USES_FILES = 5 # Variable c_long
STG_E_OLDDLL = -2147286779 # Variable c_long

tagFUNCKIND = c_int # enum
FUNC_VIRTUAL = 0
FUNC_PUREVIRTUAL = 1
FUNC_NONVIRTUAL = 2
stc6 = 1093 # Variable c_int
DS_SETFOREGROUND = 512 # Variable c_long
SERVICE_SYSTEM_START = 1 # Variable c_int
EN_MAXTEXT = 1281 # Variable c_int
MCIERR_DEVICE_OPEN = 265 # Variable c_int
S_SERDCC = -7 # Variable c_int
NLS_KATAKANA = 131072 # Variable c_int
COLOR_MENU = 4 # Variable c_int
TPM_CENTERALIGN = 4 # Variable c_long
NDR_LOCAL_DATA_REPRESENTATION = 16 # Variable c_ulong
FORM_PRINTER = 2 # Variable c_int
SM_CYSMSIZE = 53 # Variable c_int
PS_DOT = 2 # Variable c_int
CLSCTX_INPROC = 3 # Variable c_int
DFCS_MONO = 32768 # Variable c_int
OUT_STROKE_PRECIS = 3 # Variable c_int
TVC_BYKEYBOARD = 2 # Variable c_int
CDS_FULLSCREEN = 4 # Variable c_int
FD_ACCEPT = 8 # Variable c_int
IMAGE_REL_MIPS_JMPADDR16 = 16 # Variable c_int
SBT_POPOUT = 512 # Variable c_int
PSINJECT_REPLACE = 1 # Variable c_int
MIXER_OBJECTF_HWAVEOUT = 2415919104 # Variable c_ulong
SP_STOPBITS = 8 # Variable c_ulong

_GET_FILEEX_INFO_LEVELS = c_int # enum
GetFileExInfoStandard = 0
GetFileExMaxInfoLevel = 1
BN_KILLFOCUS = 7 # Variable c_int
SPI_SETNONCLIENTMETRICS = 42 # Variable c_int
CACHE_S_SOMECACHES_NOTUPDATED = 262514 # Variable c_long
PSD_NONETWORKBUTTON = 2097152 # Variable c_int
SUBLANG_PORTUGUESE_BRAZILIAN = 1 # Variable c_int
ERROR_CONNECTION_COUNT_LIMIT = 1238 # Variable c_long
CB_INSERTSTRING = 330 # Variable c_int
ORD_LANGDRIVER = 1 # Variable c_int
LOCALE_ILANGUAGE = 1 # Variable c_int
MIDIMAPPER = 4294967295 # Variable c_uint
SB_GETRECT = 1034 # Variable c_int
PRINTER_ATTRIBUTE_DO_COMPLETE_FIRST = 512 # Variable c_int
IPM_SETRANGE = 1127 # Variable c_int
MCIERR_INVALID_DEVICE_ID = 257 # Variable c_int
VFT2_DRV_LANGUAGE = 3 # Variable c_long
CBS_AUTOHSCROLL = 64 # Variable c_long
USER_MARSHAL_FC_USMALL = 4 # Variable c_int
TB_HITTEST = 1093 # Variable c_int
CAL_SLONGDATE = 6 # Variable c_int
NCBDGRECV = 33 # Variable c_int
TBSTATE_PRESSED = 2 # Variable c_int
RPC_C_OPT_MQ_JOURNAL = 3 # Variable c_int
INET_E_CANNOT_LOCK_REQUEST = -2146697194 # Variable c_long
HDN_ENDTRACKW = 4294966969 # Variable c_uint
RPC_S_SERVER_TOO_BUSY = 1723 # Variable c_long
IMAGE_SYM_CLASS_BLOCK = 100 # Variable c_int
AC_SRC_NO_PREMULT_ALPHA = 1 # Variable c_int
TVM_GETITEMA = 4364 # Variable c_int
SET_ARC_DIRECTION = 4102 # Variable c_int
TPM_VERTICAL = 64 # Variable c_long
CADV_LATEACK = 65535 # Variable c_int
SC_SEPARATOR = 61455 # Variable c_int
edt5 = 1156 # Variable c_int
FOREGROUND_INTENSITY = 8 # Variable c_int
MCI_MODE_STOP = 525 # Variable c_int
CTLCOLOR_MSGBOX = 0 # Variable c_int
ERROR_WMI_INSTANCE_NOT_FOUND = 4201 # Variable c_long
TVIF_INTEGRAL = 128 # Variable c_int
PRINTER_STATUS_PAUSED = 1 # Variable c_int
NRC_NORES = 9 # Variable c_int
LVNI_SELECTED = 2 # Variable c_int
SERVICE_ENUMERATE_DEPENDENTS = 8 # Variable c_int
PO_RENAME = 20 # Variable c_int
WH_KEYBOARD = 2 # Variable c_int
STATUS_FLOAT_DIVIDE_BY_ZERO = 3221225614 # Variable c_ulong
IMAGE_REL_IA64_RESERVED_11 = 17 # Variable c_int
NF_QUERY = 3 # Variable c_int
FACILITY_NULL = 0 # Variable c_int
CC_ENABLETEMPLATE = 32 # Variable c_int
IMAGE_SCN_ALIGN_2BYTES = 2097152 # Variable c_int
FOURCC_RIFF = 1179011410 # Variable c_ulong
RUNDLGORD = 1545 # Variable c_int
FW_NORMAL = 400 # Variable c_int
URLACTION_DOWNLOAD_CURR_MAX = 4100 # Variable c_int
SPAPI_E_DI_POSTPROCESSING_REQUIRED = -2146500058 # Variable c_long
IO_REPARSE_TAG_MOUNT_POINT = 2684354563 # Variable c_uint
PSNRET_NOERROR = 0 # Variable c_int
CAT_E_FIRST = 2147746144 # Variable c_ulong
REG_OPTION_RESERVED = 0 # Variable c_long
FLUSHOUTPUT = 6 # Variable c_int
DRAGDROP_S_DROP = 262400 # Variable c_long
PERF_100NSEC_TIMER_INV = 558957824 # Variable c_int
DLGC_WANTMESSAGE = 4 # Variable c_int
SE_LOAD_DRIVER_NAME = 'SeLoadDriverPrivilege' # Variable POINTER(c_wchar)
STD_PROPERTIES = 10 # Variable c_int
TTM_WINDOWFROMPOINT = 1040 # Variable c_int
RPC_E_CANTTRANSMIT_CALL = -2147418102 # Variable c_long
IDLFLAG_FLCID = 4 # Variable c_int
MEVT_F_CALLBACK = 1073741824 # Variable c_long
SUBLANG_FRENCH_LUXEMBOURG = 5 # Variable c_int
RESOURCEUSAGE_ATTACHED = 16 # Variable c_int
IDB_VIEW_SMALL_COLOR = 4 # Variable c_int
STARTF_USESIZE = 2 # Variable c_int
NCBSEND = 20 # Variable c_int
CDIS_SELECTED = 1 # Variable c_int
ERROR_NO_DATA = 232 # Variable c_long
DC_MINEXTENT = 4 # Variable c_int
FADF_UNKNOWN = 512 # Variable c_int
MCI_VD_GETDEVCAPS_SLOW_RATE = 16388 # Variable c_long
NTE_BAD_UID = -2146893823 # Variable c_long
STATUS_CONTROL_C_EXIT = 3221225786 # Variable c_ulong
WT_EXECUTEDELETEWAIT = 8 # Variable c_int
ERROR_ALREADY_REGISTERED = 1242 # Variable c_long
CP_RECTANGLE = 1 # Variable c_int
ERROR_BAD_DATABASE_VERSION = 1613 # Variable c_long
STANDARD_RIGHTS_EXECUTE = 131072 # Variable c_long
MCI_ANIM_OPEN_NOSTATIC = 262144 # Variable c_long
EMR_SETCOLORADJUSTMENT = 23 # Variable c_int
WSAEINPROGRESS = 10036 # Variable c_int
ERROR_DISK_CHANGE = 107 # Variable c_long
RPC_C_BIND_TO_ALL_NICS = 1 # Variable c_int
LB_INSERTSTRING = 385 # Variable c_int
WAVERR_BADFORMAT = 32 # Variable c_int
HDF_OWNERDRAW = 32768 # Variable c_int
WNNC_NET_PROTSTOR = 2162688 # Variable c_int
CTRL_BREAK_EVENT = 1 # Variable c_int
EMARCH_ENC_I17_IMM9D_VAL_POS_X = 7 # Variable c_int
PRNSETUPDLGORD = 1539 # Variable c_int
HDN_BEGINTRACKA = 4294966990 # Variable c_uint
STD_FILEOPEN = 7 # Variable c_int
CBENF_RETURN = 2 # Variable c_int
SERVICES_ACTIVE_DATABASEW = 'ServicesActive' # Variable POINTER(c_wchar)
DMTT_BITMAP = 1 # Variable c_int
SC_KEYMENU = 61696 # Variable c_int
PRINTER_CHANGE_PRINTER = 255 # Variable c_int
MCIERR_WAVE_OUTPUTUNSPECIFIED = 324 # Variable c_int
ANSI_NULL = 0 # Variable c_char
CP_OEMCP = 1 # Variable c_int
IMAGE_DIRECTORY_ENTRY_TLS = 9 # Variable c_int
DMICMMETHOD_DEVICE = 4 # Variable c_int
PSINJECT_PAGEORDER = 22 # Variable c_int
LBN_SELCANCEL = 3 # Variable c_int
URLPOLICY_AUTHENTICATE_CLEARTEXT_OK = 0 # Variable c_int
CC_STYLED = 32 # Variable c_int
SW_OTHERUNZOOM = 4 # Variable c_int
WAVECAPS_PLAYBACKRATE = 2 # Variable c_int
WNNC_NET_NETWARE = 196608 # Variable c_int
NTE_PROV_TYPE_ENTRY_BAD = -2146893800 # Variable c_long
IMAGE_SCN_GPREL = 32768 # Variable c_int
ERROR_DELETING_ICM_XFORM = 2309 # Variable c_long
SID_MAX_SUB_AUTHORITIES = 15 # Variable c_int
TT_POLYGON_TYPE = 24 # Variable c_int
SEC_IMAGE = 16777216 # Variable c_int
EMARCH_ENC_I17_IMM41b_INST_WORD_POS_X = 24 # Variable c_int
MK_E_EXCEEDEDDEADLINE = -2147221023 # Variable c_long
LVM_ARRANGE = 4118 # Variable c_int
LVBKIF_SOURCE_URL = 2 # Variable c_int
TBS_BOTTOM = 0 # Variable c_int
RIGHT_CTRL_PRESSED = 4 # Variable c_int
NUMPRS_LEADING_WHITE = 1 # Variable c_int
OBJ_FONT = 6 # Variable c_int
RPC_S_UNKNOWN_AUTHN_LEVEL = 1748 # Variable c_long
GCL_STYLE = -26 # Variable c_int
RB_GETBANDINFOW = 1052 # Variable c_int
LOCALE_IMONLZERO = 39 # Variable c_int
MB_RTLREADING = 1048576 # Variable c_long
IMAGE_DEBUG_TYPE_FIXUP = 6 # Variable c_int
PAN_XHEIGHT_CONSTANT_STD = 3 # Variable c_int
IMAGE_ARCHIVE_START = '!<arch>\n' # Variable POINTER(c_char)
RPCFLG_INPUT_SYNCHRONOUS = 536870912 # Variable c_ulong
COLOR_3DLIGHT = 22 # Variable c_int
CAT_E_CATIDNOEXIST = -2147221152 # Variable c_long
ERROR_RMODE_APP = 1153 # Variable c_long
TC_NORMAL = 0 # Variable c_int
DMICMMETHOD_SYSTEM = 2 # Variable c_int
TIMER_ALL_ACCESS = 2031619 # Variable c_long
MB_MODEMASK = 12288 # Variable c_long
CLIPBRD_S_FIRST = 262608 # Variable c_long
IMPLINK_LOWEXPER = 156 # Variable c_int
JOB_OBJECT_UILIMIT_SYSTEMPARAMETERS = 8 # Variable c_int
RPC_C_AUTHN_LEVEL_NONE = 1 # Variable c_int
HOTKEYF_EXT = 8 # Variable c_int

tagTYMED = c_int # enum
TYMED_GDI = 16
TYMED_MFPICT = 32
TYMED_ENHMF = 64
TYMED_NULL = 0
EMR_MIN = 1 # Variable c_int
OFN_NOLONGNAMES = 262144 # Variable c_int
TTM_GETMARGIN = 1051 # Variable c_int
EVENT_MODIFY_STATE = 2 # Variable c_int
SUBLANG_SWEDISH_FINLAND = 2 # Variable c_int
APD_COPY_NEW_FILES = 8 # Variable c_int
PAN_CONTRAST_INDEX = 4 # Variable c_int
RPC_S_NOTHING_TO_EXPORT = 1754 # Variable c_long
DESKTOP_HOOKCONTROL = 8 # Variable c_long
WSAEAFNOSUPPORT = 10047 # Variable c_int
HELP_INDEX = 3 # Variable c_long
ICM_SETDEFAULTPROFILE = 4 # Variable c_int
ERROR_BAD_NET_NAME = 67 # Variable c_long
LOCALE_SABBREVMONTHNAME9 = 76 # Variable c_int
STG_E_SHAREVIOLATION = -2147287008 # Variable c_long
OLE_E_CANT_GETMONIKER = -2147221495 # Variable c_long
SB_VERT = 1 # Variable c_int
ACL_REVISION1 = 1 # Variable c_int
DMPAPER_ENV_C65 = 32 # Variable c_int
stc9 = 1096 # Variable c_int
ERROR_OBJECT_NOT_FOUND = 4312 # Variable c_long
APD_STRICT_DOWNGRADE = 2 # Variable c_int
SUBLANG_GERMAN_LUXEMBOURG = 4 # Variable c_int
SPAPI_E_DI_NOFILECOPY = -2146500081 # Variable c_long
MIXERLINE_TARGETTYPE_WAVEOUT = 1 # Variable c_int
WVR_VALIDRECTS = 1024 # Variable c_int
RPC_S_NO_MORE_MEMBERS = 1757 # Variable c_long
JOY_BUTTON29 = 268435456 # Variable c_long
rad9 = 1064 # Variable c_int
NMTVCUSTOMDRAW_V3_SIZE = 56 # Variable c_uint
MCI_OVLY_WINDOW_DISABLE_STRETCH = 2097152 # Variable c_long
COMMON_LVB_UNDERSCORE = 32768 # Variable c_int
RBS_TOOLTIPS = 256 # Variable c_int
ID_DEFAULTINST = -2 # Variable c_int
LVM_GETCOLUMNA = 4121 # Variable c_int
CTRY_FINLAND = 358 # Variable c_int
SPI_GETKEYBOARDPREF = 68 # Variable c_int
CDN_LAST = 4294966597 # Variable c_uint
DO_DROPFILE = 1162627398 # Variable c_long
TB_ADDBUTTONSA = 1044 # Variable c_int
EVENTLOG_AUDIT_SUCCESS = 8 # Variable c_int
SUBLANG_GERMAN_AUSTRIAN = 3 # Variable c_int
CTRY_PHILIPPINES = 63 # Variable c_int
TBM_SETBUDDY = 1056 # Variable c_int
CBES_EX_NOEDITIMAGE = 1 # Variable c_int
META_SETVIEWPORTEXT = 526 # Variable c_int
IDI_QUESTION = 206100 # Variable POINTER(c_wchar)
VIEW_SORTNAME = 4 # Variable c_int
TAPE_SELECT_PARTITIONS = 1 # Variable c_long
LOCALE_ILDATE = 34 # Variable c_int
DC_EXTRA = 9 # Variable c_int
URLACTION_CREDENTIALS_USE = 6656 # Variable c_int
TB_SETROWS = 1063 # Variable c_int
COLOR_INACTIVEBORDER = 11 # Variable c_int
FRAME_FPO = 0 # Variable c_int
JOB_WRITE = 131088 # Variable c_long
STGM_READWRITE = 2 # Variable c_long
TVM_SETINSERTMARKCOLOR = 4389 # Variable c_int
LBS_OWNERDRAWFIXED = 16 # Variable c_long
TTF_RTLREADING = 4 # Variable c_int
TCS_MULTILINE = 512 # Variable c_int
SBS_BOTTOMALIGN = 4 # Variable c_long
PSD_INTHOUSANDTHSOFINCHES = 4 # Variable c_int
META_SETPOLYFILLMODE = 262 # Variable c_int
SERVICE_STOP = 32 # Variable c_int
MIXERCONTROL_CT_UNITS_UNSIGNED = 196608 # Variable c_long
AC_LINE_BACKUP_POWER = 2 # Variable c_int
MCI_SET_OFF = 16384 # Variable c_long
ERROR_INTERNAL_ERROR = 1359 # Variable c_long
ERROR_INVALID_NAME = 123 # Variable c_long
ENUM_E_LAST = 2147746239 # Variable c_ulong
WNNC_NET_FRONTIER = 1507328 # Variable c_int
STGM_SHARE_DENY_WRITE = 32 # Variable c_long
URLACTION_NETWORK_MIN = 6656 # Variable c_int
TTM_GETCURRENTTOOLW = 1083 # Variable c_int
LOCALE_SLIST = 12 # Variable c_int
PERF_COUNTER_MULTI_TIMER = 574686464 # Variable c_int
FILE_TYPE_UNKNOWN = 0 # Variable c_int
ERROR_INVALID_FUNCTION = 1 # Variable c_long
IMAGE_SYM_CLASS_LABEL = 6 # Variable c_int
ERROR_ACCESS_DENIED = 5 # Variable c_long
DMPAPER_STATEMENT = 6 # Variable c_int
DT_NOPREFIX = 2048 # Variable c_int
EMR_FLATTENPATH = 65 # Variable c_int
HDF_STRING = 16384 # Variable c_int
URLACTION_ACTIVEX_CONFIRM_NOOBJECTSAFETY = 4612 # Variable c_int
PSN_WIZFINISH = 4294967088 # Variable c_uint
CTRY_FAEROE_ISLANDS = 298 # Variable c_int
VOS__BASE = 0 # Variable c_long
IDCANCEL = 2 # Variable c_int
ES_RIGHT = 2 # Variable c_long
CDDS_POSTERASE = 4 # Variable c_int
DFCS_BUTTONPUSH = 16 # Variable c_int
LB_GETITEMDATA = 409 # Variable c_int
CALL_PENDING = 2 # Variable c_int
OUTPUT_DEBUG_STRING_EVENT = 8 # Variable c_int
CO_E_OBJSRV_RPC_FAILURE = -2146959354 # Variable c_long
EMARCH_ENC_I17_IMM41b_INST_WORD_X = 1 # Variable c_int
GL_ID_INPUTREADING = 36 # Variable c_int
cmb9 = 1144 # Variable c_int
STATUS_IN_PAGE_ERROR = 3221225478 # Variable c_ulong
QS_POSTMESSAGE = 8 # Variable c_int
DM_COLLATE = 32768 # Variable c_long
FILE_ATTRIBUTE_ARCHIVE = 32 # Variable c_int
IMAGE_SYM_DEBUG = -2 # Variable c_short
ERROR_PRIVATE_DIALOG_INDEX = 1415 # Variable c_long
NCBSSTAT = 52 # Variable c_int
JOB_OBJECT_SECURITY_RESTRICTED_TOKEN = 2 # Variable c_int
IMAGE_REL_MIPS_SECRELHI = 13 # Variable c_int
FACILITY_SSPI = 9 # Variable c_int
OUT_TT_ONLY_PRECIS = 7 # Variable c_int
RBS_DBLCLKTOGGLE = 32768 # Variable c_int
OUT_SCREEN_OUTLINE_PRECIS = 9 # Variable c_int
REG_OPTION_NON_VOLATILE = 0 # Variable c_long
EXCEPTION_NONCONTINUABLE = 1 # Variable c_int
RPC_X_ENUM_VALUE_OUT_OF_RANGE = 1781 # Variable c_long
SECURITY_DIALUP_RID = 1 # Variable c_long
URLACTION_SCRIPT_OVERRIDE_SAFETY = 5121 # Variable c_int
ERROR_INVALID_ADDRESS = 487 # Variable c_long
WSAENOTSOCK = 10038 # Variable c_int
DTM_FIRST = 4096 # Variable c_int
MMIO_GLOBALPROC = 268435456 # Variable c_int
CBF_FAIL_EXECUTES = 32768 # Variable c_int
DI_READ_SPOOL_JOB = 3 # Variable c_int
BF_BOTTOMLEFT = 9 # Variable c_int
MCI_WINDOW = 2113 # Variable c_int
WSAEHOSTUNREACH = 10065 # Variable c_int
SS_ETCHEDVERT = 17 # Variable c_long
ICC_INTERNET_CLASSES = 2048 # Variable c_int
LOCALE_INEGSYMPRECEDES = 86 # Variable c_int
LVN_GETINFOTIPW = 4294967138 # Variable c_uint
JOY_CAL_READALWAYS = 65536 # Variable c_long
MCI_OPEN_TYPE_ID = 4096 # Variable c_long
EMR_CREATEPEN = 38 # Variable c_int
ERROR_PRINTER_DRIVER_IN_USE = 3001 # Variable c_long
PGB_BOTTOMORRIGHT = 1 # Variable c_int
ERROR_DEVICE_REINITIALIZATION_NEEDED = 1164 # Variable c_long
EMR_SETTEXTALIGN = 22 # Variable c_int
IP_DROP_MEMBERSHIP = 6 # Variable c_int
TYPE_E_INVALIDSTATE = -2147319767 # Variable c_long
NI_SETCANDIDATE_PAGESIZE = 23 # Variable c_int
DLGWINDOWEXTRA = 30 # Variable c_int
CCM_FIRST = 8192 # Variable c_int
GCL_MENUNAME = -8 # Variable c_int
CBS_NOINTEGRALHEIGHT = 1024 # Variable c_long
HDF_IMAGE = 2048 # Variable c_int
TVHT_ONITEMRIGHT = 32 # Variable c_int
PD_NOWARNING = 128 # Variable c_int
MCIERR_CUSTOM_DRIVER_BASE = 512 # Variable c_int
MMIO_TOUPPER = 16 # Variable c_int
ERROR_WINDOW_NOT_DIALOG = 1420 # Variable c_long
TBSTYLE_ALTDRAG = 1024 # Variable c_int
META_EXCLUDECLIPRECT = 1045 # Variable c_int
CTRY_DEFAULT = 0 # Variable c_int
NFS_LISTCOMBO = 4 # Variable c_int
MM_MIM_CLOSE = 962 # Variable c_int
DCE_C_ERROR_STRING_LEN = 256 # Variable c_int
CTRY_LATVIA = 371 # Variable c_int
ERROR_WRITE_PROTECT = 19 # Variable c_long
CD_LBSELADD = 2 # Variable c_int
TOKEN_ADJUST_DEFAULT = 128 # Variable c_int
STG_S_CONSOLIDATIONFAILED = 197125 # Variable c_long
CRYPT_E_ISSUER_SERIALNUMBER = -2146889715 # Variable c_long
LBS_NODATA = 8192 # Variable c_long
TVN_SELCHANGINGA = 4294966895 # Variable c_uint
TAPE_DRIVE_ERASE_LONG = 32 # Variable c_int
EMARCH_ENC_I17_IMM5C_INST_WORD_X = 3 # Variable c_int
TBM_GETSELEND = 1042 # Variable c_int
PRINTER_STATUS_IO_ACTIVE = 256 # Variable c_int
CAL_SABBREVDAYNAME7 = 20 # Variable c_int
TB_GETEXTENDEDSTYLE = 1109 # Variable c_int
TB_GETTEXTROWS = 1085 # Variable c_int
ATTR_FIXEDCONVERTED = 5 # Variable c_int
URLPOLICY_JAVA_MEDIUM = 131072 # Variable c_int
TAPE_DRIVE_SEQUENTIAL_SMKS = 2149580800 # Variable c_uint
SM_CYSMICON = 50 # Variable c_int
CACHE_S_FORMATETC_NOTSUPPORTED = 262512 # Variable c_long
CTRY_QATAR = 974 # Variable c_int
ERROR_CANNOT_DETECT_PROCESS_ABORT = 1081 # Variable c_long
INVALID_ATOM = 0 # Variable c_ushort
CRYPT_E_NO_SIGNER = -2146885618 # Variable c_long
STGM_DELETEONRELEASE = 67108864 # Variable c_long
XCLASS_DATA = 8192 # Variable c_int
HSHELL_GETMINRECT = 5 # Variable c_int
LVN_DELETEALLITEMS = 4294967192 # Variable c_uint
SUBLANG_ARABIC_IRAQ = 2 # Variable c_int
URLACTION_MIN = 4096 # Variable c_int
CB_GETEDITSEL = 320 # Variable c_int
PDERR_PRINTERNOTFOUND = 4107 # Variable c_int
SM_DEBUG = 22 # Variable c_int
ERROR_PATH_BUSY = 148 # Variable c_long
SS_TYPEMASK = 31 # Variable c_long
IMAGE_REL_I386_DIR32NB = 7 # Variable c_int
RDW_VALIDATE = 8 # Variable c_int
SHGFI_TYPENAME = 1024 # Variable c_int
C1_BLANK = 64 # Variable c_int
edt6 = 1157 # Variable c_int
RESOURCEUSAGE_CONNECTABLE = 1 # Variable c_int
DM_PAPERLENGTH = 4 # Variable c_long
OUT_RASTER_PRECIS = 6 # Variable c_int
R2_MERGEPEN = 15 # Variable c_int
MMIOM_WRITEFLUSH = 5 # Variable c_int
PAN_LETT_NORMAL_BOXED = 4 # Variable c_int
LB_OKAY = 0 # Variable c_int
lst3 = 1122 # Variable c_int
META_DIBBITBLT = 2368 # Variable c_int
BSM_ALLCOMPONENTS = 0 # Variable c_int
MCI_PASTE = 2131 # Variable c_int
FADF_AUTO = 1 # Variable c_int
ERROR_DIRECT_ACCESS_HANDLE = 130 # Variable c_long
SERVICE_INACTIVE = 2 # Variable c_int
IMAGE_REL_IA64_RESERVED_16 = 22 # Variable c_int
WGL_SWAP_UNDERLAY8 = 8388608 # Variable c_int
JOY_RETURNALL = 255 # Variable c_long
IPPORT_SMTP = 25 # Variable c_int
SO_SNDLOWAT = 4099 # Variable c_int
SETKERNTRACK = 770 # Variable c_int
IPPORT_SYSTAT = 11 # Variable c_int
LOCALE_INEGSIGNPOSN = 83 # Variable c_int
CO_E_BAD_SERVER_NAME = -2147467244 # Variable c_long
WSAENETRESET = 10052 # Variable c_int
RB_GETBARINFO = 1027 # Variable c_int
GW_HWNDLAST = 1 # Variable c_int
LVIF_IMAGE = 2 # Variable c_int
SO_BROADCAST = 32 # Variable c_int
DMCOLLATE_TRUE = 1 # Variable c_int
ILLUMINANT_F2 = 8 # Variable c_int
SERVICE_QUERY_STATUS = 4 # Variable c_int
CC_NONE = 0 # Variable c_int
DT_WORD_ELLIPSIS = 262144 # Variable c_int
FORM_BUILTIN = 1 # Variable c_int
MCHT_TITLEMONTH = 65537 # Variable c_int
DMBIN_LOWER = 2 # Variable c_int
SBS_TOPALIGN = 2 # Variable c_long
ERROR_TOO_MANY_NAMES = 68 # Variable c_long
ERROR_CANCELLED = 1223 # Variable c_long
ERROR_KEY_DELETED = 1018 # Variable c_long
LPD_TYPE_RGBA = 0 # Variable c_int
META_SELECTPALETTE = 564 # Variable c_int
SERVICE_ACCEPT_NETBINDCHANGE = 16 # Variable c_int
MCI_VD_STATUS_SPEED = 16386 # Variable c_long
COLOROKSTRINGA = 'commdlg_ColorOK' # Variable POINTER(c_char)
RPC_E_UNEXPECTED = -2147352577 # Variable c_long
WSAEFAULT = 10014 # Variable c_int
C1_XDIGIT = 128 # Variable c_int
IPN_FIRST = 4294966436 # Variable c_uint
EM_SETHANDLE = 188 # Variable c_int
MIDISTRM_ERROR = -2 # Variable c_int
DMICMMETHOD_USER = 256 # Variable c_int
IS_TEXT_UNICODE_REVERSE_CONTROLS = 64 # Variable c_int
PSH_USEHBMWATERMARK = 65536 # Variable c_int
REGDB_E_KEYMISSING = -2147221166 # Variable c_long
RPC_C_EP_MATCH_BY_OBJ = 2 # Variable c_int
IPPROTO_PUP = 12 # Variable c_int
INADDR_BROADCAST = 4294967295 # Variable c_ulong
STG_LAYOUT_SEQUENTIAL = 0 # Variable c_long
SID_REVISION = 1 # Variable c_int
SHGFI_ICON = 256 # Variable c_int
CS_NOCLOSE = 512 # Variable c_int
IN_CLASSA_NET = 4278190080 # Variable c_uint
NRC_SCLOSED = 10 # Variable c_int
FSB_FLAT_MODE = 2 # Variable c_int
IMAGE_SCN_MEM_SHARED = 268435456 # Variable c_int
TYPE_E_BADMODULEKIND = -2147317571 # Variable c_long
ERROR_NOT_EXPORT_FORMAT = 6008 # Variable c_long
CLIPBRD_E_CANT_EMPTY = -2147221039 # Variable c_long
TAPE_SHORT_FILEMARKS = 2 # Variable c_long
SM_CXMINIMIZED = 57 # Variable c_int
CO_E_INIT_TLS = -2147467258 # Variable c_long
IMAGE_REL_MIPS_LITERAL = 7 # Variable c_int
ODT_BUTTON = 4 # Variable c_int
ABS_AUTOHIDE = 1 # Variable c_int
SERVER_ALL_ACCESS = 983043 # Variable c_long
OBJ_METAFILE = 9 # Variable c_int
DMPAPER_LETTER_EXTRA = 50 # Variable c_int
TTF_TRANSPARENT = 256 # Variable c_int
SERVICES_FAILED_DATABASEW = 'ServicesFailed' # Variable POINTER(c_wchar)
RBBS_NOVERT = 16 # Variable c_int
PERF_TYPE_COUNTER = 1024 # Variable c_int
DM_LOGPIXELS = 131072 # Variable c_long
ENHANCED_KEY = 256 # Variable c_int
LVS_ALIGNLEFT = 2048 # Variable c_int
LOCALE_FONTSIGNATURE = 88 # Variable c_int
EV_RX80FULL = 1024 # Variable c_int
HKEY_CURRENT_CONFIG = 2147483653 # Variable POINTER(HKEY__)
TBSTYLE_SEP = 1 # Variable c_int
URLPOLICY_QUERY = 1 # Variable c_int
WAVE_FORMAT_DIRECT = 8 # Variable c_int
MOD_SYNTH = 2 # Variable c_int
RPC_S_FP_DIV_ZERO = 1769 # Variable c_long
PIDSI_CREATE_DTM = 12 # Variable c_long
IME_SMODE_CONVERSATION = 16 # Variable c_int
CBR_57600 = 57600 # Variable c_int
PRINTER_NOTIFY_FIELD_AVERAGE_PPM = 21 # Variable c_int
FR_ENABLETEMPLATE = 512 # Variable c_int
EPS_SIGNATURE = 1179865157 # Variable c_int
OLE_E_CLASSDIFF = -2147221496 # Variable c_long
PD_RETURNDC = 256 # Variable c_int
LMEM_MODIFY = 128 # Variable c_int
JOB_OBJECT_UI_VALID_FLAGS = 255 # Variable c_int
ACM_PLAY = 1125 # Variable c_int
ERROR_INVALID_DWP_HANDLE = 1405 # Variable c_long
DCB_SET = 3 # Variable c_int
FILE_FLAG_NO_BUFFERING = 536870912 # Variable c_int
ILCF_MOVE = 0 # Variable c_int
SB_SETICON = 1039 # Variable c_int
I_CHILDRENCALLBACK = -1 # Variable c_int
HDATA_APPOWNED = 1 # Variable c_int
SPLREG_DS_PRESENT = 'DsPresent' # Variable POINTER(c_wchar)
IMAGE_REL_ARM_BRANCH11 = 4 # Variable c_int
MCI_WAVE_STATUS_LEVEL = 16391 # Variable c_long
FILE_ADD_FILE = 2 # Variable c_int
DMPAPER_ENV_INVITE = 47 # Variable c_int
CRYPT_E_INVALID_IA5_STRING = -2146885598 # Variable c_long
ENABLE_PROCESSED_OUTPUT = 1 # Variable c_int
DRAGDROP_E_INVALIDHWND = -2147221246 # Variable c_long
IMAGE_FILE_BYTES_REVERSED_HI = 32768 # Variable c_int
RB_GETBANDINFOA = 1053 # Variable c_int
AF_CHAOS = 5 # Variable c_int
ACTIVEOBJECT_STRONG = 0 # Variable c_int
TF_DISCONNECT = 1 # Variable c_int
OLE_E_CANT_BINDTOSOURCE = -2147221494 # Variable c_long
HS_BDIAGONAL = 3 # Variable c_int
TT_PRIM_QSPLINE = 2 # Variable c_int
JOB_CONTROL_DELETE = 5 # Variable c_int
WIN_CERT_TYPE_X509 = 1 # Variable c_int
DM_DITHERTYPE = 67108864 # Variable c_long
MIDI_CACHE_QUERY = 3 # Variable c_int
ERROR_INVALID_MSGBOX_STYLE = 1438 # Variable c_long
ILD_MASK = 16 # Variable c_int
RT_FONT = 8 # Variable POINTER(c_wchar)
IMAGE_DEBUG_TYPE_EXCEPTION = 5 # Variable c_int
LVM_SETITEMCOUNT = 4143 # Variable c_int
IME_PROP_UNICODE = 524288 # Variable c_int
EN_ERRSPACE = 1280 # Variable c_int
SB_SETTEXTA = 1025 # Variable c_int
MDMVOLFLAG_LOW = 1 # Variable c_int
CTRY_LITHUANIA = 370 # Variable c_int
IMAGE_SCN_MEM_16BIT = 131072 # Variable c_int
EMR_SETSTRETCHBLTMODE = 21 # Variable c_int
CB_RESETCONTENT = 331 # Variable c_int
SLE_ERROR = 1 # Variable c_int
OLE_S_MAC_CLIPFORMAT = 262146 # Variable c_long
MCIERR_BAD_TIME_FORMAT = 293 # Variable c_int
MIDIERR_NOMAP = 66 # Variable c_int
MIXERCONTROL_CONTROLTYPE_PERCENT = 805634048 # Variable c_long
SPAPI_E_SECTION_NOT_FOUND = -2146500351 # Variable c_long
CRYPT_E_NO_VERIFY_USAGE_CHECK = -2146885592 # Variable c_long
ERROR_NO_SYSTEM_MENU = 1437 # Variable c_long
KEY_READ = 131097 # Variable c_long
PRINTER_ENUM_ICON3 = 262144 # Variable c_int
NIM_DELETE = 2 # Variable c_int
PAN_SERIF_OBTUSE_SANS = 12 # Variable c_int
MOUSE_WHEELED = 4 # Variable c_int
FILE_NOTIFY_CHANGE_LAST_ACCESS = 32 # Variable c_int
WS_EX_ACCEPTFILES = 16 # Variable c_long
ANSI_VAR_FONT = 12 # Variable c_int
MCI_CDA_STATUS_TYPE_TRACK = 16385 # Variable c_long
FADF_FIXEDSIZE = 16 # Variable c_int
INVALID_HANDLE_VALUE = 4294967295 # Variable c_void_p
PST_NETWORK_BRIDGE = 256 # Variable c_ulong
NRC_BRIDGE = 35 # Variable c_int
TAPE_FILEMARKS = 1 # Variable c_long
TVIS_EXPANDED = 32 # Variable c_int
MCM_GETSELRANGE = 4101 # Variable c_int
MCI_NOTIFY = 1 # Variable c_long
DMMEDIA_STANDARD = 1 # Variable c_int
FNERR_SUBCLASSFAILURE = 12289 # Variable c_int
STATUS_GUARD_PAGE_VIOLATION = 2147483649 # Variable c_ulong
SEE_MASK_NOCLOSEPROCESS = 64 # Variable c_int
GCPGLYPH_LINKBEFORE = 32768 # Variable c_int
URLPOLICY_JAVA_CUSTOM = 8388608 # Variable c_int
RPC_C_NS_SYNTAX_DCE = 3 # Variable c_int
PRINTER_NOTIFY_FIELD_START_TIME = 16 # Variable c_int
ERROR_NO_ASSOCIATION = 1155 # Variable c_long
UDM_GETACCEL = 1132 # Variable c_int
PHYSICALOFFSETX = 112 # Variable c_int
FINDDLGORD = 1540 # Variable c_int
VTBIT_CY = 64 # Variable c_int
TPM_RIGHTALIGN = 8 # Variable c_long
IMAGE_SYM_CLASS_UNDEFINED_STATIC = 14 # Variable c_int

tagUSERCLASSTYPE = c_int # enum
USERCLASSTYPE_FULL = 1
USERCLASSTYPE_SHORT = 2
MCI_ANIM_STATUS_HWND = 16387 # Variable c_long
MF_UNHILITE = 0 # Variable c_long
SBS_RIGHTALIGN = 4 # Variable c_long
CAL_SERASTRING = 4 # Variable c_int
RES_CURSOR = 2 # Variable c_int
DC_INBUTTON = 16 # Variable c_int
OUT_DEVICE_PRECIS = 5 # Variable c_int
SPI_GETHIGHCONTRAST = 66 # Variable c_int
LMEM_INVALID_HANDLE = 32768 # Variable c_int
WS_POPUPWINDOW = 2156396544 # Variable c_ulong
BF_FLAT = 16384 # Variable c_int
PT_BEZIERTO = 4 # Variable c_int
NDR_FLOAT_REP_MASK = 65280 # Variable c_ulong
WAVE_MAPPED = 4 # Variable c_int
VS_FF_PRIVATEBUILD = 8 # Variable c_long
NUMPRS_TRAILING_MINUS = 32 # Variable c_int
BS_DEFPUSHBUTTON = 1 # Variable c_long
KLF_SUBSTITUTE_OK = 2 # Variable c_int
ACTFLG_Published = 8 # Variable c_int
LR_MONOCHROME = 1 # Variable c_int
ERROR_MENU_ITEM_NOT_FOUND = 1456 # Variable c_long
DLGC_RADIOBUTTON = 64 # Variable c_int
SP_BAUD = 2 # Variable c_ulong
SPI_GETSCREENSAVETIMEOUT = 14 # Variable c_int
ENABLE_ECHO_INPUT = 4 # Variable c_int
STATUS_NO_MEMORY = 3221225495 # Variable c_ulong
ISC_SHOWUIALLCANDIDATEWINDOW = 15 # Variable c_int
SC_HSCROLL = 61568 # Variable c_int
RPC_C_BINDING_INFINITE_TIMEOUT = 10 # Variable c_int
ERROR_RESOURCE_TYPE_NOT_FOUND = 1813 # Variable c_long
LVCF_FMT = 1 # Variable c_int
CONVERT10_E_STG_FMT = -2147221052 # Variable c_long
OBJ_MEMDC = 10 # Variable c_int
CONTEXT_CONTROL = 65537 # Variable c_long
DI_ROPS_READ_DESTINATION = 2 # Variable c_int
EMARCH_ENC_I17_SIGN_SIZE_X = 1 # Variable c_int
MIXERLINE_COMPONENTTYPE_DST_WAVEIN = 7 # Variable c_long
PD_ENABLEPRINTHOOK = 4096 # Variable c_int
LOCALE_IDATE = 33 # Variable c_int
MCI_ANIM_REALIZE_BKGD = 131072 # Variable c_long
IS_TEXT_UNICODE_REVERSE_MASK = 240 # Variable c_int
JOB_OBJECT_LIMIT_AFFINITY = 16 # Variable c_int
ERROR_RESMON_CREATE_FAILED = 5017 # Variable c_long
CBR_BLOCK = 4294967295 # Variable POINTER(HDDEDATA__)
MWMO_WAITALL = 1 # Variable c_int
CTRY_LUXEMBOURG = 352 # Variable c_int
GL_LEVEL_ERROR = 2 # Variable c_int
CRYPT_E_NO_MATCH = -2146885623 # Variable c_long
HDITEMA_V1_SIZE = 28 # Variable c_uint
PRINTER_ATTRIBUTE_LOCAL = 64 # Variable c_int
cmb8 = 1143 # Variable c_int
PD_NONETWORKBUTTON = 2097152 # Variable c_int
OFN_SHAREWARN = 0 # Variable c_int
NETINFO_DISKRED = 4 # Variable c_int
TBN_ENDDRAG = 4294966594 # Variable c_uint
ERROR_DUP_NAME = 52 # Variable c_long
PWR_CRITICALRESUME = 3 # Variable c_int
SUBLANG_ENGLISH_PHILIPPINES = 13 # Variable c_int
ERROR_HWNDS_HAVE_DIFF_PARENT = 1441 # Variable c_long
SPAPI_E_NO_DEFAULT_DEVICE_INTERFACE = -2146500070 # Variable c_long
DISP_CHANGE_RESTART = 1 # Variable c_int
WNNC_NET_FARALLON = 1179648 # Variable c_int
GETEXTENDEDTEXTMETRICS = 256 # Variable c_int
EM_LINELENGTH = 193 # Variable c_int
MEVT_F_LONG = 2147483648 # Variable c_ulong
VARIANT_CALENDAR_HIJRI = 8 # Variable c_int
REGDB_S_FIRST = 262480 # Variable c_long
IDCLOSE = 8 # Variable c_int
DMICMMETHOD_NONE = 1 # Variable c_int
ERROR_MEMBER_NOT_IN_GROUP = 1321 # Variable c_long
UDN_LAST = 4294966556 # Variable c_uint
URLACTION_ACTIVEX_OVERRIDE_SCRIPT_SAFETY = 4611 # Variable c_int
XTYP_ADVSTART = 4144 # Variable c_int
SMTO_BLOCK = 1 # Variable c_int

_SRVINFO = c_int # enum
RESOURCETYPE_ANY = 0 # Variable c_int
LVIF_PARAM = 4 # Variable c_int
psh10 = 1033 # Variable c_int
CRYPT_E_EXISTS = -2146885627 # Variable c_long
TBM_SETTIPSIDE = 1055 # Variable c_int
TB_GETDISABLEDIMAGELIST = 1079 # Variable c_int
JOB_NOTIFY_FIELD_SUBMITTED = 16 # Variable c_int
PERF_COUNTER_VALUE = 0 # Variable c_int
SERVICE_USER_DEFINED_CONTROL = 256 # Variable c_int
JOB_STATUS_DELETED = 256 # Variable c_int
IME_REGWORD_STYLE_USER_FIRST = 2147483648 # Variable c_uint
MKF_MOUSEKEYSON = 1 # Variable c_int
GGL_LEVEL = 1 # Variable c_int
ERROR_BAD_EXE_FORMAT = 193 # Variable c_long
SUBLANG_ENGLISH_UK = 2 # Variable c_int
BSM_VXDS = 1 # Variable c_int
NCBCHAINSEND = 23 # Variable c_int
LVS_EX_FULLROWSELECT = 32 # Variable c_int
PST_RS232 = 1 # Variable c_ulong
WVR_ALIGNTOP = 16 # Variable c_int
PAN_STRAIGHT_ARMS_VERT = 4 # Variable c_int
LANG_TATAR = 68 # Variable c_int
ERROR_LB_WITHOUT_TABSTOPS = 1434 # Variable c_long

tagLOCKTYPE = c_int # enum
PERF_COUNTER_ELAPSED = 262144 # Variable c_int
CBEM_SETEXTENDEDSTYLE = 1038 # Variable c_int
PSN_GETOBJECT = 4294967086 # Variable c_uint
MCI_VD_PLAY_SPEED = 262144 # Variable c_long
SZDDESYS_ITEM_STATUS = 'Status' # Variable POINTER(c_wchar)
EMR_GDICOMMENT = 70 # Variable c_int
FR_FINDNEXT = 8 # Variable c_int
CERT_E_REVOCATION_FAILURE = -2146762482 # Variable c_long
IMAGE_REL_PPC_TOCDEFN = 2048 # Variable c_int
ERROR_INVALID_PROFILE = 2301 # Variable c_long
ERROR_REQUEST_ABORTED = 1235 # Variable c_long
OUT_OUTLINE_PRECIS = 8 # Variable c_int
ERROR_UNRECOGNIZED_MEDIA = 1785 # Variable c_long
PERF_COUNTER_QUEUELEN = 327680 # Variable c_int
APPCLASS_STANDARD = 0 # Variable c_long
PCF_16BITMODE = 512 # Variable c_ulong
ERROR_UNKNOWN_REVISION = 1305 # Variable c_long
TYPE_E_UNDEFINEDTYPE = -2147319769 # Variable c_long
META_EXTTEXTOUT = 2610 # Variable c_int
CAL_SABBREVDAYNAME6 = 19 # Variable c_int
ERROR_INVALID_PRINTER_NAME = 1801 # Variable c_long
SEM_NOGPFAULTERRORBOX = 2 # Variable c_int
RPC_S_UNKNOWN_IF = 1717 # Variable c_long
THREAD_SET_INFORMATION = 32 # Variable c_int
DL_MOVECURSOR = 3 # Variable c_int
SE_SACL_AUTO_INHERITED = 2048 # Variable c_int
ACM_STOP = 1126 # Variable c_int
MM_DRVM_DATA = 978 # Variable c_int
CDERR_LOADRESFAILURE = 7 # Variable c_int
HEAP_GENERATE_EXCEPTIONS = 4 # Variable c_int
LOCALE_S1159 = 40 # Variable c_int
ERROR_NOT_EMPTY = 4307 # Variable c_long
CO_E_ACNOTINITIALIZED = -2147220965 # Variable c_long
DMLERR_EXECACKTIMEOUT = 16389 # Variable c_int
CBEIF_SELECTEDIMAGE = 4 # Variable c_int
FILE_NOTIFY_CHANGE_SECURITY = 256 # Variable c_int
JOB_OBJECT_ASSIGN_PROCESS = 1 # Variable c_int
MB_PRECOMPOSED = 1 # Variable c_int
LANG_MACEDONIAN = 47 # Variable c_int
ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY = 1922 # Variable c_long
SO_CONNOPT = 28673 # Variable c_int
PM_NOREMOVE = 0 # Variable c_int
S_SERDMD = -10 # Variable c_int
PD_ENABLESETUPTEMPLATE = 32768 # Variable c_int
IPPROTO_MAX = 256 # Variable c_int
SUBLANG_SPANISH_PANAMA = 6 # Variable c_int
TAPE_REWIND = 0 # Variable c_long
AF_INET = 2 # Variable c_int
LVM_GETHEADER = 4127 # Variable c_int
OLEIVERB_UIACTIVATE = -4 # Variable c_long
MCM_GETFIRSTDAYOFWEEK = 4112 # Variable c_int
edt7 = 1158 # Variable c_int
LVS_SORTDESCENDING = 32 # Variable c_int
MK_LBUTTON = 1 # Variable c_int
RPC_S_INVALID_STRING_UUID = 1705 # Variable c_long
BACKUP_EA_DATA = 2 # Variable c_int
DCB_ACCUMULATE = 2 # Variable c_int
ERROR_INVALID_FLAG_NUMBER = 186 # Variable c_long
lst4 = 1123 # Variable c_int
MM_STREAM_CLOSE = 981 # Variable c_int
RT_MENU = 4 # Variable POINTER(c_wchar)
RPC_X_SS_CONTEXT_DAMAGED = 1777 # Variable c_long
MCM_SETTODAY = 4108 # Variable c_int
IMAGE_REL_IA64_RESERVED_14 = 20 # Variable c_int
LOCALE_SMONTHNAME9 = 64 # Variable c_int
MCHT_CALENDARWEEKNUM = 131075 # Variable c_int
CRYPT_E_NOT_FOUND = -2146885628 # Variable c_long
BSF_IGNORECURRENTTASK = 2 # Variable c_int
IMN_SETCANDIDATEPOS = 9 # Variable c_int
IPPORT_FTP = 21 # Variable c_int
SPLDS_PRINT_LANGUAGE = 'printLanguage' # Variable POINTER(c_wchar)
SCS_DOS_BINARY = 1 # Variable c_int
HDN_LAST = 4294966897 # Variable c_uint
PSINJECT_BEGINSETUP = 7 # Variable c_int
TIME_CALLBACK_FUNCTION = 0 # Variable c_int
LOCALE_SMONTHNAME13 = 4110 # Variable c_int
SPI_GETICONMETRICS = 45 # Variable c_int
OLEIVERB_OPEN = -2 # Variable c_long
TMPF_DEVICE = 8 # Variable c_int
DDL_SYSTEM = 4 # Variable c_int
DMPAPER_TABLOID = 3 # Variable c_int
CFSTR_MIME_RAWDATASTRM = 'application/octet-stream' # Variable POINTER(c_wchar)
CTRY_URUGUAY = 598 # Variable c_int
USER_CALL_IS_ASYNC = 256 # Variable c_int
TARGET_IS_NT351_OR_WIN95_OR_LATER = 1 # Variable c_int
ERROR_CANT_EVICT_ACTIVE_NODE = 5009 # Variable c_long
INET_E_INVALID_URL = -2146697214 # Variable c_long
FF_SCRIPT = 64 # Variable c_int
JOB_OBJECT_SECURITY_VALID_FLAGS = 15 # Variable c_int
BS_BOTTOM = 2048 # Variable c_long
STGM_READ = 0 # Variable c_long
RT_GROUP_ICON = 20 # Variable POINTER(c_wchar)
SM_CYBORDER = 6 # Variable c_int
CO_E_RUNAS_SYNTAX = -2147467241 # Variable c_long
SWP_HIDEWINDOW = 128 # Variable c_int
OFN_NOCHANGEDIR = 8 # Variable c_int
VIETNAMESE_CHARSET = 163 # Variable c_int
RPC_C_BINDING_MAX_TIMEOUT = 9 # Variable c_int
SACL_SECURITY_INFORMATION = 8 # Variable c_long
TVM_SORTCHILDRENCB = 4373 # Variable c_int
CBN_DROPDOWN = 7 # Variable c_int
ILD_IMAGE = 32 # Variable c_int
ICC_NATIVEFNTCTL_CLASS = 8192 # Variable c_int
ERROR_INVALID_USER_BUFFER = 1784 # Variable c_long
BATTERY_FLAG_LOW = 2 # Variable c_int
MIXERCONTROL_CONTROLTYPE_FADER = 1342373888 # Variable c_long

tagOLEUPDATE = c_int # enum
REGDB_E_INVALIDVALUE = -2147221165 # Variable c_long
TVIF_TEXT = 1 # Variable c_int
LR_DEFAULTSIZE = 64 # Variable c_int
IE_BADID = -1 # Variable c_int
ERROR_LOCK_VIOLATION = 33 # Variable c_long
RPC_BUFFER_PARTIAL = 8192 # Variable c_int
MCI_WAVE_OUTPUT = 8388608 # Variable c_long
ERROR_SEVERITY_WARNING = 2147483648 # Variable c_uint
ERROR_BAD_LOGON_SESSION_STATE = 1365 # Variable c_long
SECURITY_DELEGATION = 196608 # Variable c_int
FW_EXTRALIGHT = 200 # Variable c_int
TCM_SETCURFOCUS = 4912 # Variable c_int
SB_LINEUP = 0 # Variable c_int
SCS_QUERYRECONVERTSTRING = 131072 # Variable c_int
HCF_HIGHCONTRASTON = 1 # Variable c_int
ETO_NUMERICSLATIN = 2048 # Variable c_int
ERROR_CANCEL_VIOLATION = 173 # Variable c_long
DMDITHER_RESERVED7 = 7 # Variable c_int
LVM_DELETECOLUMN = 4124 # Variable c_int
MCI_WAVE_INPUT = 4194304 # Variable c_long
IMAGE_ARCHIVE_LONGNAMES_MEMBER = '//              ' # Variable POINTER(c_char)
C2_OTHERNEUTRAL = 11 # Variable c_int
DCTT_DOWNLOAD = 2 # Variable c_long
CC_INTERIORS = 128 # Variable c_int
IMAGE_SIZEOF_FILE_HEADER = 20 # Variable c_int
AF_FIREFOX = 19 # Variable c_int
LVM_HITTEST = 4114 # Variable c_int
CONNECT_CURRENT_MEDIA = 512 # Variable c_int
CF_PRIVATELAST = 767 # Variable c_int
SETRGBSTRINGW = 'commdlg_SetRGBColor' # Variable POINTER(c_wchar)
ERROR_FILEMARK_DETECTED = 1101 # Variable c_long
SM_MENUDROPALIGNMENT = 40 # Variable c_int
LVM_GETCALLBACKMASK = 4106 # Variable c_int
MKF_HOTKEYSOUND = 16 # Variable c_int
CO_E_FAILEDTOGETTOKENINFO = -2147220989 # Variable c_long
CF_LIMITSIZE = 8192 # Variable c_long
QUERYDIBSUPPORT = 3073 # Variable c_int
IGP_UI = 16 # Variable c_int
DM_PROMPT = 4 # Variable c_int
TAPE_ABSOLUTE_POSITION = 0 # Variable c_long
MM_ANISOTROPIC = 8 # Variable c_int
MB_COMPOSITE = 2 # Variable c_int
PERF_SIZE_ZERO = 512 # Variable c_int
XTYP_EXECUTE = 16464 # Variable c_int
PD_NOSELECTION = 4 # Variable c_int
lst12 = 1131 # Variable c_int
MMSYSERR_BASE = 0 # Variable c_int
ERROR_SOME_NOT_MAPPED = 1301 # Variable c_long
GRAY_BRUSH = 2 # Variable c_int
WSAEPROCLIM = 10067 # Variable c_int
HSHELL_WINDOWCREATED = 1 # Variable c_int
EXIT_PROCESS_DEBUG_EVENT = 5 # Variable c_int
CBEIF_OVERLAY = 8 # Variable c_int
TCHT_ONITEMICON = 2 # Variable c_int
MCI_FREEZE = 2116 # Variable c_int
CTRY_ALBANIA = 355 # Variable c_int
REG_FULL_RESOURCE_DESCRIPTOR = 9 # Variable c_int
SPI_GETBORDER = 5 # Variable c_int
TB_SETSTYLE = 1080 # Variable c_int
ERROR_SEVERITY_ERROR = 3221225472 # Variable c_uint
SE_BACKUP_NAME = 'SeBackupPrivilege' # Variable POINTER(c_wchar)
SWP_SHOWWINDOW = 64 # Variable c_int
SMTO_ABORTIFHUNG = 2 # Variable c_int
SMTO_NOTIMEOUTIFNOTHUNG  = 0x0008
POLYGONALCAPS = 32 # Variable c_int
TCN_LAST = 4294966716 # Variable c_uint
MCI_SYSINFO = 2064 # Variable c_int
IDC_APPSTARTING = 206416 # Variable POINTER(c_wchar)
SEVERITY_ERROR = 1 # Variable c_int
MIXER_OBJECTF_WAVEOUT = 268435456 # Variable c_long
MCHT_CALENDAR = 131072 # Variable c_int
FKF_AVAILABLE = 2 # Variable c_int
TBS_HORZ = 0 # Variable c_int
ILD_BLEND50 = 4 # Variable c_int
PBM_SETSTEP = 1028 # Variable c_int
CBEM_GETIMAGELIST = 1027 # Variable c_int
WNNC_NET_LIFENET = 917504 # Variable c_int
HIST_BACK = 0 # Variable c_int
FILE_NOTIFY_CHANGE_FILE_NAME = 1 # Variable c_int
CBR_256000 = 256000 # Variable c_int
IPPORT_MTP = 57 # Variable c_int
REGDB_E_IIDNOTREG = -2147221163 # Variable c_long
CLIENTSITE_S_FIRST = 262544 # Variable c_long
SPAPI_E_INVALID_REFERENCE_STRING = -2146500065 # Variable c_long
XTYP_ADVSTOP = 32832 # Variable c_int
IMAGE_SCN_MEM_EXECUTE = 536870912 # Variable c_int
LVS_EX_UNDERLINECOLD = 4096 # Variable c_int
MDMVOL_LOW = 0 # Variable c_int
CDS_SET_PRIMARY = 16 # Variable c_int
LOCALE_SABBREVMONTHNAME2 = 69 # Variable c_int
MK_E_NOTBINDABLE = -2147221016 # Variable c_long
JOB_OBJECT_SECURITY_ONLY_TOKEN = 4 # Variable c_int
ERROR_GRACEFUL_DISCONNECT = 1226 # Variable c_long
STREAM_CONTAINS_SECURITY = 2 # Variable c_int
ERROR_SERVICE_DATABASE_LOCKED = 1055 # Variable c_long
STG_E_INVALIDPARAMETER = -2147286953 # Variable c_long
CBF_FAIL_ALLSVRXACTIONS = 258048 # Variable c_int
ERROR_META_EXPANSION_TOO_LONG = 208 # Variable c_long
CO_E_APPNOTFOUND = -2147221003 # Variable c_long
FILE_ALL_ACCESS = 2032639 # Variable c_long
WPF_SETMINPOSITION = 1 # Variable c_int
DMDITHER_ERRORDIFFUSION = 5 # Variable c_int
CF_DSPTEXT = 129 # Variable c_int
RPC_C_EP_MATCH_BY_BOTH = 3 # Variable c_int
LOCALE_SMONTHOUSANDSEP = 23 # Variable c_int
OPENCHANNEL = 4110 # Variable c_int
ERROR_INSTALL_SUSPEND = 1604 # Variable c_long
LOAD_WITH_ALTERED_SEARCH_PATH = 8 # Variable c_int
FACILITY_NT_BIT = 268435456 # Variable c_int
LF_FACESIZE = 32 # Variable c_int
ERROR_PRIVILEGE_NOT_HELD = 1314 # Variable c_long
LVM_GETSTRINGWIDTHA = 4113 # Variable c_int
ST_TERMINATED = 32 # Variable c_int
LB_FINDSTRINGEXACT = 418 # Variable c_int
SM_CXEDGE = 45 # Variable c_int
WSB_PROP_HBKGCOLOR = 128 # Variable c_long
ERROR_ACCOUNT_DISABLED = 1331 # Variable c_long
SUBLANG_ENGLISH_CARIBBEAN = 9 # Variable c_int
TAPE_DRIVE_FIXED_BLOCK = 1024 # Variable c_int
SPLDS_PRINTER_MODEL = 'printerModel' # Variable POINTER(c_wchar)
CONNECT_UPDATE_PROFILE = 1 # Variable c_int
FILE_ATTRIBUTE_SYSTEM = 4 # Variable c_int
PAN_CONTRAST_VERY_LOW = 3 # Variable c_int
SECURITY_PROXY_RID = 8 # Variable c_long
SERVICE_STOPPED = 1 # Variable c_int
IMAGE_ENHMETAFILE = 3 # Variable c_int
IN_CLASSA_MAX = 128 # Variable c_int
TVN_GETDISPINFOW = 4294966844 # Variable c_uint
DLL_THREAD_DETACH = 3 # Variable c_int
IS_TEXT_UNICODE_STATISTICS = 2 # Variable c_int
ID_PSRESTARTWINDOWS = 2 # Variable c_int
TB_ADDBUTTONSW = 1092 # Variable c_int
GCPCLASS_LOCALNUMBER = 4 # Variable c_int
SPAPI_E_GENERAL_SYNTAX = -2146500605 # Variable c_long
MB_SYSTEMMODAL = 4096 # Variable c_long
SERVER_ACCESS_ENUMERATE = 2 # Variable c_int
ACCESS_ALLOWED_OBJECT_ACE_TYPE = 5 # Variable c_int
CF_APPLY = 512 # Variable c_long
PAN_STROKE_RAPID_HORZ = 7 # Variable c_int
FIXED_PITCH = 1 # Variable c_int
CRYPT_E_NOT_CHAR_STRING = -2146885596 # Variable c_long
DMBIN_USER = 256 # Variable c_int
GCPCLASS_LATINNUMERICSEPARATOR = 7 # Variable c_int
IMAGE_SCN_MEM_PURGEABLE = 131072 # Variable c_int
MM_MIM_MOREDATA = 972 # Variable c_int
PD_USEDEVMODECOPIESANDCOLLATE = 262144 # Variable c_int
PROCESS_VM_READ = 16 # Variable c_int
ERROR_INVALID_AT_INTERRUPT_TIME = 104 # Variable c_long
TCM_DESELECTALL = 4914 # Variable c_int
MCI_ANIM_WINDOW_ENABLE_STRETCH = 1048576 # Variable c_long
SPLDS_USER_KEY = 'DsUser' # Variable POINTER(c_wchar)
STGM_NOSCRATCH = 1048576 # Variable c_long
DISPID_EVALUATE = -5 # Variable c_int
LVM_INSERTITEMW = 4173 # Variable c_int
MCI_OVLY_OPEN_WS = 65536 # Variable c_long
SPLDS_DESCRIPTION = 'description' # Variable POINTER(c_wchar)
ERROR_SEM_NOT_FOUND = 187 # Variable c_long
SECURITY_STATIC_TRACKING = 0 # Variable c_int
CDDS_SUBITEM = 131072 # Variable c_int
ERROR_DOMAIN_LIMIT_EXCEEDED = 1357 # Variable c_long
TBM_GETTICPOS = 1039 # Variable c_int
RESOURCEDISPLAYTYPE_ROOT = 7 # Variable c_int
WSAENOTCONN = 10057 # Variable c_int
MCI_WAVE_OFFSET = 1152 # Variable c_int
CBEM_GETCOMBOCONTROL = 1030 # Variable c_int
MIXER_GETLINECONTROLSF_ALL = 0 # Variable c_long
MDMSPKRFLAG_CALLSETUP = 8 # Variable c_int
ODA_DRAWENTIRE = 1 # Variable c_int
WSB_PROP_CYVTHUMB = 32 # Variable c_long
RPC_E_SERVERCALL_REJECTED = -2147417845 # Variable c_long
MCIERR_SEQ_PORT_NONEXISTENT = 338 # Variable c_int
SC_GROUP_IDENTIFIERA = '+' # Variable c_char
NTE_BAD_VER = -2146893817 # Variable c_long
PAN_CONTRAST_HIGH = 8 # Variable c_int
DEVICEDATA = 19 # Variable c_int
ERROR_NETWORK_BUSY = 54 # Variable c_long
DTN_FORMATQUERYA = 4294966541 # Variable c_uint
MAXUINT_PTR = 4294967295 # Variable c_ulong
IPPORT_WHOIS = 43 # Variable c_int
IMAGE_REL_PPC_SECRELHI = 20 # Variable c_int
ERROR_INVALID_SEGDPL = 198 # Variable c_long
ERROR_FLOPPY_BAD_REGISTERS = 1125 # Variable c_long

_CM_SERVICE_NODE_TYPE = c_int # enum
DriverType = 1
CF_MAX = 17 # Variable c_int
TBN_FIRST = 4294966596 # Variable c_uint
MCI_OPEN_ELEMENT_ID = 2048 # Variable c_long
GDTR_MAX = 2 # Variable c_int
IME_CHOTKEY_IME_NONIME_TOGGLE = 16 # Variable c_int
SW_NORMAL = 1 # Variable c_int
HWND_BOTTOM = 1 # Variable POINTER(HWND__)
JOB_NOTIFY_FIELD_DOCUMENT = 13 # Variable c_int
TCM_ADJUSTRECT = 4904 # Variable c_int
STARTDOC = 10 # Variable c_int
RPC_E_RETRY = -2147417847 # Variable c_long
JOB_CONTROL_PAUSE = 1 # Variable c_int
LVS_NOLABELWRAP = 128 # Variable c_int
CDRF_NOTIFYITEMDRAW = 32 # Variable c_int
PRF_CHECKVISIBLE = 1 # Variable c_long
PRINTER_NOTIFY_OPTIONS_REFRESH = 1 # Variable c_int
RPC_C_AUTHN_NONE = 0 # Variable c_int
WS_CAPTION = 12582912 # Variable c_long
PAN_PROP_MONOSPACED = 9 # Variable c_int
LANG_URDU = 32 # Variable c_int
UDM_GETBASE = 1134 # Variable c_int

tagURLTEMPLATE = c_int # enum
URLTEMPLATE_CUSTOM = 0
LOCALE_SISO3166CTRYNAME = 90 # Variable c_int
IDB_STD_LARGE_COLOR = 1 # Variable c_int
MMIOERR_CANNOTSEEK = 263 # Variable c_int
RB_SETBKCOLOR = 1043 # Variable c_int
ERROR_PASSWORD_RESTRICTION = 1325 # Variable c_long
IPN_FIELDCHANGED = 4294966436 # Variable c_uint
META_STRETCHDIB = 3907 # Variable c_int
ERROR_HANDLE_EOF = 38 # Variable c_long
SET_SCREEN_ANGLE = 4105 # Variable c_int
NRC_ILLNN = 19 # Variable c_int
CE_RXOVER = 1 # Variable c_int
psh11 = 1034 # Variable c_int
TAPE_DRIVE_WRITE_SETMARKS = 2164260864 # Variable c_uint
GW_MAX = 5 # Variable c_int
SETICMPROFILE_EMBEDED = 1 # Variable c_int
GCL_HICON = -14 # Variable c_int
CMB_MASKED = 2 # Variable c_int
ERROR_NOT_A_REPARSE_POINT = 4390 # Variable c_long
RBBIM_CHILDSIZE = 32 # Variable c_int
MEVT_VERSION = 132 # Variable c_ubyte
ERROR_NET_WRITE_FAULT = 88 # Variable c_long
OBJ_PEN = 1 # Variable c_int
IDB_HIST_LARGE_COLOR = 9 # Variable c_int
PS_ENDCAP_MASK = 3840 # Variable c_int
META_SAVEDC = 30 # Variable c_int
DD_DEFDRAGMINDIST = 2 # Variable c_int
PAN_WEIGHT_THIN = 4 # Variable c_int
SEF_DEFAULT_DESCRIPTOR_FOR_OBJECT = 4 # Variable c_int
ERROR_CAN_NOT_DEL_LOCAL_WINS = 4001 # Variable c_long
SPLDS_PORT_NAME = 'portName' # Variable POINTER(c_wchar)
DATABITS_16 = 16 # Variable c_ushort
SPAPI_E_NO_DEVICE_SELECTED = -2146500079 # Variable c_long
ERROR_OUTOFMEMORY = 14 # Variable c_long
WSAENETUNREACH = 10051 # Variable c_int
MAX_COMPUTERNAME_LENGTH = 15 # Variable c_int
SRCINVERT = 6684742 # Variable c_ulong
FILE_TYPE_PIPE = 3 # Variable c_int
IMAGE_REL_MIPS_SECREL = 11 # Variable c_int
LOCALE_SABBREVDAYNAME6 = 54 # Variable c_int
MCI_ANIM_PLAY_REVERSE = 131072 # Variable c_long
PSBTN_NEXT = 1 # Variable c_int
SB_PAGELEFT = 2 # Variable c_int
IME_CAND_UNKNOWN = 0 # Variable c_int
PERF_COUNTER_LARGE_RAWCOUNT_HEX = 256 # Variable c_int
SORT_CHINESE_PRCP = 0 # Variable c_int
CBR_300 = 300 # Variable c_int
SYNCHRONIZE = 1048576 # Variable c_long
ERROR_EXCL_SEM_ALREADY_OWNED = 101 # Variable c_long
AF_VOICEVIEW = 18 # Variable c_int
PC_WINDPOLYGON = 4 # Variable c_int
MSGF_SCROLLBAR = 5 # Variable c_int
URLACTION_SHELL_INSTALL_DTITEMS = 6144 # Variable c_int
DM_BITSPERPEL = 262144 # Variable c_long
CAL_SABBREVDAYNAME5 = 18 # Variable c_int
PAN_MIDLINE_LOW_SERIFED = 13 # Variable c_int
CO_E_LAST = 2147746303 # Variable c_ulong
NTE_BAD_LEN = -2146893820 # Variable c_long
IMM_ERROR_GENERAL = -2 # Variable c_int
ERROR_INVALID_WINDOW_STYLE = 2002 # Variable c_long
MMIOERR_ACCESSDENIED = 268 # Variable c_int
FORMAT_MESSAGE_IGNORE_INSERTS = 512 # Variable c_int
MAX_PERF_OBJECTS_IN_QUERY_FUNCTION = 8 # Variable c_long
REG_MULTI_SZ = 7 # Variable c_int
HDI_FORMAT = 4 # Variable c_int
STG_E_SHAREREQUIRED = -2147286778 # Variable c_long
MK_S_LAST = 262639 # Variable c_long
DISPID_UNKNOWN = -1 # Variable c_int
GETPHYSPAGESIZE = 12 # Variable c_int
TBM_SETTIC = 1028 # Variable c_int
SS_ETCHEDHORZ = 16 # Variable c_long
ERROR_STACK_OVERFLOW = 1001 # Variable c_long
JOY_CAL_READVONLY = 134217728 # Variable c_long
SC_MONITORPOWER = 61808 # Variable c_int
CO_E_SCM_RPC_FAILURE = -2146959357 # Variable c_long
TVGN_CARET = 9 # Variable c_int
WSAEDISCON = 10101 # Variable c_int
lst10 = 1129 # Variable c_int
ASPECTXY = 44 # Variable c_int
PT_CLOSEFIGURE = 1 # Variable c_int
TVM_SETIMAGELIST = 4361 # Variable c_int
NFS_STATIC = 2 # Variable c_int
RIGHTMOST_BUTTON_PRESSED = 2 # Variable c_int
edt1 = 1152 # Variable c_int
UDM_SETRANGE = 1125 # Variable c_int
JOB_OBJECT_MSG_END_OF_PROCESS_TIME = 2 # Variable c_int
USER_MARSHAL_FC_WCHAR = 5 # Variable c_int
DMPAPER_LEGAL = 5 # Variable c_int
MCI_ANIM_STATUS_STRETCH = 16389 # Variable c_long
TCM_SETTOOLTIPS = 4910 # Variable c_int
lst5 = 1124 # Variable c_int
LANG_INDONESIAN = 33 # Variable c_int
MIXERCONTROL_CT_CLASS_LIST = 1879048192 # Variable c_long
IDANI_CLOSE = 2 # Variable c_int
ERROR_PRINT_CANCELLED = 63 # Variable c_long
MMIOM_SEEK = 2 # Variable c_int
IMAGE_REL_IA64_RESERVED_15 = 21 # Variable c_int
SORT_CHINESE_PRC = 2 # Variable c_int
APD_COPY_ALL_FILES = 4 # Variable c_int
ERROR_INVALID_LEVEL = 124 # Variable c_long
_MAX_EXT = 256 # Variable c_int
TBBF_LARGE = 1 # Variable c_int
LANG_BENGALI = 69 # Variable c_int
NCBHANGUP = 18 # Variable c_int
JOY_CAL_READUONLY = 67108864 # Variable c_long
ERROR_UNABLE_TO_LOCK_MEDIA = 1108 # Variable c_long
IDC_IBEAM = 206099 # Variable POINTER(c_wchar)
SESSION_ESTABLISHED = 3 # Variable c_int
CFSTR_MIME_BMP = 'image/bmp' # Variable POINTER(c_wchar)
IMAGE_FILE_DLL = 8192 # Variable c_int
ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND = 5032 # Variable c_long
CBR_110 = 110 # Variable c_int
DDE_FDEFERUPD = 16384 # Variable c_int
CFS_DEFAULT = 0 # Variable c_int
FILE_ACTION_ADDED = 1 # Variable c_int
EM_SETWORDBREAKPROC = 208 # Variable c_int
CDERR_LOADSTRFAILURE = 5 # Variable c_int
WINSTA_WRITEATTRIBUTES = 16 # Variable c_long
grp2 = 1073 # Variable c_int
RPC_E_INVALID_PARAMETER = -2147418096 # Variable c_long
LVM_SETITEMTEXTW = 4212 # Variable c_int
ERROR_DISK_RESET_FAILED = 1128 # Variable c_long
DFCS_SCROLLRIGHT = 3 # Variable c_int
DMPAPER_11X17 = 17 # Variable c_int
CUDR_NOSNAPTOGRID = 1 # Variable c_int
MIXERCONTROL_CT_CLASS_TIME = 1610612736 # Variable c_long
MCI_GETDEVCAPS_ITEM = 256 # Variable c_long
SPI_GETSCREENREADER = 70 # Variable c_int
TVGN_PARENT = 3 # Variable c_int
IMAGE_DIRECTORY_ENTRY_BASERELOC = 5 # Variable c_int
SHIFTJIS_CHARSET = 128 # Variable c_int
MIXERCONTROL_CONTROLTYPE_LOUDNESS = 536936452 # Variable c_long
FILE_LIST_DIRECTORY = 1 # Variable c_int
URLACTION_DOWNLOAD_MIN = 4096 # Variable c_int
PAN_XHEIGHT_DUCKING_STD = 6 # Variable c_int
SUBLANG_ENGLISH_JAMAICA = 8 # Variable c_int
ERROR_BOOT_ALREADY_ACCEPTED = 1076 # Variable c_long
ILLUMINANT_C = 3 # Variable c_int
DMPAPER_ENV_14 = 23 # Variable c_int
DRIVERVERSION = 0 # Variable c_int
SE_CREATE_TOKEN_NAME = 'SeCreateTokenPrivilege' # Variable POINTER(c_wchar)
MF_SENDMSGS = 33554432 # Variable c_int
CLASS_E_NOAGGREGATION = -2147221232 # Variable c_long
IMAGE_REL_PPC_ADDR14 = 5 # Variable c_int
MCIERR_NULL_PARAMETER_BLOCK = 297 # Variable c_int
RPC_C_MQ_AUTHN_LEVEL_PKT_PRIVACY = 16 # Variable c_int
NOPARITY = 0 # Variable c_int
CO_E_ACCESSCHECKFAILED = -2147220985 # Variable c_long
TB_GETSTATE = 1042 # Variable c_int
BM_GETIMAGE = 246 # Variable c_int
SE_INC_BASE_PRIORITY_NAME = 'SeIncreaseBasePriorityPrivilege' # Variable POINTER(c_wchar)
XST_CONNECTED = 2 # Variable c_int
TTF_DI_SETITEM = 32768 # Variable c_int
TBS_ENABLESELRANGE = 32 # Variable c_int
PAN_WEIGHT_BOLD = 8 # Variable c_int
STATUS_FLOAT_INEXACT_RESULT = 3221225615 # Variable c_ulong
DMDITHER_RESERVED6 = 6 # Variable c_int
MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER = 4103 # Variable c_long
SC_MAXIMIZE = 61488 # Variable c_int
TB_GETBUTTON = 1047 # Variable c_int
PST_SCANNER = 34 # Variable c_ulong
MAXIMUM_PROCESSORS = 32 # Variable c_int
RPC_S_SEC_PKG_ERROR = 1825 # Variable c_long
MK_E_NOPREFIX = -2147221010 # Variable c_long
DEBUG_ONLY_THIS_PROCESS = 2 # Variable c_int
ATF_TIMEOUTON = 1 # Variable c_int
WH_MSGFILTER = -1 # Variable c_int
IME_THOTKEY_SYMBOL_TOGGLE = 114 # Variable c_int
DOMAIN_GROUP_RID_SCHEMA_ADMINS = 518 # Variable c_long
DI_COMPAT = 4 # Variable c_int
WT_EXECUTEDEFAULT = 0 # Variable c_int
IME_ESC_PRIVATE_HOTKEY = 4106 # Variable c_int
SPLDS_PRINT_SPOOLING = 'printSpooling' # Variable POINTER(c_wchar)
IMAGE_DEBUG_TYPE_COFF = 1 # Variable c_int
LOCALE_SABBREVCTRYNAME = 7 # Variable c_int
IOC_OUT = 1073741824 # Variable c_int
SIMPLEREGION = 2 # Variable c_int
PURGE_RXCLEAR = 8 # Variable c_int
SORT_HUNGARIAN_TECHNICAL = 1 # Variable c_int
OBJ_DC = 3 # Variable c_int
PSH_USECALLBACK = 256 # Variable c_int
WVR_ALIGNRIGHT = 128 # Variable c_int
MMSYSERR_BADERRNUM = 9 # Variable c_int
HTGROWBOX = 4 # Variable c_int
DC_PAPERS = 2 # Variable c_int
MIXER_GETLINEINFOF_TARGETTYPE = 4 # Variable c_long
lst13 = 1132 # Variable c_int
PS_ALTERNATE = 8 # Variable c_int
CTRY_TAIWAN = 886 # Variable c_int
DEFAULT_QUALITY = 0 # Variable c_int
RT_BITMAP = 2 # Variable POINTER(c_wchar)
EMR_FRAMERGN = 72 # Variable c_int
SO_ACCEPTCONN = 2 # Variable c_int
IDI_WINLOGO = 206103 # Variable POINTER(c_wchar)
DMBIN_FORMSOURCE = 15 # Variable c_int
MCI_PAUSE = 2057 # Variable c_int
IMAGE_FILE_MACHINE_MIPS16 = 614 # Variable c_int
PC_TRAPEZOID = 4 # Variable c_int
LANG_AFRIKAANS = 54 # Variable c_int
IME_CMODE_FULLSHAPE = 8 # Variable c_int
REBARBANDINFOA_V3_SIZE = 56 # Variable c_uint
CTRY_MONACO = 33 # Variable c_int
LEFT_ALT_PRESSED = 2 # Variable c_int
DDD_NO_BROADCAST_SYSTEM = 8 # Variable c_int
SZDDESYS_TOPIC = 'System' # Variable POINTER(c_wchar)
STG_E_INCOMPLETE = -2147286527 # Variable c_long
FILE_FLAG_OPEN_NO_RECALL = 1048576 # Variable c_int
POSTSCRIPT_PASSTHROUGH = 4115 # Variable c_int
FOCUS_EVENT = 16 # Variable c_int
TARGET_IS_NT40_OR_LATER = 0 # Variable c_int
BATTERY_FLAG_NO_BATTERY = 128 # Variable c_int
SO_DONTROUTE = 16 # Variable c_int
TCS_OWNERDRAWFIXED = 8192 # Variable c_int
RPC_S_INVALID_AUTH_IDENTITY = 1749 # Variable c_long
SBM_SETPOS = 224 # Variable c_int
MCI_ANIM_PLAY_SCAN = 1048576 # Variable c_long
MF_ENABLED = 0 # Variable c_long
WAVE_FORMAT_2S16 = 128 # Variable c_int
frm2 = 1077 # Variable c_int
IDH_NO_HELP = 28440 # Variable c_int
IMAGE_SYM_DTYPE_POINTER = 1 # Variable c_int
ALTERNATE = 1 # Variable c_int
CTRY_IRAQ = 964 # Variable c_int
TC_UA_ABLE = 2048 # Variable c_int
LOCALE_SABBREVMONTHNAME3 = 70 # Variable c_int
NOTSRCCOPY = 3342344 # Variable c_ulong
IME_CAND_READ = 1 # Variable c_int
TBSTATE_ENABLED = 4 # Variable c_int
C2_EUROPENUMBER = 3 # Variable c_int
JOY_BUTTON15 = 16384 # Variable c_long
WAVECAPS_SAMPLEACCURATE = 32 # Variable c_int
GL_LEVEL_INFORMATION = 4 # Variable c_int
MA_ACTIVATE = 1 # Variable c_int
META_PAINTREGION = 299 # Variable c_int
BACKUP_REPARSE_DATA = 8 # Variable c_int
LC_WIDE = 16 # Variable c_int
ERROR_NOT_QUORUM_CLASS = 5025 # Variable c_long
PS_INSIDEFRAME = 6 # Variable c_int
META_REALIZEPALETTE = 53 # Variable c_int
HSHELL_TASKMAN = 7 # Variable c_int
SPAPI_E_NO_CLASSINSTALL_PARAMS = -2146500075 # Variable c_long
MCI_WAVE_OPEN_BUFFER = 65536 # Variable c_long
IMAGE_DEBUG_TYPE_MISC = 4 # Variable c_int
MCIERR_CREATEWINDOW = 347 # Variable c_int
rad1 = 1056 # Variable c_int
SP_USERABORT = -3 # Variable c_int
DEFAULT_CHARSET = 1 # Variable c_int
MCI_WAVE_STATUS_BLOCKALIGN = 16389 # Variable c_long
LVM_GETSUBITEMRECT = 4152 # Variable c_int
RPC_C_PROTECT_LEVEL_NONE = 1 # Variable c_int
PROCESSOR_MIPS_R4000 = 4000 # Variable c_int
BS_AUTORADIOBUTTON = 9 # Variable c_long
CRYPT_E_INVALID_INDEX = -2146889720 # Variable c_long
TVSIL_STATE = 2 # Variable c_int
RPC_C_OPT_MQ_PRIORITY = 2 # Variable c_int
LVHT_ONITEMLABEL = 4 # Variable c_int
VIF_DIFFTYPE = 32 # Variable c_long
ERROR_DUP_DOMAINNAME = 1221 # Variable c_long
TB_SETPADDING = 1111 # Variable c_int
CF_NOSIMULATIONS = 4096 # Variable c_long
CLIENTSITE_E_LAST = 2147746207 # Variable c_ulong
SE_ERR_DDEFAIL = 29 # Variable c_int
MCI_ANIM_STEP_FRAMES = 131072 # Variable c_long
DM_SPECVERSION = 1024 # Variable c_int
CERT_E_CHAINING = -2146762486 # Variable c_long
SO_USELOOPBACK = 64 # Variable c_int
LVIS_DROPHILITED = 8 # Variable c_int
ERROR_PROFILE_NOT_FOUND = 2306 # Variable c_long
GMEM_NOTIFY = 16384 # Variable c_int
VIEW_NETCONNECT = 9 # Variable c_int
IO_REPARSE_TAG_NSS = 2147483653 # Variable c_uint
CRYPT_E_DELETED_PREV = -2146885624 # Variable c_long
RPC_C_AUTHN_LEVEL_PKT_PRIVACY = 6 # Variable c_int
DTR_CONTROL_DISABLE = 0 # Variable c_int
SPI_GETMINIMIZEDMETRICS = 43 # Variable c_int
SOL_SOCKET = 65535 # Variable c_int
NCBRECV = 21 # Variable c_int
MCI_VD_STATUS_SIDE = 16389 # Variable c_long
SPAPI_E_DI_DO_DEFAULT = -2146500082 # Variable c_long
ES_DISPLAY_REQUIRED = 2 # Variable c_ulong
EMARCH_ENC_I17_IMM41a_INST_WORD_POS_X = 14 # Variable c_int
MM_ISOTROPIC = 7 # Variable c_int
MCI_ANIM_WHERE_SOURCE = 131072 # Variable c_long
C2_BLOCKSEPARATOR = 8 # Variable c_int
CRYPT_E_NO_DECRYPT_CERT = -2146885620 # Variable c_long
CFSTR_MIME_MPEG = 'video/mpeg' # Variable POINTER(c_wchar)
SSTF_DISPLAY = 3 # Variable c_int
MMIOERR_FILENOTFOUND = 257 # Variable c_int
ERROR_RING2_STACK_IN_USE = 207 # Variable c_long
TRACKBAR_CLASSW = 'msctls_trackbar32' # Variable POINTER(c_wchar)
OLE_E_OLEVERB = -2147221504 # Variable c_long
ERROR_EXCEPTION_IN_SERVICE = 1064 # Variable c_long
ERROR_UNABLE_TO_UNLOAD_MEDIA = 1109 # Variable c_long
PERF_COUNTER_NODATA = 1073742336 # Variable c_int
RESOURCEUSAGE_NOLOCALDEVICE = 4 # Variable c_int
TB_GETMAXSIZE = 1107 # Variable c_int
LVM_EDITLABELW = 4214 # Variable c_int
GCP_JUSTIFYIN = 2097152 # Variable c_long
PROCESS_SET_SESSIONID = 4 # Variable c_int
DATA_E_LAST = 2147746111 # Variable c_ulong
BS_SOLID = 0 # Variable c_int
RPC_E_INVALID_IPID = -2147417837 # Variable c_long
NONZEROLPTR = 0 # Variable c_int
C3_KATAKANA = 16 # Variable c_int
MCI_OVLY_STATUS_HWND = 16385 # Variable c_long
ERROR_NOT_CHILD_WINDOW = 1442 # Variable c_long
PAN_PROP_VERY_EXPANDED = 7 # Variable c_int
MCI_WAVE_GETDEVCAPS_OUTPUTS = 16386 # Variable c_long
CP_MACCP = 2 # Variable c_int
MDM_ERROR_CONTROL = 2 # Variable c_int
MM_LOMETRIC = 2 # Variable c_int
RBBIM_STYLE = 1 # Variable c_int
PS_JOIN_MASK = 61440 # Variable c_int
BAUD_9600 = 2048 # Variable c_ulong
IDYES = 6 # Variable c_int
TVN_LAST = 4294966797 # Variable c_uint
IME_ESC_MAX_KEY = 4101 # Variable c_int
EMR_RESTOREDC = 34 # Variable c_int
ERROR_INDEX_ABSENT = 1611 # Variable c_long
TVS_FULLROWSELECT = 4096 # Variable c_int
TRANSPORT_TYPE_LPC = 4 # Variable c_int
CLIENTSITE_S_LAST = 262559 # Variable c_long
PFD_DOUBLEBUFFER = 1 # Variable c_int
NTE_FAIL = -2146893792 # Variable c_long
PIPE_TYPE_BYTE = 0 # Variable c_int
DOMAIN_ALIAS_RID_BACKUP_OPS = 551 # Variable c_long
CERT_E_WRONG_USAGE = -2146762480 # Variable c_long
SM_NETWORK = 63 # Variable c_int
SB_GETTEXTA = 1026 # Variable c_int
TVN_GETINFOTIPA = 4294966883 # Variable c_uint
JOY_BUTTON30 = 536870912 # Variable c_long
DMBIN_ENVMANUAL = 6 # Variable c_int
PRINTER_NOTIFY_FIELD_SECURITY_DESCRIPTOR = 12 # Variable c_int
DT_RASPRINTER = 2 # Variable c_int
RPC_C_BINDING_MIN_TIMEOUT = 0 # Variable c_int
SETRTS = 3 # Variable c_int
SET_BACKGROUND_COLOR = 4103 # Variable c_int
ARW_BOTTOMLEFT = 0 # Variable c_long
FF_ROMAN = 16 # Variable c_int
STATUS_STACK_OVERFLOW = 3221225725 # Variable c_ulong
URLACTION_INFODELIVERY_NO_ADDING_SUBSCRIPTIONS = 7427 # Variable c_int
CS_PARENTDC = 128 # Variable c_int
NRC_ILLCMD = 3 # Variable c_int
LBS_WANTKEYBOARDINPUT = 1024 # Variable c_long
MCI_SEEK_TO_END = 512 # Variable c_long
stc31 = 1118 # Variable c_int
SYSTEM_FONT = 13 # Variable c_int
LVS_EX_MULTIWORKAREAS = 8192 # Variable c_int
JOB_OBJECT_UILIMIT_NONE = 0 # Variable c_int
FILE_NOTIFY_CHANGE_LAST_WRITE = 16 # Variable c_int
HELP_WM_HELP = 12 # Variable c_int
TA_UPDATECP = 1 # Variable c_int
DOMAIN_ALIAS_RID_USERS = 545 # Variable c_long
PERF_DISPLAY_PER_SEC = 268435456 # Variable c_int
LOCALE_SABBREVDAYNAME2 = 50 # Variable c_int
IMAGE_DEBUG_TYPE_FPO = 3 # Variable c_int
IMM_ERROR_NODATA = -1 # Variable c_int
LHND = 66 # Variable c_int
PERF_COUNTER_MULTI_BASE = 1107494144 # Variable c_int
TAPE_DRIVE_CLEAN_REQUESTS = 33554432 # Variable c_int
CC_RGBINIT = 1 # Variable c_int
DATETIMEPICK_CLASSW = 'SysDateTimePick32' # Variable POINTER(c_wchar)
ILD_BLEND25 = 2 # Variable c_int
REG_NONE = 0 # Variable c_int
DRVCNF_RESTART = 2 # Variable c_int
MWT_LEFTMULTIPLY = 2 # Variable c_int
ERROR_CONNECTION_REFUSED = 1225 # Variable c_long
PDERR_NODEFAULTPRN = 4104 # Variable c_int
RESOURCEDISPLAYTYPE_FILE = 4 # Variable c_int
TCM_GETROWCOUNT = 4908 # Variable c_int
ERROR_NO_SECURITY_ON_OBJECT = 1350 # Variable c_long
ERROR_INVALID_SEPARATOR_FILE = 1799 # Variable c_long
EMR_EXCLUDECLIPRECT = 29 # Variable c_int
CLIENTSITE_E_FIRST = 2147746192 # Variable c_ulong
MB_TASKMODAL = 8192 # Variable c_long
RPC_C_QOS_IDENTITY_STATIC = 0 # Variable c_int
UPDFCACHE_ALL = 2147483647 # Variable c_ulong
JOB_STATUS_RESTART = 2048 # Variable c_int
CS_E_CLASS_NOTFOUND = -2147221146 # Variable c_long
CRYPT_E_ALREADY_DECRYPTED = -2146889719 # Variable c_long
DRAGDROP_S_FIRST = 262400 # Variable c_long
ERROR_RXACT_COMMIT_FAILURE = 1370 # Variable c_long
WNNC_NET_BMC = 1572864 # Variable c_int
CDM_GETFILEPATH = 1125 # Variable c_int
PRF_CLIENT = 4 # Variable c_long
C2_WHITESPACE = 10 # Variable c_int
S_SERQFUL = -4 # Variable c_int
CAL_SABBREVDAYNAME4 = 17 # Variable c_int
PROFILE_SERVER = 1073741824 # Variable c_int
FALT = 16 # Variable c_int
MSGF_DIALOGBOX = 0 # Variable c_int
IMC_OPENSTATUSWINDOW = 34 # Variable c_int
SECURITY_LOCAL_RID = 0 # Variable c_long
ERROR_NOT_SAME_DEVICE = 17 # Variable c_long
RB_SETPALETTE = 1061 # Variable c_int
TCM_SETCURSEL = 4876 # Variable c_int
STG_S_MULTIPLEOPENS = 197124 # Variable c_long
PID_SECURITY = 2147483650 # Variable c_uint
HELP_SETPOPUP_POS = 13 # Variable c_int
RT_STRING = 6 # Variable POINTER(c_wchar)
TBIF_TEXT = 2 # Variable c_int
TTF_TRACK = 32 # Variable c_int
ERROR_DEPENDENT_RESOURCE_EXISTS = 5001 # Variable c_long
RPC_C_OPT_MQ_TIME_TO_BE_RECEIVED = 8 # Variable c_int
CDRF_DODEFAULT = 0 # Variable c_int
JOYCAPS_HASZ = 1 # Variable c_int
CB_INITSTORAGE = 353 # Variable c_int
TVGN_ROOT = 0 # Variable c_int
WH_SHELL = 10 # Variable c_int
PRINTER_STATUS_TONER_LOW = 131072 # Variable c_int
SORT_JAPANESE_XJIS = 0 # Variable c_int
IMC_CLOSESTATUSWINDOW = 33 # Variable c_int
CBR_4800 = 4800 # Variable c_int
HWND_NOTOPMOST = 4294967294 # Variable POINTER(HWND__)
REG_NO_LAZY_FLUSH = 4 # Variable c_long
PFD_MAIN_PLANE = 0 # Variable c_int
ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE = 5031 # Variable c_long
LB_SETCARETINDEX = 414 # Variable c_int
RPC_X_NULL_REF_POINTER = 1780 # Variable c_long
edt2 = 1153 # Variable c_int
ES_LOWERCASE = 16 # Variable c_long
PHYSICALHEIGHT = 111 # Variable c_int
MCI_ANIM_PLAY_SLOW = 524288 # Variable c_long
PSM_ISDIALOGMESSAGE = 1141 # Variable c_int
FF_DECORATIVE = 80 # Variable c_int
CFSTR_MIME_PJPEG = 'image/pjpeg' # Variable POINTER(c_wchar)
lst6 = 1125 # Variable c_int
ROTFLAGS_ALLOWANYCLIENT = 2 # Variable c_int
LVM_GETITEMA = 4101 # Variable c_int
PERF_NUMBER_DECIMAL = 65536 # Variable c_int
VAR_FORMAT_NOSUBSTITUTE = 32 # Variable c_ulong
PERF_NUMBER_DEC_1000 = 131072 # Variable c_int
DRV_RESERVED = 2048 # Variable c_int
DMPAPER_ENV_B6 = 35 # Variable c_int
CO_E_WRONGOSFORAPP = -2147220998 # Variable c_long
SPI_SETSCREENSAVEACTIVE = 17 # Variable c_int
WGL_SWAP_UNDERLAY4 = 524288 # Variable c_int
DT_EXTERNALLEADING = 512 # Variable c_int
RPC_X_SS_CANNOT_GET_CALL_HANDLE = 1779 # Variable c_long
TAPE_DRIVE_REWIND_IMMEDIATE = 2147483656 # Variable c_uint
IMN_GUIDELINE = 13 # Variable c_int
URLACTION_DOWNLOAD_MAX = 4607 # Variable c_int
DSS_MONO = 128 # Variable c_int
RPC_S_NOT_LISTENING = 1715 # Variable c_long
NDR_ASCII_CHAR = 0 # Variable c_ulong
BS_PATTERN8X8 = 7 # Variable c_int
ERROR_PRINT_MONITOR_ALREADY_INSTALLED = 3006 # Variable c_long
PSINJECT_ENDPAGESETUP = 11 # Variable c_int
WGL_SWAP_OVERLAY4 = 16 # Variable c_int
TVS_RTLREADING = 64 # Variable c_int
LB_ADDSTRING = 384 # Variable c_int
IMC_SETSTATUSWINDOWPOS = 16 # Variable c_int
BI_RLE4 = 2 # Variable c_long
ICC_PAGESCROLLER_CLASS = 4096 # Variable c_int
PERF_DELTA_COUNTER = 4194304 # Variable c_int
DV_E_FORMATETC = -2147221404 # Variable c_long
TCM_DELETEALLITEMS = 4873 # Variable c_int
IMAGE_REL_PPC_SECREL16 = 15 # Variable c_int
SHGNLI_PREFIXNAME = 2 # Variable c_int
PDERR_DNDMMISMATCH = 4105 # Variable c_int
SE_SACL_PROTECTED = 8192 # Variable c_int
RPC_FLAGS_VALID_BIT = 32768 # Variable c_int
ENABLERELATIVEWIDTHS = 768 # Variable c_int
HEAP_PSEUDO_TAG_FLAG = 32768 # Variable c_int
PC_EXPLICIT = 2 # Variable c_int
PAN_MIDLINE_LOW_POINTED = 12 # Variable c_int
IMAGE_WEAK_EXTERN_SEARCH_ALIAS = 3 # Variable c_int
LOCALE_ICENTURY = 36 # Variable c_int
TBM_SETRANGE = 1030 # Variable c_int
PHYSICALOFFSETY = 113 # Variable c_int
VFFF_ISSHAREDFILE = 1 # Variable c_int
ENCAPSULATED_POSTSCRIPT = 4116 # Variable c_int
CS_IME = 65536 # Variable c_int
DMPAPER_A_PLUS = 57 # Variable c_int
DC_BINS = 6 # Variable c_int
RPC_S_UNKNOWN_AUTHZ_SERVICE = 1750 # Variable c_long
ILLUMINANT_B = 2 # Variable c_int
CTRY_LIBYA = 218 # Variable c_int
EM_GETTHUMB = 190 # Variable c_int
CO_E_CLSREG_INCONSISTENT = -2147467233 # Variable c_long
WAVE_FORMAT_2M16 = 64 # Variable c_int
CTRY_BULGARIA = 359 # Variable c_int
MARSHAL_S_FIRST = 262432 # Variable c_long
PROCESSOR_INTEL_486 = 486 # Variable c_int
ERROR_INVALID_LOGON_TYPE = 1367 # Variable c_long
TAPE_DRIVE_LOAD_UNLD_IMMED = 2147483680 # Variable c_uint
SND_ALIAS_SYSTEMEXIT = 17747 # Variable c_ulong
META_ELLIPSE = 1048 # Variable c_int
C3_ALPHA = 32768 # Variable c_int
KEYEVENTF_EXTENDEDKEY = 1 # Variable c_int
ERROR_PRINTER_DELETED = 1905 # Variable c_long
NRC_GOODRET = 0 # Variable c_int
TB_SETHOTITEM = 1096 # Variable c_int
ERROR_CANT_DISABLE_MANDATORY = 1310 # Variable c_long
MCI_VD_MODE_PARK = 1025 # Variable c_int
IMAGE_DIRECTORY_ENTRY_RESOURCE = 2 # Variable c_int
DDL_HIDDEN = 2 # Variable c_int
ERROR_INC_BACKUP = 4003 # Variable c_long
TVE_EXPAND = 2 # Variable c_int
DCB_RESET = 1 # Variable c_int
ERROR_REM_NOT_LIST = 51 # Variable c_long
PIDSI_PAGECOUNT = 14 # Variable c_long
LVM_SETCOLUMNWIDTH = 4126 # Variable c_int
SUBLANG_ARABIC_LEBANON = 12 # Variable c_int
PAGE_EXECUTE_READ = 32 # Variable c_int
RPC_C_USE_INTRANET_PORT = 2 # Variable c_int
MCI_CUT = 2129 # Variable c_int
WGL_SWAP_UNDERLAY14 = 536870912 # Variable c_int
ERROR_POINT_NOT_FOUND = 1171 # Variable c_long
ERROR_COUNTER_TIMEOUT = 1121 # Variable c_long
TTM_POP = 1052 # Variable c_int
PF_MMX_INSTRUCTIONS_AVAILABLE = 3 # Variable c_int
THREAD_PRIORITY_BELOW_NORMAL = -1 # Variable c_int
GETPRINTINGOFFSET = 13 # Variable c_int
META_SETTEXTCHAREXTRA = 264 # Variable c_int
PAN_LETT_OBLIQUE_SQUARE = 15 # Variable c_int
SS_WORDELLIPSIS = 49152 # Variable c_long
IMAGE_SYM_TYPE_UINT = 14 # Variable c_int
SEEK_CUR = 1 # Variable c_int
HTTOP = 12 # Variable c_int
ICM_ON = 2 # Variable c_int
WAVECAPS_PITCH = 1 # Variable c_int
PRINTER_CHANGE_ADD_PORT = 1048576 # Variable c_int
TB_SETPARENT = 1061 # Variable c_int
PERF_TYPE_ZERO = 3072 # Variable c_int
MIXERR_BASE = 1024 # Variable c_int
ERROR_SYSTEM_TRACE = 150 # Variable c_long
PRINTER_CHANGE_FAILED_CONNECTION_PRINTER = 8 # Variable c_int
IS_TEXT_UNICODE_REVERSE_STATISTICS = 32 # Variable c_int
DCX_CLIPCHILDREN = 8 # Variable c_long
SPI_SETDRAGHEIGHT = 77 # Variable c_int
MCI_ANIM_STATUS_FORWARD = 16386 # Variable c_long
PRINTER_NOTIFY_FIELD_UNTIL_TIME = 17 # Variable c_int
CO_E_SERVER_START_TIMEOUT = -2147467234 # Variable c_long
SPLDS_PRINT_NETWORK_ADDRESS = 'printNetworkAddress' # Variable POINTER(c_wchar)
WSAESHUTDOWN = 10058 # Variable c_int
IMAGE_SYM_CLASS_MEMBER_OF_UNION = 11 # Variable c_int
ERROR_EAS_NOT_SUPPORTED = 282 # Variable c_long
SETXON = 2 # Variable c_int
CRYPT_E_NOT_IN_REVOCATION_DATABASE = -2146885612 # Variable c_long
RPC_X_WRONG_ES_VERSION = 1828 # Variable c_long
SYSTEM_AUDIT_ACE_TYPE = 2 # Variable c_int
RPC_ASYNC_VERSION_1_0 = 68 # Variable c_uint
DT_RTLREADING = 131072 # Variable c_int
TVN_SINGLEEXPAND = 4294966881 # Variable c_uint
ERROR_TOO_MANY_SEMAPHORES = 100 # Variable c_long
META_SETMAPPERFLAGS = 561 # Variable c_int
RC_SAVEBITMAP = 64 # Variable c_int
TYPE_E_FIELDNOTFOUND = -2147319785 # Variable c_long
ERROR_ADAP_HDW_ERR = 57 # Variable c_long
ERROR_BAD_NET_RESP = 58 # Variable c_long
SC_MANAGER_QUERY_LOCK_STATUS = 16 # Variable c_int
MCI_STATUS_NUMBER_OF_TRACKS = 3 # Variable c_long
IMAGE_SCN_MEM_LOCKED = 262144 # Variable c_int
MIXERCONTROL_CONTROLTYPE_MIXER = 1895890945 # Variable c_long
ENUM_E_FIRST = 2147746224 # Variable c_ulong
PDERR_GETDEVMODEFAIL = 4101 # Variable c_int
SO_RCVTIMEO = 4102 # Variable c_int
SERVICE_BOOT_START = 0 # Variable c_int
COLOROKSTRINGW = 'commdlg_ColorOK' # Variable POINTER(c_wchar)
WRITE_DAC = 262144 # Variable c_long
MKF_AVAILABLE = 2 # Variable c_int
LOCALE_SABBREVMONTHNAME1 = 68 # Variable c_int
ERROR_HOOK_TYPE_NOT_ALLOWED = 1458 # Variable c_long
ERROR_SETCOUNT_ON_BAD_LB = 1433 # Variable c_long
WS_EX_OVERLAPPEDWINDOW = 768 # Variable c_long
BDR_RAISEDOUTER = 1 # Variable c_int
PAN_FAMILYTYPE_INDEX = 0 # Variable c_int
ETO_CLIPPED = 4 # Variable c_int
PERF_TIMER_TICK = 0 # Variable c_int
DFCS_PUSHED = 512 # Variable c_int
CO_E_PATHTOOLONG = -2147220974 # Variable c_long
RPC_S_UNSUPPORTED_NAME_SYNTAX = 1737 # Variable c_long
SERVICE_ADAPTER = 4 # Variable c_int
ERROR_WMI_UNRESOLVED_INSTANCE_REF = 4205 # Variable c_long
PROCESSOR_ARCHITECTURE_INTEL = 0 # Variable c_int
CREATE_THREAD_DEBUG_EVENT = 2 # Variable c_int
ERROR_DIR_NOT_ROOT = 144 # Variable c_long
CCM_SETUNICODEFORMAT = 8197 # Variable c_int
STD_UNDO = 3 # Variable c_int

STUB_PHASE = c_int # enum
ERROR_PIPE_CONNECTED = 535 # Variable c_long
LANG_FRENCH = 12 # Variable c_int
DDE_FAPPSTATUS = 255 # Variable c_int
UDM_GETRANGE = 1126 # Variable c_int
TCM_SETITEMSIZE = 4905 # Variable c_int
__RPCNDR_H_VERSION__ = 450 # Variable c_int
PGF_SCROLLRIGHT = 8 # Variable c_int
CCS_LEFT = 129 # Variable c_long
MIXERLINE_TARGETTYPE_MIDIOUT = 3 # Variable c_int
RB_GETBKCOLOR = 1044 # Variable c_int
CTRY_THAILAND = 66 # Variable c_int
LCMAP_TRADITIONAL_CHINESE = 67108864 # Variable c_int
PRINTER_STATUS_USER_INTERVENTION = 1048576 # Variable c_int
CONNDLG_PERSIST = 16 # Variable c_int
ERROR_INVALID_ACCOUNT_NAME = 1315 # Variable c_long
ERROR_CANNOT_IMPERSONATE = 1368 # Variable c_long
BDR_SUNKENOUTER = 2 # Variable c_int
ERROR_DS_BUSY = 1927 # Variable c_long
INET_E_NO_SESSION = -2146697213 # Variable c_long
SUBLANG_ENGLISH_CAN = 4 # Variable c_int
IME_PROP_COMPLETE_ON_UNSELECT = 1048576 # Variable c_int
TVIF_DI_SETITEM = 4096 # Variable c_int
TCS_SCROLLOPPOSITE = 1 # Variable c_int
DMDITHER_GRAYSCALE = 10 # Variable c_int
IMAGE_SYM_CLASS_END_OF_STRUCT = 102 # Variable c_int
DDE_FRELEASE = 8192 # Variable c_int
DMICM_COLORIMETRIC = 3 # Variable c_int
BSF_FORCEIFHUNG = 32 # Variable c_int
PSN_LAST = 4294966997 # Variable c_uint
LOCALE_SCURRENCY = 20 # Variable c_int
MCI_SEQ_MIDI = 16387 # Variable c_int
WS_POPUP = 2147483648 # Variable c_ulong
META_SETPALENTRIES = 55 # Variable c_int
WAVE_MAPPER = 4294967295 # Variable c_uint
PRINTER_ENUM_EXPAND = 16384 # Variable c_int
MCI_WAVE_SET_CHANNELS = 131072 # Variable c_long
VIEW_DETAILS = 3 # Variable c_int
SESSION_ABORTED = 6 # Variable c_int
WC_PAGESCROLLERA = 'SysPager' # Variable POINTER(c_char)
DRAWPATTERNRECT = 25 # Variable c_int
DKGRAY_BRUSH = 3 # Variable c_int
PERF_SIZE_LARGE = 256 # Variable c_int
IMAGE_REL_PPC_ABSOLUTE = 0 # Variable c_int
RBIM_IMAGELIST = 1 # Variable c_int
IMAGE_REL_ALPHA_LITERAL = 4 # Variable c_int
META_SETWINDOWEXT = 524 # Variable c_int
DMLERR_LAST = 16401 # Variable c_int
GETFACENAME = 513 # Variable c_int
ACS_AUTOPLAY = 4 # Variable c_int
CTRY_CANADA = 2 # Variable c_int
PRINTER_STATUS_OFFLINE = 128 # Variable c_int
USER_MARSHAL_FC_HYPER = 11 # Variable c_int
WMSZ_BOTTOM = 6 # Variable c_int
VARIANT_TRUE = -1 # Variable c_short
DMCOLOR_COLOR = 2 # Variable c_int
CALLBACK_TYPEMASK = 458752 # Variable c_long
DM_PANNINGHEIGHT = 268435456 # Variable c_long
SE_DACL_AUTO_INHERIT_REQ = 256 # Variable c_int
MMIOERR_CANNOTREAD = 261 # Variable c_int
DISPID_NEWENUM = -4 # Variable c_int
PAGE_EXECUTE_READWRITE = 64 # Variable c_int
ERROR_REGISTRY_CORRUPT = 1015 # Variable c_long
MMSYSERR_HANDLEBUSY = 12 # Variable c_int
IME_PROP_CANDLIST_START_FROM_1 = 262144 # Variable c_int
MARSHAL_E_LAST = 2147746095 # Variable c_ulong
PSINJECT_BEGINSTREAM = 0 # Variable c_int
DISP_E_OVERFLOW = -2147352566 # Variable c_long
RPC_C_PARM_MAX_PACKET_LENGTH = 1 # Variable c_int
DESKTOP_CREATEWINDOW = 2 # Variable c_long
SUBLANG_ENGLISH_SOUTH_AFRICA = 7 # Variable c_int
SUBLANG_ARABIC_UAE = 14 # Variable c_int
HDM_GETITEMW = 4619 # Variable c_int

TTN_NEEDTEXTW = TTN_GETDISPINFOW

MIIM_STATE      = 0x00000001
MIIM_ID         = 0x00000002
MIIM_SUBMENU    = 0x00000004
MIIM_CHECKMARKS = 0x00000008
MIIM_TYPE       = 0x00000010
MIIM_DATA       = 0x00000020
MIIM_STRING     = 0x00000040
MIIM_BITMAP     = 0x00000080
MIIM_FTYPE      = 0x00000100

INPUT_MOUSE     = 0
INPUT_KEYBOARD  = 1
INPUT_HARDWARE  = 2

OBJID_CLIENT  = 0xFFFFFFFC
KEYEVENTF_UNICODE    = 0x0004
KEYEVENTF_SCANCODE   = 0x0008


MN_GETHMENU        = 0x01E1

MNS_NOCHECK        = 0x80000000
MNS_MODELESS       = 0x40000000
MNS_DRAGDROP       = 0x20000000
MNS_AUTODISMISS    = 0x10000000
MNS_NOTIFYBYPOS    = 0x08000000
MNS_CHECKORBMP     = 0x04000000

MIM_MAXHEIGHT             =  0x00000001
MIM_BACKGROUND            =  0x00000002
MIM_HELPID                =  0x00000004
MIM_MENUDATA              =  0x00000008
MIM_STYLE                 =  0x00000010
MIM_APPLYTOSUBMENUS       =  0x80000000

SPIF_UPDATEINIFILE    = 0x0001
SPIF_SENDWININICHANGE = 0x0002
SPIF_SENDCHANGE       = SPIF_SENDWININICHANGE


SPI_GETMOUSEHOVERWIDTH   =  98
SPI_SETMOUSEHOVERWIDTH   =  99
SPI_GETMOUSEHOVERHEIGHT  = 100
SPI_SETMOUSEHOVERHEIGHT  = 101
SPI_GETMOUSEHOVERTIME    = 102
SPI_SETMOUSEHOVERTIME    = 103
SPI_GETWHEELSCROLLLINES  = 104
SPI_SETWHEELSCROLLLINES  = 105

SPI_GETSHOWIMEUI         = 110
SPI_SETSHOWIMEUI         = 111


SPI_GETMOUSESPEED        = 112
SPI_SETMOUSESPEED        = 113
SPI_GETSCREENSAVERRUNNING= 114

SPI_GETACTIVEWINDOWTRACKING        = 0x1000
SPI_SETACTIVEWINDOWTRACKING        = 0x1001
SPI_GETMENUANIMATION               = 0x1002
SPI_SETMENUANIMATION               = 0x1003
SPI_GETCOMBOBOXANIMATION           = 0x1004
SPI_SETCOMBOBOXANIMATION           = 0x1005
SPI_GETLISTBOXSMOOTHSCROLLING      = 0x1006
SPI_SETLISTBOXSMOOTHSCROLLING      = 0x1007
SPI_GETGRADIENTCAPTIONS            = 0x1008
SPI_SETGRADIENTCAPTIONS            = 0x1009
SPI_GETMENUUNDERLINES              = 0x100A
SPI_SETMENUUNDERLINES              = 0x100B
SPI_GETACTIVEWNDTRKZORDER          = 0x100C
SPI_SETACTIVEWNDTRKZORDER          = 0x100D
SPI_GETHOTTRACKING                 = 0x100E
SPI_SETHOTTRACKING                 = 0x100F
SPI_GETFOREGROUNDLOCKTIMEOUT       = 0x2000
SPI_SETFOREGROUNDLOCKTIMEOUT       = 0x2001
SPI_GETACTIVEWNDTRKTIMEOUT         = 0x2002
SPI_SETACTIVEWNDTRKTIMEOUT         = 0x2003
SPI_GETFOREGROUNDFLASHCOUNT        = 0x2004
SPI_SETFOREGROUNDFLASHCOUNT        = 0x2005

TBIF_BYINDEX           = 0x80000000
NM_LDOWN               = (NM_FIRST-20)

PBM_GETBARCOLOR = 0x040F
PBM_GETBKCOLOR = 0x040E
PBM_GETSTATE = 0x0411
PBM_GETSTEP = 0x040D # Windows Vista ~
PBM_SETBKCOLOR = 0x2001
PBM_SETMARQUEE = 0x040A # Windows XP ~
PBM_SETPOS = 0x0402
PBM_SETRANGE = 0x0401
PBM_SETRANGE32 = 0x0406 # Internet Explorer 3.0 ~
PBM_SETSTATE = 0x0410

PBS_MARQUEE = 0x08 # Windows XP ~
PBS_SMOOTHREVERSE = 0x10 # Windows Vista ~

PBST_NORMAL = 0x0001 # Windows Vista ~
PBST_ERROR = 0x0002
PBST_PAUSED = 0x0003

WAIT_TIMEOUT = 258 # dderror
MCM_GETCURRENTVIEW  = 4118
MCM_GETCALENDARCOUNT = 4119
MCM_SIZERECTTOMIN = 4125
MCM_SETCALENDARBORDER = 4126 # Variable c_int
MCM_GETCALENDARBORDER = 4127 # Variable c_int
MCM_SETCURRENTVIEW = 4128
MCMV_MONTH = 0
MCMV_YEAR = 1
MCMV_DECADE = 2
MCMV_CENTURY = 3

MCM_SETCALID = 4122 # Variable c_int
MCM_GETCALID = 4121 # Variable. c_int

CAL_GREGORIAN = 1 # Variable c_int
CAL_GREGORIAN_US = 2 # Variable c_int
CAL_JAPAN = 3 # Variable c_int
CAL_TAIWAN = 4 # Variable c_int
CAL_KOREA = 5 # Variable c_int
CAL_HIJRI = 6 # Variable c_int
CAL_THAI = 7 # Variable c_int
CAL_HEBREW = 8 # Variable c_int
CAL_GREGORIAN_ME_FRENCH = 9 # Variable c_int
CAL_GREGORIAN_ARABIC = 10 # Variable c_int
CAL_GREGORIAN_XLIT_ENGLISH = 11 # Variable c_int
CAL_GREGORIAN_XLIT_FRENCH = 12 # Variable c_int
CAL_UMALQURA = 23 # Variable c_int

MCSC_BACKGROUND = 0 # Variable c_int
MCSC_TEXT = 1 # Variable c_int
MCSC_TITLEBK = 2 # Variable c_int
MCSC_TITLETEXT = 3 # Variable c_int
MCSC_MONTHBK = 4 # Variable c_int
MCSC_TRAILINGTEXT = 5 # Variable c_int

# TODO: remove these duplicates and use definitions from win32con
VK_LBUTTON = 1  # Variable c_int
VK_RBUTTON = 2  # Variable c_int
VK_CANCEL = 3  # Variable c_int
VK_MBUTTON = 4  # Variable c_int
VK_BACK = 8  # Variable c_int
VK_TAB = 9  # Variable c_int
VK_CLEAR = 12  # Variable c_int
VK_RETURN = 13  # Variable c_int
VK_SHIFT = 16  # Variable c_int
VK_CONTROL = 17  # Variable c_int
VK_MENU = 18  # Variable c_int
VK_PAUSE = 19  # Variable c_int
VK_CAPITAL = 20  # Variable c_int
VK_HANGUL = 21  # Variable c_int
VK_HANGEUL = 21  # Variable c_int
VK_KANA = 21  # Variable c_int
VK_JUNJA = 23  # Variable c_int
VK_FINAL = 24  # Variable c_int
VK_KANJI = 25  # Variable c_int
VK_HANJA = 25  # Variable c_int
VK_ESCAPE = 27  # Variable c_int
VK_CONVERT = 28  # Variable c_int
VK_NONCONVERT = 29  # Variable c_int
VK_ACCEPT = 30  # Variable c_int
VK_MODECHANGE = 31  # Variable c_int
VK_SPACE = 32  # Variable c_int
VK_PRIOR = 33  # Variable c_int
VK_NEXT = 34  # Variable c_int
VK_END = 35  # Variable c_int
VK_HOME = 36  # Variable c_int
VK_LEFT = 37  # Variable c_int
VK_UP = 38  # Variable c_int
VK_RIGHT = 39  # Variable c_int
VK_DOWN = 40  # Variable c_int
VK_SELECT = 41  # Variable c_int
VK_PRINT = 42  # Variable c_int
VK_EXECUTE = 43  # Variable c_int
VK_SNAPSHOT = 44  # Variable c_int
VK_INSERT = 45  # Variable c_int
VK_DELETE = 46  # Variable c_int
VK_HELP = 47  # Variable c_int
VK_LWIN = 91  # Variable c_int
VK_RWIN = 92  # Variable c_int
VK_APPS = 93  # Variable c_int
VK_NUMPAD0 = 96  # Variable c_int
VK_NUMPAD1 = 97  # Variable c_int
VK_NUMPAD2 = 98  # Variable c_int
VK_NUMPAD3 = 99  # Variable c_int
VK_NUMPAD4 = 100  # Variable c_int
VK_NUMPAD5 = 101  # Variable c_int
VK_NUMPAD6 = 102  # Variable c_int
VK_NUMPAD7 = 103  # Variable c_int
VK_NUMPAD8 = 104  # Variable c_int
VK_NUMPAD9 = 105  # Variable c_int
VK_MULTIPLY = 106  # Variable c_int
VK_ADD = 107  # Variable c_int
VK_SEPARATOR = 108  # Variable c_int
VK_SUBTRACT = 109  # Variable c_int
VK_DECIMAL = 110  # Variable c_int
VK_DIVIDE = 111  # Variable c_int
VK_F1 = 112  # Variable c_int
VK_F2 = 113  # Variable c_int
VK_F3 = 114  # Variable c_int
VK_F4 = 115  # Variable c_int
VK_F5 = 116  # Variable c_int
VK_F6 = 117  # Variable c_int
VK_F7 = 118  # Variable c_int
VK_F8 = 119  # Variable c_int
VK_F9 = 120  # Variable c_int
VK_F10 = 121  # Variable c_int
VK_F11 = 122  # Variable c_int
VK_F12 = 123  # Variable c_int
VK_F13 = 124  # Variable c_int
VK_F14 = 125  # Variable c_int
VK_F15 = 126  # Variable c_int
VK_F16 = 127  # Variable c_int
VK_F17 = 128  # Variable c_int
VK_F18 = 129  # Variable c_int
VK_F19 = 130  # Variable c_int
VK_F20 = 131  # Variable c_int
VK_F21 = 132  # Variable c_int
VK_F22 = 133  # Variable c_int
VK_F23 = 134  # Variable c_int
VK_F24 = 135  # Variable c_int
VK_NUMLOCK = 144  # Variable c_int
VK_SCROLL = 145  # Variable c_int
VK_LSHIFT = 160  # Variable c_int
VK_RSHIFT = 161  # Variable c_int
VK_LCONTROL = 162  # Variable c_int
VK_RCONTROL = 163  # Variable c_int
VK_LMENU = 164  # Variable c_int
VK_RMENU = 165  # Variable c_int
VK_PROCESSKEY = 229  # Variable c_int
VK_PACKET = 231
VK_ATTN = 246  # Variable c_int
VK_CRSEL = 247  # Variable c_int
VK_EXSEL = 248  # Variable c_int
VK_EREOF = 249  # Variable c_int
VK_PLAY = 250  # Variable c_int
VK_ZOOM = 251  # Variable c_int
VK_NONAME = 252  # Variable c_int
VK_PA1 = 253  # Variable c_int
VK_OEM_CLEAR = 254  # Variable c_int

# TODO: remove these duplicates and use definitions from win32con
WM_NULL = 0  # Variable c_int
WM_CREATE = 1  # Variable c_int
WM_DESTROY = 2  # Variable c_int
WM_MOVE = 3  # Variable c_int
WM_SIZE = 5  # Variable c_int
WM_ACTIVATE = 6  # Variable c_int
WM_SETFOCUS = 7  # Variable c_int
WM_KILLFOCUS = 8  # Variable c_int
WM_ENABLE = 10  # Variable c_int
WM_SETREDRAW = 11  # Variable c_int
WM_SETTEXT = 12  # Variable c_int
WM_GETTEXT = 13  # Variable c_int
WM_GETTEXTLENGTH = 14  # Variable c_int
WM_PAINT = 15  # Variable c_int
WM_CLOSE = 16  # Variable c_int
WM_QUERYENDSESSION = 17  # Variable c_int
WM_QUIT = 18  # Variable c_int
WM_QUERYOPEN = 19  # Variable c_int
WM_ERASEBKGND = 20  # Variable c_int
WM_SYSCOLORCHANGE = 21  # Variable c_int
WM_ENDSESSION = 22  # Variable c_int
WM_SHOWWINDOW = 24  # Variable c_int
WM_WININICHANGE = 26  # Variable c_int
WM_DEVMODECHANGE = 27  # Variable c_int
WM_ACTIVATEAPP = 28  # Variable c_int
WM_FONTCHANGE = 29  # Variable c_int
WM_TIMECHANGE = 30  # Variable c_int
WM_CANCELMODE = 31  # Variable c_int
WM_SETCURSOR = 32  # Variable c_int
WM_MOUSEACTIVATE = 33  # Variable c_int
WM_CHILDACTIVATE = 34  # Variable c_int
WM_QUEUESYNC = 35  # Variable c_int
WM_GETMINMAXINFO = 36  # Variable c_int
WM_PAINTICON = 38  # Variable c_int
WM_ICONERASEBKGND = 39  # Variable c_int
WM_NEXTDLGCTL = 40  # Variable c_int
WM_SPOOLERSTATUS = 42  # Variable c_int
WM_DRAWITEM = 43  # Variable c_int
WM_MEASUREITEM = 44  # Variable c_int
WM_DELETEITEM = 45  # Variable c_int
WM_VKEYTOITEM = 46  # Variable c_int
WM_CHARTOITEM = 47  # Variable c_int
WM_SETFONT = 48  # Variable c_int
WM_GETFONT = 49  # Variable c_int
WM_SETHOTKEY = 50  # Variable c_int
WM_GETHOTKEY = 51  # Variable c_int
WM_QUERYDRAGICON = 55  # Variable c_int
WM_COMPAREITEM = 57  # Variable c_int
WM_COMPACTING = 65  # Variable c_int
WM_COMMNOTIFY = 68  # Variable c_int
WM_WINDOWPOSCHANGING = 70  # Variable c_int
WM_WINDOWPOSCHANGED = 71  # Variable c_int
WM_POWER = 72  # Variable c_int
WM_COPYDATA = 74  # Variable c_int
WM_CANCELJOURNAL = 75  # Variable c_int
WM_NOTIFY = 78  # Variable c_int
WM_INPUTLANGCHANGEREQUEST = 80  # Variable c_int
WM_INPUTLANGCHANGE = 81  # Variable c_int
WM_TCARD = 82  # Variable c_int
WM_HELP = 83  # Variable c_int
WM_USERCHANGED = 84  # Variable c_int
WM_NOTIFYFORMAT = 85  # Variable c_int
WM_CONTEXTMENU = 123  # Variable c_int
WM_STYLECHANGING = 124  # Variable c_int
WM_STYLECHANGED = 125  # Variable c_int
WM_DISPLAYCHANGE = 126  # Variable c_int
WM_GETICON = 127  # Variable c_int
WM_SETICON = 128  # Variable c_int
WM_NCCREATE = 129  # Variable c_int
WM_NCDESTROY = 130  # Variable c_int
WM_NCCALCSIZE = 131  # Variable c_int
WM_NCHITTEST = 132  # Variable c_int
WM_NCPAINT = 133  # Variable c_int
WM_NCACTIVATE = 134  # Variable c_int
WM_GETDLGCODE = 135  # Variable c_int
WM_SYNCPAINT = 136  # Variable c_int
WM_NCMOUSEMOVE = 160  # Variable c_int
WM_NCLBUTTONDOWN = 161  # Variable c_int
WM_NCLBUTTONUP = 162  # Variable c_int
WM_NCLBUTTONDBLCLK = 163  # Variable c_int
WM_NCRBUTTONDOWN = 164  # Variable c_int
WM_NCRBUTTONUP = 165  # Variable c_int
WM_NCRBUTTONDBLCLK = 166  # Variable c_int
WM_NCMBUTTONDOWN = 167  # Variable c_int
WM_NCMBUTTONUP = 168  # Variable c_int
WM_NCMBUTTONDBLCLK = 169  # Variable c_int
WM_KEYDOWN = 256  # Variable c_int
WM_KEYFIRST = 256  # Variable c_int
WM_KEYUP = 257  # Variable c_int
WM_CHAR = 258  # Variable c_int
WM_DEADCHAR = 259  # Variable c_int
WM_SYSKEYDOWN = 260  # Variable c_int
WM_SYSKEYUP = 261  # Variable c_int
WM_SYSCHAR = 262  # Variable c_int
WM_SYSDEADCHAR = 263  # Variable c_int
WM_KEYLAST = 264  # Variable c_int
WM_IME_STARTCOMPOSITION = 269  # Variable c_int
WM_IME_ENDCOMPOSITION = 270  # Variable c_int
WM_IME_KEYLAST = 271  # Variable c_int
WM_IME_COMPOSITION = 271  # Variable c_int
WM_INITDIALOG = 272  # Variable c_int
WM_COMMAND = 273  # Variable c_int
WM_SYSCOMMAND = 274  # Variable c_int
WM_TIMER = 275  # Variable c_int
WM_HSCROLL = 276  # Variable c_int
WM_VSCROLL = 277  # Variable c_int
WM_INITMENU = 278  # Variable c_int
WM_INITMENUPOPUP = 279  # Variable c_int
WM_MENUSELECT = 287  # Variable c_int
WM_MENUCHAR = 288  # Variable c_int
WM_ENTERIDLE = 289  # Variable c_int
WM_MENUCOMMAND = 294
WM_CTLCOLORMSGBOX = 306  # Variable c_int
WM_CTLCOLOREDIT = 307  # Variable c_int
WM_CTLCOLORLISTBOX = 308  # Variable c_int
WM_CTLCOLORBTN = 309  # Variable c_int
WM_CTLCOLORDLG = 310  # Variable c_int
WM_CTLCOLORSCROLLBAR = 311  # Variable c_int
WM_CTLCOLORSTATIC = 312  # Variable c_int
WM_MOUSEMOVE = 512   # Variable c_int
WM_MOUSEFIRST = 512  # Variable c_int
WM_LBUTTONDOWN = 513  # Variable c_int
WM_LBUTTONUP = 514  # Variable c_int
WM_LBUTTONDBLCLK = 515  # Variable c_int
WM_RBUTTONDOWN = 516  # Variable c_int
WM_RBUTTONUP = 517  # Variable c_int
WM_RBUTTONDBLCLK = 518  # Variable c_int
WM_MBUTTONDOWN = 519  # Variable c_int
WM_MBUTTONUP = 520  # Variable c_int
WM_MBUTTONDBLCLK = 521  # Variable c_int
WM_MOUSELAST = 521  # Variable c_int
WM_MOUSEWHEEL = 522
WM_PARENTNOTIFY = 528  # Variable c_int
WM_ENTERMENULOOP = 529  # Variable c_int
WM_EXITMENULOOP = 530  # Variable c_int
WM_NEXTMENU = 531  # Variable c_int
WM_SIZING = 532  # Variable c_int
WM_CAPTURECHANGED = 533  # Variable c_int
WM_MOVING = 534  # Variable c_int
WM_POWERBROADCAST = 536  # Variable c_int
WM_DEVICECHANGE = 537  # Variable c_int
WM_MDICREATE = 544  # Variable c_int
WM_MDIDESTROY = 545  # Variable c_int
WM_MDIACTIVATE = 546  # Variable c_int
WM_MDIRESTORE = 547  # Variable c_int
WM_MDINEXT = 548  # Variable c_int
WM_MDIMAXIMIZE = 549  # Variable c_int
WM_MDITILE = 550  # Variable c_int
WM_MDICASCADE = 551  # Variable c_int
WM_MDIICONARRANGE = 552  # Variable c_int
WM_MDIGETACTIVE = 553  # Variable c_int
WM_MDISETMENU = 560  # Variable c_int
WM_ENTERSIZEMOVE = 561  # Variable c_int
WM_EXITSIZEMOVE = 562  # Variable c_int
WM_DROPFILES = 563  # Variable c_int
WM_MDIREFRESHMENU = 564  # Variable c_int
WM_IME_SETCONTEXT = 641  # Variable c_int
WM_IME_NOTIFY = 642  # Variable c_int
WM_IME_CONTROL = 643  # Variable c_int
WM_IME_COMPOSITIONFULL = 644  # Variable c_int
WM_IME_SELECT = 645  # Variable c_int
WM_IME_CHAR = 646  # Variable c_int
WM_IME_KEYDOWN = 656  # Variable c_int
WM_IME_KEYUP = 657  # Variable c_int
WM_MOUSEHOVER = 673  # Variable c_int
WM_MOUSELEAVE = 675  # Variable c_int
WM_CUT = 768  # Variable c_int
WM_COPY = 769  # Variable c_int
WM_PASTE = 770  # Variable c_int
WM_CLEAR = 771  # Variable c_int
WM_UNDO = 772  # Variable c_int
WM_RENDERFORMAT = 773  # Variable c_int
WM_RENDERALLFORMATS = 774  # Variable c_int
WM_DESTROYCLIPBOARD = 775  # Variable c_int
WM_DRAWCLIPBOARD = 776  # Variable c_int
WM_PAINTCLIPBOARD = 777  # Variable c_int
WM_VSCROLLCLIPBOARD = 778  # Variable c_int
WM_SIZECLIPBOARD = 779  # Variable c_int
WM_ASKCBFORMATNAME = 780  # Variable c_int
WM_CHANGECBCHAIN = 781  # Variable c_int
WM_HSCROLLCLIPBOARD = 782  # Variable c_int
WM_QUERYNEWPALETTE = 783  # Variable c_int
WM_PALETTEISCHANGING = 784  # Variable c_int
WM_PALETTECHANGED = 785  # Variable c_int
WM_HOTKEY = 786  # Variable c_int
WM_PRINT = 791  # Variable c_int
WM_PRINTCLIENT = 792  # Variable c_int
WM_HANDHELDFIRST = 856  # Variable c_int
WM_HANDHELDLAST = 863  # Variable c_int
WM_AFXFIRST = 864  # Variable c_int
WM_AFXLAST = 895  # Variable c_int
WM_PENWINFIRST = 896  # Variable c_int
WM_PENWINLAST = 911  # Variable c_int
WM_DDE_INITIATE = 992  # Variable c_int
WM_DDE_FIRST = 992  # Variable c_int
WM_DDE_TERMINATE = 993  # Variable c_int
WM_DDE_ADVISE = 994  # Variable c_int
WM_DDE_UNADVISE = 995  # Variable c_int
WM_DDE_ACK = 996  # Variable c_int
WM_DDE_DATA = 997  # Variable c_int
WM_DDE_REQUEST = 998  # Variable c_int
WM_DDE_POKE = 999  # Variable c_int
WM_DDE_EXECUTE = 1000  # Variable c_int
WM_DDE_LAST = 1000  # Variable c_int
WM_PSD_PAGESETUPDLG = 1024  # Variable c_int
WM_USER = 1024  # Variable c_int
WM_CHOOSEFONT_GETLOGFONT = 1025  # Variable c_int
WM_PSD_FULLPAGERECT = 1025  # Variable c_int
WM_PSD_MINMARGINRECT = 1026  # Variable c_int
WM_PSD_MARGINRECT = 1027  # Variable c_int
WM_PSD_GREEKTEXTRECT = 1028  # Variable c_int
WM_PSD_ENVSTAMPRECT = 1029  # Variable c_int
WM_PSD_YAFULLPAGERECT = 1030  # Variable c_int
WM_CHOOSEFONT_SETLOGFONT = 1125  # Variable c_int
WM_CHOOSEFONT_SETFLAGS = 1126  # Variable c_int
WM_APP = 32768  # Variable c_int
PROCESS_STILL_ACTIVE = 259 # Variable c_int
